#!/path/to/Rscript --vanilla
#################################################################################
# R-Script for the graphical output of the posterior probabilities              #
#  for each subtype for the given query sequence(s)                             #
#  if the sequence is located in a part with consensus columns in the alignment #
#                                                                               #
# author: Anne-Kathrin Schultz, email: anne@gobics.de                           #
#################################################################################


#################################################################################
# program call (Linux command line):                                            #
#    R --no-save < R_posterior_probabilities_plot.R --args al_type >/dev/null   #
# e.g. R --no-save < R_posterior_probabilities_plot.R --args jpHMM >/dev/null   #
#################################################################################

#################################################################################
# needed input files:                                                           #
# in the default format as described in the DOCUMENTATION of the jpHMM program  #
#  - posterior_probabilities_for_seq_*.txt                                      #
#      (* = current considered sequence, = numSeq)                              #
#  - recombination_without_positions.txt                                        #
#################################################################################
# parameter:                                                                    #
# al_type = type of alignment that should be used for plotting                  #
#           the posterior probabilities based on reference numbering:           #
#  - 'jpHMM' (alignment of query to ref seq determined by jpHMM (df) OR         #
#  - 'SLT' (alignment of query to ref seq determined by LANL                    #
#           HIV Sequence Locator Tool; if tool is given and                     #
#           run by the jpHMM pipeline!)                                         #
#           -> filename can be changed in l.88)                                 #
#                                                                               #
# If no ref seq is given the posterior probabilities are plotted based on       #
# raw sequence positions !!                                                     #
#################################################################################

#################################################################################
# Ghostscript must be installed !
# for bitmap
# for png X11 needed
#################################################################################	

########################
#    Initialization    #
########################
library(seqinr);
args    <- commandArgs(TRUE);
al_type <- "jpHMM";             	# default
if(length(args) == 1) {
	al_type <- args[1];
}

# Assignment of numbers to the subtypes
subtypes        <- c(1:16);
names(subtypes) <- c("5'-Insertion", "A1", "A2", "B", "C", "D", "F1", "F2", "G", "H", "J", "K", "O", "CPZ", "01_AE", "3'-Insertion")

# number of considered subtypes, incl. 2 Insert regions
num_subtypes = 16;

# Color palette for the 'real' subtypes (can be changed) (no Insertions): 
col <- colorRampPalette(c("#fc4128", "#fc4128", "#00aaff", "#b08155", "#faacd5", "#d0f000", "#d0f000", "#6bc772", "#ffd400", "#00faff", "#b95fff", "#cc9900", "#969696", "#ffff00"))
r    = col(num_subtypes-2); # do not count 5'- and 3'-insertion as real subtype

#color for Insertion: "#000000" (could be added to col at the beginning and end of col -> change also plot command (from 1 to 16) )

#col <- c([0 0 0; 0.99 0.25 0.16; 0.99 0.25 0.16; 0 0.67 1; 0.69 0.51 0.33; 0.98 0.67 0.84; 0.82 1 0; 0.82 1 0; 0.42 0.78 0.45; 1 0.83 0; 0 0.98 1; 0.73 0.37 1; 0.7 0.6 0; 0.5 0.5 0.5; 1 1 0; 0 0 0];)
#     N_I     A1      A2      B       C       D       F1     F2     G       H         J         K       O      CPZ    CRF_01  C_I
#    black        red        blue   brown lightred   yellowgreen  green ocher/yellow turquoise violet  ocher   gray  yellow   black

# Input files #

RECOMB <- file("recombination_without_positions.txt","r");
line    = scan(RECOMB, what=list("character","character","character"), sep="\t", nlines=-1, fill = TRUE, quiet = TRUE, comment.char = "#");
len     = length(line[[1]]);

# format: line[[1]] = sequence names
#         line[[2]] = parameters
#         line[[3]] = predicted subtypes (list)


# File containing the alignments of the query sequences to the ref sequence
# user can choose between the 'jpHMM predicted alignment' ("query_to_ref_alignments.txt") 
#                         and alignment derived from the 'sequence locator tool' (LANL) ("query_to_ref_by_SLT.txt") (if tool is provided)
# If no ref seq is contained in the given multiple sequence alignment, these files cannot be produced
# so, if no ref seq is given, the posterior probabilities are plotted based on raw sequence positions

al_filename <- "query_to_ref_alignments.txt";
if (identical(al_type, "SLT") == TRUE) {
	al_filename <- "query_to_ref_by_SLT.txt";
}

# check if alignment file not empty
ALIGNMENT <- file(al_filename,"r");  # default
al <- read.fasta(file = ALIGNMENT, set.attributes = TRUE); # vector of single characters

al_type <- "RAW";

numSeq       = 1;   # num tested sequences
seq_name     = 0;
numFailedSeq = 0;   # num tested sequences without predicted recombination (because of missing information ('N/A'))

for (row in 1:len) {
	if (line[[3]][row] == 'N/A') { 
		# -> no recombination predicted and no posterior probabilities calculated
		numSeq <- numSeq + 1;
		numFailedSeq <- numFailedSeq + 1;
    	}
    	else {
		seq_name = line[[1]][row];

		# Read in the matrix of the posterior probabilities for the 'numSeq' different sequences
		filename = paste("posterior_probabilities_for_seq_",numSeq,".txt", sep="");
		m        = read.table(filename,sep=' ');
		names    = colnames(m);

		# Query sequence positions:
		seq_size = length(m[,names[1]]);
		y        = 1:seq_size;
	
		if (identical(al_type, "RAW") == TRUE) {
			plot_pos = y;
		}
		else {
			# Query seq positions based on ref seq numbering

			cur_seq_no <- 0;    # no of seq in alignment file (1st: ref seq, 2nd: seq1, 3rd: ref seq, 4th: seq2,...)
			if (identical(al_type, "SLT") == TRUE) {
				cur_seq_no <- numSeq*2;
			}
			else {
				cur_seq_no <- (numSeq-numFailedSeq)*2;
			}
		
			seq        <- al[[cur_seq_no]];
			seq_info   <- strsplit(attr(seq, "Annot"), " ");
		
			ref_seq       <- al[[cur_seq_no-1]];
			ref_seq_info  <- strsplit(attr(ref_seq, "Annot"), " ");
			ref_seq_begin <- as.numeric(ref_seq_info[[1]][2]);      # ref_seq_info[[1]][1] = name = ref_seq
			ref_seq_end   <- as.numeric(ref_seq_info[[1]][3]);

	
			plot_pos      <- rep(0, seq_size);
			cur_plot_pos  <- 1;
			cur_ref_seq_pos  <- ref_seq_begin;
			count_inserts <- 0;

			for (pos in 1:length(seq)) {
				# run through alignment
				if (identical(seq[pos], '-') == FALSE && identical(seq[pos], '*') == FALSE) {
					if (identical(ref_seq[pos], '-') == FALSE && identical(ref_seq[pos], '*') == FALSE) { # alignment/match of seq to ref seq
						# check for preceding insertions in query
		
						if (count_inserts > 0) {
							# insert data points between preceding (prec_pos) and current (cur_ref_seq_pos) ref_seq position
							if (identical(cur_plot_pos, 1) == TRUE) {
								prec_pos <- cur_ref_seq_pos-1;
							}
							else {
 								prec_pos = plot_pos[cur_plot_pos-1];
							}
							distance = (cur_ref_seq_pos-prec_pos)/(count_inserts+1);
							for (d in 1:count_inserts) {
								plot_pos[cur_plot_pos] <- (prec_pos + d*distance);
								cur_plot_pos           <- cur_plot_pos + 1;
							}
						}
						plot_pos[cur_plot_pos] <- cur_ref_seq_pos;
						cur_plot_pos           <- cur_plot_pos + 1;

						cur_ref_seq_pos  <- cur_ref_seq_pos + 1;
						count_inserts <- 0;
					}
					else { # insert in seq
						count_inserts <- count_inserts + 1;
						# mapping when next match occurs
					}
				}
				else {
					if (identical(ref_seq[pos], '-') == FALSE && identical(ref_seq[pos], '*') == FALSE) { # delete in seq
						# to current ref seq position no posterior probabilities are assigned to	
						# -> current ref seq position won't be included in vector 'plot_pos'
						# increase current ref seq pos by 1
						cur_ref_seq_pos <- cur_ref_seq_pos + 1;
					}
					# else: both '-' => nothing happens
				}
			}
			# check for insertions at the end of the query
			if (count_inserts > 0) {
				# insert data points between preceding (prec_pos) and current (cur_ref_seq_pos) ref seq position
				prec_pos <- cur_ref_seq_pos;
				dist = 1/(count_inserts+1);
				for (d in 1:count_inserts) {
					plot_pos[cur_plot_pos] <- (prec_pos + d*dist);
					cur_plot_pos <- cur_plot_pos + 1;
				}
			}
		}
		##############################################################
		#
		# Graphical output
		#
		##############################################################
		# Open png-file for graphical output: 
	
		filename = paste("posterior_prob_plot_for_seq_",numSeq,".png", sep="");
		bitmap(file=filename, type = "png256", width=13, height=5, res = 72, pointsize=12);

	   
		##############################################################
		# plot of the posterior probabilities
		#     for each query sequence position
		#     and each subtype
		##############################################################
		op <- par(lend=1, ljoin=2,xpd=FALSE,cex.main=1,pin=c(6,3),bty="n", pty="m", xaxs="r", usr=c(plot_pos[1],plot_pos[length(plot_pos)], 0, 1));
	
		xlabel=paste("Sequence position", "( len=", seq_size, ")", "based on reference sequence (based on jpHMM output):", floor(plot_pos[1]), "-", ceiling(plot_pos[length(plot_pos)]), sep=" ");

		y <- plot_pos;

		par(op);

		# plot only 'real' subtypes, no insertions (=N/A) (names[2] - names[num_subtypes-1])
		plot(y, m[,names[2]], type='l', las=1, lwd=3, ylab='Posterior probability', xlab=xlabel, col=r[1], xlim=range((plot_pos[1]*10):((plot_pos[length(plot_pos)]+(seq_size/10))*10))/10, ylim=range(0:11)/10);

		for (j in 3:num_subtypes-1) {
			if (j == 3 || j == 8) {
				lines(y, m[,names[j]], col=r[j-1], lty="dashed", lwd=3);
			}
			else {
				lines(y, m[,names[j]], col=r[j-1], lwd=3);
			}
		}
		legend(plot_pos[length(plot_pos)]+1, 1.15, c('A1','A2','B','C','D','F1','F2','G','H','J','K','O','CPZ','01_AE'),col=c(r), lty=c(1,2,1,1,1,1,2,1,1,1,1,1,1,1),lwd=3);
			
		dev.off();
		numSeq <- numSeq + 1;
	}
}