// ~/jpHMM/src/alignment.cpp
/**************************************************************
 File name	: alignment.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Read in of information about 
  - subtypes
  - sequences 
  - reference genome sequence 
 in the alignment
***************************************************************/

#include "std_source.h"
#include "error.h"
#include "const_def.h"
#include "alignment.h"


using namespace std;


/*******************************************************\

 *                 class Alignment                     *

\*******************************************************/

//------------------------------------------------------------------------------
// Alignment():
//------------------------------------------------------------------------------
Alignment::Alignment(string infile, 
		     bool   is_extended, 
		     int    orig_len) {

  ifstream ifstrm(infile.c_str(), ios::in);
  if (!ifstrm) {
    stringstream ss;
    ss << "Failure of the alignment input file: ";
    ss <<  infile;
    throw Error(ss.str());
  }
  
  m_is_extended = is_extended;
  m_orig_len = orig_len;
  m_subs.clear();
  m_seqs.clear();

  m_has_ref_seq            = false;
  m_ref_seq                = 0; //! reference sequence
  m_num_residues           = 0; //! number of residues in alignment
  m_num_considered_columns = 0;


  Subtype* cur_sub      = 0;
  string   cur_sub_name = "";
  string   cur_seq_str  = "";
  string   cur_seq_name = "";
  bool     has_cur_sub  = false;
  bool     has_cur_seq  = false;
  string   ref_seq_name = Const_Def::virus_ref_seq_name(Const_Def::virus_type());
  int      pos_ref_seq  = -1;
  string   str;

  while (!ifstrm.eof()) {
    getline(ifstrm, str);
    if (str.substr(0, 2) == ">>") {
      has_cur_sub  = true;
      cur_sub_name = str.substr(2);
      cur_sub      = new Subtype(cur_sub_name);
      m_subs      += cur_sub;
    }
    else
      if (str.substr(0, 1) == ">") {
	has_cur_seq  = true;
	cur_seq_name = str.substr(1);
	cur_seq_str  = "";

	//! reference sequence exists?:
	if (cur_seq_name.find(ref_seq_name) != string::npos) {
	  m_has_ref_seq = true;
	  pos_ref_seq   = m_seqs.size();
	}
      }
      else
	if (!has_cur_sub || !has_cur_seq)
	  throw Error("Alignment in illegal state. No subtype name or sequence name given.");
	else {
	  cur_seq_str += str;
	  bool end_of_seq_reached = false;
	  if (!ifstrm.eof()) {
	    char c;
	    ifstrm >> c;
	    if (ifstrm.eof() || c == '>') { //  || c == '\n' || c == '\r') {
	      end_of_seq_reached = true;
	    }
	    if (!ifstrm.eof())
	      ifstrm.unget();
	  }
	  else
	    end_of_seq_reached = true;
	  if (end_of_seq_reached) {
	    //! save raw sequence without gaps:
	    for (unsigned int i=0; i<cur_seq_str.length(); ++i)
              if (cur_seq_str[i] != '-' && cur_seq_str[i] != '*' && cur_seq_str[i] != '.')
                ++m_num_residues;

            Seq* cur_seq     = new Seq(cur_seq_name, cur_sub, cur_seq_str);
	    m_seqs          += cur_seq;
	    cur_sub->seqs() += cur_seq;
	  }
	}
  }

  //! length of each subtype = number of columns (sequences have all same length):
  int len = m_seqs[0]->str().length();

  for (unsigned s=0; s<m_subs.size(); ++s) {
    Subtype* st = m_subs[s];
    st->len() = len;
  }

  if (!m_is_extended) {
    //! original length of alignment is current length (e.g. alignment is not extended for linear sequences)
    m_orig_len = len;
  }

  //! reference sequence:
  if (m_has_ref_seq && pos_ref_seq != -1)
    m_ref_seq = m_seqs[pos_ref_seq];
  
  check();
}

//------------------------------------------------------------------------------
// ~Alignment():
//------------------------------------------------------------------------------
Alignment::~Alignment(){
  FOREACH (Subtype* st, m_subs)
    delete st;
}

//------------------------------------------------------------------------------
// void check():
//
//------------------------------------------------------------------------------
void Alignment::check(){

  int    num_seqs     = 0;
  string cur_sub_name = "";

  if (m_subs.size() == 0)
    throw Error("Please define subtypes in the alignment in the following way:\n>>A\n>seq1\nacgtaag\n");

  FOREACH (Subtype* st, m_subs) {
    st->check();
    cur_sub_name          = st->name();
    vector<Seq*> cur_seqs = st->seqs();
    num_seqs             += cur_seqs.size();

    FOREACH (Seq* s, cur_seqs)
      if (s->sub()->name() != cur_sub_name)
        throw Error("Alignment in illegal state");
  }
  if ((unsigned int) num_seqs != m_seqs.size())
    throw Error("Alignment in illegal state");

  FOREACH (Seq* s, m_seqs) {
    s->check();
    if (s->str().length() != m_seqs[0]->str().length())
      throw Error("Seqs in alignment have different lengths");
  }
  // check ref seq:
  if (m_has_ref_seq && !m_ref_seq)
    throw Error("Alignment in illegal state");
}

//------------------------------------------------------------------------------
// check if alignment contains a reference genome sequence
//------------------------------------------------------------------------------
bool Alignment::has_ref_seq() const {

  if (!m_has_ref_seq) 
    return false; 
  return true;
}

//------------------------------------------------------------------------------
// output of reference sequence (if available):
//------------------------------------------------------------------------------
Seq* Alignment::ref_seq() const {

  if (m_has_ref_seq) 
    return m_ref_seq; 
  else 
    throw Error("no reference sequence available");
}

//--------------------------------------------------------------------------------------------
// output of the corresponding alignment column of the pos-th residue in the reference sequence:
// input:  based on real numbering (starting at 1)
// output: based on real numbering (starting at 1)
//--------------------------------------------------------------------------------------------
int Alignment::ref_seq_pos_to_al_col(int pos) const {
  
  int col_in_al = 0;
  
  if (!m_has_ref_seq)  
    throw Error("no reference sequence available");
  else {
    string str = m_ref_seq->str();
    int    len = (int)str.length();
    if (pos <= 0)
      col_in_al = 1;
    else if (pos > len)
      col_in_al = len;
    else {
      int num_res = 0;
      while (col_in_al < len && num_res < pos) {
        if (str[col_in_al] != '-' && str[col_in_al] != '*' && str[col_in_al] != '.')
           ++num_res;
	++col_in_al;
      }
    }
  }
  return col_in_al;
}

//------------------------------------------------------------------------------
// output:
//------------------------------------------------------------------------------
ostream& operator<<(ostream& s, const Alignment& al) {

  FOREACH (Subtype* st, al.m_subs) {
    s << ">>" << st->name() << endl;
    FOREACH (Seq* seq, st->seqs()) {
      s << ">" << seq->name() << endl;
      s << seq->str() << endl;
    }
  }
  if (al.m_has_ref_seq) {
    s << "reference sequence: " << endl;
    s << *al.m_ref_seq << endl;
  }
  return s;
}


/*******************************************************\

 *                 class Seq                      *

\*******************************************************/

//------------------------------------------------------------------------------
// Seq():
//
//------------------------------------------------------------------------------
Seq::Seq(string name, Subtype* sub, string str) {
  m_name = name;
  m_sub  = sub;
  m_str  = str;
}

//------------------------------------------------------------------------------
// void check():
//
//------------------------------------------------------------------------------
void Seq::check() {
  if (!m_name.length() || !m_sub || !m_str.length())
    throw Error("Sequence in illegal state");
}

//------------------------------------------------------------------------------
// output:
//------------------------------------------------------------------------------
ostream& operator<<(ostream& s, const Seq& seq) {

      s << "name: " << seq.m_name << endl;
      s << "characters: " << seq.m_str << endl;
      s << "subtype: " << seq.m_sub->name() << endl;
      return s;
}


/*******************************************************\

 *                 class Subtype                       *

\*******************************************************/

//------------------------------------------------------------------------------
// Subtype():
//
//------------------------------------------------------------------------------
Subtype::Subtype(string cur_name) {
  m_name = cur_name;
}

Subtype::~Subtype(){
  FOREACH(Seq* s, m_seqs)
    delete s;
}

//------------------------------------------------------------------------------
// void check():
//
//------------------------------------------------------------------------------
void Subtype::check() {
  if (!m_name.length() || !m_seqs.size())
    throw Error("Subtype in illegal state");
}

//------------------------------------------------------------------------------
// output:
//------------------------------------------------------------------------------
ostream& operator<<(ostream& s, const Subtype& st) {

      s << "name: " << st.m_name << endl;
      s << "length: " << st.m_len << endl;
      s << "sequences: " << endl;
      FOREACH (Seq* seq, st.m_seqs) {
        s << ">" << seq->name() << endl;
        s << seq->str() << endl;
      }
      return s;
}
