// ~/jpHMM/src/alphabet.cpp
/**************************************************************
 File name	: alphabet.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Determination of nucleotide/amino acid and transition alphabet:
  = allowed nucleotides in alignment
  and allowed transitions (e.g. M -> M)
***************************************************************/

#include "std_source.h"
#include "error.h"
#include "alphabet.h"


using namespace std;

/*******************************************************\

 *                 class Alphabet                      *

\*******************************************************/

//--------------------------------------------------------------------------------------------------
// init(vector<char>):
//
// initialize with characters
//--------------------------------------------------------------------------------------------------
void Alphabet::init(vector<char> characters, const vector<char> all_characters) {
  m_characters = characters;

  for (unsigned i=0; i<m_characters.size(); ++i) {
    char c = m_characters[i];
    m_char_to_idx[c]          = i;
    m_char_to_idx[tolower(c)] = i;
  }

  FOREACH (char c, all_characters) {
    if (!is_legal_character(c)) {
      if (is_legal_character('X')) {
        m_char_to_idx[c]          = m_char_to_idx['X'];
        m_char_to_idx[tolower(c)] = m_char_to_idx['X'];
      }
      else
        throw Error("Alphabet in illegal state, special character X not defined");
    }
  }
  m_is_init = true;
}

//--------------------------------------------------------------------------------------------------
// character(int i):
//
// convert index (int) to nucleotide (char)
//--------------------------------------------------------------------------------------------------
char Alphabet::character(int i) const {
  check();
  try {
    return m_characters[i];
  }
  catch (out_of_range) {
    stringstream ss;
    ss << "Trying to get character for illegal index: \'" << i << "\'";
    throw Error(ss.str());
  }
}

//--------------------------------------------------------------------------------------------------
// characters(vector<int> v):
//
// convert vector of indices (int) to sequence string
//--------------------------------------------------------------------------------------------------
string Alphabet::characters(vector<int> v) const {
  check();
  string s;
  FOREACH (int i, v)
    s = s + character(i);
  return s;
}

//--------------------------------------------------------------------------------------------------
// idx(char c):
//
// convert nucleotide (char) to vector of indices (int)
//--------------------------------------------------------------------------------------------------
int Alphabet::idx(char c) const {
  check();
  map<char, int>::const_iterator it = m_char_to_idx.find(c);

  if (it == m_char_to_idx.end()) {
    stringstream ss;
    ss << "Trying to get index of character not in the alphabet: \'" << c << "\'";
    throw Error(ss.str());
  }
  return it->second;
}

//--------------------------------------------------------------------------------------------------
// indices(string s):
//
// convert sequence string to vector of indices (int)
//--------------------------------------------------------------------------------------------------
vector<int> Alphabet::indices(string s) const {
  check();
  vector<int> v;
  FOREACH (char c, s)
    v += idx(c);
  return v;
}

//--------------------------------------------------------------------------------------------------
// is_legal_indices(vector<int> v):
//
// check whether vector of indices only contains legal values
//--------------------------------------------------------------------------------------------------
bool Alphabet::is_legal_indices(vector<int> v) const {
  FOREACH (int i, v)
    if (!is_legal_idx(i))
      return false;
  return true;
}

//--------------------------------------------------------------------------------------------------
// remove_illegal_characters(string str):
//
// remove illegal characters from string
//--------------------------------------------------------------------------------------------------
string Alphabet::remove_illegal_characters(string str) const {
  string ret;
  for (unsigned i = 0; i < str.size(); ++i)
    if (is_legal_character(str[i]))
      ret += str[i];
  return ret;
}

//--------------------------------------------------------------------------------------------------
// check():
//
// check integrity
//--------------------------------------------------------------------------------------------------
void Alphabet::check() const {
  if (!m_is_init)
  throw Error("Alphabet in illegal state");
}

//--------------------------------------------------------------------------------------------------
// output:
//
//--------------------------------------------------------------------------------------------------
ostream& operator<<(ostream& s, const Alphabet& al) {
  s << "characters: " << al.m_characters << endl;
  s << "char_to_idx: " << al.m_char_to_idx << endl;
  s << "is_init: " << boolalpha << al.m_is_init << endl;
  return s;
}


/*******************************************************\

 *                 class TransAlphabet                 *

\*******************************************************/

//--------------------------------------------------------------------------------------------------
// init(vector<string>):
//
// initialize with strings
//--------------------------------------------------------------------------------------------------
void TransAlphabet::init(vector< pair<StateType, StateType> > transitions) {
  m_transitions = transitions;

  for (unsigned i=0; i<m_transitions.size(); ++i)
    m_pair_to_idx[m_transitions[i]] = i;

  m_is_init = true;
}


//--------------------------------------------------------------------------------------------------
// transition(int i):
//
// convert index (int) to transition (string)
//--------------------------------------------------------------------------------------------------
pair<StateType, StateType> TransAlphabet::transition(int i) const {
  check();
  try {
    return m_transitions[i];
  }
  catch (out_of_range) {
    stringstream ss;
    ss << "Trying to get transition for illegal index: \'" << i << "\'";
    throw Error(ss.str());
  }
}

//--------------------------------------------------------------------------------------------------
// idx(char c):
//
// convert transition (string) to int
//--------------------------------------------------------------------------------------------------
int TransAlphabet::idx(StateType s, StateType t) const {
  check();

  pair<StateType, StateType> p = make_pair(s,t);
  map< pair<StateType, StateType>, int>::const_iterator it = m_pair_to_idx.find(p);

  if (it == m_pair_to_idx.end()) {
    stringstream ss;
    ss << "Trying to get index of transition not in the transition alphabet: \'" << s << "\'";
    throw Error(ss.str());
  }

  return it->second;
}

//--------------------------------------------------------------------------------------------------
// is_legal_pair():
//
// check whether vector of indices only contains legal values
//--------------------------------------------------------------------------------------------------
bool TransAlphabet::is_legal_pair(StateType s, StateType t) const {

  return ((m_pair_to_idx.find(make_pair(s,t))!=m_pair_to_idx.end()) ? true:false);
}

//--------------------------------------------------------------------------------------------------
// is_legal_indices(vector<int> v):
//
// check whether vector of indices only contains legal values
//--------------------------------------------------------------------------------------------------
bool TransAlphabet::is_legal_indices(vector<int> v) const {
  FOREACH (int i, v)
    if (!is_legal_idx(i))
      return false;
  return true;
}

//--------------------------------------------------------------------------------------------------
// check():
//
// check integrity
//--------------------------------------------------------------------------------------------------
void TransAlphabet::check() const {
  if (!m_is_init)
    throw Error("TransAlphabet in illegal state");
}

//--------------------------------------------------------------------------------------------------
// output:
//
//--------------------------------------------------------------------------------------------------
ostream& operator<<(ostream& s, const TransAlphabet& t_al) {
  s << "transition alphabet: " << endl;

  for (map< pair<StateType, StateType>, int >::const_iterator it = t_al.m_pair_to_idx.begin(); it != t_al.m_pair_to_idx.end(); ++it) {
    pair<StateType, StateType> p = it->first;
    s << p.first << " " << p.second << "\tidx: " << it->second << endl;
  }
  s << "is_init: " << boolalpha << t_al.m_is_init << endl;
  return s;
}
