// ~/jpHMM/src/const_def.cpp
/**************************************************************
 File name	: const_def.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Definition of all 
 - jpHMM variables
 - nucleotide/amino acid/gap characters
 - allowed virus types
 - reference genome name for different viruses
 - file names
 - script names
 - input/output directory names
 - possible state transitions
 
 Output of output file headers
***************************************************************/

#include "std_source.h"
#include "const_def.h"
#include "error.h"
#include "alphabet.h"

Const_Def* Const_Def::s_instance = 0;

using namespace std;


/*******************************************************\

 *                 class Const_Def                     *

\*******************************************************/

/**************************************//**
   Const_Def.

   Constant definitions for jpHMM.
*******************************************/
Const_Def::Const_Def() {

  //! jump prob:
  //! probability of a jump from one subtype to another subtype. jp=10^-09 for HIV, jp=10^-07 for HBV
  m_jump_prob            = 1e-9;
  m_virus_type_to_jump_prob[HIV] = 1e-9;
  m_virus_type_to_jump_prob[HCV] = 1e-9;
  m_virus_type_to_jump_prob[HBV] = 1e-7;

  m_epsilon              = 0.01;               // 1-epsilon = transProb(N_Insert->N_Insert) = TransProb(C_Insert->C_Insert)
  m_gap_init             = 0.01;               // gap_init = jumpProb(MATCH->C_DELETE) taking into account the corresponding consensus column of the MATCH
  m_gap_ext              = 0.99;               // gap_ext = transProb(N_DELETE->N_DELETE) = TransProb(C_DELETE->C_DELETE)
  m_beam_width           = 1e-20;              // beam_width for beam-search algorithm in Viterbi algorithm
  m_consensus_percentage = 0.5;                // percentage of non-gaps in a column
  m_consensus_threshold  = 5;                  // number of non-gaps in a column  (if e.g. =1000 => no effect)
  
  m_threshold_long_deletion = 50;              // threshold/length of deletion in query considered as a "(too) long deletion" (maybe no analysis possible)
  
  m_query_ext = 500;                           // extension of query for circular genomes (in nt)
  m_msa_ext   = 0.25;                          // extension factor for msa for circular genomes

  m_virus_ref_seq_name[HIV] = "HXB2";
  m_virus_ref_seq_name[HBV] = "AM282986";

  m_alphabet       = new Alphabet();
  m_trans_alphabet = new TransAlphabet();

  m_virus_type   = HIV;
  m_genome_type  = non_circular;               // type of genomes: circular or non-circular
  m_prog_version = beam_search;

  m_adapt_jump_probs = false;
  
  m_has_stamp = false;
  
  m_input_dir_name 		       = "~/jpHMM/input/";
  m_priors_dir_name 		       = "~/jpHMM/priors/";
  m_output_dir_name 		       = "./output/";
  
  //! input files:
  //! 1) alignment:
  m_hiv_msa_file_name    = "HIV_alignment.fas";
  m_hcv_msa_file_name    = "HCV_alignment.fas";
  m_hbv_msa_file_name    = "HBV_alignment.fas";
  
  //! 2) pseudo counts for emission probabilities:
  m_hiv_priors_file_name = "emissionPriors_HIV.txt";
  m_hcv_priors_file_name = "emissionPriors_HCV.txt";
  m_hbv_priors_file_name = "emissionPriors_HBV.txt";

  //! 3) pseudo counts for transition probabilities:
  m_transition_priors_file_name = "transition_priors.txt";
  
  //! output files:
  m_recomb_file_basename           = "recombination";
  m_recomb_incl_gaps_file_basename = "recombination_incl_gaps";
  m_pred_subtypes_file_name    	   = "recombination_without_positions.txt";
  m_viterbi_gff_file_name      	   = "viterbi_path.gff";
  m_al_to_msa_file_name            = "alignment_to_msa.txt";
  m_query_to_ref_seq_al_file_name  = "query_to_ref_alignments.txt";
  m_variable_regions_file_name     = "variable_regions.txt";
  m_posterior_path_file_name       = "posterior_recombination_path.txt";
  m_posterior_prob_file_name       = "posterior_probabilities_for_seq_";
  m_active_alignment_columns_file_name = "active_alignment_columns.txt";
  
  //! test output files:
  m_hmm_analysis_file_name                    = "viterbi_analysis.html";
  m_active_state_intervals_file_name          = "active_state_intervals.txt";                      // columns of most left and most right active M, I or D state.
  m_hmm_output_file_name                      = "hmm_output.txt";
  m_active_state_intervals_for_plot_file_name = "active_state_intervals_for_plot.txt";    // columns of most left and most right active M, I or D state.
  
  //! additional programs and scripts
  //! 1) for extending circular genomes (MSA and queries) 
  //! 2) for calling and evaluating the blat program (for blat version).
  //! 3) blat programs

  m_ext_script_name            = "~/jpHMM/src/extend_circ_gen.pl";
  m_blat_script_name           = "~/jpHMM/src/eval_blat_output.pl";
  m_blat_program_name          = "~/Blat/blat";

  vector<char>   gap_chars;
  vector<char>   nucleic_chars;
  vector<char>   amino_chars;
  vector<char>   all_characters;

  gap_chars      += '-';
  nucleic_chars  += 'A','C','G','T','X';
  amino_chars    += 'A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y','X';
  all_characters += 'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z';


  vector<char> s;
  vector<char> characters = nucleic_chars;

  FOREACH (char c, characters)
    s += c;
  FOREACH (char c, gap_chars)
    s += c;


  m_alphabet->init(s, all_characters);

  vector< pair<StateType, StateType> > state_transitions;

  for (int wherefrom = (int)Match; wherefrom <= (int)Delete; ++wherefrom) {
    for (int whereto = (int)Match; whereto <= (int)Delete; ++whereto) {
      if (wherefrom == Match || whereto == (int)Match || whereto == (int) wherefrom) {
        pair<StateType, StateType> p = make_pair((StateType) wherefrom, (StateType) whereto);
        state_transitions += p;
      }
    }
  }

  m_trans_alphabet->init(state_transitions);
}

/****************************//**
   Const_Def destructor.
*********************************/
Const_Def::~Const_Def() {
  delete m_alphabet;
  delete m_trans_alphabet;
}

/****************************//**
   Const_Def instance.
*********************************/
Const_Def* Const_Def::instance()
{
	if (!s_instance)
		s_instance = new Const_Def();
	return s_instance;
}

/******************************************************************************//**
   Virus ref seq name.
   
   Name of reference genome/sequence for given virus type.

   @param[in] vt virus type.
   @param[out] ref_seq_name Name of reference genome/sequence for given virus type.
***********************************************************************************/
string Const_Def::virus_ref_seq_name(Virus_Type vt) { 
  
  string ref_seq_name = "";
  map<Virus_Type, string>::const_iterator it = instance()->m_virus_ref_seq_name.find(vt);
  if (it != instance()->m_virus_ref_seq_name.end()) {
    ref_seq_name = instance()->m_virus_ref_seq_name[vt];
  }
  else {
    //! else: no reference sequence exists for this virus type
    ref_seq_name = "mKskaKmvnkmsn";
  }
  return ref_seq_name; 
};

/******************************************************************************//**
  Recomb file name.
   
   Name of output file for predicted recombination.
***********************************************************************************/
string Const_Def::recomb_file_name()           
{ 
	string str = output_dir_name() + "/" + instance()->m_recomb_file_basename;
	if (has_stamp())
		str += "_" + stamp();
	str += ".txt";
	return str; 
}

/******************************************************************************//**
  Recomb file name.
   
   Name of output file for predicted recombination if query sequence contained gaps.
***********************************************************************************/
string Const_Def::recomb_incl_gaps_file_name()           
{ 
	string str = output_dir_name() + "/" + instance()->m_recomb_incl_gaps_file_basename;
	if (has_stamp())
		str += "_" + stamp();
	str += ".txt";
	return str; 
}

/******************************************************************************//**
  Init out file.
   
   Initialization of output files.
***********************************************************************************/
void Const_Def::init_out_files() {
  ofstream RECOMB(recomb_file_name().c_str(), ios::out);
  ofstream RECOMB_GAPS(recomb_incl_gaps_file_name().c_str(), ios::out);
  ofstream RECOMB_WO_POS(pred_subtypes_file_name().c_str(), ios::out);
  ofstream VITERBI_GFF(viterbi_gff_file_name().c_str(), ios::out);
  ofstream AL_OUT (query_to_ref_seq_al_file_name().c_str(), ios::out);
  ofstream AL_TO_MSA_OUT(al_to_msa_file_name().c_str(), ios::out);
  ofstream VAR_REG (variable_regions_file_name().c_str(), ios::out);
  ofstream POST_PATH (posterior_path_file_name().c_str(), ios::out);

  RECOMB << "# Output of the jpHMM. " << endl
         << "#" << endl
	 << "#   chosen parameters: " << endl
	 << "#" << "\t" << "jump probability: " << "\t" << jump_prob() << endl
	 << "#" << endl
	 << "#   format: " << endl
	 << "#   sequence_name (bw=beam-width)" << endl
	 << "#   start_position\t end_position\t predicted_subtype " << endl
	 << "#" << endl
	 << "# predicted recombination with breakpoints based on raw sequence positions " << endl
	 << "# of the following sequence(s): " << endl << endl;

  RECOMB_GAPS << "# Output of the jpHMM. " << endl
	      << "#" << endl
		<< "#   chosen parameters: " << endl
		<< "#" << "\t" << "jump probability: " << "\t" << jump_prob() << endl
		<< "#" << endl
		<< "#   format: " << endl
		<< "#   sequence_name " << endl
		<< "#   start_position\t end_position\t predicted_subtype " << endl
		<< "#" << endl
		<< "# predicted recombination with breakpoints based on sequence positions with gaps " << endl
		<< "# of the following sequence(s): " << endl
		<< "# (only if the sequence contains gaps; if not, please see 'recombination.txt' for the results)" << endl << endl;

  RECOMB_WO_POS << "# Output of the jpHMM. " << endl
		     << "#" << endl
		     << "#   format: " << endl
		     << "#   sequence_name\t chosen_parameters\t predicted_subtypes" << endl
		     << "#" << endl
		     << "#   (bw=beam-width, jp=jump probability)" << endl
		     << "#" << endl
		     << "# predicted recombination of the following sequence(s): " << endl << endl;

  VITERBI_GFF << "# Output of the jpHMM. " << endl
	      << "#" << endl
	      << "# Predicted recombination for the following query sequences:" << endl;
  VITERBI_GFF.close();

  AL_TO_MSA_OUT << "# Output of the alignment of the query sequence to the given multiple sequence alignment " << endl
         << "# defined by the most probable path through the jphmm generating the query sequence" << endl
         << "#" << endl
	 << "#   format: " << endl
	 << "#   sequence_name (bw=beam-width)" << endl
	 << "#   for each query sequence position the aligned msa column " << endl << endl;
  AL_TO_MSA_OUT.close();

  VAR_REG << "# Output of the variable regions" << endl
	  << "# in the following sequences (based on raw sequence positions)" << endl
	  << "#" << endl
	  << "# format:" << endl
	  << "#   sequence_name" << endl
	  << "#   start_position\t end_position\t" << endl
	  << "#" << endl << endl;
  VAR_REG.close();

  POST_PATH << "# Output of the 'path' defined by the posterior probabilities of the query sequence(s). " << endl
	    << "#" << endl
	    << "#   format: " << endl
	    << "#   sequence_name " << endl
	    << "#   start_position\t end_position\t subtype with highest posterior prob " << endl
	    << "#" << endl
	    << "# output based on raw sequence positions " << endl
	    << "# of the following sequence(s): " << endl << endl;
  POST_PATH.close();

}

