// ~/jpHMM/src/const_def.h
/**************************************************************
 File name	: const_def.h
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Classes: Const_Def

 Definition of all 
 - jpHMM variables
 - nucleotide/amino acid/gap characters
 - allowed virus types
 - reference genome name for different viruses
 - file names
 - script names
 - input/output directory names
 - possible state transitions
 
 Output of output file headers
***************************************************************/


#ifndef _CONST_DEF_H
#define _CONST_DEF_H
#include "std_header.h"
#include "std_source.h"


class Alphabet;
class TransAlphabet;

#define NUM_AA 20                     //! number of amino acids
#define NUM_NUCL 4                    //! number of nucleotides
#define NUM_GAPS 1                    //! gaps: '-'
#define NUM_STATES 3                  //! active state types of a HMM:  Match, Insert, Delete
#define NUM_SPEC_STATES 6             //! special state types of a HMM: Begin, N_Delete, N_Insert, C_Insert, C_Delete, End
#define NUM_JUMP_STATES 2             //! special jump state types: jumps to Match, jump to Delete
#define NUM_MIN_ACT_COL_BLAT 0        //! extension of aligned BLAT intervals (if Blat version chosen)
#define NUM_MIN_ACT_COL_PER_ST_BLAT 3 //! number of active columns for each subtype  if Blat version chosen

#define MSA_EXT    0                  //! extension of start and end of query in MSA, if start and end positions are given in query file, only if not Blat chosen

#define NUM_TRANS 7                   //! number of possible state transitions within a profile HMM for one column (M->M, M->I, M->D, I->M, I->I, D->M, D->D)
#define INF DBL_MAX                   //! if mytype==float => #define INF FLT_MAX


using namespace std;

enum Program_Version {beam_search, blat};
enum Jump_Version {linear, quadratic};
enum Virus_Type {HIV, HCV, HBV};
enum Genome_Type {circular, non_circular};

typedef double mytype;                //! if float mytype change INF DBL_MAX to FLT_MAX

class Const_Def {
public:
  static void    jump_prob(Virus_Type vt) { instance()->m_jump_prob = instance()->m_virus_type_to_jump_prob[vt];};
  static void    double_jump_prob()       { instance()->m_jump_prob = 2*instance()->m_jump_prob;};
  static mytype &jump_prob()              { return instance()->m_jump_prob;};
  static mytype &epsilon()                { return instance()->m_epsilon; };
  static mytype &gap_init()               { return instance()->m_gap_init; };
  static mytype &gap_ext()                { return instance()->m_gap_ext; };
  static mytype &beam_width()             { return instance()->m_beam_width; };

  static mytype &consensus_percentage()   { return instance()->m_consensus_percentage; };
  static int    &consensus_threshold()    { return instance()->m_consensus_threshold; };

  static int    &threshold_long_deletion() { return instance()->m_threshold_long_deletion; };
  
  static int    &query_ext()              { return instance()->m_query_ext; };
  static mytype &msa_ext()                { return instance()->m_msa_ext; };

  static Alphabet*      alphabet()        { return instance()->m_alphabet; };
  static TransAlphabet* trans_alphabet()  { return instance()->m_trans_alphabet; };

  static Virus_Type      &virus_type()    { return instance()->m_virus_type; };
  static Program_Version &prog_version()  { return instance()->m_prog_version; };
  static Genome_Type     &genome_type()   { return instance()->m_genome_type; };
  static Jump_Version    &jump_version()  { return instance()->m_jump_version; };
  
  static bool            &adapt_jump_probs() {return instance()->m_adapt_jump_probs; };
  
  static string virus_ref_seq_name(Virus_Type vt);
  
  static string& stamp()     { return instance()->m_stamp; };
  static bool&   has_stamp() { return instance()->m_has_stamp; };

  static string& input_dir_name()            { return instance()->m_input_dir_name; };
  static string& priors_dir_name()           { return instance()->m_priors_dir_name; };
  static string& output_dir_name()           { return instance()->m_output_dir_name; };

  static string hiv_msa_file_name()    { return input_dir_name() + "/" + instance()->m_hiv_msa_file_name; };
  static string hcv_msa_file_name()    { return input_dir_name() + "/" + instance()->m_hcv_msa_file_name; };
  static string hbv_msa_file_name()    { return input_dir_name() + "/" + instance()->m_hbv_msa_file_name; };
  static string hiv_priors_file_name() { return priors_dir_name() + "/" + instance()->m_hiv_priors_file_name; };
  static string hcv_priors_file_name() { return priors_dir_name() + "/" + instance()->m_hcv_priors_file_name; };
  static string hbv_priors_file_name() { return priors_dir_name() + "/" + instance()->m_hbv_priors_file_name; };
  static string transition_priors_file_name() { return priors_dir_name() + "/" + instance()->m_transition_priors_file_name; };

  static string recomb_file_name();
  static string recomb_incl_gaps_file_name();
  static string pred_subtypes_file_name()            { return output_dir_name() + "/" + instance()->m_pred_subtypes_file_name; };
  static string viterbi_gff_file_name()              { return output_dir_name() + "/" + instance()->m_viterbi_gff_file_name; };
  static string al_to_msa_file_name()                { return output_dir_name() + "/" + instance()->m_al_to_msa_file_name; };
  static string query_to_ref_seq_al_file_name()      { return output_dir_name() + "/" + instance()->m_query_to_ref_seq_al_file_name; };
  static string variable_regions_file_name()         { return output_dir_name() + "/" + instance()->m_variable_regions_file_name; };
  static string posterior_path_file_name()           { return output_dir_name() + "/" + instance()->m_posterior_path_file_name; };
  static string posterior_prob_file_name()           { return output_dir_name() + "/" + instance()->m_posterior_prob_file_name; };
  static string active_alignment_columns_file_name() { return instance()->m_active_alignment_columns_file_name; };
  static string active_state_intervals_file_name()   { return output_dir_name() + "/" + instance()->m_active_state_intervals_file_name; };

  static string active_state_intervals_for_plot_file_name()   { return output_dir_name() + "/" + instance()->m_active_state_intervals_for_plot_file_name; };
  static string hmm_output_file_name()                        { return output_dir_name() + "/" + instance()->m_hmm_output_file_name; };
  static string hmm_analysis_file_name()                      { return output_dir_name() + "/" + instance()->m_hmm_analysis_file_name; };
  static void   init_out_files();

  static string ext_script_name()            { return instance()->m_ext_script_name; };
  static string blat_script_name()           { return instance()->m_blat_script_name; };
  static string blat_program_name()          { return instance()->m_blat_program_name; };
  
private:
  Const_Def();
  ~Const_Def();
  static Const_Def* instance();
  static Const_Def* s_instance;

  mytype m_jump_profile_to_row;  // -x
  mytype m_jump_row_to_profile;  // -y
  map<Virus_Type, mytype> m_virus_type_to_jump_prob;  //! default jump prob for each virus type
  
  mytype m_jump_prob;            // -j
  mytype m_epsilon;
  mytype m_gap_init;             // -c
  mytype m_gap_ext;              // -d
  mytype m_beam_width;           // -B
  mytype m_consensus_percentage; // -p
  int    m_consensus_threshold;  // -t
  int    m_threshold_long_deletion;
  int    m_query_ext;
  mytype m_msa_ext;
  
  Alphabet*      m_alphabet;
  TransAlphabet* m_trans_alphabet;
  
  Virus_Type      m_virus_type;
  Genome_Type     m_genome_type;
  Program_Version m_prog_version;
  Jump_Version    m_jump_version;
  bool            m_adapt_jump_probs;  //! to compare 'linear' and 'quadratic' jump model: same jump probs
  
  string m_stamp;
  bool   m_has_stamp;
  
  map<Virus_Type, string> m_virus_ref_seq_name;  //! names of virus reference sequence  
  
  string m_input_dir_name;
  string m_priors_dir_name;
  string m_output_dir_name;
  
  string m_hiv_msa_file_name;
  string m_hcv_msa_file_name;
  string m_hbv_msa_file_name;
  string m_hiv_priors_file_name;
  string m_hcv_priors_file_name;
  string m_hbv_priors_file_name;
  string m_transition_priors_file_name;

  
  string m_recomb_file_basename;
  string m_recomb_incl_gaps_file_basename;
  string m_pred_subtypes_file_name;
  string m_viterbi_gff_file_name;
  string m_al_to_msa_file_name;
  string m_query_to_ref_seq_al_file_name;
  string m_variable_regions_file_name;
  string m_posterior_path_file_name;
  string m_posterior_prob_file_name;
  string m_active_alignment_columns_file_name;
  
  string m_active_state_intervals_file_name;
  string m_active_state_intervals_for_plot_file_name;
  string m_hmm_output_file_name;
  string m_hmm_analysis_file_name;
  
  string m_ext_script_name;   //! name of script for extension of circular genomes (MSA and queries)
  string m_blat_script_name;  //! name of script that calls BLAT and evaluates the aligned columns
  string m_blat_program_name;
  
};

#endif        // _CONST_DEF_H
