#!/usr/bin/perl

# Determination of uncertainty regions (UR)
# and interval estimates for breakpoints (BPI)
#  based on the posterior probabilities of the subtypes
#
# author: Anne-Kathrin Schultz, anne@gobics.de
# ------------------------------------------------------

use warnings;
use strict;

########################################################################################
# command line program call: 
#    ./determine_UR.pl [perc_var_UR] [perc_non_var_UR] [perc_var_BPR] [perc_nonvar_BPR]
#  e.g. ./determine_UR.pl 99 99 99 99
########################################################################################


#################################################################################
# input parameters and files:                                                   #
#################################################################################
# input parameters  ( in [0,100] ):                                             #
#                                                                               #
# 1) thresholds for the determination of uncertainty regions:                   #
#    perc_var_UR:     threshold for URs in variable regions in the genome       #
#    perc_nonvar_UR:  threshold for URs in non-variable regions                 #
# 2) thresholds for the determination of breakpoint regions:                    #
#    perc_var_BPR:    threshold for BPRs in variable regions in the genome      #
#    perc_nonvar_BPR: threshold for BPRs in non-variable regions                #
#################################################################################
#################################################################################
# needed input files:                                                           #
#                                                                               #
# in the default format as described in the README file of the jpHMM program    #
#    recombination.txt                                                          #
#    variable_regions.txt (both with a blank line at the end)                   #
#    posterior_probabilities_for_seq_*.txt                                      #
#      (* = 1,...,no_of_query_sequences_considered_in_recombination.txt)        #
#################################################################################


my $perc_var_UR     = 0.99;  # default thresholds
my $perc_nonvar_UR  = 0.99;
my $perc_var_BPR    = 0.99;
my $perc_nonvar_BPR = 0.99;

#if (scalar(@ARGV) == 0) {
#  print "program call:\n";
#  print "[determine_UR.pl perc_var_UR perc_nonvar_UR perc_var_BPR perc_nonvar_BPR modified_recomb_output_file UR_and_BPR_output_file]\n\n";
#  print "parameters:\n";
#  print "perc_var_UR:                 threshold for URs in variable regions in the genome (in %)\n";
#  print "perc_nonvar_UR:              threshold for URs in non-variable regions (in %)\n";
#  print "perc_var_BPR:                threshold for BPRs in variable regions in the genome (in %)\n";
#  print "perc_nonvar_BPR:             threshold for BPRs in non-variable regions (in %)\n";
#  print "modified_recomb_output_file: output file for the recombination prediction containing uncertainty and breakpoint regions\n";
#  print "UR_and_BPR_output_file:      output file for the uncertainty and breakpoint regions in the recombination prediction\n\n";
#  print "default parameters used:\n";
#  print "[determine_UR.pl 99 99 99 99 recombination_incl_UR_and_BPR.txt uncertainty_and_BP_regions.txt]\n\n";
#}


if (scalar(@ARGV) > 1) {
  $perc_var_UR    = $ARGV[0]/100;
  $perc_nonvar_UR = $ARGV[1]/100;
}
if (scalar(@ARGV) > 3) {
  $perc_var_BPR    = $ARGV[2]/100;
  $perc_nonvar_BPR = $ARGV[3]/100;
}

my $recomb_out_file = "./recombination_incl_UR_and_BPR.txt"; # default modified recombination output file
if (scalar(@ARGV) > 4) {
  $recomb_out_file = $ARGV[4];
}
my $UR_out_file = "./uncertainty_and_BP_regions.txt";        # default uncertainty/breakpoint region output file
if (scalar(@ARGV) > 5) {
  $UR_out_file = $ARGV[5];
}

################
# input files: #
################
my $recomb_in_file = "./recombination.txt";    # file with the viterbi path/predicted recombination
my $var_reg_file   = "./variable_regions.txt"; # file with the variable regions in the query sequences

open(VIT_RECOMB, $recomb_in_file) or die "can't find file for jpHMM recombination prediction\n";

my $var_reg_exist = 0;
if (-z $var_reg_file) {
  open(VAR_REG, $var_reg_file);
}
else {
  $var_reg_exist = 1;
}

#################
# output files: #
#################
# file for the output of UR and BPI: #
open(UR_REG, ">$UR_out_file");
my $old_fh = select(UR_REG);
$| = 1;
select($old_fh);

print UR_REG "# Output of the uncertainty and the breakpoint regions of the following sequences\n#\n";
print UR_REG "#   based on the posterior probabilities (calculated by the jpHMM)\n";
print UR_REG "#   Uncertainty regions are determined by the following thresholds:\n";
print UR_REG "#       in variable regions:\t", $perc_var_UR, "\n";
print UR_REG "#       in non-variable regions:\t", $perc_nonvar_UR, "\n#\n";
print UR_REG "#   Breakpoint regions are determined by the following thresholds:\n";
print UR_REG "#       in variable regions:\t", $perc_var_BPR, "\n";
print UR_REG "#       in non-variable regions:\t", $perc_nonvar_BPR, "\n#\n";
print UR_REG "#   format: \n";
print UR_REG "#   sequence_name \n";
print UR_REG "#   start_position\tend_position\n\n";

# file for the output of the predicted recombination including uncertainty and the breakpoint regions: #
open(PRED_RECOMB, ">$recomb_out_file");

# to allow output before end of program run:
my $old_recomb = select(PRED_RECOMB);
$| = 1;
select($old_recomb);

print PRED_RECOMB "# Output of the recombination predicted by the jpHMM\n";
print PRED_RECOMB "# including posterior evaluation of uncertainty and breakpoint regions\n";
print PRED_RECOMB "#   based on the posterior probabilities (calculated by the jpHMM)\n#\n";
print PRED_RECOMB "#   Uncertainty regions are determined by the following thresholds:\n";
print PRED_RECOMB "#       in variable regions:\t", $perc_var_UR, "\n";
print PRED_RECOMB "#       in non-variable regions:\t", $perc_nonvar_UR, "\n#\n";
print PRED_RECOMB "#   Breakpoint regions are determined by the following thresholds:\n";
print PRED_RECOMB "#       in variable regions:\t", $perc_var_BPR, "\n";
print PRED_RECOMB "#       in non-variable regions:\t", $perc_nonvar_BPR, "\n#\n";
print PRED_RECOMB "#   format: \n";
print PRED_RECOMB "#   sequence_name \n";
print PRED_RECOMB "#   start_position\tend_position\tpredicted_subtype\n\n";

#########################################################
# variables:
#
# abbreviations in variable names:
# ST = subtype, UR = uncertainty region, VAR = variable region
#########################################################
# Initialization

# variables for each single sequence:
# (reinitialized for each sequence)
my @seq_name         = "";
my @seq_names        = ();
my @seq_names_in_VAR = ();
my $seq_length       = 0;
my @post_probs       = ();
my %subtypes;
my @subtype_names    = ();
my @VAR_pos          = ();
my @recomb_data      = ();
my @ST_seq           = ();
my @viterbi_Path_BP  = ();
my @viterbi_Path_ST  = ();

my $genome_type = "non_circular";
my $seq_type    = "full_length";

until (eof(VIT_RECOMB)){
  #############################################################################
  # Read in the predicted recombination (Viterbi path) of the actual sequence:
  # including the predicted recombination breakpoints
  #############################################################################
  my $recomb_line = <VIT_RECOMB>;
  chomp($recomb_line);
  if ($recomb_line =~ m/^#/) {
    next;
  }
  elsif ($recomb_line =~ /^\s*$/ && scalar(@seq_names) == 0) { # blank line above 1st seq
    next;
  }
  elsif ($recomb_line =~ m/^\>/i) { # seq name
    chomp($recomb_line);
    if ($recomb_line =~ m/circular/g) {
      $genome_type = "circular";
    }
    else {
      $genome_type = "non_circular";
    }
    if ($recomb_line =~ m/fragment/g) {
      $seq_type = "fragment";
    }
    else {
      $seq_type = "full_length";
    }

    @seq_name = split(" ", $recomb_line);
    push(@seq_names, $seq_name[0]);

    print UR_REG $seq_name[0], "\n";

    my $tmp_name = $seq_name[0];
    print PRED_RECOMB $recomb_line, "\n";

    #######################################################
    # Read in posterior probabilities
    # for current query seq
    #######################################################

    for (my $i=0; $i<scalar(@subtype_names); $i++){
      delete $subtypes{$i};
    }
    @subtype_names = ();
    @post_probs   = ();

    open(POST_PROB, "./posterior_probabilities_for_seq_".(scalar(@seq_names)).".txt") or die "can't find file for posterior probabilities\n";	

    my $post_line = <POST_PROB>; # read in comment line: contains subtype names
    my @post_line = split(" ", $post_line);

    for (my $i=1; $i<scalar(@post_line); $i++) { # i=0: seq name
      push(@subtype_names, $post_line[$i]);
    }

    for (my $i=0; $i<scalar(@subtype_names); $i++) {
      $subtypes{$subtype_names[$i]} = $i;
    }

    my @post_data = ();
    until (eof(POST_PROB)) {
      $post_line = <POST_PROB>;
      chomp($post_line);
      @post_data = split(" ", $post_line);
      push(@post_probs, [@post_data]);   # Add array with posterior probabilities for current pos to matrix of all posterior probabilities
    }
    close(POST_PROB);
    next;
  }
  elsif ($recomb_line =~ m/^[0-9]/i) {
    #######################################################
    # Read in predicted recombination for actual sequence:
    # format:
    # begin_of_reg    end_of_reg    predicted_Subtyp
    #######################################################
    push(@recomb_data, $recomb_line);

    if (scalar(@recomb_data) > 1) { # (1st) breakpoint
      my @ST_region = split( "\t", $recomb_data[scalar(@recomb_data)-2]);
      push(@viterbi_Path_BP, $ST_region[0]);
      push(@viterbi_Path_ST, $ST_region[2]);

      ############################################################
      # storage of subtype sequence:
      # = for each position of the sequence the predicted subtype
      ############################################################
      for (my $i = $ST_region[0]; $i<$ST_region[1]+1; $i++) {
	if ($ST_region[2] eq 'N/A') {
	  push(@ST_seq, -1);
	}
	else {
	  push(@ST_seq, $subtypes{$ST_region[2]});
	}
      }
    }
    next;
  }
  else {
    #################################################################
    # blank line after 'recombination block' in file
    # => new sequence
    # => evaluation for preceeding sequence:
    #################################################################
    # Evaluation of the last row (is missing in preceding 'elsif'):
    #################################################################
    my @ST_region  = split( "\t", $recomb_data[scalar(@recomb_data)-1]);
    push(@viterbi_Path_BP, $ST_region[0]);
    push(@viterbi_Path_ST, $ST_region[2]);
    
    ############################################################
    # storage of subtype sequence
    # = for each position of the sequence the predicted subtype
    ############################################################
    for (my $i = $ST_region[0]; $i<$ST_region[1]+1; $i++) {
      if ($ST_region[2] eq 'N/A') {
	push(@ST_seq, -1);
      }
      else {
	push(@ST_seq, $subtypes{$ST_region[2]});
      }
    }
    $seq_length = $ST_region[1];
    push(@viterbi_Path_BP, $seq_length);
    
    if (scalar(@viterbi_Path_BP) == 2 && $ST_seq[0] eq '-1') {
      # -> only 1 and seq_length and N/A
      # -> no file containing the posterior prob. for this sequence exist
      
      print UR_REG "# uncertainty regions:\n";
      print UR_REG "1\t", $seq_length, "\tN/A", "\n";
      print UR_REG "# breakpoint regions:\n\n";
      
      print PRED_RECOMB "1\t", $seq_length, "\tN/A\n\n";
    }
    else {
      
      if ($var_reg_exist eq 0) {
	###########################################################
	# Read in the variable regions for the current sequence:
	# (otherwise continue with posterior probabilities)
	###########################################################
      LINE: my $var_line = <VAR_REG>;
	if ($var_line =~ m/^#/) {
	  goto LINE;
	}
	elsif ($var_line =~ /^\s*$/ && scalar(@seq_names_in_VAR) == 0) { # blank line above 1st sequence
	  goto LINE;
	}
	elsif ($var_line =~ m/^\>/i) { # sequence name
	  chomp($var_line);
	  my @var_seq_name = split(" ", $var_line);
	  if ($var_seq_name[0] eq $seq_name[0]) {
	    push(@seq_names_in_VAR, $var_seq_name[0]);
	  }
	  else {
	    die "sequences in $var_reg_file are not in the same order as in $recomb_in_file\n";
	  }
	  goto LINE;
	}
	elsif ($var_line =~ m/^[0-9]/i) {
	  chomp($var_line);
	  
	  ###################################################
	  # Read in variable regions for current sequence:
	  # format:
	  # begin_of_reg    end_of_reg
	  ###################################################
	  # begin and end belong to the variable region!
	  ###################################################
	  my @act_VAR = split( "\t", $var_line);
	  for (my $i=$act_VAR[0]; $i<$act_VAR[1]+1; $i++) {
	    push(@VAR_pos, $i);
	  }
	  goto LINE;
	}
	else {
	  chomp($var_line);
	  ###############################################
	  # blank line after 'variable block' in file
	  # => new sequence
	  # => evaluation for preceding sequence:
	  ###############################################
	  
	}
      }
      ##########################################################################################################
      # Evaluation: Determine for each sequence position the posterior probability of the predicted subtype
      #             if this posterior probability < threshold => uncertainty region
      ##########################################################################################################
      my @UR_path    = ();
      my @BPR_path   = ();
      for (my $i=0; $i<$seq_length; $i++) {
	push(@UR_path,0);
	push(@BPR_path,0);
      }
      
      my $other_ST           = 0;
      my @UR_perc_threshold  = ();
      my @BPR_perc_threshold = ();

      my $pos = 1;

      # Determine for each position, if it lies inside a variable region:
      # assign corresponding UR and BPI threshold
      while ($pos < $seq_length+1) {
	if (grep /^$pos$/,@VAR_pos) {
	  push(@UR_perc_threshold, $perc_var_UR);
	  push(@BPR_perc_threshold, $perc_var_BPR);
	}
	else {
	  push(@UR_perc_threshold, $perc_nonvar_UR);
	  push(@BPR_perc_threshold, $perc_nonvar_BPR);
	}
	$pos = $pos+1;
      }

      # general determination of uncertainty regions
      # for those regions x that are later defined as a BP region:
      # UR_path(x) = 0;
      
      for ($pos=0; $pos<$seq_length; $pos++) {
	if ($ST_seq[$pos] ne $subtypes{$subtype_names[0]} && $ST_seq[$pos] ne $subtypes{$subtype_names[scalar(@subtype_names)-1]}) {
          if ($post_probs[$pos][$ST_seq[$pos]] < $UR_perc_threshold[$pos]) {
	    $UR_path[$pos] = 1;
	  }
	}
      }
      
      ##################################################################
      # Determination of uncertainty (UR) and breakpoint regions (BPR) #
      ##################################################################

      my $num_real_BPs = scalar(@viterbi_Path_BP)-2;

      # in a circular genome: if ST of first and last segment are different, "real" BP at end of Seq!
      if ($genome_type eq "circular" && $seq_type ne "fragment") {
        if ($viterbi_Path_ST[0] ne $viterbi_Path_ST[-1]) {
          $num_real_BPs = scalar(@viterbi_Path_BP)-1; # also last BP as true BP (= seq_length)
        }
      }

      for (my $i=1; $i<=$num_real_BPs; $i++) { # viterbi_Path_BP[0] = 1     BP = BP-1/BP
	my $prev_ST = $subtypes{$viterbi_Path_ST[$i-1]};
        my $succ_ST = "";
        if ($i >= scalar(@viterbi_Path_ST)) {
          # restart at first segment
	  $succ_ST = $subtypes{$viterbi_Path_ST[0]};
	}
        else {
	  $succ_ST = $subtypes{$viterbi_Path_ST[$i]};
        }

        my $cur_BP  = $viterbi_Path_BP[$i];	
        my $prev_BP = $viterbi_Path_BP[$i-1];
        my $succ_BP = 0;

        if ($i+1 >= scalar(@viterbi_Path_BP)) {
          $succ_BP = $viterbi_Path_BP[1];
	}
        else {
          $succ_BP = $viterbi_Path_BP[$i+1];
	}

        if ($genome_type eq "circular" && $seq_type ne "fragment") {
          if ($i == 1) {
             if ($ST_seq[-1] == $prev_ST) {
               $prev_BP = $viterbi_Path_BP[-2]; # last "real" BP in recomb prediction
             }
          }
          if ($i == scalar(@viterbi_Path_BP)-2) {
             if ($ST_seq[0] == $succ_ST) { # first position, else len/1 is real predicted BP
               $succ_BP = $viterbi_Path_BP[1];  # first "real" BP in recomb prediction
             }
          }
        }
	if ($prev_ST ne $subtypes{$subtype_names[0]} && $succ_ST ne $subtypes{$subtype_names[-1]}) { # != *_Insertion -> TODO
	  $other_ST   = 0;
	  
	  # 1. to the left to preceding BP
	  my $j = $cur_BP-1;
          my $cur_prev_BP = $prev_BP;
          if ($prev_BP > $j) {
            $cur_prev_BP -= $seq_length;
          }
	  # in viterbi_path real positions -> start counting at 1
#	  while (($j > 0) && ($post_probs[$j-1][$prev_ST] < $UR_perc_threshold[$j-1]) && ($j > $prev_BP)) {
	  while ($j > $cur_prev_BP && ($post_probs[$j%$seq_length-1][$prev_ST] < $UR_perc_threshold[$j%$seq_length-1])) {
	    for (my $a=0; $a < scalar(keys(%subtypes)); $a++) {
	      if ($a != $prev_ST && $a != $succ_ST) {
		if ($post_probs[$j%$seq_length-1][$a] >= $post_probs[$j%$seq_length-1][$prev_ST] && $post_probs[$j%$seq_length-1][$a] >= $post_probs[$j%$seq_length-1][$succ_ST]) {
		  $other_ST = 1;
		  $j        = $j-1;
                  goto OTHER_ST1;
		}
	      }
	    }
	    $j = $j-1;
            if ($j <= 0) {
              $j = $seq_length-$j;
              $cur_prev_BP = $prev_BP; # wieder zurueck gesetzt, da Beginn unterschritten
            }
	  }
          OTHER_ST1:
	  if ($j != $prev_BP-1) {
	    # => threshold exceeded within this segment or there's another subtype with a higher/the same posterior probability
	    # => breakpoint region"
	    #    can be smaller than possible UR, because of another threshold
	    #    BP region: region in which the two ST have a posterior prob. smaller than the threshold for a BPR, extended to predicted BP
	    # start at j+1 (=1st pos with a post prob < threshold for UR),
	    # determine 1st pos. with a post prob < threshold for BPR
	    # => rest up to BP is BP region
	    
	    # "exact" breakpoints don't define a breakpoint "region" !!
	    my $k=($j+1);
            if ($k > $cur_BP) {
              $k = $k-$seq_length; # %seq_length statt - ??
            }
	    if ($other_ST == 0) {
	      while ($k < $cur_BP && $post_probs[$k%$seq_length-1][$prev_ST] >= $BPR_perc_threshold[$k%$seq_length]) {
		$k = ($k+1);#%$seq_length;
	      }
	      for (my $l=$k-1; $l<$cur_BP-1;$l++) { # BP region, no UR
		$BPR_path[$l%$seq_length] = 1;
		$UR_path[$l%$seq_length]  = 0;
	      }
	      # preceding region is not a UR any more
	      for (my $l=$j; $l<$k-1; $l++) { # alt: l<k-1 !!!
		$UR_path[$l%$seq_length] = 0;
	      }
	    }
	    else {
	      ## post prob of pred. ST not high enough, and other ST has a higher prob
              # other ST has a higher prob
	      for (my $l=$k-1; $l<$cur_BP-1;$l++) { # l=k, # alt: l=k-1 !!! s.o.
		if ($BPR_path[$l%$seq_length] == 0) {   # Bedg. hier richtig ???
		  $UR_path[$l%$seq_length] = 1;
		}
	      }
	    }
	  }
	  else {
	    # current subtype has at no position a post. prob > threshold => UR
            if ($prev_BP > $cur_BP) {
              $prev_BP = $prev_BP%$seq_length;
            }
	    for (my $a=$prev_BP-1; $a<$cur_BP-1; $a++) {
	      if ($BPR_path[$a%$seq_length] == 0) {
		$UR_path[$a%$seq_length] = 1;
	      }
	    }
	  }
	  
	  # 2. to the right to successive BP
	  $j        = $cur_BP;
	  $other_ST = 0;

          my $cur_succ_BP = $succ_BP;
          
          #todo??
          if ($j > $succ_BP) {
            $cur_succ_BP += $seq_length;
          }
	  while ($j < $cur_succ_BP && ($post_probs[$j%$seq_length-1][$succ_ST] < $UR_perc_threshold[$j%$seq_length-1])) {
	    for (my $a=0; $a<scalar(keys(%subtypes)); $a++) {
	      if ($a != $prev_ST && $a != $succ_ST) {
		if ($post_probs[$j%$seq_length-1][$a] >= $post_probs[$j%$seq_length-1][$prev_ST] && $post_probs[$j%$seq_length-1][$a] >= $post_probs[$j%$seq_length-1][$succ_ST]) {
		  $other_ST = 1;
		  $j        = $j+1;
		  #last;
                  goto OTHER_ST2;
		}
	      }
	    }
	    $j = $j+1;
            if ($j > $viterbi_Path_BP[-1]) {
              $j = $j-$seq_length;
              $cur_succ_BP = $succ_BP; # wieder zurueck gesetzt, da Ende ueberschritten
            }
	  }
          OTHER_ST2:
	  if ($j != $succ_BP) {
	    # => threshold exceeded within this segment or there's no other subtype with a higher/the same posterior probability
	    # => breakpoint region"
	    #    can be smaller than possible UR, because of another threshold
	    #    BP region: region in which the two ST have a posterior prob. smaller than the threshold for a BPR, extended to predicted BP
	    # start at j-1 (= last pos with a post prob < threshold for UR),
	    # determine last pos. with a post prob < threshold for BPR
	    # => rest back to BP is BP region

	    my $k=$j-1; #last position with post_prob < threshold
            if ($k < $cur_BP-1) {
              $k = $k+$seq_length;
            }
	    if ($other_ST == 0) {
	      while ($k > $cur_BP-1 && $post_probs[$k%$seq_length-1][$succ_ST] >= $BPR_perc_threshold[$k%$seq_length]) {
		$k = $k-1;
	      }

	      for (my $l=$cur_BP-1; $l<$k; $l++) {
		$BPR_path[$l%$seq_length] = 1;
		$UR_path[$l%$seq_length]  = 0;
	      }
	      # following region is not a UR any more
	      for (my $l=$k; $l<$j-1; $l++) {
		$UR_path[$l%$seq_length] = 0;
	      }
	    }
	    else {
	      # post prob of pred. ST not high enough, and other ST has a higher prob
	      for (my $l=$cur_BP-1; $l<$k; $l++) {
		if ($BPR_path[$l%$seq_length] == 0) {
		  $UR_path[$l%$seq_length] = 1;
		}
	      }
	    }
	  }
	  else {
	    # current subtype has at no position a post. prob > threshold => UR
            if ($succ_BP < $cur_BP) {
              $succ_BP += $seq_length;
            }
	    for (my $a=$cur_BP-1; $a<$succ_BP-1; $a++) {
	      if ($BPR_path[$a%$seq_length] == 0) {
		$UR_path[$a%$seq_length] = 1;
	      }
	    }
	  }
	}
      }
      # Test calculation of UR and BPR (are not allowed to overlap)
      for (my $j=0; $j<$seq_length; $j++) {
	if ($UR_path[$j] == 1 && $BPR_path[$j] == 1) {
	  die "at pos. $j BPR and UR\n";
	}
      }
      
      #########################################
      # Storage and output of the UR and BPR: #
      #########################################
      # breakpoints between last and first position of the query sequence
      # always define a breakpoint interval independently from posterior probabilities
      # comprising the last query sequence position
      # => if only this position is defined as breakpoint interval, it will be set to 0
      if  ($BPR_path[-1] == 1  && $BPR_path[-2] == 0 && $BPR_path[0] == 0) {
        $BPR_path[-1] = 0;
      }

      # Attention: UR cannot be followed by a BP region and vice versa
      # => replace each BPR by a UR if this happens
      for (my $j=0; $j<$seq_length; $j++) {
	if ($BPR_path[$j] == 1 && $UR_path[$j-1] == 1) {
	  $BPR_path[$j] = 0;
	  $UR_path[$j]  = 1;
	}
      }
      if ($genome_type eq "circular" && $seq_type ne "fragment") {
        my $j=0;
        while ($j < $seq_length && $BPR_path[$j] == 1 && $UR_path[$j-1] == 1) {
	  $BPR_path[$j] = 0;
	  $UR_path[$j]  = 1;
          $j++;
        }
      }

      for (my $j=0; $j<$seq_length; $j++) {
	my $k = ($seq_length-1)-($j+1);
	if ($BPR_path[$k] == 1 && $UR_path[($k+1)%$seq_length] == 1) {
	  $BPR_path[$k] = 0;
	  $UR_path[$k]  = 1;
	}
      }

      if ($genome_type eq "circular" && $seq_type ne "fragment") {
        my $j=$seq_length-1;
        while ($j >= 0 && $BPR_path[$j] == 1 && $UR_path[($j+1)%$seq_length] == 1) {
          $BPR_path[$j] = 0;
	  $UR_path[$j]  = 1;
          $j--;
        }
      }

      ##########################
      # Storage of UR and BPR: #
      ##########################
      # to store correct positions: j=1:seq_length
      # to find correct entry: j-1, since counting in perl starts at 0
      my @begin_UR  = ();
      my @end_UR    = ();
      my @begin_BPR = ();
      my @end_BPR   = ();
      
      for (my $j=1; $j<$seq_length+1; $j++) {  # in arrays: 0 - seq_length-1
	if ($j == 1) {
	  if ($UR_path[$j-1] == 1) { # UR start at 1st pos.
	    push(@begin_UR, $j);
	  }
	  elsif ($BPR_path[$j-1] == 1) {
	    push(@begin_BPR, $j);
	  }
	}
	elsif ($j == $seq_length) {
	  # can be begin and end of a UR
	  if ($UR_path[$j-1] == 1) { # UR end
	    if ($UR_path[$j-2] == 0) { # UR start
	      push(@begin_UR, $j);
	    }
	    push(@end_UR, $j);
	  }
	  else {#$UR_path[$j-1] == 0
	    if ($UR_path[$j-2] == 1) { # UR end
	      push(@end_UR, $j-1);
	    }
	  }
	  if ($BPR_path[$j-1] == 1) {
	    if ($BPR_path[$j-2] == 0) { # BPR start
	      push(@begin_BPR, $j);
	    }
	    push(@end_BPR, $j);
	  }
	  else {
	    if ($BPR_path[$j-2] == 1) { # BPR end at j-1
	      push(@end_BPR, $j-1);
	    }
	  }
	}
	else {
	  if ($UR_path[$j-1] == 1) {
	    if ($UR_path[$j-2] == 0) { # UR start
	      push(@begin_UR, $j);
	    }
	  }
	  else { #$UR_path[$j-1]==0
	    if ($UR_path[$j-2] == 1) { # UR end
	      push(@end_UR, $j-1);
	    }
	  }
	  if ($BPR_path[$j-1] == 1) {
	    if ($BPR_path[$j-2] == 0) { # BPR start
	      push(@begin_BPR, $j);
	    }
	  }
	  else {
	    if ($BPR_path[$j-2] == 1) { # BPR end at j-1
	      push(@end_BPR, $j-1);
	    }
	  }
	}
      }

      ######################################################
      # Output of the UR and BPR in text-format
      #       and the predicted recombination
      #         including uncertainty and breakpoint regions
      #         based on the information from the post. probs
      ######################################################

      if (scalar(@begin_UR) != scalar(@end_UR)) {
	print $seq_name[0], "\n";
	print "begin UR: \n";
	for (my $i=0; $i<scalar(@begin_UR); $i++) {
	  print $begin_UR[$i], "\t";
	}
	print "\n end UR:\n";
	for (my $i=0; $i<scalar(@end_UR); $i++) {
	  print $end_UR[$i], "\t";
	}
	print "\n";
	
	print "begin BPR: \n";
	for (my $i=0; $i<scalar(@begin_BPR); $i++) {
	  print $begin_BPR[$i], "\t";
	}
	print "\n end BPR:\n";
	for (my $i=0; $i<scalar(@end_BPR); $i++) {
	  print $end_BPR[$i], "\t";
	}
	print "\n";
	#die "error in determining the uncertainty regions; the number of begin and end positions is not the same!\n";
      }
      if (scalar(@begin_BPR) != scalar(@end_BPR)) {
	die "error in determining the breakpoint regions; the number of begin and end positions is not the same!\n";
      }
      
      my $akt_UR  = $UR_path[0];
      my $akt_BPR = $BPR_path[0];
      my $z0      = 1;
      my $z1      = 1;   # Pos in viterbi_path_breakpoints
      my %uncertainty_regions;
      my %breakpoint_intervals;
      while ($z0 < $seq_length+1) {
	my $x = $z0;
	while ($x < $seq_length+1 && $UR_path[$x-1] == $akt_UR && $BPR_path[$x-1] == $akt_BPR) {
	  #if ($z1 < scalar(@viterbi_Path_BP)-1 && $x == $viterbi_Path_BP[$z1]-1) {
	  if ($z1 < scalar(@viterbi_Path_BP)-1 && $x == $viterbi_Path_BP[$z1]-1) {
	    $z1 = $z1+1;
	  }
#	  if ($BPR_path[$x-1] == 0 && $z1 < scalar(@viterbi_Path_BP)-1 && $x == $viterbi_Path_BP[$z1-1]-1) {
	  if ($BPR_path[$x-1] == 0 && $x == $viterbi_Path_BP[$z1-1]-1) {
	    $x = $x+1;
	    last;
	  }
	  $x = $x+1;
	}
	
	# $z0 = begin of the current region
	# $x = already belongs to new region
	# $y = current end of considered region:
	
	my $y = $x-1;
	if ($x == $seq_length) {
	  $y = $x-1;
	}
	my $z = $z0;
	
	if ($akt_UR == 0) {
	  # no UR
	  if ($akt_BPR == 0) {
	    # no BPR => 'regular' prediction
	    print PRED_RECOMB $z, "\t", $y, "\t",  $subtype_names[$ST_seq[$z-1]], "\n";
	  }
	  else {
	    # BPR => not clear where breakpoint between the two ST => print both

            my $prev_ST = "";
            my $succ_ST = "";
            my $BPI_start = $z;
            my $BPI_end   = $y;

            if ($z > 1) { # z in real pos, here start at 0 => -1 everywhere
              $prev_ST = $subtype_names[$ST_seq[$z-1-1]];
	    }
            else {
              # determine corresponding BP of BPI
              # e.g.    can be      or
              # A       A/C         C/A
              # C       C           A
              # A       C/A         A/C
              #         A           C
              if ($BPI_start <= $viterbi_Path_BP[1] && $viterbi_Path_BP[1] <= $BPI_end+1) { # BPI around first BP
                $prev_ST = $subtype_names[$ST_seq[0]];
                $succ_ST = $viterbi_Path_ST[1];
	      }
              else { # BPI around last BP
                #$prev_ST = $ST_seq[-1];
                if ($ST_seq[0] == $ST_seq[-1]) {
                  $prev_ST = $viterbi_Path_ST[-2]; #vorletzter ST
                }
                else {
                  $prev_ST = $subtype_names[$ST_seq[-1]];
	        }
                $succ_ST = $subtype_names[$ST_seq[0]];
	      }
	    }
            if ($y+1-1 >= scalar(@ST_seq)) {
              if ($BPI_start <= $viterbi_Path_BP[-2] && $viterbi_Path_BP[-2] <= $BPI_end+1) { # BPI around last BP (BP = BP-1/BP -> desh. BPI_end+1)
                $prev_ST = $viterbi_Path_ST[-2];
                $succ_ST = $subtype_names[$ST_seq[-1]];
	      }
              else { # BPI around first BP
                #$prev_ST = $ST_seq[-1];
                if ($ST_seq[0] == $ST_seq[-1]) {
                  $succ_ST = $viterbi_Path_ST[1]; #zweiter ST
                }
                else {
                  $succ_ST = $subtype_names[$ST_seq[0]];
	        }
                $prev_ST = $subtype_names[$ST_seq[-1]];
	      }
            }
            else {
              $succ_ST = $subtype_names[$ST_seq[$y+1-1]];
	    }
	    my $cur_BPI = $prev_ST."/".$succ_ST;
	    print PRED_RECOMB $z, "\t", $y, "\t",  $cur_BPI, "\n";

	    my @cur_list = ($y, $cur_BPI);
	    $breakpoint_intervals{$z} = \@cur_list;
	  }
	}
	else {
	  # UR => define region as U, but also give predicted subtype for these positions
	  print PRED_RECOMB $z, "\t", $y, "\t?/", $subtype_names[$ST_seq[$z-1]], "\n";
	  my $cur_UR = "?/".$subtype_names[$ST_seq[$z-1]];
	  my @cur_list = ($y, $cur_UR);
	  $uncertainty_regions{$z} = \@cur_list;
	}
	
	$akt_UR  = $UR_path[$x-1];
	$akt_BPR = $BPR_path[$x-1];
	$z0       = $x;
      }

      print UR_REG "# uncertainty regions:\n";
      foreach my $start (sort {$a <=> $b} keys(%uncertainty_regions)) {
	print UR_REG $start, "\t", $uncertainty_regions{$start}->[0], "\t", $uncertainty_regions{$start}->[1], "\n";
      }
      print UR_REG "# breakpoint regions:\n";
      foreach my $start (sort {$a <=> $b} keys(%breakpoint_intervals)) {
	print UR_REG $start, "\t", $breakpoint_intervals{$start}->[0], "\t", $breakpoint_intervals{$start}->[1], "\n";
      }
      
      print PRED_RECOMB "\n";
      print UR_REG "\n";
      #####################################
      # reinitialization for next sequence
      #####################################

      $seq_length      = 0;
      @seq_name        = ();
      @recomb_data     = ();
      @ST_seq          = ();
      @VAR_pos         = ();
      @viterbi_Path_BP = ();
      @viterbi_Path_ST = ();
    }
  }
} # end until

close(VIT_RECOMB);
if (-e $var_reg_exist) {
  close(VAR_REG);
}
close(UR_REG);
close(PRED_RECOMB);
