// ~/jpHMM/src/emTransProfile.cpp
/**************************************************************
 File name	: emTransProfile.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Profile of each column in each subtype:
 = emission and transition profile (see resp. files)

 - Determination of consensus and non-consensus columns in each subtype in the given alignment
   by determination of number of observed nucleotides in each column
 - For each consensus column, determination of preceding and successive consensus column
 - Determination of first and last consensus column in whole model
 - Determination of number of special characters in the alignment
***************************************************************/

#include "std_source.h"
#include "error.h"
#include "alphabet.h"
#include "const_def.h"
#include "alignment.h"
#include "emissionProfile.h"
#include "transitionProfile.h"
#include "emTransProfile.h"
#include "extEmTrProbs.h"

using namespace std;

const bool OUTPUT_DEBUG = false;


/*******************************************************\

 *                 class EmTransProfile                *

\*******************************************************/

EmTransProfile::EmTransProfile(Alignment *al, bool use_eetp, const ExtEmTrProbs* eetp) {

  vector<Subtype*> subs           = al->subs();
  Alphabet*        alphabet       = Const_Def::alphabet();
  TransAlphabet*   trans_alphabet = Const_Def::trans_alphabet();

  map<SubtypeProfile*, vector<bool> > sub_profs_cons_cols;
  
  int num_spec_char = 1;
  int num_all_char  = 0;

  FOREACH (Subtype* st, subs) {
    SubtypeProfile* cur_sub_prof   = new SubtypeProfile(st);
    vector<Seq*>    cur_seqs       = st->seqs();
    int             sub_len        = st->len();
    vector<bool>    cur_cons_cols;

    for (int cur_col=0; cur_col<sub_len; ++cur_col) {
      ColumnProfile* cur_col_prof = new ColumnProfile(cur_seqs, cur_col);
      cur_sub_prof->col_profs()  += cur_col_prof;
      cur_cons_cols              += use_eetp ? eetp->is_cons_col(cur_col) : cur_col_prof->is_cons_col();
      vector<int> frequencies     = cur_col_prof->frequencies();
      vector<int>::iterator it    = --(--frequencies.end());

      num_spec_char += (*it); // elem. an vorletzter Pos.
      num_all_char  += cur_seqs.size()-frequencies.back();
    }
    m_sub_profs += cur_sub_prof;
    sub_profs_cons_cols[cur_sub_prof] = cur_cons_cols;
  }

  //! determine first and last consensus column that is present in ALL subtypes
  int sub_len = subs[0]->len();
  m_first_cons_col = sub_len-1;
  for (int cur_col=0; cur_col<sub_len; ++cur_col) {
    int num_cons_cols = 0;
    FOREACH (SubtypeProfile* cur_sub_prof, m_sub_profs) {
      if (cur_sub_prof->col_profs()[cur_col]->is_cons_col())
        num_cons_cols++;
    }
    //! The first common consensus column must be a consensus column in ALL subtypes.
    if (num_cons_cols == (int)m_sub_profs.size()) {
      m_first_cons_col = cur_col;
      break;
    }
  }
  m_last_cons_col = 0;
  for (int cur_col=sub_len-1; cur_col >= m_first_cons_col; --cur_col) {
    int num_cons_cols = 0;
    FOREACH (SubtypeProfile* cur_sub_prof, m_sub_profs) {
      if (cur_sub_prof->col_profs()[cur_col]->is_cons_col())
        num_cons_cols++;
    }
    //! The last common consensus column must be a consensus column in ALL subtypes.
    if (num_cons_cols == (int)m_sub_profs.size()) {
      m_last_cons_col = cur_col;
      break;
    }
  }
  if (m_first_cons_col > m_last_cons_col) 
    throw Error("No column exists that is a consensus column in all subtypes");

    //! determine consensus column profiles in each subtype (within the first and last consensus column)

  FOREACH (SubtypeProfile* cur_sub_prof, m_sub_profs) {
    vector<bool> cur_cons_cols = sub_profs_cons_cols[cur_sub_prof];
    //! ConsColumnProfile for Begin state (N-terminal state)
    ConsColumnProfile* cur_cons_prof = new ConsColumnProfile(cur_sub_prof, *alphabet, *trans_alphabet, -1, m_first_cons_col);
    cur_sub_prof->nterminal_cons_col_prof() = cur_cons_prof;
    
    
    for (int cur_col=m_first_cons_col; cur_col<=m_last_cons_col; ++cur_col) {
      if (cur_cons_cols[cur_col] == true) {
        int succ_cons_col = cur_col+1;
        while (succ_cons_col <= m_last_cons_col && cur_cons_cols[succ_cons_col] == false)
          ++succ_cons_col;
        ConsColumnProfile* cur_cons_prof = new ConsColumnProfile(cur_sub_prof, *alphabet, *trans_alphabet, cur_col, succ_cons_col);
        cur_sub_prof->cons_col_profs()[cur_col] = cur_cons_prof;
      }
    }
  }

  if (num_all_char > 0)
    m_rel_freq_spec_chars = ((mytype)num_spec_char)/((mytype)num_all_char);
  else
    m_rel_freq_spec_chars = (mytype)num_spec_char/((mytype)1000); // arbitrarily chosen;
  if (OUTPUT_DEBUG)
	  cout << "num X " << num_spec_char << " all res: " << num_all_char << " ";
  check();
}

//------------------------------------------------------------------------------
EmTransProfile::~EmTransProfile(){

  FOREACH (SubtypeProfile* sub_prof, m_sub_profs)
    delete sub_prof;
}

//------------------------------------------------------------------------------
// void check():
//------------------------------------------------------------------------------
void EmTransProfile::check(){

  if (m_sub_profs.size() == 0)
    throw Error("No subtype profile could be determined\n");

  FOREACH (SubtypeProfile* sub_prof, m_sub_profs)
    sub_prof->check();
}

//------------------------------------------------------------------------------
// output:
//------------------------------------------------------------------------------
ostream& operator<<(ostream& s, const EmTransProfile& pr) {

  FOREACH (SubtypeProfile* prof, pr.m_sub_profs)
    s << *prof << endl;
  return s;
}

/*******************************************************\

 *                 class SubtypeProfile                *

\*******************************************************/

SubtypeProfile::SubtypeProfile(Subtype* cur_sub) {
  m_sub = cur_sub;
}

SubtypeProfile::~SubtypeProfile() {

  FOREACH (ColumnProfile* c, m_col_profs)
     delete c;

  for(map<int, ConsColumnProfile*>::const_iterator it=m_cons_col_profs.begin(); it != m_cons_col_profs.end(); ++it)
    delete it->second;

  delete m_nterminal_cons_col_prof;
}

//------------------------------------------------------------------------------
// void check():
//------------------------------------------------------------------------------
void SubtypeProfile::check() {
  if (!m_sub)
    throw Error("Subtype profile in illegal state");
  if (!m_cons_col_profs.size())
    throw Error("No consensus column for at least one subtype");
}

//--------------------------------------------------------------------------------------------------
// has_pred_cons_col(int i):
//
// current state has a preceding consensus column?
//--------------------------------------------------------------------------------------------------
bool SubtypeProfile::has_pred_cons_col(int i) const {
  // TODO: beachte ConsColProfile fuer Begin => in emTransProfile aendern
  map<int, ConsColumnProfile*>::const_iterator it = m_cons_col_profs.lower_bound(i); // greater or equal i

  if (it == m_cons_col_profs.begin() || m_cons_col_profs.size() == 0)
    return false;
  return true;
}

//--------------------------------------------------------------------------------------------------
// has_succ_cons_col(int i):
//
// current state has a successive consensus column?
//--------------------------------------------------------------------------------------------------
bool SubtypeProfile::has_succ_cons_col(int i) const {

  map<int, ConsColumnProfile*>::const_iterator it = m_cons_col_profs.upper_bound(i);

  if (it == m_cons_col_profs.end())
    return false;
  return true;
}

//--------------------------------------------------------------------------------------------------
// pred_cons_col(int i):
//
// return the preceding consensus column for a certain column in the subtype
//--------------------------------------------------------------------------------------------------
int SubtypeProfile::pred_cons_col(int i) const {

  map<int, ConsColumnProfile*>::const_iterator it = m_cons_col_profs.lower_bound(i); // greater or equal i

  if (it == m_cons_col_profs.begin() || m_cons_col_profs.size() == 0) {
    stringstream ss;
    ss << "In this subtype profile no successive consensus column exists for column: \'" << i << "\'";
    throw Error(ss.str());
  }

  --it;
  return it->first;
}

//--------------------------------------------------------------------------------------------------
// succ_cons_col(int i):
//
// return the following consensus column for a certain column in the subtype
//--------------------------------------------------------------------------------------------------
int SubtypeProfile::succ_cons_col(int i) const {

  map<int, ConsColumnProfile*>::const_iterator it = m_cons_col_profs.upper_bound(i);

  if (it == m_cons_col_profs.end()) {
    stringstream ss;
    ss << "In this subtype profile no successive consensus column exists for column: \'" << i << "\'";
    throw Error(ss.str());
  }
  return it->first;
}

//--------------------------------------------------------------------------------------------------
// num_pred_cons_col(int i):
//
// return the number of preceding consensus columns for a certain column in the subtype
//--------------------------------------------------------------------------------------------------
int SubtypeProfile::num_pred_cons_cols(int i) const {

  int num_pred_cons_cols = 0;

  for (map<int, ConsColumnProfile*>::const_iterator it = m_cons_col_profs.begin(); it != m_cons_col_profs.lower_bound(i); ++it)
    ++num_pred_cons_cols;

  return num_pred_cons_cols;
}

//--------------------------------------------------------------------------------------------------
// num_succ_cons_col(int i):
//
// return the number of following consensus columns for a certain column in the subtype
//--------------------------------------------------------------------------------------------------
int SubtypeProfile::num_succ_cons_cols(int i) const {

  int num_succ_cons_cols = 0;

  for (map<int, ConsColumnProfile*>::const_iterator it = m_cons_col_profs.upper_bound(i); it != m_cons_col_profs.end(); ++it)
    ++num_succ_cons_cols;

  return num_succ_cons_cols;
}

//--------------------------------------------------------------------------------------------------
// is_cons_col(int i):
//--------------------------------------------------------------------------------------------------
bool SubtypeProfile::is_cons_column(int i) const {

  map<int, ConsColumnProfile*>::const_iterator it = m_cons_col_profs.find(i);
  if (it == m_cons_col_profs.end())
    return false;
  else
    return true;
}

//--------------------------------------------------------------------------------------------------
// cons_col_prof(int i):
//
// return ConsColumnProfile for a certain column in the subtype
//--------------------------------------------------------------------------------------------------
ConsColumnProfile* SubtypeProfile::cons_col_prof(int i) const {

  map<int, ConsColumnProfile*>::const_iterator it = m_cons_col_profs.find(i);

  if (it == m_cons_col_profs.end()) {
    stringstream ss;
    ss << "Trying to get consensus column profile for a non-consensus column: \'" << i << "\'";
    throw Error(ss.str());
  }
  return it->second;
}

//------------------------------------------------------------------------------
// output:
//------------------------------------------------------------------------------
ostream& operator<<(ostream& s, const SubtypeProfile& st_pr) {

  s << "subtype: " << st_pr.m_sub->name() << endl;
  s << "n_terminal: " << endl;
  s << *st_pr.m_nterminal_cons_col_prof << endl;
  s << "consensus columns: " << endl;
  
  for(map<int, ConsColumnProfile*>::const_iterator it=st_pr.m_cons_col_profs.begin(); it != st_pr.m_cons_col_profs.end(); ++it) {
    s << "column " << it->first+1 << ": " << endl;
    s << *it->second;
  }

  return s;
}


/*******************************************************\

 *                 class ConsColumnProfile             *

\*******************************************************/

ConsColumnProfile::ConsColumnProfile(SubtypeProfile* sub_prof, const Alphabet& al, const TransAlphabet& t_al, int cons_col, int succ_cons_col) {
  m_sub_prof    = sub_prof;
  m_column      = cons_col;
  m_succ_column = succ_cons_col;
  m_em_prof     = new EmissionProfile(this, al);
  m_trans_prof  = new TransitionProfile(this, t_al);
}

ConsColumnProfile::~ConsColumnProfile() {
  delete m_em_prof;
  delete m_trans_prof;
}

//------------------------------------------------------------------------------
// void check():
//------------------------------------------------------------------------------
void ConsColumnProfile::check() {
  if (!m_sub_prof || !m_em_prof || !m_trans_prof)
    throw Error("Consensus column profile in illegal state");
}

//------------------------------------------------------------------------------
// output:
//------------------------------------------------------------------------------
ostream& operator<<(ostream& s, const ConsColumnProfile& c) {

  s << *c.m_em_prof << endl;
  s << *c.m_trans_prof << endl;
  return s;
}


/*******************************************************\

 *                 class ColumnProfile                 *

\*******************************************************/

ColumnProfile::ColumnProfile(vector<Seq*> seqs, int col) {
  m_column   = col;

  Alphabet* al = Const_Def::alphabet();
  int size = al->num_chars();

  m_frequencies = list_of(0).repeat(size-1, 0);
  FOREACH (Seq* s, seqs) {
    char cur_res   = s->str(m_column);
    int  cur_index = al->idx(cur_res);

    ++m_frequencies[cur_index];      // A, C, G, T, X and gaps (-)
  }

  int num_residues = seqs.size() - m_frequencies.back(); // m_frequencies.back() = gaps
  if ((num_residues >= Const_Def::consensus_percentage() * seqs.size()) || (num_residues >= Const_Def::consensus_threshold()))
    m_is_cons_col = true;
  else
    m_is_cons_col = false;
}


//------------------------------------------------------------------------------
// void check():
//------------------------------------------------------------------------------
void ColumnProfile::check() {
  if (!m_frequencies.size())
    throw Error("Column profile in illegal state");
}

//------------------------------------------------------------------------------
// output:
//------------------------------------------------------------------------------
ostream& operator<<(ostream& s, const ColumnProfile& c) {

  s << "column:" << c.m_column << endl;
  if (c.m_is_cons_col) 
    s << "is a consensus column" << endl;
  else
    s << "is not a consensus column" << endl;
  s << "nucl. frequencies: " << c.m_frequencies << endl;
  return s;
}

