// ~/jpHMM/src/emTransProfile.h
/**************************************************************
 File name	: emTransProfile.h
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Classes: EmTransPofile
          SubtypeProfile
	  ConsColumnProfile
	  ColumnProfile

 Profile of each column in each subtype:
 = emission and transition profile (see resp. files)

 - Determination of consensus and non-consensus columns in each subtype in the given alignment
   by determination of number of observed nucleotides in each column
 - For each consensus column, determination of preceding and successive consensus column
 - Determination of first and last consensus column in whole model
 - Determination of number of special characters in the alignment
***************************************************************/

#ifndef _EMTRANSPROFILE_H
#define _EMTRANSPROFILE_H


using namespace std;

class Alphabet;
class TransAlphabet;
class Alignment;
class Subtype;
class Seq;
class EmissionProfile;
class TransitionProfile;
class ColumnProfile;
class ConsColumnProfile;
class SubtypeProfile;
class ExtEmTrProbs;

class EmTransProfile
{
   friend ostream& operator<<(ostream& s, const EmTransProfile& pr);
   public:
      EmTransProfile(Alignment* al, bool use_eetp, const ExtEmTrProbs* eetp = 0);
      ~EmTransProfile();
      void check();
      vector<SubtypeProfile*> sub_profs()           const {return m_sub_profs;};
      mytype                  rel_freq_spec_chars() const {return m_rel_freq_spec_chars;};
      int                     first_cons_col()      const {return m_first_cons_col;};
      int                     last_cons_col()       const {return m_last_cons_col;};
      
  private:
      vector<SubtypeProfile*> m_sub_profs;
      mytype                  m_rel_freq_spec_chars;
      int                     m_first_cons_col;  //! first common consensus column of all subtypes
      int                     m_last_cons_col;   //! last common consensus column of all subtypes
};

class SubtypeProfile
{
   friend ostream& operator<<(ostream& s, const SubtypeProfile& st_pr);
   public:
      SubtypeProfile(){};
      SubtypeProfile(Subtype *sub);
      ~SubtypeProfile();
      void check();
      bool has_pred_cons_col(int i)  const;
      bool has_succ_cons_col(int i)  const;
      int  pred_cons_col(int i)      const;
      int  succ_cons_col(int i)      const;
      int  num_pred_cons_cols(int i) const;
      int  num_succ_cons_cols(int i) const;


      bool is_cons_column(int i) const;
      ConsColumnProfile* cons_col_prof(int i) const;

      Subtype*                      sub()                     const { return m_sub;};
      map<int, ConsColumnProfile*> &cons_col_profs()                { return m_cons_col_profs;};
      map<int, ConsColumnProfile*>  cons_col_profs()          const { return m_cons_col_profs;};
      vector<ColumnProfile*>        col_profs()               const { return m_col_profs;};
      vector<ColumnProfile*>       &col_profs()                     { return m_col_profs;};
      ConsColumnProfile*            nterminal_cons_col_prof() const { return m_nterminal_cons_col_prof;};
      ConsColumnProfile*           &nterminal_cons_col_prof()       { return m_nterminal_cons_col_prof;};

   private:
      Subtype*                     m_sub;
      map<int, ConsColumnProfile*> m_cons_col_profs;
      ConsColumnProfile*           m_nterminal_cons_col_prof;
      vector<ColumnProfile*>       m_col_profs;
};

class ConsColumnProfile
{
   friend ostream& operator<<(ostream& s, const ConsColumnProfile& c);
   public:
     ConsColumnProfile(SubtypeProfile* sub_prof, const Alphabet& al, const TransAlphabet& t_al, int cons_col, int succ_cons_col);
     ~ConsColumnProfile();
     void check();

     SubtypeProfile*     sub_prof()    const {return m_sub_prof;};
     EmissionProfile*    em_prof()     const {return m_em_prof;};
     TransitionProfile*  trans_prof()  const {return m_trans_prof;};
     int                 column()      const {return m_column;};
     int                 succ_column() const {return m_succ_column;};
   private:
     SubtypeProfile*    m_sub_prof;
     EmissionProfile*   m_em_prof;
     TransitionProfile* m_trans_prof;
     int                m_column;
     int                m_succ_column;
};

class ColumnProfile
{
   friend ostream& operator<<(ostream& s, const ColumnProfile& c);
   public:
     ColumnProfile(vector<Seq*> seqs, int col);
     ~ColumnProfile(){};
     void check();

     int             column()      const {return m_column;};
     vector<int>     frequencies() const {return m_frequencies;};
     bool            is_cons_col() const {return m_is_cons_col;};

   private:
     int             m_column;               //! corresponding column in alignment
     vector<int>     m_frequencies;          //! A, C, G, T, X (sonstige), gaps - frequencies
     bool            m_is_cons_col;          //! column is consensus column?
};

ostream& operator<<(ostream& s, const EmTransProfile& pr);
ostream& operator<<(ostream& s, const SubtypeProfile& st_pr);
ostream& operator<<(ostream& s, const ConsColumnProfile& c);
ostream& operator<<(ostream& s, const ColumnProfile& c);

#endif  // _EMTRANSPROFILE_H
