// ~/jpHMM/src/emissionProfile.cpp
/**************************************************************
 File name	: emissionProfile.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Determination of observed nucleotide and gap frequencies in a certain 
  consensus column (for match state) and successive non-consensus columns (insert state)
  for a certain subtype in the given alignment
***************************************************************/

#include "std_source.h"
#include "error.h"
#include "alphabet.h"
#include "const_def.h"
#include "emTransProfile.h"
#include "emissionProfile.h"

using namespace std;

/*******************************************************\

 *                 class EmissionProfile               *

\*******************************************************/

EmissionProfile::EmissionProfile(ConsColumnProfile* cons_prof, const Alphabet& al) {

  m_column = cons_prof->column();
  SubtypeProfile* cur_sub_prof  = cons_prof->sub_prof();
  int             succ_cons_col = cons_prof->succ_column();

  m_match_frequencies.assign(al.num_chars(), 0);
  m_insert_frequencies.assign(al.num_chars(), 0);

  if (m_column > -1) {
    ColumnProfile* cur_col_prof = cur_sub_prof->col_profs()[m_column];
    vector<int> col_frequencies = cur_col_prof->frequencies();

    for (unsigned i=0; i<col_frequencies.size(); ++i)
      m_match_frequencies[i] += col_frequencies[i];
  }
  for (int col=m_column+1; col<succ_cons_col; ++col) {
    ColumnProfile* cur_col_prof = cur_sub_prof->col_profs()[col];
    vector<int> col_frequencies = cur_col_prof->frequencies();

    for (unsigned i=0; i<col_frequencies.size(); ++i)
      m_insert_frequencies[i] += col_frequencies[i];
  }
}

//------------------------------------------------------------------------------
// void check():
//------------------------------------------------------------------------------
void EmissionProfile::check(){

  if (m_column > -1 && !m_match_frequencies.size())
    throw Error("No match emission frequencies could be determined\n");
  if (!m_insert_frequencies.size())
    throw Error("No insert emission frequencies could be determined\n");
}


//------------------------------------------------------------------------------
// match_frequency(int i):
//  observed (nucleic) frequency in consensus column:
//------------------------------------------------------------------------------
int EmissionProfile::match_frequency(int i) const {
  try {
    return m_match_frequencies[i];
  }
  catch(out_of_range) {
    throw Error("illegal index");
  }
}

//------------------------------------------------------------------------------
// match_frequency(int i):
//  observed (nucleic) frequency in consensus column:
//------------------------------------------------------------------------------
int EmissionProfile::insert_frequency(int i) const {
  try {
    return m_insert_frequencies[i];
  }
  catch(out_of_range) {
    throw Error("illegal index");
  }
}


//------------------------------------------------------------------------------
// output:
//------------------------------------------------------------------------------
ostream& operator<<(ostream& s, const EmissionProfile& e) {
  s << "match emission frequencies (A,C,G,T,X,-): " << e.m_match_frequencies << endl;
  s << "insert emission frequencies (A,C,G,T,X,-): " << e.m_insert_frequencies;
  return s;
}

