#!/usr/bin/perl

use warnings;
use strict;
use POSIX;
use List::Util qw[min max];
use List::Util qw[sum];

# default path for BLAT program and BLAT parameters
my $blatSource  = "~/Programme/Blat/blat";
my $minIdentity = 80;    # min. seq. identity (%), df = 90
my $minScore    = 30;    # min. score: 2xmatches - mismatches - gap penalty, df = 30
my $oneOff      = 1;     # = 1 allows 1 mismatch in tile (len=11), df = 0

# default input files
my $database    = "HIV-alignment_pureSubtypes.fas";

# default output file
my $output  = "blat_output.psl";

# tmp file for raw MSA sequences (for BLAT)
my $rawSeqs = "raw_sequences.fas";

# active alignment columns for each query sequence position (for jpHMM)
my $alignmentColumns   = "active_alignment_columns.txt";
my $query              = "query_sequences.fas";

my $isCircGen = 0;
my $msaExt = 2;  # for circular genomes at least doubled!
my $origMSALen = 0;

my $numChosenSeqs   = 100;   # number of sequences of MSA that will be chosen for Blat alignment (in $rawSeqs)
my $alSeqsPerc      = 0.5;   # number of sequences each query position must be aligned to to be considered as "aligned"
my $gapPerc         = 0.8;   # at least 80% non gaps OR
my $consColThresh   = 0.8;   # threshold for conserved columns
my $smoothCrit      = 2;
my $intervalExt     = 3;     # extension of active interval

# for each alignment column save conserved, gap and active columns
my @conservedCol    = ();
my @gapCol          = ();
my @nonActCol = ();

# non active intervals in alignment
my @nonActInt = ();


# in the case of several possible pairwise alignments of a query sequence to a database sequence
my $accScoreDistPerc    = 20/100;
my $accAlignedQueryPerc = 90/100;       # percentage of Query that must be aligned (last-first+1 aligned position)
my $accStretchPerc      = 20/100;       # length of aligned target Seq = last-first+1 aligned column
my $accColDistPerc      = 90/100;       # allowed distance (percentage of orig MSA len) of 2 aligned columns in alignment (can be >= origMSALen for circular genomes -> not allowed)


if (scalar(@ARGV) < 4) {
    die "please give the following parameters: [blat_source_dir] [msa_file] [query_file] [active_msa_cols_outfile] [circular](opt.if it's a circular genome)\n";
}
else {
  $blatSource       = $ARGV[0];
  $database         = $ARGV[1];
  $query            = $ARGV[2];
  $alignmentColumns = $ARGV[3];

  if (scalar(@ARGV) > 4) {
      if ($ARGV[4] =~ m/circular/) {
          $isCircGen = 1;
      }
      else {
          die "the fifth argument must be [circular]\n";
      }
      if (scalar(@ARGV) == 5) {
          die "Please give the [MSA_extension] for circular genomes:\n [blat_source_dir] [msa_file] [query_file] [active_msa_cols_outfile] [cicular] [orig_MSA_Length]\n";
      }
      else {
          $msaExt = $msaExt + $ARGV[5]; # msa doubled + certain part/factor of alignment
      }
  }
}


##############################################################################################################
# Read in database source (multiple sequence alignment)
# save raw sequences for BLAT
#
# Save for each raw sequence positions the position in the sequence in the alignment (including gaps)

##############################################################################################################
# Save each msa sequence once as query sequence and evaluate the accuracy of the Blat Alignment:
# for each msa sequence position the position in the msa is known
# => compare these positions to the interquartile range of the positions defined by the Blat alignment
##############################################################################################################


open(DATABASE, "$database") or die "no database file available\n";

my @seqs         = <DATABASE>;
my @seqNames     = ();
my %msaPositions;
my %msaSeqs;
my %assignedST;
my %subtypeNames;
my %allSeqs;
my $curSeqName   = "";
my $numSeqs      = 0;
my $numColumns   = 0;
my $numSubtypes  = 0;
my @numSeqsPerST = ();   # number of sequences for each subtype
my @colFreqs     = ();
my @numGaps      = ();
my %queryNames   = ();

my $seq = "";

for (my $i=0; $i<scalar(@seqs); $i++) {
  chomp($seqs[$i]);

  if ($seqs[$i] =~ m/^>>/) { #subtype
    $seqs[$i] =~ s/\>>//;
    $subtypeNames{$numSubtypes} = $seqs[$i];
    if ($numSubtypes > 0) { push @numSeqsPerST, $numSeqs; }
    $numSubtypes++;
    $numSeqs = 0;
  }
  else {
    if ($seqs[$i] =~ m/>/) { # sequence name
      if (scalar(@seqNames) > 0) {
        my @curSeq    = split("", $seq);  # current sequence in alignment
        $numColumns   = scalar(@curSeq);
        my @curMsaPos = ();

        if (scalar(@seqNames) == 1) {
	    $origMSALen = int($numColumns/$msaExt); # ~ orig. msa length (can slightly differ due to rounding in extend_circ_gen.pl
          for (my $j=0; $j<$numColumns; $j++) {
            push(@numGaps, 0);
            my @nuclFreqs = ();
            for (my $k=0; $k<4; $k++) { push(@nuclFreqs, 0); }
            push @colFreqs, \@nuclFreqs;
          }
        }

        # length of MSA (for subroutine (interval of active columns))
        if ($origMSALen == 0) {
	    $origMSALen = $numColumns;
	}

        for (my $j=0; $j<$numColumns; $j++) {
          if ($curSeq[$j] ne '-') {
	    # save pos. of cur seq pos in MSA
	    push @curMsaPos, $j;

	    # nucleotide frequencies:
	    if    ($curSeq[$j] eq "A" || $curSeq[$j] eq "a") { $colFreqs[$j][0]++; }
	    elsif ($curSeq[$j] eq "C" || $curSeq[$j] eq "c") { $colFreqs[$j][1]++; }
	    elsif ($curSeq[$j] eq "G" || $curSeq[$j] eq "g") { $colFreqs[$j][2]++; }
	    elsif ($curSeq[$j] eq "T" || $curSeq[$j] eq "t") { $colFreqs[$j][3]++; }
	  }
          else {
	    $numGaps[$j]++;
	  }
        }
        $msaPositions{$curSeqName} = \@curMsaPos;
        $allSeqs{$curSeqName}      = \@curSeq;
      }
      $curSeqName = $seqs[$i];
      $curSeqName =~ s/\>//;
      push @seqNames, $curSeqName;
      $assignedST{$curSeqName} = $numSubtypes-1;
      $numSeqs++;
      $seq = "";
    }
    else {
     $seq .= $seqs[$i];
    }
  }
}
my @curSeq    = split("", $seq);
$numColumns   = scalar(@curSeq);
my @curMsaPos = ();

if ($numSubtypes == 1) {
  die "Please define at least two subtypes in the alignment\n";
}

for (my $j=0; $j<$numColumns; $j++) {
  if ($curSeq[$j] ne '-') {
    # save pos. of cur seq pos in MSA
    push @curMsaPos, $j;

    # nucleotide frequencies:
    if    ($curSeq[$j] eq "A" || $curSeq[$j] eq "a") { $colFreqs[$j][0]++; }
    elsif ($curSeq[$j] eq "C" || $curSeq[$j] eq "c") { $colFreqs[$j][1]++; }
    elsif ($curSeq[$j] eq "G" || $curSeq[$j] eq "g") { $colFreqs[$j][2]++; }
    elsif ($curSeq[$j] eq "T" || $curSeq[$j] eq "t") { $colFreqs[$j][3]++; }
  }
  else {
    $numGaps[$j]++;
  }
}
$msaPositions{$seqNames[scalar(@seqNames)-1]} = \@curMsaPos;
$allSeqs{$seqNames[scalar(@seqNames)-1]}      = \@curSeq;

# for last ST:
push @numSeqsPerST, $numSeqs;
close(DATABASE);

$numSeqs = 0;
foreach (@numSeqsPerST) {
  $numSeqs += $_;
}


# choose $numChosenSeqs sequences in alignment (for each subtypes $numChosenSeqs/scalar(@numSeqsPerST))
open(RAW_SEQS, ">$rawSeqs");

my $numChosenSeqsPerST = &ceil($numChosenSeqs/scalar(@numSeqsPerST));
my $numSeqsToAlign     = 0;
for (my $i=0; $i<scalar(@numSeqsPerST);$i++) {
  if ($numSeqsPerST[$i] <= $numChosenSeqsPerST) {
    foreach my $curSeqName (@seqNames) {
      if ($assignedST{$curSeqName} == $i) {
        my $curSeq = join("", @{$allSeqs{$curSeqName}});
        $curSeq =~ s/-//g;

        print RAW_SEQS ">", $curSeqName, "\n";
        print RAW_SEQS $curSeq, "\n";
      }
    }
    $numSeqsToAlign += $numSeqsPerST[$i];
  }
  else {
    my %index;
    for (my $j=0; $j<$numChosenSeqsPerST; $j++) {
      my $number = int(rand(scalar(@seqNames)));
      redo while (exists $index{$number} || $assignedST{$seqNames[$number]} != $i);
      $index{$number} = 1;
      my $curSeq = join("", @{$allSeqs{$seqNames[$number]}});
      $curSeq =~ s/-//g;

      print RAW_SEQS ">", $seqNames[$number], "\n";
      print RAW_SEQS $curSeq, "\n";
    }
    $numSeqsToAlign += $numChosenSeqsPerST;
  }
}
close(RAW_SEQS);

# Frequencies of the most frequent nucleotide
my @nuclOfMax = ();
my @freqOfMax = ();
my @percOfMax = ();
my $nuclOfMax = 0;
my $curMax    = 0;

for (my $i=0; $i<$numColumns; $i++) {
  $nuclOfMax = 0;
  $curMax    = 0;
  for (my $j=0; $j<4; $j++) {
    if ($colFreqs[$i][$j] > $curMax) {
      $curMax    = $colFreqs[$i][$j];
      $nuclOfMax = $j;
    }
  }
  push @nuclOfMax, $nuclOfMax;
  push @freqOfMax, $curMax;
  push @percOfMax, ($curMax/$numSeqs*100);
}
$numSeqs = 0;  # sum of all seqs
foreach (@numSeqsPerST) { $numSeqs += $_; }


#############################################################################################
# active/non-active columns in MSA:x
#############################################################################################
# 1) Definition of conserved/variable columns in the MSA:
#    conserved column: frequency of one nucl >= $consColThresh*100%
#############################################################################################

for (my $i=0; $i<$numColumns; $i++) {
  if ($percOfMax[$i] >= $consColThresh * 100) { push @conservedCol, 0; } # conserved column
  else                                        { push @conservedCol, 1; } # not conserved
}

#############################################################################################
# 2) Definition of gap columns
#     MSA column = gap column if gap column in ALL ST
#     gap column in ST: > $gapPerc% gaps (frueher auch: OR < $gapThresh residues)
#############################################################################################
for (my $i=0; $i<$numColumns; $i++) {
  push(@gapCol, 0);     # default: gap
  push(@nonActCol, 0);  # default: non-active col

  if ($numGaps[$i] <= (1-$gapPerc) * $numSeqs) { $gapCol[$i] = 1; }  # non-gap
  else                                         { $gapCol[$i] = 0; }  # gap column
}

#############################################################################################
# Merge information about gap and conserved columns for active/non-active columns:
#                      smoothing: conserved column criterion fulfilled for +- 2 columns
#############################################################################################
for (my $i=0; $i<$numColumns; $i++) {
  if ($gapCol[$i] == 1) { # non-gap column
    if ($conservedCol[$i] == 0) { # conserved Col
      my $smooth = 1;
      for (my $j=$i-$smoothCrit; $j<$i; $j++) {
        if ($j > -1 && $conservedCol[$j] == 0) { $smooth++; }
      }
      for (my $j=$i+1; $j<$i+$smoothCrit+1; $j++) {
        if ($j < $numColumns && $conservedCol[$j] == 0) { $smooth++; }
      }
      if ($smooth == 2*$smoothCrit + 1) { $nonActCol[$i] = 1; } # active
      else                              { $nonActCol[$i] = 0; } # non-active column
    }
    else { $nonActCol[$i] = 0;} # not conserved column -> non-active column
  }
  else {
    $nonActCol[$i] = 0;  # gap column -> non-active column
  }
}
for (my $i=0; $i<$numColumns; $i++) {
  if ($nonActCol[$i] == 0) {
    # non-active column
    if ($i == 0 || $i == $numColumns-1 || ($i >0 && $i < $numColumns-1 && $nonActCol[$i-1] == 1)) {
      # 1./3.:start of non active interval
      # 2.:   end ...
      push(@nonActInt, $i);
    }
  }
  else {
    if ($i>0 && $nonActCol[$i-1] == 0) {
      # end of non active interval at prec. pos
      push(@nonActInt, $i-1);
    }
  }
}


# Merge overlapping non-active intervals
my $i = 2;
while ($i<scalar(@nonActInt)) {
  if ($nonActInt[$i] < $nonActInt[$i-1]) { # begin_2 < end_1
    if ($nonActInt[$i+1] > $nonActInt[$i-1]) { $nonActInt[$i-1] = $nonActInt[$i+1]; }
    splice @nonActInt, $i, 2;
  }
  else {
    $i = $i+2;
  }
}
for (my $i=0; $i<scalar(@nonActInt)-1; $i++) {
  if ($i%2 == 0) { # non active interval start
    for (my $j=$nonActInt[$i]; $j<=$nonActInt[$i+1]; $j++) { $nonActCol[$j] = 0; }
  }
}
###########################################################################################


####################
# query sequences:
####################
open(QUERY1, "$query") or die "no query sequence file available\n";
my @query1     = <QUERY1>;
my @queryNames = ();
my @querySizes = ();
my @curQuery   = ();
my @queries    = ();

for (my $i=0; $i<scalar(@query1); $i++) {
  chomp($query1[$i]);
  if ($query1[$i] =~ m/>/){
    $query1[$i] =~ s/\>//;
    @curQuery   = split(" ", $query1[$i]);
    push @queryNames, $curQuery[0];
  }
  else {
    @curQuery = split("", $query1[$i]);
    push @querySizes, scalar(@curQuery);
  }
}
close(QUERY1);
############################################################################

##############################################################################################################
# BLAT:
# 1) run Blat
# 2) sortiere die paarweisen Alignments
#    es kann passieren, dass eine Anfragesequenz mehrmals zu einer MSA-Sequenz (verschiedene Art und Weise) aligniert wird
#    => nimm besten match
#       deshalb sortiere nach 1. Spalte (gebunden an Querynamen und MSA-Sequenznamen)
#       und suche dann besten Treffer
#
# sort:
# -n numeric  -r reverse result of comparisons  -k start a key at POS1 (end at POS2)
# -u unique -k
#############################################################################################

system("$blatSource $rawSeqs $query -noHead -minIdentity=$minIdentity -minScore=$minScore -oneOff=$oneOff $output");
my @blatOutput  = qx(cat $output | cut -f 1,10,11,12,13,14,19,20,21); # | sort -k 2,2 | sort -s -k 6,6);
# extracted output: score, q_name, q_len, q_1st_al_pos, q_last_al_pos, t_name, blockSize, q_starts, t_starts

##############################################################################################################
# save for each query sequence position all position in raw sequence it has been aligned to
#
# map these positions to the original columns in the alignment (using @msaPositions)
#
# calculate for each query sequence position the IQR of the columns in the alignment
#############################################################################################

my @scores          = ();
my @blockSizes      = ();
my @queryStarts     = ();
my @targetStarts    = ();
my $blockSize       = 0;
my $queryPos        = 0;
my $targetPos       = 0;
my $targetName      = "";
my $msaPos          = 0;
my $curLine         = 0;
my @alignedColPerST = ();
my %activeCols;
my %rawActiveCols;  # for test output
@curSeq = ();
my $queryName = "";
my $queryLen  = "";

my $curNumQuery = 0;
open(ACTIVE_COLUMNS, ">$alignmentColumns");

if (scalar(@blatOutput) == 0) {
  for (my $i=0; $i<scalar(@queryNames); $i++) {
    print ACTIVE_COLUMNS ">", $queryNames[$i], "\n";
    for (my $j=0; $j<$querySizes[$i]; $j++) {
      print ACTIVE_COLUMNS "0 ", $numColumns-1, " ";
    }
    print ACTIVE_COLUMNS "\n";
  }
}
else {
  while ($curLine < scalar(@blatOutput)) {
    if ($curLine < scalar(@blatOutput)) { # ???
      @curSeq          = split("\t", $blatOutput[$curLine]);
      $queryName       = $curSeq[1];
      $queryLen        = $curSeq[2];
      $msaPos          = 0;
      @scores          = ();
      @alignedColPerST = ();   # save alignments for each subtype separately

      for (my $i=0; $i<$numSubtypes; $i++) {
        my @M = ();
        for (my $j=0; $j<$queryLen; $j++) {
	  my @A = ();
	  push @M, \@A;
        }
        push @alignedColPerST, \@M;
      }

      while ($curLine < scalar(@blatOutput) && $curSeq[1] eq $queryName) {
        $targetName = $curSeq[5];

        # several matches -> compare position of match, length of match and score
        my @scores    = ();
        my @lengths   = (); # length of aligned part of query
        my @stretches = (); # distance between first and last aligned column in MSA
        my @starts    = ();

        my $succLine = $curLine;
        my $samePair = 1;

        while ($succLine < scalar(@blatOutput) && $samePair == 1) {
	    my @succSeq  = split("\t", $blatOutput[$succLine]); # init
	    if ($succSeq[1] eq $queryName && $succSeq[5] eq $targetName) {
		$samePair = 1;
		my @succBlockSizes   = split(",", $succSeq[6]);
		my $curLen = $succSeq[4]-$succSeq[3]+1; # last_aligned_pos - $first_aligned_pos + 1
		my @succQueryStarts  = split(",", $succSeq[7]);
		my @succTargetStarts = split(",", $succSeq[8]);
		
		push(@scores, $succSeq[0]);
		push(@lengths, $curLen);
		push(@starts, $succTargetStarts[0]);
		pop(@succTargetStarts); # last entry is "\n", (if chomp(), than empty="") empty due to , in blat output

		my $curStretch = 0 + $succTargetStarts[-1];
		$curStretch = $curStretch + $succBlockSizes[-1];
		$curStretch = $curStretch - $succTargetStarts[0];
		push(@stretches, $curStretch);
		$succLine++;
	    }
	    else {
		$samePair = 0;
	    }
        }

        # determine best match of all matches:
        my $maxLen   = max(@lengths);
        my $maxScore = max(@scores);
        my @bestMatches = ();

        for (my $i=0; $i<scalar(@scores); $i++) {
          if ($scores[$i] >= $maxScore*(1-$accScoreDistPerc) && $lengths[$i] >= $maxLen*$accAlignedQueryPerc && $stretches[$i] <= $maxLen*(1+$accStretchPerc)) {
            push(@bestMatches, $i);
          }
        }
        if (scalar(@bestMatches) > 0) {
          # for all best matches search minTargetStart
          my $bestMatch      = 0;
          my $maxTargetStart = 0; #max(@starts)+1;
          my $max            = 0;
          for (my $i=0; $i<scalar(@bestMatches); $i++) {
            if ($starts[$bestMatches[$i]] > $maxTargetStart) { # most right begin
              $bestMatch = $bestMatches[$i];
              $maxTargetStart = $starts[$bestMatches[$i]];
	    }
          }

          my $curTargetLine = $curLine+$bestMatch;
          @curSeq  = split("\t", $blatOutput[$curTargetLine]);

	  @blockSizes   = split(",", $curSeq[6]);
          @queryStarts  = split(",", $curSeq[7]);
	  @targetStarts = split(",", $curSeq[8]);

	  for (my $j=0; $j<scalar(@blockSizes); $j++) {
	    $blockSize = $blockSizes[$j];
	    $queryPos  = $queryStarts[$j];
	    $targetPos = $targetStarts[$j];
	
	    for (my $k=0; $k<$blockSize; $k++) {
	      # save for each query Pos, that is involved in this pairwise alignment,
	      # the corresponding position in the target sequence (current sequence of the MSA)
	
	      $queryPos  = $queryStarts[$j]  + $k;
	      $targetPos = $targetStarts[$j] + $k;
	
	      # blat output startet bei 0!!!
	      # determine target position in the original multiple alignment (including gaps)
	
	      $msaPos = $msaPositions{$targetName}->[$targetPos];
	      push @{$alignedColPerST[$assignedST{$targetName}][$queryPos]}, $msaPos;
            }
	  }
	}
        $curLine = $succLine; # for next query
        if ($curLine < scalar(@blatOutput)) {
	  @curSeq = split("\t", $blatOutput[$curLine]);
        }
      }

      # for each query sequence position determine active columns in the multiple alignment:
      # for each position a list of target alignment columns is saved in a list
      # determine interquartile range of these columns
      # => part within are the active columns for the considered query sequence position
      #
      # -> for each query sequence position save two columns (begin and end)!


      ##########################
      # active columns in MSA:
      ##########################

      my @activeCols  = ();
      my @numAlSeqs   = (); # number of aligned sequences in a subtype
      my $numAlSeqs   = 0;  # number of aligned sequences in MSA

      for (my $j=0; $j < $queryLen; $j++) {
        my @curActiveCol = ();
        my $isAligned    = 1;
        # number of aligned sequences (=columns) in each subtype
        my @alignedCol = ();

        for (my $l=0; $l<$numSubtypes; $l++) {
	  push @alignedCol, @{$alignedColPerST[$l][$j]};
	  push @numAlSeqs, scalar(@{$alignedColPerST[$l][$j]});
        }
        $numAlSeqs = scalar(@alignedCol); # for each aligned seq one column

        # Definition of an aligned query sequence position:
        #   a seq pos is 'aligned' to the MSA, if it is
        #   to at least $gapPerc % of the or $gapTresh sequences in at least one subtype

        # if definition of 'aligned' is fulfilled, the active columns for this pos are determined (IQR)
        # else: defined as 'unaligned'
        if ($numAlSeqs > $alSeqsPerc * $numSeqsToAlign) { # only $numChosenSeqs sequences chosen for Blat alignment
	  @curActiveCol = @{ &active_columns(@alignedCol)};
	  $isAligned    = 0;
        }

        if ($isAligned == 1) { @curActiveCol = (-1,-1); }

        push @activeCols, $curActiveCol[0];
        push @activeCols, $curActiveCol[1];

        for (my $k=0; $k<$numSubtypes; $k++) { @{$alignedColPerST[$k][$j]} = (); }
      }

      #####################################################
      # Assignment of active columns for 'unaligned pos.:
      #   for 'unaligned' seq positions:
      #    assign active columns of prec. and succ. pos.
      #####################################################

      # determine min. and max. assigned MSA column
      my $minCol = $numColumns;
      my $maxCol = 0;
      my $minPos = -1;
      my $maxPos = -1;

      for (my $j=0; $j < scalar(@activeCols); $j++) {
        if ($j%2 == 0) {
          if ($activeCols[$j] != -1 && $activeCols[$j] < $minCol) {
            $minCol = $activeCols[$j];
            $minPos = $j/2;
          }
          if ($activeCols[$j+1] != -1 && $activeCols[$j+1] > $maxCol) {
            $maxCol = $activeCols[$j+1];
            $maxPos = $j/2;
          }
        }
      }

      # for all pos. in front of $minPos:
      # extend active interval by $minPos active columns;
      # same for $maxPos
      my $pos        = $minCol-1;
      my $numActCols = 0;

      while ($pos > -1 && $numActCols < $minPos+1) { # begin at 0
        if ($nonActCol[$pos] == 1) { $numActCols++; } # active Col (consensus and non-gap)
        $pos--;
      }
      $minCol     = $pos+1;
      $pos        = $maxCol+1;
      $numActCols = 0;

      while ($pos < $numColumns && $numActCols < ($queryLen-$maxPos-1)) { # begin at 0
        if ($nonActCol[$pos] == 1) { $numActCols++; } # active Col (consensus and non-gap)
        $pos++;
      }
      $maxCol = $pos-1;

      # begin of interval:
      if ($maxCol < 0 || $maxCol >= $numColumns) { $maxCol = $numColumns-1; }
      if ($minCol < 0 || $minCol >= $numColumns) { $minCol = 0;  }
      for (my $j=0; $j < scalar(@activeCols); $j++) {
        if ($activeCols[$j] == -1) {
	  if ($j%2 == 0) {
	    if ($j-2 >= 0) { $activeCols[$j] = $activeCols[$j-2]; }
	    else           { $activeCols[$j] = $minCol; }
	  }
        }
      }

      # end of interval
      my $dummy = 0;
      for (my $j=scalar(@activeCols)-1; $j>=0; $j--) {
        if ($j%2 == 1) {
	  if ($activeCols[$j] == -1) {
	    if ($j+2 < scalar(@activeCols)) { $activeCols[$j] = $activeCols[$j+2]; }
	    else                            { $activeCols[$j] = $maxCol; }
	  }
	  if ($activeCols[$j-1] > $activeCols[$j]) { #if begin > end
	    $dummy            = $activeCols[$j-1];
	    $activeCols[$j-1] = $activeCols[$j];
	    $activeCols[$j]   = $dummy;
	  }
        }
      }

      #################################################
      # Extension of active intervals:
      #  1) each interval must at least contain 1 'active' columns
      #  2) artificial extension: +/- $intervalExt 'active' columns
      #  3) if gap between two succ. active intervals => merge them
      #################################################

      my $j=0;
      while ($j < scalar(@activeCols)) {
        my $actCol        = 0;
        my $numCurActCols = 0;
        for (my $a=$activeCols[$j]; $a <= $activeCols[$j+1]; $a++) {
           if ($nonActCol[$a] == 1) { # active col
              $numCurActCols++;
           }
        }
        # each interval must contain at least 1 'active' column
        # -> extend active intervals, if no active column included
        while ($numCurActCols == 0 && ($activeCols[$j] > 0 || $activeCols[$j+1] < $numColumns-1)) {
          if ($activeCols[$j] > 0) {
            $activeCols[$j]--;
            # if now active column:
            if ($nonActCol[$activeCols[$j]] == 1) {
              $numCurActCols++;
            }
          }
          if ($activeCols[$j+1] < $numColumns-1) {
            $activeCols[$j+1]++;
            # if now active column:
            if ($nonActCol[$activeCols[$j+1]] == 1) {
              $numCurActCols++;
            }
          }
        }
        # extend active intervals by += $intervalExt active columns:
        $numCurActCols = 0;
        # extend left interval end by $intervalExt active columns
        while ($numCurActCols < $intervalExt && $activeCols[$j] > 0) {
          $activeCols[$j]--;
          # if cur start of active interval is active column:
          if ($nonActCol[$activeCols[$j]] == 1) {
            $numCurActCols++;
          }
        }
        $numCurActCols = 0;
        # extend right interval end by $intervalExt active columns
        while ($numCurActCols < $intervalExt && $activeCols[$j+1] < $numColumns-1) {
          $activeCols[$j+1]++;
          # if cur end of active interval is active column:
          if ($nonActCol[$activeCols[$j+1]] == 1) {
            $numCurActCols++;
          }
        }
        $j += 2;
      }
      # compare successive intervals and check for gaps in between
      $j=2;
      while ($j < scalar(@activeCols)-1) {
        # 1) extend successive intervals if gap between them, such that the two intervals are merged
        if ($activeCols[$j] > $activeCols[$j-1]+1) {
          if ($activeCols[$j] - $activeCols[$j-1] -1 > 50) {
            print $queryName, " contains a deletion of > 50bp\n";
          }
	  $activeCols[$j]   = $activeCols[$j-2];     # cur begin = pred begin
	  $activeCols[$j-1] = $activeCols[$j+1];     # pred end  = cur end
        }
        # 2) if end(i) < end(i-1) : end(i) = end(i-1) -> succ. int. end >= prec. int. end
        if ($activeCols[$j+1] < $activeCols[$j-1]) { 
          $activeCols[$j+1] = $activeCols[$j-1]; 
        }
        $j += 2;
      }

      $j = scalar(@activeCols)-2;
      while ($j>=2) {
        # 3) if start(i) < start(i-1) : start(i-1) = start(i) -> succ. int. start >= prec. int. start
        if ($activeCols[$j] < $activeCols[$j-2])   { 
          $activeCols[$j-2] = $activeCols[$j]; 
        }
        $j -= 2;
      }
      $activeCols{$queryName}    = \@activeCols;

      $| = 1;
      if ($queryName ne $queryNames[$curNumQuery]) {
	while ($queryNames[$curNumQuery] ne $queryName) {
	  print "\n", $queryNames[$curNumQuery], "no blat alignment possible\n";
	  print ACTIVE_COLUMNS ">", $queryNames[$curNumQuery], "\n";
	  for (my $j=0; $j<$querySizes[$curNumQuery]; $j++) {
	    print ACTIVE_COLUMNS "0 ", $numColumns-1, " ";
	  }
	  print ACTIVE_COLUMNS "\n";
	  $curNumQuery++;
	}
      }
      if ($queryName ne $queryNames[$curNumQuery]) {
        die "something's wrong\n";
      }
      print ACTIVE_COLUMNS ">", $queryName, "\n";
      print ACTIVE_COLUMNS join(" ", @{$activeCols{$queryName}}), "\n";
      my @curActCols = @{$activeCols{$queryName}};
      $curNumQuery++;
    }
  }
}
close(ACTIVE_COLUMNS);


#########################################################################
# Calculate Median of Positions;

sub active_columns {
  my @targetPos     = sort {$a <=> $b} @_;
  my $n             = scalar(@targetPos);
  my @iqr           = ();
  my $firstQuartile = 0;
  my $thirdQuartile = 0;

  if ($isCircGen == 1) {
      # for a circular genome, this sequence position is aligned to different parts in alignment
      # decide for the part with the most frequent pairwise alignments 
      # problem: position independent...
      my $i=0;
      while ($i<scalar(@targetPos)-1 && ($targetPos[$i+1]-$targetPos[$i] < $origMSALen*$accColDistPerc)) {
	  $i++;
      }
      $i++;
      my @tmpTargetPos = ();
      if ($i < scalar(@targetPos)/2) {
          for (my $j = $i; $j<scalar(@targetPos); $j++) {
	      push(@tmpTargetPos, $targetPos[$j]);
	  }
      }
      else {
	  for (my $j=0; $j<$i; $j++) {
	      push(@tmpTargetPos, $targetPos[$j]);
	  }
      }
      $n = scalar(@tmpTargetPos);

      # 1st quartile
      if ($n%4 == 0) { $firstQuartile = ($tmpTargetPos[$n/4-1] + $tmpTargetPos[$n/4])/2; }
      else           { $firstQuartile = $tmpTargetPos[&ceil($n/4)-1]; }
      
      # 3rd quartile
      if (($n*3)%4 == 0) { $thirdQuartile = ($tmpTargetPos[$n*3/4-1] + $tmpTargetPos[$n*3/4])/2; }
      else               { $thirdQuartile = $tmpTargetPos[&ceil(($n*3)/4)-1]; }
  }
  else {
    # 1st quartile
    if ($n%4 == 0) { $firstQuartile = ($targetPos[$n/4-1] + $targetPos[$n/4])/2; }
    else           { $firstQuartile = $targetPos[&ceil($n/4)-1]; }

    # 3rd quartile
    if (($n*3)%4 == 0) { $thirdQuartile = ($targetPos[$n*3/4-1] + $targetPos[$n*3/4])/2; }
    else               { $thirdQuartile = $targetPos[&ceil(($n*3)/4)-1]; }
  }

  push @iqr, &floor($firstQuartile);
  push @iqr, &ceil($thirdQuartile);
  return \@iqr;
}
