#!/usr/bin/perl

use warnings;
use strict;
use POSIX;
use POSIX qw(ceil floor);
use List::Util qw[min max];


# default input files
my $database    = "HIV-alignment_pureSubtypes.fas";
my $queries     = "query_sequences.fas";


# parameters:
# extend MSA: 2xMSALen + $extMSA*MSALen:
my $extMSA = 0.25;
# extend queries by # nucleotides
my $extQ = 500;

# a query is extended if it is a full-length genome

my $dist   = 30;
my $minLen = 3000; #3182 - $dist; # subgenotype D (HBV length: 3182nt(HBV/D) - 3248nt (HBV/G))

if (scalar(@ARGV) > 0) {
  $database   = $ARGV[0];
  if (scalar(@ARGV) > 1) {
    $queries = $ARGV[1];
    if (scalar(@ARGV) > 2) {
      $extMSA = $ARGV[2];
      if ($extMSA > 1) {
	  die "MSA extension must be given as a factor, e.g. 0.25\n";
      }
      if (scalar(@ARGV) > 3) {
        $extQ = $ARGV[3];
        if ($extQ < 1) {
	    die "query extension must be given in nucleotides, e.g. 500\n";
	}
      }
    }
  }
}

# output file
my $outMSA = $database.".ext";
my $outQ   = $queries.".ext";

##############################################################################################################
# Read in database source (multiple sequence alignment)

open(DATABASE, "$database") or die "no database file available\n";
open(OUT, ">$outMSA");

my @seqs         = <DATABASE>;
my $curSeqName = "";
my $numSeqs = 0;
my $numSeqsPerST = 0;
my $seq = "";

my $origMSALen = 0;
# read in MSA
for (my $i=0; $i<scalar(@seqs); $i++) {
  chomp($seqs[$i]);
  $seqs[$i] =~ s/\r+//g;
  $seqs[$i] =~ s/\n+//g;


  if ($seqs[$i] =~ m/^>>/) { #subtype
    if ($numSeqsPerST > 0) {
      print OUT ">", $curSeqName, "\n";
      my $len = length($seq);
      my $ext = &ceil($extMSA*$len);
      my $subSeq = substr($seq, 0, $ext);
      my $curSeq = $seq.$seq.$subSeq;
      print OUT $curSeq, "\n";
      $origMSALen = length($seq);
    }
    print OUT $seqs[$i], "\n";
    $numSeqsPerST = 0;
  }
  else {
    if ($seqs[$i] =~ m/>/) { # sequence name
      if ($numSeqsPerST > 0) {
        print OUT ">", $curSeqName, "\n";
        my $len = length($seq);
        my $ext = &ceil($extMSA*$len);
        my $subSeq = substr($seq, 0, $ext);
        my $curSeq = $seq.$seq.$subSeq;
        print OUT $curSeq, "\n";
      }
      $curSeqName = $seqs[$i];
      $curSeqName =~ s/\>//;
      $numSeqs++;
      $numSeqsPerST++;
      $seq = "";
    }
    else {
      $seq .= $seqs[$i];
    }
  }
}
print OUT ">", $curSeqName, "\n";
my $len = length($seq);
my $ext = &ceil($extMSA*$len);
my $subSeq = substr($seq, 0, $ext);
my $curSeq = $seq.$seq.$subSeq;
print OUT $curSeq, "\n";

$seq =~  s/-//g;

close(OUT);
close(DATABASE);
##############################################################################################################
# Read in query sequences

open(QUERY, "$queries") or die "no query sequence file available\n";
open(OUT, ">$outQ");

@seqs       = <QUERY>;
$curSeqName = "";
$numSeqs    = 0;
$seq        = "";

my $orig_start = 1;
my $orig_end   = 1;

for (my $i=0; $i<scalar(@seqs); $i++) {
  chomp($seqs[$i]);
  $seqs[$i] =~ s/\r+//g;
  $seqs[$i] =~ s/\n+//g;

  if ($seqs[$i] =~ m/>/) { # sequence name
    if ($numSeqs > 0) {
      $len = length($seq);

      # if full-length genome
      if ($len >= $minLen) {
        my $curExtQ  = min($extQ, &floor($len/2)); # rounded down
        my $seqBegin = substr($seq, 0, $curExtQ);
        my $seqEnd   = substr($seq, $len-$curExtQ, $curExtQ);

        $curSeq     = $seqEnd.$seq.$seqBegin;
        $orig_start = $curExtQ+1;
        $orig_end   = length($curSeq)-$curExtQ;
      }
      else {
        $curSeq     = $seq;
        $orig_start = 1;
        $orig_end   = length($curSeq);
      }
      print OUT ">", $curSeqName, "\t", $orig_start, "\t", $orig_end, "\n";
      print OUT $curSeq, "\n";
    }
    my @curSeqName = split("\t", $seqs[$i]);
    $curSeqName = $curSeqName[0];
    $curSeqName =~ s/\>//;
    $numSeqs++;
    $seq = "";
  }
  else {
    # erase gaps: "-"
    $seqs[$i] =~ s/-//g;
    $seq .= $seqs[$i];
  }
}
$len = length($seq);

if ($len >= $minLen) {
  my $curExtQ  = min($extQ, &floor($len/2));
  my $seqBegin = substr($seq, 0, $curExtQ);
  my $seqEnd   = substr($seq, $len-$curExtQ, $curExtQ);

  $curSeq     = $seqEnd.$seq.$seqBegin;
  $orig_start = $curExtQ+1;
  $orig_end   = length($curSeq)-$curExtQ;
}
else {
  $curSeq     = $seq;
  $orig_start = 1;
  $orig_end   = length($curSeq);
}
print OUT ">", $curSeqName, "\t", $orig_start, "\t", $orig_end, "\n";
print OUT $curSeq, "\n";

close(OUT);
close(QUERY);

# print out length of original input alignment (not extended):
print $origMSALen, "\n";
