// ~/jpHMM/src/hmm.cpp
/**************************************************************
 File name	: hmm.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 - Construction of HMM by combination of profile HMMs for each subtype
 - Definition of special states - at beginning and end of HMM
                              - for jumps between different profile HMMs (if this option is chosen)
              and respective emission and transition probabilities
 - Definition of local HMM for certain region in the alignment/HMM
 - Definition of first and last common consensus column in all subtypes
 - Calculation of jump probabilities between states of different subtypes

 Class ActiveStates to define states that are 'active' in Viterbi recursion
***************************************************************/

#include "std_source.h"
#include "error.h"
#include "const_def.h"
#include "alignment.h"
#include "emissionPriors.h"
#include "transitionPriors.h"
#include "emissionProfile.h"
#include "emTransProfile.h"
#include "profileHMM.h"
#include "hmm.h"

using namespace std;

const bool OUTPUT_DEBUG = false;


/**************************************************
 *
 *                class HMM.
 *
****************************************************/

/**************************************************************************************//**
   HMM.

   Calls ProfileHMM constructor to create profile HMMs of different subtypes.
   These profile HMMs are combined to a jumping profile HMM.\n
   Special states (Begin, N_Insert, N_Delete, C_Insert, C_Delete, End) are added.\n
   The emission probabilities of the N_Insert and C_Insert state are determined.\n
   Different functions are called to add transitions from and to these special states
   and jumps between the profile HMMs, and to calculate the corresponding transition probabilities.

   @param[in] al Alignment.
   @param[in] e Emission priors.
   @param[in] t Transition priors.
   @param[in] etpr Emission and transition profile of alignment.
   @param[in] use_eetp Use external emission and transition probabilities?.
   @param[in] eetp External emission and transition probabilities.

  //! \todo {Implement jpHMM version, that contains profileHMM for whole alignment.}
******************************************************************************************/
HMM::HMM(Alignment* al,
	 const EmissionPriors&   e,
	 const TransitionPriors& t,
	 const EmTransProfile&   etpr,
	 bool use_eetp, 
	 const ExtEmTrProbs* eetp) {

  if (OUTPUT_DEBUG)
	  cout << "HMM: " << endl;
  m_len = al->subs()[0]->len();
  m_is_local_hmm = false;
  
  //! Special hmm states.
  /*!  - create special hmm states.
   *     -# begin:    BEGIN state of the HMM (5'-end).
   *     -# n_delete: N-terminal DELETE state (5'-end).
   *     -# n_insert: N-terminal unaligned sequence state (INSERT) (5'-end).
   *     -# c_insert: C-terminal unaligned sequence state (INSERT) (3'-end).
   *     -# c_delete: C-terminal DELETE state (3'-end).
   *     -# end:      END state of the HMM (3'-end).
   *  - add transitions between special states.
   *     -# begin    -> n_delete.
   *     -# begin    -> n_insert.
   *     -# n_insert -> n_insert.
   *     -# c_insert -> end.
   *     -# c_delete -> end.
   *     -# c_insert -> c_insert.
  */
  begin    = new State(0, 0, Begin);
  n_delete = new State(0, 0, N_Delete);
  n_insert = new State(0, 0, N_Insert);
  c_insert = new State(0, 0, C_Insert);
  c_delete = new State(0, 0, C_Delete);
  end      = new State(0, 0, End);

  m_spec_state_to_idx = list_of< pair<State*, int> >(begin, 0)(n_delete, 1)(n_insert, 2)(c_insert, 3)(c_delete, 4)(end, 5);
  m_spec_state_to_str = list_of< pair<State*, string> >(begin, "Begin")(n_delete, "5'-Deletion")(n_insert, "5'-Insertion")(c_insert, "3'-Insertion")(c_delete, "3'-Deletion")(end, "End");
  m_str_to_spec_state = list_of< pair<string, State*> >("Begin", begin)("5'-Deletion", n_delete)("5'-Insertion", n_insert)("3'-Insertion", c_insert)("3'-Deletion", c_delete)("End", end);

  const mytype GAPINIT = Const_Def::gap_init();
  const mytype EPSILON = Const_Def::epsilon();

  Transition* cur_transition = new Transition(begin, n_delete);
  cur_transition->log_p() = log(GAPINIT);
  begin->succ()    +=cur_transition;
  n_delete->pred() +=cur_transition;

  cur_transition = new Transition(begin, n_insert);
  cur_transition->log_p() = log((mytype)1-EPSILON-GAPINIT);
  begin->succ()    +=cur_transition;
  n_insert->pred() +=cur_transition;

  cur_transition = new Transition(n_insert, n_insert);
  cur_transition->log_p() = log((mytype)1-EPSILON);
  n_insert->succ() +=cur_transition;
  n_insert->pred() +=cur_transition;

  cur_transition = new Transition(c_insert, end);
  cur_transition->log_p() = log(EPSILON);
  c_insert->succ() +=cur_transition;
  end->pred()      +=cur_transition;

  cur_transition = new Transition(c_delete, end);
  cur_transition->log_p() = log(1);
  c_delete->succ() +=cur_transition;
  end->pred()      +=cur_transition;

  cur_transition = new Transition(c_insert, c_insert);
  cur_transition->log_p() = log((mytype)1-EPSILON);
  c_insert->succ() +=cur_transition;
  c_insert->pred() +=cur_transition;

  if (OUTPUT_DEBUG)
	  cout << "Profile HMMs: " << endl;
  //! Profile HMM for each subtype.
  TransAlphabet  *trans_alphabet  = Const_Def::trans_alphabet();
  mytype          prob_spec_chars = etpr.rel_freq_spec_chars();

  FOREACH (SubtypeProfile* sub_prof, etpr.sub_profs()) {
    ProfileHMM* phmm = new ProfileHMM(sub_prof, t, e, *trans_alphabet, prob_spec_chars, use_eetp, eetp);
    m_phmms         += phmm;
    if (OUTPUT_DEBUG)
    	cout << m_phmms.size() << endl;
  }
  m_num_phmms = (int)m_phmms.size();
  //! Subtype index of each profile HMM.
  for (int s=0; s<m_num_phmms; ++s)
     m_phmm_to_sub_idx[m_phmms[s]] = s;


  //! First and last consensus column in hmm.
  /*!  first common and last common consensus column in all subtypes or any specified column. */
  m_first_cons_col = first_cons_col(0, m_len-1);
  m_last_cons_col  = last_cons_col(m_first_cons_col, m_len-1);	

  if (m_first_cons_col < 0 || m_last_cons_col >= m_len || m_first_cons_col > m_last_cons_col) {
    throw Error("Something is wrong with the consensus columns");
  }
  else {
    if (OUTPUT_DEBUG)
      cout << "m_first_cons_col: " << m_first_cons_col << endl << "m_last_cons_col: " << m_last_cons_col << endl;

    if (Const_Def::jump_version() == linear) {
      /*!** special mute states for jumps between profile HMMs: ***/
      //! mute jump state only in those columns where at least one subtype has a consensus column, i.e. a match state
      //! and not in last consensus column (-> no jump possible)
      m_jump_states.assign (m_len, make_pair((State*) 0, (State*) 0));
      for (int j=m_first_cons_col; j<m_last_cons_col; ++j) {
	int num_cons_cols = 0;
	FOREACH (ProfileHMM* phmm, m_phmms) {
	  if (phmm->has_state(j, (int)Match)) {
	    ++num_cons_cols;
	    break;
	  }
	}
	if (num_cons_cols > 0) {
          State* cur_jump_state_match = new State(0, j, Jump_State_Match);
          State* cur_jump_state_delete = new State(0, j, Jump_State_Delete);

          m_jump_states[j] = make_pair(cur_jump_state_match, cur_jump_state_delete);
	}
      }
      m_type_to_jump_type = list_of< pair<StateType, JumpStateType> >(Jump_State_Match, Jump_Match)(Jump_State_Delete, Jump_Delete);
      m_jump_state_to_str = list_of< pair<JumpStateType, string> >(Jump_Match, "Jump_Match")(Jump_Delete, "Jump_Delete");
    }
    
    //! Initialization for emission probabilities of N- and C-Insert.
    vector<mytype> insert_em_priors  = e.insert_priors();
    vector<int>    n_insert_frequencies(insert_em_priors.size(), 0);
    vector<int>    c_insert_frequencies(insert_em_priors.size(), 0);

    //! A jumping profile HMM.
    /*! Combine all profile HMMs and special states to form a jpHMM by
        adding jumps between states of different profile HMMs and from and to special states.
    */
    int num_jumps = 0;

    FOREACH (ProfileHMM* phmm, m_phmms) {
      if (OUTPUT_DEBUG)    	
        cout << "Phmm: " << m_phmm_to_sub_idx[phmm] << endl;
      for (int i=m_first_cons_col; i<m_last_cons_col; i++) { // for last_col Insert not active
	if (phmm->has_state(i, (int)Match)) {
          for (int type=(int)Match; type<=(int)Delete; type++) {
            State* cur_state = phmm->state(i, type);
	    if (Const_Def::jump_version() == linear) {
	      //! Calculation of jumps/transitions to (mute) jumps states.
	      calc_jumps_to_jump_states(cur_state);
	    }
            else {
	      //! Calculation of jumps and jump probabilities to other subtypes.
              calc_jumps(cur_state, &num_jumps);
              calc_jump_probs(cur_state, num_jumps);
	    }

	    //! Calculation of jumps and jump probabilities to special states.
	    calc_spec_jumps(cur_state);
            calc_spec_jump_probs(cur_state);
          }
        }
      }
      if (OUTPUT_DEBUG)
        cout << "last col: " << endl;
      int cur_types[2] = {(int)Match, (int)Delete};
      if (phmm->has_state(m_last_cons_col, (int)Match)) {
        for (int i=0; i<2; ++i) {
          State* cur_state = phmm->state(m_last_cons_col, cur_types[i]);
          //! only jumps to special states possible.
	  calc_spec_jumps(cur_state);
	  calc_spec_jump_probs(cur_state);
	}
      }

      //! N_INSERT emission frequencies.
      /*!
       *  Nucleotides in all columns in front of first hmm consensus column taken into account.
       *  (Some subtypes can have consensus column in front of first hmm consensus column.
       *  So, add up also MATCH frequencies for some subtypes.
       *  Nucleotides in front of first consensus column in a certain subtype are saves in nterminal_cons_col_prof()).
      */
      if (OUTPUT_DEBUG)
        cout << "N_I em counts: " << endl;
      SubtypeProfile* cur_sub_prof = phmm->sub_prof();
      ConsColumnProfile* cur_cons_col_prof = cur_sub_prof->nterminal_cons_col_prof();
      EmissionProfile*   cur_em_prof       = cur_cons_col_prof->em_prof();
      int counted_cols = 1;
      for (unsigned i=0; i<insert_em_priors.size(); ++i)
        n_insert_frequencies[i] += cur_em_prof->insert_frequency(i);

      map<int, ConsColumnProfile*>::const_iterator it = cur_sub_prof->cons_col_profs().begin();
      while (it != cur_sub_prof->cons_col_profs().end() && (it->first) < m_first_cons_col) {
	counted_cols++;
        cur_cons_col_prof = cur_sub_prof->cons_col_prof(it->first);
        cur_em_prof       = cur_cons_col_prof->em_prof();
        for (unsigned i=0; i<insert_em_priors.size(); ++i) {
          n_insert_frequencies[i] += cur_em_prof->match_frequency(i) + cur_em_prof->insert_frequency(i);
        }
        ++it;
      }

      //! C_INSERT emission frequencies.
      /*!  Nucleotides in all columns after last hmm consensus column taken into account.
       *  (some subtypes ... (see N_INSERT emission frequencies)).
       */
      if (OUTPUT_DEBUG)
        cout << "C_I em counts: " << endl;
      map<int, ConsColumnProfile*>::reverse_iterator rit = cur_sub_prof->cons_col_profs().rbegin();
      while (rit != cur_sub_prof->cons_col_profs().rend() && (rit->first) > m_last_cons_col){
        cur_cons_col_prof = cur_sub_prof->cons_col_prof(rit->first);
        cur_em_prof       = cur_cons_col_prof->em_prof();
        for (unsigned i=0; i<insert_em_priors.size(); ++i) {
          c_insert_frequencies[i] += cur_em_prof->match_frequency(i) + cur_em_prof->insert_frequency(i);
        }
        ++rit;
      }
      cur_cons_col_prof = cur_sub_prof->cons_col_prof(m_last_cons_col);
      cur_em_prof       = cur_cons_col_prof->em_prof();
      for (unsigned i=0; i<insert_em_priors.size(); ++i) {
        c_insert_frequencies[i] += cur_em_prof->insert_frequency(i);
      }
      if (OUTPUT_DEBUG)
        cout << "C_I em counts end: " << endl;
    }
    if (OUTPUT_DEBUG)
      cout << "N_I em probs: " << endl;
    //! N_INSERT and C_INSERT emission probabilities.
    /*!   Calculate emission probabilities by adding pseudocounts (insert_em_priors) to observed emission frequencies. */
    ProfileHMM* p;
    vector<mytype> cur_ep = p->emission_probs(insert_em_priors, n_insert_frequencies);
    for (vector<mytype>::iterator it=cur_ep.begin(); it < cur_ep.end(); ++it)
      (*it) = ((mytype)1 - prob_spec_chars)*(*it);
    cur_ep            += prob_spec_chars;
    n_insert->log_ep() = p->log_of_prob(cur_ep);

    if (OUTPUT_DEBUG)
      cout << "C_I em probs: " << endl;
    cur_ep = p->emission_probs(insert_em_priors, c_insert_frequencies);
    for (vector<mytype>::iterator it=cur_ep.begin(); it < cur_ep.end(); ++it)
      (*it) = ((mytype)1 - prob_spec_chars)*(*it);
    cur_ep            += prob_spec_chars;
    c_insert->log_ep() = p->log_of_prob(cur_ep);


    //! Transition probabilities for special states.
    calc_spec_jump_probs(n_insert);
    calc_spec_jump_probs(n_delete);
    calc_spec_jump_probs(begin);

    if (OUTPUT_DEBUG)
      cout << "Standardization: " << endl;
    //! Standardization of all transition probabilities.
    //! and addition of id to each state

    begin->id() = 0;
    m_idx_to_state[0] = begin;
    n_delete->id() = 1;
    m_idx_to_state[1] = n_delete;
    n_insert->id() = 2;
    m_idx_to_state[2] = n_insert;
    int cur_id = 3;

    if (Const_Def::jump_version() == linear) {
      //! Jumps/transitions from special mute states to states in all subtypes.
      //! standardization and assignment of state id.
      for (int j=m_first_cons_col; j<m_last_cons_col; ++j) {
        if (m_jump_states[j].first) {
          calc_jumps_from_jump_states(j);

          State* cur_state = m_jump_states[j].first;
          if (!cur_state->succ().empty())
            cur_state->succ_standardization(m_first_cons_col, m_last_cons_col);
          cur_state->id() = cur_id;
          m_idx_to_state[cur_id] = cur_state;
          ++cur_id;

          cur_state = m_jump_states[j].second;
          if (!cur_state->succ().empty())
            cur_state->succ_standardization(m_first_cons_col, m_last_cons_col);
          cur_state->id() = cur_id;
          m_idx_to_state[cur_id] = cur_state;
          ++cur_id;
	}
      }
    }
    
    //! standardization and assignment of state id for all states in all subtypes.
    FOREACH (ProfileHMM* phmm, m_phmms) {
      for (int i=m_first_cons_col; i<m_last_cons_col; ++i) {
        if (phmm->has_state(i, (int)Match)) {
	  for (int type=(int)Match; type<=(int)Delete; ++type) {
            State* cur_state = phmm->state(i, type);
            cur_state->succ_standardization(m_first_cons_col, m_last_cons_col);
	    cur_state->id() = cur_id;
	    m_idx_to_state[cur_id] = cur_state;
	    ++cur_id;
          }
        }
      }
      //! Insert in last cons col doesn't exist:
      for (int type=(int)Match; type<=(int)Delete; ++type) {
        if (phmm->has_state(m_last_cons_col, type)) {
          State* cur_state = phmm->state(m_last_cons_col, type);
          cur_state->succ_standardization(m_first_cons_col, m_last_cons_col);
          cur_state->id() = cur_id;
	  m_idx_to_state[cur_id] = cur_state;
	  ++cur_id;
        }
      }
    }
    
    begin->succ_standardization(m_first_cons_col, m_last_cons_col);
    n_delete->succ_standardization(m_first_cons_col, m_last_cons_col);
    n_insert->succ_standardization(m_first_cons_col, m_last_cons_col);
    c_insert->succ_standardization(m_first_cons_col, m_last_cons_col);
    c_delete->succ_standardization(m_first_cons_col, m_last_cons_col);
    c_insert->id()           = cur_id;
    m_idx_to_state[cur_id]   = c_insert;
    c_delete->id()           = cur_id+1;
    m_idx_to_state[cur_id+1] = c_delete;
    end->id()                = cur_id+2;
    m_idx_to_state[cur_id+2] = end;

    m_num_states = end->id()+1;
    
    if (OUTPUT_DEBUG)
      hmm_out();
  }
}

/****************************************************************************************//**
   HMM copy constructor.

   Creates a copy of the hmm for a local hmm.
   @param[in] hmm Original (global) HMM for whole alignment.
   @param[in] first_cons_col First consensus column in (local) alignment = start of HMM.
   @param[in] last_cons_col Last consensus column in (local) alignment = end of HMM.
********************************************************************************************/
HMM::HMM(const HMM& hmm,
	 int        first_cons_col,
	 int        last_cons_col) {

  m_len               = hmm.m_len;
  m_num_phmms         = hmm.m_num_phmms;
  m_first_cons_col    = hmm.m_first_cons_col;
  m_last_cons_col     = hmm.m_last_cons_col;
  m_first_cons_col    = first_cons_col;
  m_last_cons_col     = last_cons_col;
  m_idx_to_state      = hmm.m_idx_to_state;
  m_phmm_to_sub_idx   = hmm.m_phmm_to_sub_idx;
  m_spec_state_to_idx = hmm.m_spec_state_to_idx;
  m_spec_state_to_str = hmm.m_spec_state_to_str;
  m_str_to_spec_state = hmm.m_str_to_spec_state;

  //! Copy special hmm states.
  /*!
   *   - special hmm states.
   *   - transitions between special states.
   *     probabilities must be initialize here again, since normalized for original hmm.
  */

  map<State*, State*> state_to_state;

  begin    = new State(*hmm.begin);
  state_to_state[begin] = hmm.begin;
  state_to_state[hmm.begin] = begin;
  
  n_delete = new State(*hmm.n_delete);
  state_to_state[n_delete] = hmm.n_delete;
  state_to_state[hmm.n_delete] = n_delete;

  n_insert = new State(*hmm.n_insert);
  state_to_state[n_insert] = hmm.n_insert;
  state_to_state[hmm.n_insert] = n_insert;

  c_insert = new State(*hmm.c_insert);
  state_to_state[c_insert] = hmm.c_insert;
  state_to_state[hmm.c_insert] = c_insert;

  c_delete = new State(*hmm.c_delete);
  state_to_state[c_delete] = hmm.c_delete;
  state_to_state[hmm.c_delete] = c_delete;

  end      = new State(*hmm.end);
  state_to_state[end] = hmm.end;
  state_to_state[hmm.end] = end;

  //! Copy necessary indices.
  m_spec_state_to_idx = list_of< pair<State*, int> >(begin, 0)(n_delete, 1)(n_insert, 2)(c_insert, 3)(c_delete, 4)(end, 5);
  m_spec_state_to_str = list_of< pair<State*, string> >(begin, "Begin")(n_delete, "5'-Deletion")(n_insert, "5'-Insertion")(c_insert, "3'-Insertion")(c_delete, "3'-Deletion")(end, "End");
  m_str_to_spec_state = list_of< pair<string, State*> >("Begin", begin)("5'-Deletion", n_delete)("5'-Insertion", n_insert)("3'-Insertion", c_insert)("3'-Deletion", c_delete)("End", end);

  const mytype GAPINIT = Const_Def::gap_init();
  const mytype EPSILON = Const_Def::epsilon();

  Transition* cur_transition = new Transition(begin, n_delete);
  cur_transition->log_p() = log(GAPINIT);
  begin->succ()    +=cur_transition;
  n_delete->pred() +=cur_transition;

  cur_transition = new Transition(begin, n_insert);
  cur_transition->log_p() = log((mytype)1-EPSILON-GAPINIT);
  begin->succ()    +=cur_transition;
  n_insert->pred() +=cur_transition;

  cur_transition = new Transition(n_insert, n_insert);
  cur_transition->log_p() = log((mytype)1-EPSILON);
  n_insert->succ() +=cur_transition;
  n_insert->pred() +=cur_transition;

  cur_transition = new Transition(c_insert, end);
  cur_transition->log_p() = log(EPSILON);
  c_insert->succ() +=cur_transition;
  end->pred()      +=cur_transition;

  cur_transition = new Transition(c_delete, end);
  cur_transition->log_p() = log(1);
  c_delete->succ() +=cur_transition;
  end->pred()      +=cur_transition;

  cur_transition = new Transition(c_insert, c_insert);
  cur_transition->log_p() = log((mytype)1-EPSILON);
  c_insert->succ() +=cur_transition;
  c_insert->pred() +=cur_transition;

  //! Copy states in each profile HMM.
  for (int i=0; i<(int)hmm.m_num_phmms; ++i) {
     ProfileHMM* phmm = new ProfileHMM(*hmm.m_phmms[i], m_first_cons_col, m_last_cons_col, &state_to_state);
     m_phmms += phmm;
  }
  m_num_phmms = m_phmms.size();

  for (int s=0; s<m_num_phmms; ++s)
     m_phmm_to_sub_idx[m_phmms[s]] = s;

  if (Const_Def::jump_version() == linear) {
    m_type_to_jump_type = hmm.m_type_to_jump_type;
    m_jump_state_to_str = hmm.m_jump_state_to_str;

    m_type_to_jump_type = list_of< pair<StateType, JumpStateType> >(Jump_State_Match, Jump_Match)(Jump_State_Delete, Jump_Delete);
    m_jump_state_to_str = list_of< pair<JumpStateType, string> >(Jump_Match, "Jump_Match")(Jump_Delete, "Jump_Delete");

    /*!** Copy special mute states for jumps between profile HMMs (in construction): ***/
    m_jump_states.assign (m_len, make_pair((State*) 0, (State*) 0));
    for (int i=m_first_cons_col; i<m_last_cons_col; ++i) {
      if (hmm.m_jump_states[i].first) {
        State* cur_jump_state_match  = new State(*hmm.m_jump_states[i].first);
        state_to_state[cur_jump_state_match] = hmm.m_jump_states[i].first;
        state_to_state[hmm.m_jump_states[i].first] = cur_jump_state_match;

        State* cur_jump_state_delete = new State(*hmm.m_jump_states[i].second);
        state_to_state[cur_jump_state_delete] = hmm.m_jump_states[i].second;
        state_to_state[hmm.m_jump_states[i].second] = cur_jump_state_delete;

        m_jump_states[i] = make_pair(cur_jump_state_match, cur_jump_state_delete);
      }
    }
  }

  //! Copy jumps between profile hmms or from and to special jump states (for 'linear' jphmm version).
  /*!
   *  - successors of states in other profile hmms.
   *    (transition from M, I and D to special states separately in local_hmm()).
   *  - predecessors of states in other profile hmms.
   *  - mute match and delete jump states (for each column) (for 'linear' jphmm version).
   */

  //! Copy transitions to other profile hmms or to special jump states (for 'linear' jphmm version).
  FOREACH (ProfileHMM* p, hmm.m_phmms) {
    for (int i=m_first_cons_col; i<m_last_cons_col; ++i) {
      if (p->has_state(i, (int)Match)){
        for (int type = (int)Match; type<=(int)Delete; ++type) {
	  State* cur_state = p->state(i, type);

	  //! Original jump prob to C_Delete.
          mytype orig_log_jp_to_c_del = 0;
          if (type == (int)Match) {
            Transition* to_c_del = transition(cur_state, hmm.c_delete);
            orig_log_jp_to_c_del = to_c_del->log_p();
          }
	  //! Copy transition probabilities.
          State* wherefrom = state_to_state[cur_state];
          FOREACH (Transition* t, cur_state->succ()) {
	    State* cur_succ = t->whereto();
            if (cur_succ->column() < m_last_cons_col || (cur_succ->column() <= m_last_cons_col && !cur_succ->is_insert())) {
	      State* whereto = state_to_state[cur_succ];
	    
              if (!whereto->is_special()) {
                Transition* cur_trans = new Transition(wherefrom, whereto);
		if (whereto->is_jump_state() || cur_succ->phmm() != cur_state->phmm()) {
		  cur_trans->log_p()    = t->log_p();
		}
		else {
		  if (type == (int)Match) {
		    cur_trans->log_p() = log((exp(t->log_p())* ((mytype) 1 / ((mytype)1-exp(orig_log_jp_to_c_del)))));
		  }
		  else {
		    cur_trans->log_p() = t->log_p();
		  }
		}
                wherefrom->succ() += cur_trans;
              }
	    }
          }
        }
      }
    }
    //! for last cons column, Match and Delete states can only have End and special Insert as successors
    //! Insert in last_cons_col is not active
  }

  if (Const_Def::jump_version() == linear) {
    /*!** Copy transitions from special mute jump states to profile HMMs: ***/
    for (int i=m_first_cons_col; i<m_last_cons_col; ++i) {
      if (hmm.m_jump_states[i].first) {
        State* cur_state = hmm.m_jump_states[i].first;
        State* wherefrom = state_to_state[cur_state];
        FOREACH (Transition* t, cur_state->succ()) {
          State* cur_succ = t->whereto();
          State* whereto = state_to_state[cur_succ];
	    
          Transition* cur_trans = new Transition(wherefrom, whereto);
          cur_trans->log_p()    = t->log_p();
          wherefrom->succ() += cur_trans;
        }

        cur_state = hmm.m_jump_states[i].second;
        wherefrom = state_to_state[cur_state];
        FOREACH (Transition* t, cur_state->succ()) {
          State* cur_succ = t->whereto();
          State* whereto = state_to_state[cur_succ];
	    
          Transition* cur_trans = new Transition(wherefrom, whereto);
          cur_trans->log_p()    = t->log_p();
          wherefrom->succ() += cur_trans;
        }
      }
    }
  }

  //! Copy predecessors in other profile hmms.
  FOREACH (ProfileHMM* p, hmm.m_phmms) {
    for (int i=m_first_cons_col; i<=m_last_cons_col; ++i) {
      for (int type = (int)Match; type<=(int)Delete; ++type) {
        if (p->has_state(i, type)) {
          State* cur_state = p->state(i, type);
	  if (i < m_last_cons_col || !cur_state->is_insert()) {
            State* whereto   = state_to_state[cur_state];
	    FOREACH (Transition* t, cur_state->pred()) {
	      State* cur_pred = t->wherefrom();
	      if (cur_pred->column() >= m_first_cons_col) {
                State*      wherefrom = state_to_state[cur_pred];
                if (!wherefrom->is_special()) {
                  Transition* cur_trans = transition(wherefrom, whereto);
                  whereto->pred() += cur_trans;
		}
              }
	    }
          }
        }
      }
    }
  }
  if (Const_Def::jump_version() == linear) {
    for (int i=m_first_cons_col; i<m_last_cons_col; ++i) {
      if (hmm.m_jump_states[i].first) {
        State* cur_state = hmm.m_jump_states[i].first;
        State* whereto   = state_to_state[cur_state];
        FOREACH (Transition* t, cur_state->pred()) {
          State*      cur_pred  = t->wherefrom();
          State*      wherefrom = state_to_state[cur_pred];
          Transition* cur_trans = transition(wherefrom, whereto);
          whereto->pred() += cur_trans;
        }

        cur_state = hmm.m_jump_states[i].second;
        whereto = state_to_state[cur_state];
        FOREACH (Transition* t, cur_state->pred()) {
          State*      cur_pred  = t->wherefrom();
          State*      wherefrom = state_to_state[cur_pred];
          Transition* cur_trans = transition(wherefrom, whereto);
          whereto->pred() += cur_trans;
        }
      }
    }
  }
}

/****************************************************************//**
   HMM destructor.
   
   The lists of active states in the Viterbi algorithm are cleared,
   profile HMMs and special states deleted.
********************************************************************/
HMM::~HMM() {

  //clear_act_states();

  FOREACH (ProfileHMM* phmm, m_phmms)
    delete phmm;

  vector<State*> spec_states;
  spec_states += begin, n_delete, n_insert, c_insert, c_delete, end;

  FOREACH (State* s, spec_states)
    delete s;

  if (Const_Def::jump_version() == linear) {
    for (int i=m_first_cons_col; i<m_last_cons_col; ++i) {
      delete m_jump_states[i].first;
      delete m_jump_states[i].second;
    }
  }
}

/****************************************************************************************************//**
   Local HMM.

   A local HMM is build for short query sequences, i.e. only a certain part of the global HMM is taken into account.
   For this local HMM the transitions to and from the special states must be adapted,
   since f. ex. N_Insert is the predecessor of the Match state in the first consensus 
   column of the current (this local) HMM.
   The probabilities of a jump from N_Delete to each Match state and from each Match state to C_Delete 
   must be recalculated since they depend on the number of consensus columns in the local HMM.
   Also the emission probabilities of N_Insert and C_Insert are reestimated.
   
   @param[in] e Emission priors.
   @param[in] etpr Emission and transition profile of alignment.
********************************************************************************************************/
void HMM::local_hmm(const EmissionPriors& e,
		    const EmTransProfile& etpr) {

  m_is_local_hmm = true;
    
  vector<mytype> insert_em_priors  = e.insert_priors();
  vector<int>    n_insert_frequencies(insert_em_priors.size(), 0);
  vector<int>    c_insert_frequencies(insert_em_priors.size(), 0);

  //! profile HMMs:
  mytype prob_spec_chars = etpr.rel_freq_spec_chars();
  FOREACH (ProfileHMM* phmm, m_phmms) {
    for (int i=m_first_cons_col; i<m_last_cons_col; i++) {
      if (phmm->has_state(i, (int)Match)) {
        for (int type=(int)Match; type<=(int)Delete; type++) {
          State* cur_state = phmm->state(i, type);
          //! Recalculation of jumps and jump probabilities to special states for local hmm.
          calc_spec_jumps(cur_state);
          calc_spec_jump_probs(cur_state);
        }
      }
    }
    if (phmm->has_state(m_last_cons_col, (int)Match)) {
      State* cur_state = phmm->state(m_last_cons_col, (int)Match);
      calc_spec_jumps(cur_state);
      calc_spec_jump_probs(cur_state);
      cur_state = phmm->state(m_last_cons_col, (int)Delete);
      calc_spec_jumps(cur_state);
      calc_spec_jump_probs(cur_state);
    }
      
    //! N_INSERT and C_INSERT emission probabilities.
    /*!
     *  Nucleotides in all columns in front of current first and behind current last consensus column (local cons col) are taken into account, resp..
        For further information about the calculation of the emission probabilities please see description in HMM constructor.
     */
    SubtypeProfile* cur_sub_prof = phmm->sub_prof();
    ConsColumnProfile* cur_cons_col_prof = cur_sub_prof->nterminal_cons_col_prof();
    EmissionProfile*   cur_em_prof       = cur_cons_col_prof->em_prof();

    //! N_INSERT emission frequencies.
    for (unsigned i=0; i<insert_em_priors.size(); ++i)
      n_insert_frequencies[i] += cur_em_prof->insert_frequency(i);

    for (int i=0; i<m_first_cons_col; ++i) {
      if (cur_sub_prof->is_cons_column(i)) {
        cur_cons_col_prof = cur_sub_prof->cons_col_prof(i);
        cur_em_prof       = cur_cons_col_prof->em_prof();
        for (unsigned j=0; j<insert_em_priors.size(); ++j) {
          n_insert_frequencies[j] += cur_em_prof->match_frequency(j);
          n_insert_frequencies[j] += cur_em_prof->insert_frequency(j);
        }
      }
    }

    //! C_INSERT emission frequencies.
    cur_cons_col_prof = cur_sub_prof->cons_col_prof(m_last_cons_col);
    cur_em_prof       = cur_cons_col_prof->em_prof();
    for (unsigned i=0; i<insert_em_priors.size(); ++i) {
      c_insert_frequencies[i] += cur_em_prof->insert_frequency(i);
    }
    for (int i=m_last_cons_col+1; i<m_len; ++i) {
      if (cur_sub_prof->is_cons_column(i)) {
	cur_cons_col_prof = cur_sub_prof->cons_col_prof(i);
        cur_em_prof       = cur_cons_col_prof->em_prof();
        for (unsigned j=0; j<insert_em_priors.size(); ++j) {
          c_insert_frequencies[j] += cur_em_prof->match_frequency(j);
          c_insert_frequencies[j] += cur_em_prof->insert_frequency(j);
        }
      }
    }
  }

  //! N_INSERT and C_INSERT emission probabilities.
  ProfileHMM* p;
  vector<mytype> cur_ep = p->emission_probs(insert_em_priors, n_insert_frequencies);
  for (vector<mytype>::iterator it=cur_ep.begin(); it < cur_ep.end(); ++it)
    (*it) = ((mytype)1 - prob_spec_chars)*(*it);
  cur_ep            += prob_spec_chars;
  n_insert->log_ep() = p->log_of_prob(cur_ep);

  cur_ep = p->emission_probs(insert_em_priors, c_insert_frequencies);
  for (vector<mytype>::iterator it=cur_ep.begin(); it < cur_ep.end(); ++it)
    (*it) = ((mytype)1 - prob_spec_chars)*(*it);

  cur_ep            += prob_spec_chars;
  c_insert->log_ep() = p->log_of_prob(cur_ep);

  //! Transition probabilities for special states.
  calc_spec_jump_probs(n_insert);
  calc_spec_jump_probs(n_delete);
  calc_spec_jump_probs(begin);

  //! Standardization of transition probabilities
  //! and assignment of state id for all states in hmm.
  begin->id() = 0;
  m_idx_to_state[0] = begin;
  n_delete->id() = 1;
  m_idx_to_state[1] = n_delete;
  n_insert->id() = 2;
  m_idx_to_state[2] = n_insert;
  int cur_id = 3;

  if (Const_Def::jump_version() == linear) {
    for (int j=m_first_cons_col; j<m_last_cons_col; ++j) {
      if (m_jump_states[j].first) {
        State* cur_state = m_jump_states[j].first;
        if (!cur_state->succ().empty())
          cur_state->succ_standardization(m_first_cons_col, m_last_cons_col);
        cur_state->id() = cur_id;
        m_idx_to_state[cur_id] = cur_state;
        ++cur_id;

        cur_state = m_jump_states[j].second;
        if (!cur_state->succ().empty())
          cur_state->succ_standardization(m_first_cons_col, m_last_cons_col);
        cur_state->id() = cur_id;
        m_idx_to_state[cur_id] = cur_state;
        ++cur_id;
      }
    }
  }

  FOREACH (ProfileHMM* phmm, m_phmms) {
    for (int i=m_first_cons_col; i<m_last_cons_col; ++i) {
      if (phmm->has_state(i, (int)Match)) {
        for (int type=(int)Match; type<=(int)Delete; ++type) {
          State* cur_state = phmm->state(i, type);
          cur_state->succ_standardization(m_first_cons_col, m_last_cons_col);
	  cur_state->id() = cur_id;
	  m_idx_to_state[cur_id] = cur_state;
	  ++cur_id;
        }
      }
    }

    //! Insert in last cons col doesn't exist:
    for (int type=(int)Match; type<=(int)Delete; ++type) {
      if (phmm->has_state(m_last_cons_col, type)) {
        State* cur_state = phmm->state(m_last_cons_col, type);
        cur_state->succ_standardization(m_first_cons_col, m_last_cons_col);
        cur_state->id() = cur_id;
	m_idx_to_state[cur_id] = cur_state;
	++cur_id;
      }
    }
  }


  begin->succ_standardization(m_first_cons_col, m_last_cons_col);
  n_delete->succ_standardization(m_first_cons_col, m_last_cons_col);
  n_insert->succ_standardization(m_first_cons_col, m_last_cons_col);
  c_insert->succ_standardization(m_first_cons_col, m_last_cons_col);
  c_delete->succ_standardization(m_first_cons_col, m_last_cons_col);

  c_insert->id()           = cur_id;
  m_idx_to_state[cur_id]   = c_insert;
  c_delete->id()           = cur_id+1;
  m_idx_to_state[cur_id+1] = c_delete;
  end->id()                = cur_id+2;
  m_idx_to_state[cur_id+2] = end;

  m_num_states = end->id()+1;

}

/*****************************************************************************************//**
   First consensus column in hmm.

   Determination of the first column that is a consensus column in all subtypes in between columns col_start and col_end.
   col_start and col_end can be any column, so this function is also valid for a local HMM.

   @param[in] col_start First possible consensus column in local HMM.
   @param[in] col_end Last possible consensus column in local HMM.
*********************************************************************************************/
int HMM::first_cons_col(int col_start, 
			int col_end) {

  int first_cons_col = 0;

  for (int i=col_start; i<=col_end; i++) {
    int num_cons_cols = 0;
    FOREACH (ProfileHMM* phmm, m_phmms){
      if (phmm->sub_prof()->is_cons_column(i))
        num_cons_cols++;
    }
    //! The first common consensus column must be a consensus column in ALL subtypes.
    if (num_cons_cols == m_num_phmms) {
      first_cons_col = i;
      break;
    }
  }
  return first_cons_col;
}

/*****************************************************************************************//**
   Last consensus column in hmm.

   Determination of the last column that is a consensus column in all subtypes in between columns col_start and col_end.
   col_start and col_end can be any column, so this function is also valid for a local HMM.
   
   @param[in] col_start First possible consensus column in local HMM.
   @param[in] col_end Last possible consensus column in local HMM.
*********************************************************************************************/
int HMM::last_cons_col(int col_start, 
		       int col_end){

  int last_cons_col = 0;
  for (int i=col_end; i>=col_start; --i) {
    int num_cons_cols = 0;
    FOREACH (ProfileHMM* phmm, m_phmms){
      if (phmm->sub_prof()->is_cons_column(i))
        ++num_cons_cols;
    }
    //! The first common consensus column must be a consensus column in ALL subtypes.
    if (num_cons_cols == m_num_phmms) {
      last_cons_col = i;
      break;
    }
  }
  return last_cons_col;
}


/**************************************************************************//**
 Calculate jumps in the HMM.

   Determination of jumps from one subtype to all other subtypes.\n
   Jumps are allowed 
     - from Match, Insert and Delete states to Match states
     - from Match states to Delete states.\n
   Probabilities for these jumps are calculated in calc_jump_probs().
   
   @param[in] cur_state Current state.
   @param[in,out] num_jumps Number of jumps from this state to other profile HMMs.
******************************************************************************/
void HMM::calc_jumps(State *cur_state,
		     int   *num_jumps) {

  int cur_column = cur_state->column();
  int cur_type   = cur_state->type();
  ProfileHMM*     cur_phmm     = cur_state->phmm();
  SubtypeProfile* cur_sub_prof = cur_phmm->sub_prof();
  *num_jumps = 0;

  //! Jumps to other subtypes.
  int succ_cons_col = cur_sub_prof->succ_cons_col(cur_column);

  FOREACH (ProfileHMM* phmm, m_phmms) {
    if (phmm != cur_phmm && cur_column < m_last_cons_col) {

      //! Column of successor state.
      /*! This column must be the next possible column to the right (and only one successive column is allowed). */
      int diff_succ_cons_col = phmm->sub_prof()->succ_cons_col(cur_column);
      if (diff_succ_cons_col <= succ_cons_col) {
        if (phmm->has_state(diff_succ_cons_col, (int)Match)) {
          //! Jump from a Match, Insert or Delete state to a Match state.
          State*      succ_state     = phmm->state(diff_succ_cons_col, (int)Match);
          Transition* cur_transition = new Transition(cur_state, succ_state);
          cur_transition->log_p() = log(0);
          cur_state->succ()      += cur_transition;
          succ_state->pred()     += cur_transition;
          ++(*num_jumps);

          //! Jump from a Match state to a Delete state.
          if (cur_type == (int)Match) {
            if (phmm->has_state(diff_succ_cons_col, (int)Delete)) {
              succ_state          = phmm->state(diff_succ_cons_col, (int)Delete);
              cur_transition      = new Transition(cur_state, succ_state);
              cur_transition->log_p() = log(0);
              cur_state->succ()      += cur_transition;
              succ_state->pred()     += cur_transition;
            }
          }
        }
      }
    }
  }
}

/********************************************************************************//**
  Calculation of jump probabilities.

  Calculation of the probabilities of jumps between different subtypes/profile HMMs.
  The transition probabilities within a profile HMM are adjusted 
  (multiplied by (1-prob of jump from this state)).

   @param[in] cur_state Current state.
   @param[in] num_jumps Number of jumps from this state to other profile HMMs.
************************************************************************************/
void HMM::calc_jump_probs(State* cur_state,
			  int    num_jumps) {

  //! \todo {Include option/jumps for jpHMM version, that contains profileHMM for whole alignment.}
  
  //! jump prob at current alignment column depends on number of possible jumps.
  mytype local_jump_prob = 0;
  if (Const_Def::adapt_jump_probs()) {
    local_jump_prob = Const_Def::jump_prob()/(num_jumps+1);
  }
  else {
    if (num_jumps > 0)
      local_jump_prob = Const_Def::jump_prob()/num_jumps;
  }

  ProfileHMM *cur_phmm = cur_state->phmm();
  if (cur_state->is_insert() || cur_state->is_delete()) {
    FOREACH (Transition* cur_trans, cur_state->succ()) {
      State* cur_succ = cur_trans->whereto();
      if (cur_succ->phmm() == cur_phmm) { //! original transition prob weighted with (1-jumpProb).
        cur_trans->log_p() = log(((mytype)1-local_jump_prob) * cur_trans->trans_prob());
      }
      else 
	if (cur_succ->is_match())    //! jump to Match state of another subtype.
          cur_trans->log_p() = log(local_jump_prob);
    }
  }
  else { //! Match
    FOREACH (Transition* cur_trans, cur_state->succ()) {
      State* cur_succ = cur_trans->whereto();

      if (!cur_succ->is_special()) {
        ProfileHMM *cur_succ_phmm = cur_succ->phmm();
        if (cur_succ_phmm == cur_phmm) { //! original transition prob weighted with (1-jumpProb).
          cur_trans->log_p() = log(cur_trans->trans_prob() * ((mytype)1-local_jump_prob));
	}
        else {
          //! jump to Match or Delete in another (target) subtype.
	  /*! jump prob to target subtype is divided according to transition prob to M and D in target subtype.
	     -> trans prob M->M and M->D in this profile HMM is determined -> jump prob determined */

          int    cur_succ_col = cur_succ->column();
          mytype trans_MM     = 0;
          mytype trans_MD     = 0;

          //! preceding cons col in target subtype.
	  int cur_pred_cons_col = cur_succ_phmm->sub_prof()->pred_cons_col(cur_succ_col);
	  //! transition prob. from Match to Match and Delete in target subtype.
          if (cur_succ_phmm->has_state(cur_pred_cons_col, (int) Match)) {
            State* cur_pred_match = cur_succ_phmm->state(cur_pred_cons_col, (int) Match);
            State* cur_match      = cur_succ_phmm->state(cur_succ_col, (int) Match); // Match in der Spalte in die gesprungen wird
            State* cur_delete     = cur_succ_phmm->state(cur_succ_col, (int) Delete);
	    trans_MM = transition(cur_pred_match, cur_match)->trans_prob();
            trans_MD = transition(cur_pred_match, cur_delete)->trans_prob();
          }
          if (trans_MM == 0 && trans_MD == 0) {
            const mytype EPSILON = Const_Def::epsilon();
            trans_MD = trans_MM = EPSILON/((mytype)2);
          }
	  //! calculation of current jump prob.
          if (cur_succ->is_match()) {
	    if (OUTPUT_DEBUG)
	      cout << cur_state->phmm()->sub_prof()->sub()->name() << " " << cur_succ->column()+1 << " " << cur_succ->phmm()->sub_prof()->sub()->name() << " " << "trans_MM: " << trans_MM << " trans_MD: " << trans_MD << endl;
            cur_trans->log_p() = log(local_jump_prob * (trans_MM/(trans_MM+trans_MD)));
	    if (Const_Def::adapt_jump_probs())
	      cur_trans->log_p() = log(local_jump_prob * (mytype)0.5);
	  }
          else {
            cur_trans->log_p() = log(local_jump_prob * (trans_MD/(trans_MM+trans_MD)));
   	    if (Const_Def::adapt_jump_probs())
	      cur_trans->log_p() = log(local_jump_prob * (mytype)0.5);
	  }
        }
      }
    }
  }
}


/**********************************************************************************//**
 Calculate special jumps/transitions.

   Determination of jumps from/to special states.\n
   Special jumps are allowed 
     - from Begin to Match and Delete states in first consensus column in all subtypes
     - from N_Delete to all Match states
     - from N_Insert to Match states in first consensus column in all subtypes
     - from Match states in last consensus column in all subtypes to C_INSERT
     - from all Match states to C_Delete     
     - from Match and Delete states in last consensus column in all subtypes to End.\n
   Probabilities for these jumps are calculated in calc_spec_jump_probs().

   @param[in] cur_state Current state.
***************************************************************************************/
void HMM::calc_spec_jumps(State *cur_state){

  int cur_column = cur_state->column();
  bool is_match = false;
  if (cur_state->is_match())
    is_match = true;
  
  if (is_match || cur_state->is_delete()) {
    if (cur_column == m_first_cons_col) {
      //! Jump from Begin to Match or Delete.
      Transition* cur_transition = new Transition(begin, cur_state);
      begin->succ()     += cur_transition;
      cur_state->pred() += cur_transition;
      if (is_match) {
        //! Jump from N_Insert to Match.
	cur_transition     = new Transition(n_insert, cur_state);
        n_insert->succ()  += cur_transition;
        cur_state->pred() += cur_transition;
      }
    }
    else {
      if (is_match) {
        //! Jump from N_Delete to Match.
	Transition *cur_transition = new Transition(n_delete, cur_state);
        cur_state->pred() += cur_transition;
        n_delete->succ()  += cur_transition;
      }
    }
    if (cur_column == m_last_cons_col) {
      //! Jump from Match or Delete to End.
      Transition* cur_transition = new Transition(cur_state, end);
      end->pred()       += cur_transition;
      cur_state->succ() += cur_transition;
      if (is_match) {
        //! Jump from Match to C_Insert.
        cur_transition     = new Transition(cur_state, c_insert);
        c_insert->pred()  += cur_transition;
        cur_state->succ() += cur_transition;
      }
    }
    else {
      if (is_match) {
	//! Jump from Match to C_Delete.
        Transition *cur_transition = new Transition(cur_state, c_delete);
        c_delete->pred()  += cur_transition;
        cur_state->succ() += cur_transition;
      }
    }
  }
}


/**************************************************************************************************************//**
  Calculation of special jump probabilities.

  Calculation of probabilities for jumps from/to special states.
  Probabilities of a transition from a special state to a special state is calculated above 
  directly after definition in HMM().

  @param[in] cur_state Current state.
  \todo{trans. prob from N_Delete to states in 1st column and from states in last column to C_Delete must be 0.
         This must be done after the program has been tested (same bug in old version).}
******************************************************************************************************************/
void HMM::calc_spec_jump_probs(State* cur_state){

  if (cur_state->is_begin()) {
    //! BEGIN->Match/Delete
    FOREACH (Transition* cur_trans, cur_state->succ()) {
      State *cur_succ = cur_trans->whereto();
      if (cur_succ->is_delete() || cur_succ->is_match()) // 
        cur_trans->log_p() = log((Const_Def::epsilon()/2)*((mytype)1/m_num_phmms));
    }
  }
  else if (cur_state->is_n_insert()) {
    //! N_INSERT->Match
    FOREACH (Transition* cur_trans, cur_state->succ()) {
      State *cur_succ = cur_trans->whereto();
      if (!cur_succ->is_n_insert())
        cur_trans->log_p() = log(Const_Def::epsilon()/m_num_phmms);
    }
  }
  else if (cur_state->is_n_delete()) { 
    //! N_DELETE->MATCH (all! match states)
    FOREACH (Transition* cur_trans, cur_state->succ()) {
      State *cur_succ = cur_trans->whereto();
      int cur_succ_col = cur_succ->column();
      if (cur_succ_col > m_first_cons_col) {
        SubtypeProfile *cur_succ_sub_prof = cur_succ->phmm()->sub_prof();
        //! number of preceding consensus columns in subtype profile of successor state in local hmm
        int lc_num_pred_cons_cols = cur_succ_sub_prof->num_pred_cons_cols(cur_succ_col) - cur_succ_sub_prof->num_pred_cons_cols(m_first_cons_col); // since local alignment but global subtype profile
        cur_trans->log_p() = (((mytype)(lc_num_pred_cons_cols))*log(Const_Def::gap_ext())) + log((mytype)1/m_num_phmms);
      }
    }
  }
  else if (cur_state->is_delete()) {
    //! DELETE->END in last cons col
    if (cur_state->column() == m_last_cons_col) {
      Transition* cur_trans = transition(cur_state, end);
      cur_trans->log_p()    = 0; //log(1);
    }
  }
  else if (cur_state->is_match()) {
    //! MATCH->C_INSERT
    //! MATCH->C_DELETE
    //! MATCH->END
    /*! \todo{Local jump prob and jump to c delete must be 0 for last consensus column.
              Has impact on trans prob to C_INSERT and C_DELETE.
              Change later, current jpHMM version must be comparable to former version.} */

    ProfileHMM     *cur_phmm              = cur_state->phmm();
    SubtypeProfile *cur_sub_prof          = cur_phmm->sub_prof();
    int             cur_col               = cur_state->column();
    int             lc_num_succ_cons_cols = 0;

    /*! Jump prob to C_DELETE = log(gap_init)+ (#num cons columns between M and m_last_cons_col)*log(gap_ext).
        For local alignments/hmms do not count cons columns behind m_last_cons_col. */
    if (cur_col < m_last_cons_col)
      lc_num_succ_cons_cols = cur_sub_prof->num_succ_cons_cols(cur_col) - cur_sub_prof->num_succ_cons_cols(m_last_cons_col); // TODO: num_succ_cons_col(m_last_cons_col) == 0??
    mytype jump_to_c_delete = exp(log(Const_Def::gap_init()) +(((mytype)(lc_num_succ_cons_cols-1))*(log(Const_Def::gap_ext())))); // TODO: aendern fuer num_succ_cons_col = 0 ??

    //! Determine the number of jumps from Match to other subtypes (not taking into account that jump to M and D are possible)
    int num_jumps = 0;
    if (cur_col < m_last_cons_col) {
      FOREACH (Transition* cur_trans, cur_state->succ()) {
        State* cur_succ  = cur_trans->whereto();
        if (cur_succ->phmm() != cur_phmm && cur_succ->is_match())
          ++num_jumps;
      }
    }
    
    //! jump prob at a certain alignment column = const jump_prob / num observed jumps
    mytype local_jump_prob = 0;
    if (Const_Def::adapt_jump_probs())
      local_jump_prob = Const_Def::jump_prob()/(num_jumps+1);
    else {
      if (num_jumps > 0)
        local_jump_prob = Const_Def::jump_prob()/num_jumps;
    }
    
    //! Assign probabilities to jumps: probs must sum up to 1
    FOREACH (Transition* cur_trans, cur_state->succ()) {
      State* cur_succ  = cur_trans->whereto();
      if (cur_succ->is_end())
        cur_trans->log_p() = log(Const_Def::epsilon());
      else if (cur_succ->is_c_insert()) {
        cur_trans->log_p() = log(((mytype)1-local_jump_prob - jump_to_c_delete - Const_Def::epsilon()));
      }
      else if (cur_succ->is_c_delete())
        cur_trans->log_p() = log(jump_to_c_delete);
      else {
        //! jumps to M and D in other subtypes
	if (cur_col < m_last_cons_col) {
          ProfileHMM *cur_succ_phmm = cur_succ->phmm();
          if (!cur_succ->is_jump_state() && cur_succ_phmm->sub_name() == cur_phmm->sub_name()) {
	    //! transition within same profile HMM
            if (m_is_local_hmm) {
	      //! local hmm: in copy constructor original jump_to_c_delete subtracted from jump prob -> add current local jump_to_c_delete
              cur_trans->log_p() = log(cur_trans->trans_prob() * ((mytype)1-jump_to_c_delete));
	    }
	    else {
	      //! jump prob is weighted by 1-jump_to_c_delete
              cur_trans->log_p() = log(cur_trans->trans_prob() * ((mytype)1-jump_to_c_delete));
	    }
          }
          else {
	    //! jump to other subtype or mute jump state
            if (m_is_local_hmm) {
	      //! local hmm: in copy constructor original jump_to_c_delete subtracted from jump prob -> add current local jump_to_c_delete
	      cur_trans->log_p() = log(cur_trans->trans_prob());
	    }
            else {
	      //! jump prob is weighted by 1-jump_to_c_delete
              cur_trans->log_p() = log(cur_trans->trans_prob());
	    }
          }
        }
      }
    }
  }
}

/*******************************************************************//**
 Calculation of mute jumps:

 Instead of jumps between subtypes 
 jumps to a mute state and from this mute state to the next possible state in each subtype.
 (not yet used/finished)
 Mute state for Matches and for Deletes.
 
 Original possible jumps: 
   M, I, D -> M
         M -> D
	 
 => two types of mute states
 mute_match_jump  : successors are successive Match states in all subtypes (in corresponding successive consensus column)
 mute_delete_jump : successors are successive Delete states in all subtypes (in corresponding successive consensus column)
 
 Possible jumps:
   M, I, D -> mute_match_jump  -> M in all subtypes
         M -> mute_delete_jump -> D in all subtypes

First part of jump.

  @param[in] cur_state Current state.
***********************************************************************/
void HMM::calc_jumps_to_jump_states(State* cur_state) {

  int         cur_col  = cur_state->column();
  ProfileHMM* cur_phmm = cur_state->phmm();
  
  if (cur_phmm->sub_prof()->has_succ_cons_col(cur_col)) {
    //! Else: path ends in mute_match oder mute_delete
    
    //! Current column of states = cur_col
    //! Successive consensus column for each subtype: succ_cons_col
    int cur_succ_cons_col = cur_phmm->sub_prof()->succ_cons_col(cur_col);

    if (cur_succ_cons_col <= m_last_cons_col && cur_phmm->has_state(cur_succ_cons_col, (int)Match)) {
      mytype jump_prob = Const_Def::jump_prob();
  
      //! log of jump prob.
      mytype cur_log_jp   = log(jump_prob);
  
      //! adjust trans probs within pHMM
      FOREACH (Transition *t, cur_state->succ()) {
	if (t->whereto()->phmm() == cur_state->phmm()) {
	  t->log_p() = log(t->trans_prob() * (((mytype)1-jump_prob)/m_num_phmms));
	}
      }

      //! log of probability of a jump to the mute_match_jump state.
      mytype cur_log_jp_to_mute_match = cur_log_jp;
      mytype cur_log_jp_to_mute_delete = cur_log_jp;
      if (cur_state->is_match()) {
	//! determine ratio of transition probabilities to successive column in same profile HMM:
/*	State* succ_match = cur_phmm->state(cur_succ_cons_col, (int)Match);
	State* succ_delete = cur_phmm->state(cur_succ_cons_col, (int)Delete);
	Transition* trans_to_succ_match = transition(cur_state, succ_match);
	Transition* trans_to_succ_delete = transition(cur_state, succ_delete);*/
	
	mytype ratio = (mytype) 0.5;
	cur_log_jp_to_mute_match  = log(jump_prob * ratio);
	cur_log_jp_to_mute_delete = log(jump_prob * ((mytype) 1-ratio));
      }

      State* cur_jump_state_match = m_jump_states[cur_col].first;

      //! Modelling of jumps from M, I, D to Match states in other subtypes.
      Transition* cur_trans = new Transition(cur_state, cur_jump_state_match);
      cur_trans->log_p()            = cur_log_jp_to_mute_match;
      cur_state->succ()            += cur_trans;
      cur_jump_state_match->pred() += cur_trans;
      if (cur_state->is_match()) {
        //! Modelling of jumps to Delete states in other subtypes only for Match states.

        //! -> jump to "cur_mute_delete_jump" (has Delete states as successors).
        State* cur_jump_state_delete     = m_jump_states[cur_col].second;
	
        cur_trans = new Transition(cur_state, cur_jump_state_delete);
        cur_trans->log_p()            = cur_log_jp_to_mute_delete;
        cur_state->succ()             += cur_trans;
        cur_jump_state_delete->pred() += cur_trans;

      }
    }
  }
}

/*******************************************************************//**
 Calculation of mute jumps:

 Instead of jumps between subtypes 
 jumps to a mute state and from this mute state to the next possible state in each subtype.
 (not yet used/finished)
 Mute state for Matches and for Deletes.
 
 Original possible jumps: 
   M, I, D -> M
         M -> D
	 
 => two types of mute states
 mute_match_jump  : successors are successive Match states in all subtypes (in corresponding successive consensus column)
 mute_delete_jump : successors are successive Delete states in all subtypes (in corresponding successive consensus column)
 
 Possible jumps:
   M, I, D -> mute_match_jump  -> M in all subtypes
         M -> mute_delete_jump -> D in all subtypes

  @param[in] cur_col Current column.
***********************************************************************/
void HMM::calc_jumps_from_jump_states(int cur_col) {

  //! Special mute states for current column.
  State* cur_jump_state_match  = m_jump_states[cur_col].first;
  State* cur_jump_state_delete = m_jump_states[cur_col].second;

  //! log of probability for a jump from one of the mute jump states to a subtype
  mytype cur_log_jp_to_st = log((mytype)1/m_num_phmms);

  FOREACH (ProfileHMM* cur_phmm, m_phmms) {
    if (cur_phmm->sub_prof()->has_succ_cons_col(cur_col)) {

      //! Successive consensus column for each subtype: succ_cons_col
      int cur_succ_cons_col = cur_phmm->sub_prof()->succ_cons_col(cur_col);

      if (cur_succ_cons_col <= m_last_cons_col && cur_phmm->has_state(cur_succ_cons_col, (int)Match)) {

        //! Jump from special mute match state to Match state in successive consensus column (in current subtype).
        State*      cur_succ_M  = cur_phmm->state(cur_succ_cons_col,(int)Match);
        Transition* cur_trans_M = new Transition(cur_jump_state_match, cur_succ_M);
        cur_trans_M->log_p()           = cur_log_jp_to_st;
        cur_succ_M->pred()            += cur_trans_M;
        cur_jump_state_match->succ() += cur_trans_M;
	  
        //! Jump from special mute delete state to Delete state in successive consensus column (in current subtype).
        State* cur_succ_D  = cur_phmm->state(cur_succ_cons_col,(int)Delete); //! ex. when M ex.
        Transition* cur_trans_D = new Transition(cur_jump_state_delete, cur_succ_D);
        cur_trans_D->log_p()           = cur_log_jp_to_st;
        cur_succ_D->pred()             += cur_trans_D;
        cur_jump_state_delete->succ() += cur_trans_D;
      }
    }
  }
}


/*****************************************************//**
 Transition.

 Returns transition between two given states if exists.
 
 @param[in] wherefrom Start state of transition.
 @param[in] whereto Target state of transition.
*********************************************************/
Transition* HMM::transition(State *wherefrom,
			    State *whereto) const {

  if (!wherefrom)
    throw Error("State wherefrom in transition doesn't exist");
  if (!whereto)
    throw Error("State whereto in transition doesn't exist");
  
  vector<Transition*> cur_succ = wherefrom->succ();
  vector<Transition*>::const_iterator it = cur_succ.begin();
  //! Look for successor (whereto) in list of successors of state wherefrom.
  while (it != cur_succ.end() && (*it)->whereto() != whereto)
    ++it;
  if (it == cur_succ.end()) {
    cout << "from " 
         << wherefrom->type() << " " << wherefrom->column() << " " << wherefrom->phmm()->sub_prof()->sub()->name()
         << " to " 
         << whereto->type() << " " << whereto->column() << " " << whereto->phmm()->sub_prof()->sub()->name()
         << endl;
    throw Error("Transition doesn't exist");
  }
  return (*it);
}


/**************************************************
 *
 *                class ActiveState.
 *
***************************************************/

/************************************************//**
   ActiveState

   Active state in Viterbi (beam search) algorithm.
   
   @param[in] state Current state.
   @param[in] best_pred Best predecessor of current state (in Viterbi).
   @param[in] vit Viterbi variable of current state.
   @param[in] fw Forward variable of current state.
   @param[in] bw Backward variable of current state.
****************************************************/

ActiveState::ActiveState(State* state,
	                 State* best_pred,
                         mytype vit,
			 mytype fw,
			 mytype bw) {
  
  m_state      = state;
  m_best_pred  = best_pred;
  m_viterbi    = vit;
  m_forward    = fw;
  m_backward   = bw;
}
