// ~/jpHMM/src/hmm.h
/**************************************************************
 File name	: hmm.h
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Classes: HMM
          ActiveState

 Corresponding files: hmm.cpp
                      hmm_algorithms.cpp
		      hmm_output.cpp

 hmm.cpp:
 - Construction of HMM by combination of profile HMMs for each subtype
 - Definition of special states - at beginning and end of HMM
                                - for jumps between different profile HMMs (if this option is chosen)
              and respective emission and transition probabilities
 - Definition of local HMM for certain region in the alignment/HMM
 - Definition of first and last common consensus column in all subtypes
 - Calculation of jump probabilities between states of different subtypes

 hmm_algorithms.cpp:
 Implementation of all jpHMM algorithms:
  - Viterbi and Forward 
    (active states determined with Beam-search algorithm or by prealignment with BLAT)
  - Viterbi backtracking
  - Backward
  - Posterior probabilities
  - Pruning of recombination prediction for extended circular sequences to original sequence
  - Call of output functions

 hmm_output.cpp
  Output of 
  - predicted recombination defined by Viterbi path
  - alignment of query sequence to multiple alignment defined by Viterbi path
  - alignment of query sequence to reference genome defined by Viterbi path
  - posterior probabilities
  - path of subtypes with highest posterior probability for each sequence position

 Test output for HMM: 
  - all states and their successors (or predecessors)
  - active states at each query sequence positoin
  - active intervals (alignment columns) for M, I and D states
  - details about Viterbi path and emission probabilities in tabular form


 ActiveStates: define states that are 'active' in Viterbi recursion
***************************************************************/

#ifndef _HMM_H
#define _HMM_H
#include "std_header.h"

using namespace std;

class Const_Def;
class Query;
class Alignment;
class EmissionPriors;
class TransitionPriors;
class State;
class ActiveState;
class ExtEmTrProbs;


/******************************************************\
 *                   class HMM                        *
\******************************************************/

class HMM {
	friend ostream& operator<<(ostream& s, const HMM& h);
public:
  HMM(Alignment* al, const EmissionPriors& e, const TransitionPriors& t, const EmTransProfile& etpr, bool use_eetp, const ExtEmTrProbs* eetp = 0);
  HMM(const HMM &hmm, int s, int e); //! copy constructor
  ~HMM();
  int  len() const {return m_len;};
  int  first_cons_col() const {return m_first_cons_col;};
  int  last_cons_col()  const {return m_last_cons_col;};

  void local_hmm(const EmissionPriors& e, const EmTransProfile& etpr);
  int  first_cons_col(int col_start, int col_end);
  int  last_cons_col(int col_start, int col_end);

  //! algorithms:
  bool        vit_and_fw_algo(Query *q, vector<pair<int, int> > act_al_columns);
  void        prune_ext_circ_genome(Query q, vector<State*> *vit_path, vector<int> *vit_path_start_pos, vector<int> *vit_path_end_pos, vector<string> *vit_path_subtype, list<string> *pred_subtypes);
  void        vit_back(Alignment *al, Query q);
  void        bw_algo(string query);
  void        post_probs(Query q);
  void        clear_act_states(Query q);

  //! output:
  void        recomb_output(Query q, vector<int> vit_path_start_pos, vector<int> vit_path_end_pos, vector<string> vit_path_subtype, list<string> pred_subtypes);
  void        post_prob_output(Query q, multi_array<mytype, 2> post_probs);
  void        post_path_output(vector<int> post_path_start_pos, vector<int> post_path_end_pos, vector<string> post_path_subtype);
  void        alignment_to_msa_output(string query_name, vector<State*>  vit_path);
  void        alignment_to_ref_seq_output(Alignment *al, Query q, vector<State*>  vit_path);
  
  //! test output:
  void hmm_out();
  void act_states_out();
  void output_details(Query query, vector<State*> vit_path_states, vector<string> vit_path_subs);

  
private:
  vector<ProfileHMM*> m_phmms;
  int                 m_num_phmms;
  map<int, State*>    m_idx_to_state;
  int                 m_num_states;
  int                 m_len;
  bool                m_is_local_hmm;

  int       m_first_cons_col;
  int       m_last_cons_col;
  State*    begin;
  State*    n_delete;
  State*    n_insert;
  State*    c_insert;
  State*    c_delete;
  State*    end;
  
  vector< pair<State*, State*> > m_jump_states;

  multi_array<ActiveState*, 2>         m_act_spec_states;
  multi_array<vector<ActiveState*>, 3> m_act_states;
  multi_array<pair<int,int>, 3>        m_act_intervals;
  multi_array<vector<ActiveState*>, 2> m_act_jump_states;
  multi_array<pair<int,int>, 2>        m_act_jump_intervals;

  
  map<ProfileHMM*, int>         m_phmm_to_sub_idx;
  map<State*, int>              m_spec_state_to_idx;
  map<State*, string>           m_spec_state_to_str;
  map<string, State*>           m_str_to_spec_state;
  map<StateType, JumpStateType> m_type_to_jump_type;
  map<JumpStateType, string>    m_jump_state_to_str;


  Transition* transition(State *wherefrom, State *whereto) const;
  void        calc_jumps(State *cur_state, int *num_jumps);
  void        calc_jump_probs(State *cur_state, int num_jumps);
  void        calc_spec_jumps(State *cur_state);
  void        calc_spec_jump_probs(State *cur_state);
  void        calc_jumps_to_jump_states(State* cur_state);
  void        calc_jumps_from_jump_states(int cur_col);
  
  //! algorithms:
  vector<pair<int, int> > ext_act_al_intervals(Query q, vector<pair<int, int> > active_al_columns);
  void                 vit_and_fw_spec_state(State *cur_state, int qp, int cur_nucl_idx, mytype *max_vit, bool ***is_cur_act_state, pair<mytype, mytype> ***vit_fw_cur_act_states, int ***best_pred_idx_cur_act_states);
  void                 vit_and_fw_algo_jump_states(JumpStateType cur_jump_state_type, Query q, int qp, int cur_nucl_idx, mytype *max_vit, mytype log_beam, bool ***is_cur_act_state, pair<mytype, mytype> ***vit_fw_cur_act_states, int ***best_pred_idx_cur_act_states);
  void                 vit_and_fw_var(State *cur_state, int qp, int cur_nucl_idx, mytype *viterbi, mytype *forward, State** best_pred, bool ***is_cur_act_state, pair<mytype, mytype> ***vit_fw_cur_act_states, int ***best_pred_idx_cur_act_states);
  multi_array<pair<int,int>, 2> calc_act_M_and_I_intervals(Query q, int qp, pair<int, int> cur_al_interval, mytype log_beam, int cur_nucl_idx, mytype *max_vit, bool ***is_cur_act_state, pair<mytype,mytype> ***vit_fw_cur_act_states, int ***best_pred_idx_cur_act_states);
  void                          calc_act_D_intervals_left_endpoint(ProfileHMM *phmm, int type, Query q, int qp, pair<int, int> cur_al_interval, int cur_nucl_idx, mytype cur_beam_thresh, bool ***is_cur_act_state, pair<mytype, mytype> ***vit_fw_cur_act_states, int ***best_pred_idx_cur_act_states);
  void                          calc_act_D_intervals_right_endpoint(ProfileHMM *phmm, int type, Query q, int qp, pair<int, int> cur_al_interval, int cur_nucl_idx, int cur_act_int_leftt, mytype cur_beam_thresh, bool ***is_cur_act_state, pair<mytype, mytype> ***vit_fw_cur_act_states, int ***best_pred_idx_cur_act_states);
  void                          calc_act_D_intervals_right_endpoint_last_qp(ProfileHMM *phmm, int type, Query q, int qp, pair<int, int> cur_al_interval, int cur_nucl_idx, int cur_act_int_left, bool ***is_cur_act_state, pair<mytype, mytype> ***vit_fw_cur_act_states, int ***best_pred_idx_cur_act_states);
  vector<ActiveState*> save_act_states (ProfileHMM* phmm, int type, pair<int, int> *cur_act_int, mytype cur_beam_thresh, bool ***is_cur_act_state, pair<mytype, mytype> ***vit_fw_cur_act_states, int ***best_pred_idx_cur_act_states);
  void                 reinit_tmp_arrays(bool ***is_cur_act_state, pair<mytype, mytype> ***vit_fw_cur_act_states, int ***best_pred_idx_cur_act_states);
  int                  most_left_act_col(Query q, int qp);
  int                  most_right_act_col(Query q, int qp);
  int                  num_active_subtypes(multi_array<bool,2> *act_subtypes, StateType max_act_type) const;
  bool                 has_active_succ(State* cur_state, int qp_idx);
  void                 bw_rec(State *cur_state, string query, int qp, int cur_nucl_idx, mytype *backward);
  map<string, mytype>  post_prob_rec(int qp, mytype fw_score);  
  //void        delete_special_states_transitions();
  void act_intervals_out(int qp, mytype max_vit, mytype log_beam);
  void act_intervals_out_for_plot(Query q);
};

ostream& operator<<(ostream& s, const HMM& h);

/******************************************************\
 *               class ActiveState                   *
\******************************************************/

class ActiveState {

public:
  ActiveState(State* state, State* best_pred, mytype vit, mytype fw, mytype bw);
  ~ActiveState(){};
  State*  state()     const { return m_state; };
  State*  best_pred() const { return m_best_pred; };
  mytype  viterbi()   const { return m_viterbi; };
  mytype  forward()   const { return m_forward; };
  mytype  backward()  const { return m_backward; };
  mytype &backward()        { return m_backward; };


private:
  State* m_state;
  State* m_best_pred;
  mytype m_viterbi;
  mytype m_forward;
  mytype m_backward;

};

#endif        // _HMM_H
