// ~/jpHMM/src/hmm_algorithms.cpp
/**************************************************************
 File name	: hmm_algorithms.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Implementation of all jpHMM algorithms:
  - Viterbi and Forward 
    (active states determined with Beam-search algorithm or by prealignment with BLAT)
  - Viterbi backtracking
  - Backward
  - Posterior probabilities
  - Pruning of recombination prediction for extended circular sequences to original sequence
  - Call of output functions

 Declaration of functions in hmm.h
***************************************************************/

#include "std_source.h"
#include "error.h"
#include "alphabet.h"
#include "const_def.h"
#include "query.h"
#include "alignment.h"
#include "emTransProfile.h"
#include "profileHMM.h"
#include "hmm.h"
#include <ios>
#include <iomanip>
using namespace std;

const bool OUTPUT_DEBUG = false;


/*********************************
 *
 *   Algorithms for class HMM.
 *
**********************************/

/*****************************************************//**
 Clear active states.

 Clear/Initialize active states for viterbi algorithm.
*********************************************************/
void HMM::clear_act_states(Query q) {

  string query = q.str();
  int    query_size = (int)query.size();
 
  //! clear:
  for (int i=-1; i<query_size; ++i) {
    for (int p=0; p<m_num_phmms; ++p) {
      for (int type = (int)Match; type<= (int)Delete; ++type) {
        int cur_size = m_act_states[i][p][type].size();
	for (int k=0; k<cur_size; ++k)
	  delete m_act_states[i][p][type][k];
        m_act_states[i][p][type].clear();
      }
    }
    for (int s=0; s<NUM_SPEC_STATES; ++s) {
      delete m_act_spec_states[i][s];
    }
  }
  multi_array<ActiveState*, 2>::extent_gen         ext_act_spec_states;
  multi_array<vector<ActiveState*>, 3>::extent_gen ext_act_states;
  multi_array<pair<int,int>, 3>::extent_gen        ext_act_intervals;

  m_act_spec_states.resize(ext_act_spec_states[1][1]);
  m_act_states.resize(ext_act_states[1][1][1]);
  m_act_intervals.resize(ext_act_intervals[1][1][1]);


  if (Const_Def::jump_version() == linear) {
    for (int i=-1; i<query_size; ++i) {
      for (int type = (int)Jump_Match; type<= (int)Jump_Delete; ++type) {
	int cur_size = m_act_jump_states[i][type].size();
	for (int k=0; k<cur_size; ++k)
	  delete m_act_jump_states[i][type][k];
	
	m_act_jump_states[i][type].clear();
      }
    }
    
    multi_array<vector<ActiveState*>, 2>::extent_gen ext_act_jump_states;
    multi_array<pair<int,int>, 2>::extent_gen        ext_act_jump_intervals;

    m_act_jump_states.resize(ext_act_jump_states[1][1]);
    m_act_jump_intervals.resize(ext_act_jump_intervals[1][1]);
  }
}

/*********************************************************************************************************************************//**
  Viterbi and forward algorithm.

  Computation of Viterbi and forward table:

  For each query sequence position (qp) only a certain set of 'active states' is considered.
    (-> beam-search algorithm (for expl. see Schultz et al. 2006)).\n
     Active states.
      - qp = -1:                   1.) BEGIN, 2.) N_DELETE, 3.) all DELETEs (except last of each subtype).
      - qp = 0:                    1.) all MATCHes, 2.) N_INSERT, 3.) all DELETEs (except first and last of each subtype).
      - qp = 1,...,(query_size-2): 1.) all MATCHes, 2.) all INSERTs, N_INSERT, C_INSERT, all DELETEs (except first and last of each subtype).
      - qp = query_size-1:         1.) all MATCHes, 2.) C_INSERT, 3.) all DELETEs (except first of each subtype), 4.) C_DELETE, 5.) END.
     Beam-search condition.\n
      - qp = -1,...,query_size-2: cur_vit >= max_vit + log_beam.
      - qp = query_size-1:         no beam-seach criterion, to avoid 'lost' pathes (not ending in END).

   Variables.\n
     Active states for each state type of each subtype at every query sequence position.
     - m_act_states:         active states (Match, Insert, Delete).
     - m_act_spec_states: special active states (Begin, N_Delete, ...).
     - m_act_intervals:     interval of active alignment columns (->first = first active state,->second = last active state)

   @param[in] q Query object
   @param[in] act_al_columns active alignment columns for each qp = columns in MSA to which qp can be aligned to (can be 1 - len(MSA))
   @param[out] true/false viterbi algorithm ran successfully?
**************************************************************************************************************************************/
bool HMM::vit_and_fw_algo(Query *q,
			  vector<pair<int, int> > act_al_columns) {
 
  if (OUTPUT_DEBUG)
    ofstream ACT_INT_OUT(Const_Def::active_state_intervals_file_name().c_str(), ios::out);
  
  Alphabet *alphabet = Const_Def::alphabet();

  //! query sequence:
  string query = q->str();
  int    query_size = (int)query.size();
  
  //! alignment columns (interval) aligned to query sequence positions 
  //! (with BLAT. if this option isn't chosen, aligned intervals are always [q->start_in_al(), q->end_in_al()])
  
  vector<pair<int, int> > al_intervals;

  if (q->prog_version() == blat) {
    al_intervals = ext_act_al_intervals((*q), act_al_columns);
  }
  else {
    pair<int, int> cur_int = make_pair (q->start_in_al(), q->end_in_al());
    al_intervals.assign(query_size+1, cur_int);
  }
  
  //! Determine if beam-search restriction is necessary for Viterbi algorithm: 
  //! max memory for Viterbi table is 1GB
  //! -> count number of active states in al_intervals (assuming that all columns in all subtypes are consensus columns):
  mytype num_act_columns = al_intervals[0].second - al_intervals[0].first + 1; //! for qp=-1
  for (int qp=0; qp<query_size; ++qp) {
    pair<mytype, mytype> cur_al_interval = al_intervals[qp];
    mytype cur_size = cur_al_interval.second - cur_al_interval.first + 1;
    num_act_columns += cur_size;
  }
  
  mytype num_act_states_per_column = NUM_STATES;
  if (Const_Def::jump_version() == linear)
    num_act_states_per_column += NUM_JUMP_STATES;
  
  mytype mem_per_act_state = 2* sizeof(State*) + 3* sizeof(mytype);
  mytype log_poss_used_mem = log(num_act_columns) + log(m_num_phmms) + log(num_act_states_per_column) + log(mem_per_act_state);
  mytype max_mem           = (mytype) 1073741824; //! 1 GB
  
  if (q->prog_version() == blat && log_poss_used_mem <= log(max_mem)) {
    q->beam_width() = 1e-100; //! no beam-search restriction
  }
  
  mytype log_beam = log(q->beam_width());
  
  //! Dimension of arrays for active states and intervals, and active special states.
  multi_array<ActiveState*, 2>::extent_gen         ext_act_spec_states;
  multi_array<vector<ActiveState*>, 3>::extent_gen ext_act_states;
  multi_array<pair<int,int>, 3>::extent_gen        ext_act_intervals;

  typedef multi_array<ActiveState*, 2>::extent_range         range_act_spec_states;
  typedef multi_array<vector<ActiveState*>, 3>::extent_range range_act_states;
  typedef multi_array<pair<int,int>, 3>::extent_range        range_act_intervals;

  m_act_spec_states.resize(ext_act_spec_states[range_act_spec_states(-1,query_size)][NUM_SPEC_STATES]);
  m_act_states.resize(ext_act_states[range_act_states(-1,query_size)][m_num_phmms][NUM_STATES]);
  m_act_intervals.resize(ext_act_intervals[range_act_intervals(-1,query_size)][m_num_phmms][NUM_STATES]);

  //! Initialization.
  for (int qp=-1; qp<query_size; ++qp) {
    for (int s=0; s<m_num_phmms; ++s)
      for (int type = (int)Match; type<= (int)Delete; ++type)
        m_act_intervals[qp][s][type] = make_pair (-1, -1);
    for (int i=0; i<NUM_SPEC_STATES;++i)
      m_act_spec_states[qp][i] = 0;
  }


  if (Const_Def::jump_version() == linear) {
    //! Dimension of arrays for active jump states and intervals.
    multi_array<vector<ActiveState*>, 2>::extent_gen ext_act_jump_states;
    multi_array<pair<int,int>, 2>::extent_gen        ext_act_jump_intervals;
    typedef multi_array<vector<ActiveState*>, 2>::extent_range range_act_jump_states;
    typedef multi_array<pair<int,int>, 2>::extent_range        range_act_jump_intervals;
    m_act_jump_states.resize(ext_act_jump_states[range_act_jump_states(-1,query_size)][NUM_JUMP_STATES]);
    m_act_jump_intervals.resize(ext_act_jump_intervals[range_act_jump_intervals(-1,query_size)][NUM_JUMP_STATES]);
    
    //! Initialization.
    for (int qp=-1; qp<query_size; ++qp) {
      for (int i=(int)Jump_Match; i<=(int)Jump_Delete;++i)
        m_act_jump_intervals[qp][i] = make_pair (-1, -1);
    }
  }
  
  //! Temporary arrays for active states for current query position cp and preceding query position qp-1:
  /*!   each state has a certain position in the list of all states.
        here: in 1st array, all active states for qp-1 are marked,
              in 2nd array, all active states for qp are marked      */
  bool **is_cur_act_state = new bool*[2];

  //! viterbi and forward values of active states (in is_cur_act_state).
  pair<mytype,mytype> **vit_fw_cur_act_states = new pair<mytype, mytype>*[2];
  
  //! for each active state (in is_cur_act_state) the index of the best predecessor in the list of all states is given.
  int **best_pred_idx_cur_act_states = new int*[2];
  
  for (int i=0; i<2; ++i) {
    is_cur_act_state[i]             = new bool[m_num_states];
    vit_fw_cur_act_states[i]        = new pair<mytype, mytype>[m_num_states];
    best_pred_idx_cur_act_states[i] = new int[m_num_states];

    //! Initialization of tmp arrays.
    for (int j=0; j<m_num_states; ++j) {
      is_cur_act_state[i][j]             = false;
      vit_fw_cur_act_states[i][j]        = make_pair (-1,-1);
      best_pred_idx_cur_act_states[i][j] = -1;
    }
  }

  const int cur_act_array_idx = 1; //! [1] = array with information about active states for current qp.


  //!---------------------------------------------------------------------------
  //! Viterbi algorithm for query pos -1 (in front of first query position).
  //!---------------------------------------------------------------------------
  int qp = -1;
  
  //! Initialization.
  mytype cur_vit = -INF;     //! current Viterbi value.
  mytype cur_fw  = -INF;     //! current forward value.
  mytype max_vit = 0;        //! maximal Viterbi value for current qp.
  State* cur_best_pred = 0;  //! current best predecessor of a certain state.

  //! aligned alignment interval for current qp (if blat chosen. else: [q->start_in_al(), q->end_in_al()])
  pair<int, int> cur_al_interval = al_intervals[qp+1];

  //! BEGIN.
  State* cur_state    = begin;
  m_act_spec_states[qp][m_spec_state_to_idx[cur_state]] = new ActiveState(cur_state, 0, 0, 0, -INF);
  int cur_state_id = cur_state->id();

  is_cur_act_state[cur_act_array_idx][cur_state_id]      = true;
  vit_fw_cur_act_states[cur_act_array_idx][cur_state_id] = make_pair (0, 0); //! both, viterbi and Fw value = 0

  //! N_DELETE.
  vit_and_fw_spec_state(n_delete, qp, 0, &max_vit, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);

  //! Current threshold for Viterbi values in beam search algorithm.
  mytype cur_beam_thresh = log_beam + max_vit;

  //! DELETEs.
  int cur_type = (int)Delete;
  //! Left endpoint of active interval is q->start_in_al().
  pair<int, int> cur_act_int =  al_intervals[qp+1];
  for (int s=0; s<m_num_phmms; ++s) {
    //! Right endpoint of active interval.
    calc_act_D_intervals_right_endpoint(m_phmms[s], cur_type, *q, qp, cur_al_interval, 0, cur_act_int.first, cur_beam_thresh, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
    cur_act_int.second = m_act_intervals[qp][s][cur_type].second;

    //! Save list of active states and active interval/
    m_act_states[qp][s][cur_type]    = save_act_states(m_phmms[s], cur_type, &cur_act_int, cur_beam_thresh, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
    m_act_intervals[qp][s][cur_type] = cur_act_int;
  }

  if (Const_Def::jump_version() == linear) {
    //! jump state type Match (successors of Matches, Inserts and Deletes (here for qp=-1 only Deletes) and predecessors of Matches).
    vit_and_fw_algo_jump_states(Jump_Match, *q, qp, 0, &max_vit, log_beam, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
  }
  
  //! Tmp arrays values for current qp set to qp-1, reinitialization of tmp arrays for qp.
  reinit_tmp_arrays(&is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
  if (OUTPUT_DEBUG)
    act_intervals_out(qp, max_vit, log_beam);
 
  //!---------------------------------------------------------------------------
  //! Viterbi algorithm for each query sequence position (0,...,query_size-1).
  //!---------------------------------------------------------------------------
  int cur_q_end_in_al = q->end_in_al();
  for (qp = 0; qp < query_size-1; ++qp) {
    //! Nucleotide at query position qp and its index in the alphabet (A, C, G, T, X).
    char cur_nucl     = query[qp];
    int  cur_nucl_idx = alphabet->idx(cur_nucl);

    //! aligned alignment interval for current qp (if blat chosen. else: [q->start_in_al(), q->end_in_al()])
    cur_al_interval = al_intervals[qp+1];

    //! Initialization.
    cur_vit = cur_fw = max_vit = -INF;
    cur_best_pred = 0;

    //! MATCHes and INSERTs.
    multi_array<pair<int,int>, 2> cur_act_M_and_I_int = calc_act_M_and_I_intervals(*q, qp, cur_al_interval, log_beam, cur_nucl_idx, &max_vit, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);

    //! Current threshold for Viterbi values in beam search algorithm (max_vit = current local max. Viterbi value)
    cur_beam_thresh = log_beam + max_vit;
    //! N_INSERT and C_INSERT.
    if (qp < query_size-1 - q->ext())
      vit_and_fw_spec_state(n_insert, qp, cur_nucl_idx, &max_vit, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
    if (qp > 0 + q->ext())
      vit_and_fw_spec_state(c_insert, qp, cur_nucl_idx, &max_vit, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);

    //! Current threshold for Viterbi values in beam search algorithm (max_vit = current local max. Viterbi value)
    cur_beam_thresh = log_beam + max_vit;

    //! Save active MATCHes and INSERTs (that satisfy beam search criterion).
    //!  for qp=0, only MATCHes are active => maximal active emitting state type = Match
    StateType max_act_type  = Insert;
    if (qp == 0)
      max_act_type  = Match;
    for (int s=0; s<m_num_phmms; ++s) {
      ProfileHMM* cur_phmm = m_phmms[s];
      for (int type=(int)Match; type<=(int)max_act_type; ++type) {
	pair<int, int> cur_act_int   = cur_act_M_and_I_int[s][type];
	m_act_states[qp][s][type]    = save_act_states(cur_phmm, type, &cur_act_int, cur_beam_thresh, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
        m_act_intervals[qp][s][type] = cur_act_int;
      }
    }

    if (Const_Def::jump_version() == linear) {
      //! jump state type Delete (successors of Matches and predecessors of Deletes).
      vit_and_fw_algo_jump_states(Jump_Delete, *q, qp, cur_nucl_idx, &max_vit, log_beam, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
    }

    //! DELETEs.
    //! (only successors of active MATCH states can be active).
    cur_type = (int)Delete;
    for (int s=0; s<m_num_phmms; ++s) {
      ProfileHMM* cur_phmm = m_phmms[s];
      
      //! Left endpoint of active interval.
      calc_act_D_intervals_left_endpoint(cur_phmm, cur_type, *q, qp, cur_al_interval, cur_nucl_idx, cur_beam_thresh, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);

      //! Right endpoint of active interval.
      calc_act_D_intervals_right_endpoint(cur_phmm, cur_type, *q, qp, cur_al_interval, cur_nucl_idx, m_act_intervals[qp][s][cur_type].first, cur_beam_thresh, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);

      //! Save Delete states that satisfy beam search criterion.
      m_act_states[qp][s][cur_type] = save_act_states(cur_phmm, cur_type, &m_act_intervals[qp][s][cur_type], cur_beam_thresh, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
    }

    if (Const_Def::jump_version() == linear) {
      //! jump state type Match (successors of Matches, Inserts and Deletes and predecessors of Matches).
      vit_and_fw_algo_jump_states(Jump_Match, *q, qp, cur_nucl_idx, &max_vit, log_beam, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
    }

    //! count number of active emitting states (Match, Insert, N_Insert, C_Insert)
    //! if none is active (and non-blat version chosen!)
    //! restart viterbi algorithm without any restriction to active intervals
    int cur_num_act_states = 0;
    if (m_act_spec_states[qp][m_spec_state_to_idx[n_insert]] && m_act_spec_states[qp][m_spec_state_to_idx[n_insert]]->viterbi() > -INF)
      cur_num_act_states++;
    else if (m_act_spec_states[qp][m_spec_state_to_idx[c_insert]] && m_act_spec_states[qp][m_spec_state_to_idx[c_insert]]->viterbi() > -INF)
      cur_num_act_states++;
    else {
      for (int s=0; s<m_num_phmms; ++s) {
        for (int type=(int)Match; type<=(int)Insert; ++type)
          cur_num_act_states += m_act_states[qp][s][type].size();
      }
      if (cur_num_act_states == 0) {
        cout << "qp: " << qp << ": no emitting state active" << endl;
	  //! memory deallocation
        for (int i=0; i<2; ++i) {
          delete[] is_cur_act_state[i];
          delete[] vit_fw_cur_act_states[i];
          delete[] best_pred_idx_cur_act_states[i];
        }
        delete[] is_cur_act_state;
        delete[] vit_fw_cur_act_states;
        delete[] best_pred_idx_cur_act_states;

        al_intervals.clear();
	  
        return false;
      }
    }
    //! Tmp arrays values for current qp set to qp-1, reinitialization of tmp arrays for qp.
    reinit_tmp_arrays(&is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
    if (OUTPUT_DEBUG)
      act_intervals_out(qp, max_vit, log_beam);
  }
    
  //!---------------------------------------------------------------------------
  //! Viterbi algorithm for last qp.
  //!   No beam search algorithm applied, all states can be active.
  //!---------------------------------------------------------------------------
  qp              = query_size-1;
  cur_beam_thresh = -INF;
  
  //! aligned alignment interval for current qp (if blat chosen. else: [q->start_in_al(), q->end_in_al()])
  cur_al_interval = al_intervals[qp+1];

  //! Nucleotide at query position qp and its index in the alphabet (A, C, G, T, X).
  char cur_nucl     = query[qp];
  int  cur_nucl_idx = alphabet->idx(cur_nucl);
  
  //! Initialization.
  cur_vit = cur_fw = max_vit = -INF;
  cur_best_pred = 0;


  //! MATCHes. (for last qp, only Match states active as emitting states, not Insert states).
  cur_type = (int)Match;
  
  vector<pair<int,int> > cur_act_M_int (m_num_phmms, make_pair (-1,-1) );
  vector<int>            num_act_states (m_num_phmms, 0); //! number of active states per subtype

  //! Define most left active column for all subtypes at preceding qp (succ. = first possible active state, since succ. M and D must be in succ. col.)
  int prec_most_left_act_col  = most_left_act_col(*q, qp-1)+1;
  int cur_col = prec_most_left_act_col;
  
  //! Matches: successive possible active states are within [prec_most_left_act_col+1, prec_most_right_act_col+1] (successors of M, I, D)
  
  //! first column to be tested must be first potential column for Insert states (see above)! (since have same column as prec. active states)
  //! last column to be tested must be last potential column for Match states (see above)! (prec_most_right_act_col+1 or cur_al_interval.second )

  while (cur_col <= cur_q_end_in_al) {
    for (int s=0; s<m_num_phmms; ++s) {
      ProfileHMM* cur_phmm = m_phmms[s];
      if (cur_phmm->has_state(cur_col, cur_type)) {
        cur_state    = cur_phmm->state(cur_col, cur_type);
        cur_state_id = cur_state->id();
        vit_and_fw_var(cur_state, qp, cur_nucl_idx, &cur_vit, &cur_fw, &cur_best_pred, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
	  
        //! Current state is active for this qp
        ++num_act_states[s];
        is_cur_act_state[cur_act_array_idx][cur_state_id]      = true;
        vit_fw_cur_act_states[cur_act_array_idx][cur_state_id] = make_pair(cur_vit, cur_fw);
        cur_act_M_int[s].second                                = cur_col;
	  
        if (num_act_states[s] == 1) {
	  //! First active state defines left endpoint of active interval.
	  cur_act_M_int[s].first = cur_col;
	}
      }
    }
    ++cur_col;
  }

  //! C_INSERT.
  vit_and_fw_spec_state(c_insert, qp, cur_nucl_idx, &max_vit, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);

  //! Save active MATCHes
  //! and count number of active MATCH states
  int cur_num_act_states = 0;
  for (int s=0; s<m_num_phmms; ++s) {
    pair<int, int> cur_act_int       = cur_act_M_int[s];
    m_act_states[qp][s][cur_type]    = save_act_states(m_phmms[s], cur_type, &cur_act_int, cur_beam_thresh, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
    m_act_intervals[qp][s][cur_type] = cur_act_int;
    cur_num_act_states += m_act_states[qp][s][cur_type].size();
  }

  if (cur_num_act_states == 0 && m_act_spec_states[qp][m_spec_state_to_idx[c_insert]]->viterbi() <= -INF) {
    //! if no MATCH state is active and C_INSERT has -INF as Viterbi value
    //! interrupt Viterbi algorithm
    //! (for Blat version restart (without Blat restrictions, only beam), else recombination prediction impossible)
    //! memory deallocation
    for (int i=0; i<2; ++i) {
      delete[] is_cur_act_state[i];
      delete[] vit_fw_cur_act_states[i];
      delete[] best_pred_idx_cur_act_states[i];
    }
    delete[] is_cur_act_state;
    delete[] vit_fw_cur_act_states;
    delete[] best_pred_idx_cur_act_states;

    al_intervals.clear();
    
    return false;
  }

  if (Const_Def::jump_version() == linear) {
    //! jump state type Delete (successors of Matches and predecessors of Deletes).
    vit_and_fw_algo_jump_states(Jump_Delete, *q, qp, cur_nucl_idx, &max_vit, log_beam, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
  }

  //! DELETEs.
  cur_type        = (int)Delete;

  for (int s=0; s<m_num_phmms; ++s) {
    ProfileHMM* cur_phmm = m_phmms[s];
    //! Left endpoint of active interval.
    calc_act_D_intervals_left_endpoint(cur_phmm, cur_type, *q, qp, cur_al_interval, cur_nucl_idx, cur_beam_thresh, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
    
    //! Right endpoint of active interval:
    calc_act_D_intervals_right_endpoint_last_qp(cur_phmm, cur_type, *q, qp, cur_al_interval, cur_nucl_idx, m_act_intervals[qp][s][cur_type].first, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
    m_act_states[qp][s][cur_type] = save_act_states(cur_phmm, cur_type, &m_act_intervals[qp][s][cur_type], cur_beam_thresh, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
  }
  
  //! last qp: no jumps to mute match jump states possible

  // jump/transition to C_DELETE only possible if Match states are active at last qp
  
  //! C_DELETE.
  vit_and_fw_spec_state(c_delete, qp, cur_nucl_idx, &max_vit, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);
  //! END.
  vit_and_fw_spec_state(end, qp, cur_nucl_idx, &max_vit, &is_cur_act_state, &vit_fw_cur_act_states, &best_pred_idx_cur_act_states);

  if (OUTPUT_DEBUG)
    act_intervals_out(qp, max_vit, log_beam);

  //! Output of active intervals (most left and most right column) for Matlab plot.
  //act_intervals_out_for_plot(q);

  //! memory deallocation
  for (int i=0; i<2; ++i) {
    delete[] is_cur_act_state[i];
    delete[] vit_fw_cur_act_states[i];
    delete[] best_pred_idx_cur_act_states[i];
  }
  delete[] is_cur_act_state;
  delete[] vit_fw_cur_act_states;
  delete[] best_pred_idx_cur_act_states;

  al_intervals.clear();

  return true;
}

/*****************************************************//**
 Extension of aligned (BLAT) alignment columns
 
 for each query position.

 If blat-option chosen
 the query is pre-aligned to the multiple alignment with BLAT
 so, each qp is assigned to a certain interval in the alignment

 Here:
 These alignments are extended such that:
   1.) each interval is extended by +- NUM_MIN_ACT_COL_BLAT
   2.) each subtype has at least NUM_MIN_ACT_COL_PER_ST_BLAT consensus/active columns in this interval
   3.) right endpoint of interval for qp must be >= right endpoint of interval for qp-1
   4.) no gaps between the intervals for successive qp are allowed.

  @param[in] q Current query object
  @param[in] active_al_columns Aligned intervals for each qp
  @param[out] cur_act_intervals Extended aligned intervals for each qp
*********************************************************/
vector<pair<int, int> > HMM::ext_act_al_intervals(Query q, 
			                          vector<pair<int, int> > active_al_columns) {

  if (active_al_columns.empty())
    throw Error ("Blat option chosen, but no aligned alignment columns available.\n");

  vector<pair<int, int> > cur_act_intervals;
  int cur_q_start_in_al = q.start_in_al();
  int cur_q_end_in_al   = q.end_in_al();

  //! for qp == -1 = qp in front of first qp (Begin, N_Deletes and Deletes are active) (Deletion in front of first qp)
  int qp = -1;
  pair<int, int> succ_al_int = active_al_columns[qp+1]; //! aligned interval for qp=0
  pair<int, int> cur_al_int  = make_pair (cur_q_start_in_al, min(cur_q_end_in_al, succ_al_int.second)); //! end of act int for qp = 0;
  cur_act_intervals += cur_al_int;
  
  int query_size = q.str().length();

  pair<int, int> prec_al_int = cur_al_int;
  
  for (qp=0; qp<query_size; ++qp) {
    //! aligned interval for current qp.
    cur_al_int  = active_al_columns[qp];
    //! extend interval by +- NUM_MIN_ACT_COL_BLAT.
    int cur_al_int_left  = cur_al_int.first - NUM_MIN_ACT_COL_BLAT;
    int cur_al_int_right = cur_al_int.second + NUM_MIN_ACT_COL_BLAT;

    //! start and end cannot be < q.start_in_al() or > q.end_in_al().
    cur_al_int.first  = max(cur_q_start_in_al, cur_al_int_left);
    cur_al_int.second = min(cur_q_end_in_al, cur_al_int_right);

    //! aligned interval for succ qp.
    succ_al_int = make_pair (-1, -1);
    if (qp < query_size-1) {
      succ_al_int = active_al_columns[qp+1];
      //! extend interval by +- NUM_MIN_ACT_COL_BLAT.
      int succ_al_int_left  = succ_al_int.first - NUM_MIN_ACT_COL_BLAT;
      int succ_al_int_right = succ_al_int.second + NUM_MIN_ACT_COL_BLAT;
      succ_al_int.first  = max(cur_q_start_in_al, succ_al_int_left);
      succ_al_int.second = min(cur_q_end_in_al, succ_al_int_right);
    }
      
    //! In current interval EACH subtype must at least contain NUM_MIN_ACT_COL_PER_ST_BLAT consensus columns -> test.
    //! check for each subtype if this criterion is satisfied
    //! and count number of subtypes that satisfy this criterion.
    vector<bool> sub_satisfies_cons_col_crit(m_num_phmms, false);
    int num_sub_satisfy_cons_col_crit = 0;
    for (int i=0; i<m_num_phmms; ++i) {
      SubtypeProfile* cur_sub_prof = m_phmms[i]->sub_prof();
      int cur_num_cons_cols = 0;
      for (int j=cur_al_int.first; j<=cur_al_int.second; ++j) {
	cur_num_cons_cols += (int)cur_sub_prof->is_cons_column(j);
        if (cur_num_cons_cols >= NUM_MIN_ACT_COL_PER_ST_BLAT) {
	  sub_satisfies_cons_col_crit[i] = true; // if crit not satisfied for cur_al_int.second, than not satisfied at all for this subtype
	  break;
	}
      }
      num_sub_satisfy_cons_col_crit += (int)sub_satisfies_cons_col_crit[i];
    }
      
    while (num_sub_satisfy_cons_col_crit != m_num_phmms && (cur_al_int.first > cur_q_start_in_al || cur_al_int.second < cur_q_end_in_al)) {
      if (cur_al_int.first > cur_q_start_in_al)
	--cur_al_int.first;
      if (cur_al_int.second < cur_q_end_in_al)
	++cur_al_int.second;

      num_sub_satisfy_cons_col_crit = 0;
      for (int i=0; i<m_num_phmms; ++i) {
	if (!sub_satisfies_cons_col_crit[i]) { //! otherwise not necessary to check since criterion already satisfied for this subtype.
	  SubtypeProfile* cur_sub_prof = m_phmms[i]->sub_prof();
	  int cur_num_cons_cols = 0;
          for (int j=cur_al_int.first; j<=cur_al_int.second; ++j) {
	    cur_num_cons_cols += (int)cur_sub_prof->is_cons_column(j);
            if (cur_num_cons_cols == NUM_MIN_ACT_COL_PER_ST_BLAT) {
	      sub_satisfies_cons_col_crit[i] = true; // if crit not satisfied for cur_al_int.second, than not satisfied at all for this subtype
	      break;
	    }
          }
	}
        num_sub_satisfy_cons_col_crit += (int)sub_satisfies_cons_col_crit[i];
      }
    }
    cur_al_int.first = max(cur_q_start_in_al, cur_al_int.first);
    cur_al_int.second = min(cur_q_end_in_al, cur_al_int.second);

    cur_act_intervals += cur_al_int;
    prec_al_int = cur_al_int;
  }

  return cur_act_intervals;
}

/**********************************************************************************//**
 Viterbi and forward for special state.

 Viterbi and forward variables for a special state (N_D, N_I, C_I, C_D, E) are computed,
 The state is set to be an active state for this qp.
 
  @param[in] cur_state Current state.
  @param[in] qp Current query position.
  @param[in] cur_nucl_idx Nucleotide in query at current qp.
  @param[in,out] max_vit Current maximal Viterbi value for this state type and qp.
  @param[in,out] is_cur_act_state Array with information about active states at qp and qp-1
  @param[in,out] vit_fw_cur_act_states Viterbi values of active states (is_cur_act_state)
  @param[in,out] best_pred_idx_cur_act_states Best predecessor of active states (is_cur_act_state)
**************************************************************************************/
void HMM::vit_and_fw_spec_state(State  *cur_state,
				     int     qp,
				     int     cur_nucl_idx, 
				     mytype *max_vit,
			             bool    ***is_cur_act_state,
			             pair<mytype, mytype> ***vit_fw_cur_act_states,
			             int    ***best_pred_idx_cur_act_states){

  int    cur_state_id  = cur_state->id();
  mytype cur_vit       = -INF;
  mytype cur_fw        = -INF;
  State* cur_best_pred = 0;
  
  const int cur_act_array_idx = 1; //! [1] = array with information about active states for current qp.

  vit_and_fw_var(cur_state, qp, cur_nucl_idx, &cur_vit, &cur_fw, &cur_best_pred, is_cur_act_state, vit_fw_cur_act_states, best_pred_idx_cur_act_states);
  
  m_act_spec_states[qp][m_spec_state_to_idx[cur_state]] = new ActiveState(cur_state, cur_best_pred, cur_vit, cur_fw, -INF);
  
  (*is_cur_act_state)[cur_act_array_idx][cur_state_id]      = true;
  (*vit_fw_cur_act_states)[cur_act_array_idx][cur_state_id] = make_pair (cur_vit, cur_fw);
  
  *max_vit = max(*max_vit, cur_vit);
}

/**********************************************************************************//**
 Viterbi and forward algorithm for jump states.

 Viterbi and forward variables for special states (N_D, N_I, C_I, C_D, E) are computed.
 
  @param[in] cur_state Current state.
  @param[in] qp Current query position.
  @param[in] cur_nucl_idx Nucleotide in query at current qp.
  @param[in,out] max_vit Current maximal Viterbi value for this state type and qp.
  @param[in,out] is_cur_act_state Array with information about active states at qp and qp-1
  @param[in,out] vit_fw_cur_act_states Viterbi values of active states (is_cur_act_state)
  @param[in,out] best_pred_idx_cur_act_states Best predecessor of active states (is_cur_act_state)
**************************************************************************************/
void HMM::vit_and_fw_algo_jump_states(JumpStateType  cur_jump_state_type,
                                      Query   q,
				      int     qp,
				      int     cur_nucl_idx, 
				      mytype *max_vit,
				      mytype  log_beam,
			              bool    ***is_cur_act_state,
			              pair<mytype, mytype> ***vit_fw_cur_act_states,
			              int    ***best_pred_idx_cur_act_states){

  mytype cur_vit       = -INF;
  mytype cur_fw        = -INF;
  State* cur_best_pred = 0;
  
  mytype cur_beam_thresh = log_beam + *max_vit;
  
  int cur_most_left_act_col  = q.end_in_al();
  int cur_most_right_act_col = q.start_in_al();
  
  const int cur_act_array_idx = 1; //! [1] = array with information about active states for current qp.
  
  if (cur_jump_state_type == Jump_Match) {
    //! predecessors are M, I, D states
    //! define most left and right active M, I or D state at qp
    for (int s=0; s<m_num_phmms; ++s) {
      for (int type = (int)Match; type<=(int)Delete; ++type) {
	pair<int, int> cur_act_int = m_act_intervals[qp][s][type];
	int cur_act_int_left = cur_act_int.first;
        int cur_act_int_right = cur_act_int.second;
	
	if (cur_act_int_left > -1)
          cur_most_left_act_col = min(cur_most_left_act_col, cur_act_int_left);
        if (cur_act_int_right > -1)
          cur_most_right_act_col = max(cur_most_right_act_col, cur_act_int_right);
      }
    }
  }
  else { //! jump_state_delete
    //! predecessors are M states
    //! define most left and most right active M state at qp
    int cur_pred_type = (int)Match;
    for (int s=0; s<m_num_phmms; ++s) {
      pair<int, int> cur_act_int = m_act_intervals[qp][s][cur_pred_type];
      int cur_act_int_left = cur_act_int.first;
      int cur_act_int_right = cur_act_int.second;

      if (cur_act_int_left > -1)
        cur_most_left_act_col = min(cur_most_left_act_col, cur_act_int_left);
      if (cur_act_int_right > -1)
        cur_most_right_act_col = max(cur_most_right_act_col, cur_act_int_right);
    }
  }
  cur_most_left_act_col  = max(cur_most_left_act_col, q.start_in_al());
  cur_most_right_act_col = min(cur_most_right_act_col, q.end_in_al());
  cur_most_right_act_col = min(cur_most_right_act_col, m_last_cons_col-1);

  if (cur_most_right_act_col < cur_most_left_act_col) //! no active state
    m_act_jump_intervals[qp][(int)cur_jump_state_type] = make_pair (-1, -1);
  else {
    m_act_jump_intervals[qp][(int)cur_jump_state_type] = make_pair (cur_most_left_act_col, cur_most_right_act_col);

    vector<ActiveState*> cur_act_states;
    for (int i=cur_most_left_act_col; i<=cur_most_right_act_col; i++) {
      ActiveState* cur_act_state = (ActiveState*)0; //! Initialization.
      if (m_jump_states[i].first) {
        State *cur_state;
        pair<State*, State*> cur_jump_states = m_jump_states[i];
        if (cur_jump_state_type == Jump_Match) {
	  cur_state = cur_jump_states.first;
        }
        else {
	  cur_state = cur_jump_states.second;
        }
      
        int          cur_state_id  = cur_state->id();
     
        vit_and_fw_var(cur_state, qp, cur_nucl_idx, &cur_vit, &cur_fw, &cur_best_pred, is_cur_act_state, vit_fw_cur_act_states, best_pred_idx_cur_act_states);
	
        if (cur_vit >= cur_beam_thresh) {
          //! Current state is active for this qp (beam search criterion satisfied).
	  cur_act_state = new ActiveState(cur_state, cur_best_pred, cur_vit, cur_fw, -INF);

	  (*is_cur_act_state)[cur_act_array_idx][cur_state_id]      = true;
	  (*vit_fw_cur_act_states)[cur_act_array_idx][cur_state_id] = make_pair (cur_vit, cur_fw);
        }
        else {
          (*is_cur_act_state)[cur_act_array_idx][cur_state_id]      = false;
          (*vit_fw_cur_act_states)[cur_act_array_idx][cur_state_id] = make_pair (-1, -1);
          (*best_pred_idx_cur_act_states)[cur_act_array_idx][cur_state_id] = -1;
        }
      }
      cur_act_states += cur_act_state;
    }
    m_act_jump_states[qp][(int)cur_jump_state_type] = cur_act_states;
  }
}

/****************************************************************//**
  Viterbi and forward variables.

  Computation of Viterbi and forward variable for a certain state.
  Recursion over all (active) predecessors.

  @param[in] cur_state Current state.
  @param[in] qp Current query position.
  @param[in] cur_nucl_idx Nucleotide in query at current qp.
  @param[in,out] viterbi Viterbi variable of current state.
  @param[in,out] forward Forward variable of current state.
  @param[in,out] cur_best_pred Best predecessor of current state.
  @param[in,out] is_cur_act_state Array with information about active states at qp and qp-1
  @param[in,out] vit_fw_cur_act_states Viterbi values of active states (is_cur_act_state)
  @param[in,out] best_pred_idx_cur_act_states Best predecessor of active states (is_cur_act_state)
********************************************************************/
void HMM::vit_and_fw_var(State  *cur_state,
                         int     qp,
			 int     cur_nucl_idx,
                         mytype *viterbi,
                         mytype *forward,
                         State  **best_pred,
			 bool    ***is_cur_act_state,
		         pair<mytype, mytype> ***vit_fw_cur_act_states,
		         int    ***best_pred_idx_cur_act_states) {

  mytype max_vit = -INF; // log(0);
  mytype log_fw  = -INF;
  *best_pred = 0;
  State* cur_best_pred;
  int    cur_prec_qp = 0;  //! preceding query sequence position (= qp for mute states, qp-1 for Matches and Inserts).
  mytype cur_log_ep  = 0;  //! log of emission probability (= 0 for mute states).

  const int cur_act_array_idx = 1; //! [1] = array with information about active states for current qp.
  int prec_act_array_idx = 0;      //! array with information about active states for preceding qp: [0] (for emitting states -> pred. in prec list) or [1] (for mute states)

  //! preceding query sequence position and emission probability for Matches and Inserts
  if (!cur_state->is_mute()) {
    cur_prec_qp = qp-1;
    cur_log_ep  = cur_state->log_ep()[cur_nucl_idx];
    prec_act_array_idx = 0;
  }
  else {
    cur_prec_qp = qp;
    cur_log_ep  = 0;
    prec_act_array_idx = 1;
  }

  //! Viterbi recursion over all predecessors.
  FOREACH( Transition* cur_transition, cur_state->pred() ) {
    State* cur_pred    = cur_transition->wherefrom();

    if (cur_pred->is_special() || (cur_pred->column() >= m_first_cons_col && cur_pred->column() <= m_last_cons_col)) { // for different states different consensus columns. start at same column
      int    cur_pred_id = cur_pred->id();
    
      if ((*is_cur_act_state)[prec_act_array_idx][cur_pred_id]) {
        pair<mytype, mytype> cur_prec_vit_fw = (*vit_fw_cur_act_states)[prec_act_array_idx][cur_pred_id];
        mytype cur_prec_vit = cur_prec_vit_fw.first;
        mytype cur_prec_fw  = cur_prec_vit_fw.second;
	
        mytype cur_log_tp   = cur_transition->log_p();

        //! Calculation of current viterbi and forward variable.
        mytype cur_log_fw = cur_prec_fw + cur_log_tp;
        mytype log_vit    = cur_prec_vit + cur_log_tp;
	
        if (log_fw > cur_log_fw)
          log_fw = log_fw + log(1 + exp(cur_log_fw-log_fw));
        else
          log_fw = cur_log_fw + log(1 + exp(log_fw-cur_log_fw));

        //! Viterbi variable and best predecessor.
        if (log_vit > max_vit) {
          max_vit       = log_vit;
          *best_pred    = cur_pred;
	  cur_best_pred = cur_pred;
        }
      }
    }
  }
  *viterbi = max_vit + cur_log_ep;
  if ((*best_pred))
    (*best_pred_idx_cur_act_states)[cur_act_array_idx][cur_state->id()] = (*best_pred)->id();

  //! addition of log(emission probability)
  log_fw  += cur_log_ep;
  *forward = log_fw;
}

/***************************************************************************//**
  Calculation of active MATCH and INSERT state intervals.
  
  For each subtype and each state type (MATCH and INSERT)
  the column of the current first and the last active state is calculated.
  Thereby, for each state the Viterbi and forward variable is computed
  and the state is active, if the beam search criterion is satisfied.
  (Beam-search criterion is reassessed later, when local Viterbi max. of all tmp. active states determined).
  
  The left endpoint of the interval is the most left active column for qp-1
  Starting in that column, for each state the Viterbi value is calculated and the beam-search criterion is checked.
  The column number is incremented until the beam-search criterion cannot be satisfied by any state in any subtype
  For each subtype and state tyoe the right endpoint of the interval is the last column where the beam-search criterion is satisfied.

  @param[in]  qp Current query position.
  @param[in]  log_beam Logarithm of beam width.
  @param[in]  cur_nucl_idx Index of current nucleotide in alphabet.
  @param[in,out]  max_vit Maximal Viterbi value for current qp.
  @param[in,out]  is_cur_act_state Array with information about active states at qp and qp-1
  @param[in,out]  vit_fw_cur_act_states Viterbi values of active states (is_cur_act_state)
  @param[in,out]  best_pred_idx_cur_act_states Best predecessor of active states (is_cur_act_state)
  @param[out] cur_act_int Calculated active intervals for M and I states for qp.
*******************************************************************************/
multi_array<pair<int,int>, 2> HMM::calc_act_M_and_I_intervals(Query q,
							      int     qp,
							      pair<int, int> cur_al_interval,
	 	                                              mytype  log_beam,
			                                      int     cur_nucl_idx,
				                              mytype *max_vit,
			                                      bool   ***is_cur_act_state,
		                                              pair<mytype, mytype> ***vit_fw_cur_act_states,
		                                              int    ***best_pred_idx_cur_act_states) {

  mytype cur_vit = -INF;
  mytype cur_fw  = -INF;
  State* cur_best_pred = 0;
  
  const int cur_act_array_idx = 1; //! [1] = array with information about active states for current qp.
  
  //! Current threshold for Viterbi values in beam search algorithm.
  mytype cur_beam_thresh = log_beam + *max_vit;
  
  //! Active emitting state types: 
  //!  usually MATCH and INSERT          => maximal active emitting state type = Insert
  //!  for qp=0, only MATCHes are active => maximal active emitting state type = Match
  StateType max_act_type  = Insert;
  int       num_act_types = 2;
  if (qp == 0) {
    max_act_type  = Match;
    num_act_types = 1;
  }

  //! Define active subtypes for each state type (MATCH/INSERT)
  multi_array<bool,2> act_subtypes (boost::extents[m_num_phmms][num_act_types]);    
  //! Number of active states per subtype.
  multi_array<int, 2> num_act_states (boost::extents[m_num_phmms][num_act_types]);
  //! Current active interval per subtype and state type.
  multi_array<pair<int,int>, 2> cur_act_int (boost::extents[m_num_phmms][num_act_types]);
  
  //! Number of active subtypes * state types:
  int num_act_subtypes = m_num_phmms*(num_act_types);

  //! Initialization.  
  for (int s=0; s<m_num_phmms; ++s) {
    for (int type=(int)Match; type<=(int)max_act_type; ++type) {
       act_subtypes[s][type]   = true;
       num_act_states[s][type] = 0;
       cur_act_int[s][type]    = make_pair (-1, -1);
    }
  }

  //!-------------------------------------------------------------------------------------------------
  //! Viterbi recursion, incl. beam search algorithm.
  /*!  for each subtype and each state type the recursion is continued until it is defined as non-active:
         a state type stays active if 
          1.) it satisfies the beamsearch criterion
              (-> saved as active state) OR
          2.) it's successor can be active for this query pos. (if succ has active predecessors)
	      (-> saved as 0/empty active state) */
  //!-------------------------------------------------------------------------------------------------

  //! Define most left active column for all subtypes at preceding qp (= column of first possible active Insert state, succ. = column of first possible active Match state)
  int prec_most_left_act_col  = most_left_act_col(q, qp-1);
  int prec_most_right_act_col = most_right_act_col(q, qp-1);
  
  //! Matches: successive potential active states are within [prec_most_left_act_col+1, prec_most_right_act_col+1]
  //! Inserts: successive potential active states are within [prec_most_left_act_col, prec_most_right_act_col] (if these cols are determined by M or I states. D->I doesn't exist)
  int cur_col = max(cur_al_interval.first, prec_most_left_act_col);
  int potential_right_endpoint = cur_al_interval.second;
  
  //! first column to be tested must be first potential column for Insert states (see above)! (since have same column as prec. active states)
  //! last column to be tested must be last potential column for Match states (see above)! (prec_most_right_act_col+1 or cur_al_interval.second )

  if (cur_col > potential_right_endpoint)
    potential_right_endpoint = q.end_in_al();
  if (cur_col > prec_most_right_act_col)
    cur_col = prec_most_left_act_col;
  while (num_act_subtypes > 0 && cur_col <= potential_right_endpoint) {
    for (int s=0; s<m_num_phmms; ++s) {
      ProfileHMM* cur_phmm = m_phmms[s];
      for (int type = (int)Match; type<=(int)max_act_type; ++type) {
        if (act_subtypes[s][type]) {
          if (cur_phmm->has_state(cur_col, type)) {
            State *cur_state = cur_phmm->state(cur_col, type);
            int cur_state_id = cur_state->id();
	    vit_and_fw_var(cur_state, qp, cur_nucl_idx, &cur_vit, &cur_fw, &cur_best_pred, is_cur_act_state, vit_fw_cur_act_states, best_pred_idx_cur_act_states);
	     
            if (cur_vit >= cur_beam_thresh) {
	      //! Current state is active for this qp (beam search criterion satisfied).
	      (*is_cur_act_state)[cur_act_array_idx][cur_state_id]    = true;
	      (*vit_fw_cur_act_states)[cur_act_array_idx][cur_state_id] = make_pair (cur_vit, cur_fw);

	      ++num_act_states[s][type];	      
	      
              //! Left endpoint of active interval = column of first active state.
	      if (num_act_states[s][type] == 1)
                cur_act_int[s][type].first = cur_col;
	      
	      //! Current right endpoint of interval is current alignment column cur_col.
	      cur_act_int[s][type].second = cur_col;

              //! Current threshold for Viterbi values in beam search algorithm (max_vit = current local max. Viterbi value)
	      *max_vit        = max(*max_vit, cur_vit);
              cur_beam_thresh = log_beam + *max_vit;
	    }
            else {
	      //! Current state not active
	      //! Check if successor has another active predecessor.
	      if (!has_active_succ(cur_state, qp) && cur_col > m_act_intervals[qp-1][s][type].second) {// && q.prog_version() != blat) {
		//! Right endpoint of active interval for qp-1 reached and successor state is not active
		//! => subtype isn't active for this state type.
		act_subtypes[s][type] = false;
	      }
	    }
	  }
        }
      }
    }
    //! Number of remaining active subtypes.
    num_act_subtypes = num_active_subtypes(&act_subtypes, max_act_type);
    ++cur_col;
  }

  //! MATCH successors of N_INSERT (can be active):
  //! column of these MATCH states = q.start_in_al().
  //! these Matches are only active if their column is within the aligned interval.
  if (cur_al_interval.first <= q.start_in_al()) {
    int cur_type = (int)Match;
    cur_col = q.start_in_al();

    for (int s=0; s<m_num_phmms; ++s) {
      ProfileHMM* cur_phmm = m_phmms[s];
      if (cur_phmm->has_state(cur_col, cur_type)) {
        State *cur_state    = cur_phmm->state(cur_col, cur_type);
        vit_and_fw_var(cur_state, qp, cur_nucl_idx, &cur_vit, &cur_fw, &cur_best_pred, is_cur_act_state, vit_fw_cur_act_states, best_pred_idx_cur_act_states);
     
        if (cur_vit >= cur_beam_thresh) {
	  //! Current state is active for this qp (beam search criterion satisfied).
          int    cur_state_id = cur_state->id();
          (*is_cur_act_state)[cur_act_array_idx][cur_state_id]      = true;
	  (*vit_fw_cur_act_states)[cur_act_array_idx][cur_state_id] = make_pair (cur_vit, cur_fw);

	  //! Left endpoint of active interval = q.start_in_al().
	  cur_act_int[s][cur_type].first = cur_col;
	
	  //! If no other MATCH state has been activated yet (see recursion above),
	  //! right endpoint of interval = q.start_in_al().
	  if (cur_act_int[s][cur_type].second == -1)
            cur_act_int[s][cur_type].second = cur_col;
	
          //! Current threshold for Viterbi values in beam search algorithm (max_vit = current local max. Viterbi value)
          *max_vit        = max(*max_vit, cur_vit);
          cur_beam_thresh = log_beam + *max_vit;
        }
      }
    }
  }
  return cur_act_int;
}

/***************************************************************************//**
  Left endpoint of active intervals for DELETEs.
 
  For each subtype the left endpoint of the active interval for the DELETE states is determined.
  This is the column of the most left active DELETE state.
  
  Since all active DELETEs must be successors of active MATCHes for the same qp,
  the algorithm starts at the column of the most left active MATCH state.
  This column is incremented while the beam search criterion is not satisfied.
  For each subtype the left endpoint of the interval
  is the column of the most left DELETE state that satisfies the beam search criterion
  
  For each state the Viterbi and forward variable is calculated
  and the state is set to be active, if the beam search criterion is satisfied.

  @param[in] phmm Current profile HMM.
  @param[in] type Current state type.
  @param[in] qp Current query position.
  @param[in] cur_nucl_idx Nucleotide in query at current qp.
  @param[in] cur_beam_thresh Current accepted threshold for Viterbi value (-INF for last qp).
  @param[in,out] is_cur_act_state Array with information about active states at qp and qp-1
  @param[in,out] vit_fw_cur_act_states Viterbi values of active states (is_cur_act_state)
  @param[in,out] best_pred_idx_cur_act_states Best predecessor of active states (is_cur_act_state)
*******************************************************************************/
void HMM::calc_act_D_intervals_left_endpoint(ProfileHMM *phmm,
			                     int         type, 
					     Query       q,
			                     int         qp,
					     pair<int, int> cur_al_interval,
			                     int         cur_nucl_idx,
			                     mytype      cur_beam_thresh,
			                     bool    ***is_cur_act_state,
			                     pair<mytype, mytype> ***vit_fw_cur_act_states,
			                     int    ***best_pred_idx_cur_act_states) {


  mytype cur_vit       = -INF;
  mytype cur_fw        = -INF;
  State* cur_best_pred = 0;
  int    cur_phmm_to_sub_idx = m_phmm_to_sub_idx[phmm];
  SubtypeProfile* cur_sub_prof = phmm->sub_prof();
  int cur_q_end_in_al = q.end_in_al();
  
  const int cur_act_array_idx = 1; //! [1] = array with information about active states for current qp.
  
  //! Determine alignment column of most left active Match state.
  int most_left_act_col = cur_q_end_in_al;
  int most_right_act_col = q.start_in_al();
  for (int s=0; s<m_num_phmms; ++s) {
    pair<int, int> cur_act_int = m_act_intervals[qp][s][(int)Match];
    int cur_act_int_left = cur_act_int.first;
    int cur_act_int_right = cur_act_int.second;
    if (cur_act_int_left > -1)
      most_left_act_col = min(most_left_act_col, cur_act_int_left);
    if (cur_act_int_right > -1)
      most_right_act_col = max(most_right_act_col, cur_act_int_right);
  }
  int cur_act_int_left = most_left_act_col+1; // D are succ of M or D -> first D must be in succ column of most left act M

  /*! Increment current left interval endpoint of current state type and profile HMM
      while beam search criterion is not satisfied. 
      For last qp, cur_beam_thresh = -INF, so left interval endpoint = most_left_act_col */

  int cur_col = cur_act_int_left; //! first successor (must be > q.start_in_al(), if qp > -1)

  //! Deletes: successors of current active Matches (and Deletes).
  //! Deletes in last cons col aren't active.
  int potential_right_endpoint = min(cur_al_interval.second, cur_q_end_in_al);//-1);
  
  //! first column to be tested must be first active Match column +1.
  //! ignore here beginning (if blat interval or by beam) -> already set up for M and I states
  //! last column to be tested must be last potential column cur_al_interval or exception:
  if (cur_col > potential_right_endpoint) //! active states start behind BLAT interval
    potential_right_endpoint = cur_q_end_in_al-1;

  do {
    cur_act_int_left = cur_col;
    if (phmm->has_state(cur_col, type)) {
      State *cur_state = phmm->state(cur_col, type);
      vit_and_fw_var(cur_state, qp, cur_nucl_idx, &cur_vit, &cur_fw, &cur_best_pred, is_cur_act_state, vit_fw_cur_act_states, best_pred_idx_cur_act_states);

      if (cur_vit < cur_beam_thresh) {
        //! Define state as inactive, if Viterbi value to small.
        int cur_state_id = cur_state->id();

        (*is_cur_act_state)[cur_act_array_idx][cur_state_id]             = false;
        (*vit_fw_cur_act_states)[cur_act_array_idx][cur_state_id]        = make_pair (-1, -1); 
        (*best_pred_idx_cur_act_states)[cur_act_array_idx][cur_state_id] = -1;
      }
    }
    cur_col = (cur_col < cur_q_end_in_al) ? cur_sub_prof->succ_cons_col(cur_col) : m_len;
  }
  while (cur_col <= potential_right_endpoint && cur_vit < cur_beam_thresh); //! only for last qp, Delete in q.end_in_al() can be active

  m_act_intervals[qp][cur_phmm_to_sub_idx][type].first = cur_act_int_left;
}

/************************************************************************************//**
  Right endpoint of active intervals for DELETEs.
 
  For each subtype the right endpoint of the active interval for the DELETE states is determined.
  This is the column of the most right active DELETE state.
  
  Starting at the left endpoint of the current active interval (cur_act_int_left)
  as potential right endpoint of the interval,
  the column is incremented while the beam search criterion is satisfied.
  For each subtype the right endpoint of the interval
  is the column of the most right DELETE state that satisfied the beam search criterion
  
  For each state the Viterbi and forward variable is calculated
  and the state is set to be active, if the beam search criterion is satisfied.

  @param[in] phmm Current profile HMM.
  @param[in] type Current state type.
  @param[in] qp Current query position.
  @param[in] cur_nucl_idx Nucleotide in query at current qp.
  @param[in] cur_act_int_right Current right endpoint of active interval.
  @param[in] cur_beam_thresh Current accepted threshold for Viterbi value.
  @param[in,out] is_cur_act_state Array with information about active states at qp and qp-1
  @param[in,out] vit_fw_cur_act_states Viterbi values of active states (is_cur_act_state)
  @param[in,out] best_pred_idx_cur_act_states Best predecessor of active states (is_cur_act_state)
****************************************************************************************/
void HMM::calc_act_D_intervals_right_endpoint(ProfileHMM *phmm,
				              int    type,
					      Query  q,
				              int    qp,
					      pair<int, int> cur_al_interval,
                                              int    cur_nucl_idx,
                                              int    cur_act_int_left,
			                      mytype cur_beam_thresh,
			                      bool    ***is_cur_act_state,
			                      pair<mytype, mytype> ***vit_fw_cur_act_states,
			                      int    ***best_pred_idx_cur_act_states) {

  mytype cur_vit       = -INF;
  mytype cur_fw        = -INF;
  State* cur_best_pred = 0;
  int    cur_phmm_to_sub_idx = m_phmm_to_sub_idx[phmm];
  SubtypeProfile* cur_sub_prof = phmm->sub_prof();
  int cur_q_end_in_al = q.end_in_al();
  
  const int cur_act_array_idx = 1; //! [1] = array with information about active states for current qp.
  
  boost::array<multi_array<pair<int, int>,3>::index,3> cur_act_int_idx = {{qp,cur_phmm_to_sub_idx,type}};

  bool is_act_sub        = true;  //! subtype is active.

  int  cur_col       = cur_act_int_left;
  int  cur_act_int_right = m_act_intervals(cur_act_int_idx).second;

  //! Deletes: successors of current active Matches (and Deletes).
  //! Deletes in last cons col aren't active.
  int potential_right_endpoint = min(cur_al_interval.second, cur_q_end_in_al-1);
  
  //! first column to be tested must be first active Match column +1.
  //! ignore here beginning (if blat interval or by beam) -> already set up for M and I states
  //! last column to be tested must be last potential column cur_al_interval or exception:
  if (cur_col > potential_right_endpoint) //! active states start behind BLAT interval
    potential_right_endpoint = cur_q_end_in_al-1;
  
  do {
    if (phmm->has_state(cur_col, type)) {
      State *cur_state    = phmm->state(cur_col, type);
      int    cur_state_id = cur_state->id();
      vit_and_fw_var(cur_state, qp, cur_nucl_idx, &cur_vit, &cur_fw, &cur_best_pred, is_cur_act_state, vit_fw_cur_act_states, best_pred_idx_cur_act_states);
      if (cur_vit >= cur_beam_thresh) {
        //! Current state is active for this qp (beam search criterion satisfied).
        (*is_cur_act_state)[cur_act_array_idx][cur_state_id]      = true;
	(*vit_fw_cur_act_states)[cur_act_array_idx][cur_state_id] = make_pair (cur_vit, cur_fw);

	is_act_sub = true;
	
        //! Current right endpoint of interval.
        cur_act_int_right    = cur_col;
      }
      else {
        if (qp == -1)// && q.prog_version() != blat)
	  is_act_sub = false;
	else {
	  //! Check if successor has another active predecessor.
	  //! if current state has active successor, subtype is set to be still active.
	  is_act_sub = has_active_succ(cur_state, qp);
          if (is_act_sub) {// || q.prog_version() == blat) {
	    //! Successor can be active, but current state is inactive.
            (*is_cur_act_state)[cur_act_array_idx][cur_state_id]      = false;
            (*vit_fw_cur_act_states)[cur_act_array_idx][cur_state_id] = make_pair (-1, -1);
	    (*best_pred_idx_cur_act_states)[cur_act_array_idx][cur_state_id] = -1;
      
	    //! Current right endpoint of interval.
	    cur_act_int_right = cur_col;
	  }
	}
      }
    }
    cur_col  = (cur_col < cur_q_end_in_al) ? cur_sub_prof->succ_cons_col(cur_col) : m_len;
  }
  while ((is_act_sub == true) && cur_col <= potential_right_endpoint);
  m_act_intervals(cur_act_int_idx).second = cur_act_int_right;
}

/************************************************************************************//**
  Right endpoint of active intervals for DELETEs for last qp.
 
  For each subtype the right endpoint of the active interval for the DELETE states is determined.
  This is the column of the most right active DELETE state.
  
  Starting at the left endpoint of the current active interval (cur_act_int_left)
  as potential right endpoint of the interval,
  the column is incremented until the end of alignment is reached (no beam search criterion at last qp).
  
  For each state the Viterbi and forward variable is calculated.

  @param[in] phmm Current profile HMM.
  @param[in] type Current state type.
  @param[in] qp Current query position.
  @param[in] cur_nucl_idx Nucleotide in query at current qp.
  @param[in] cur_act_int_right Current right endpoint of active interval.
  @param[in,out] is_cur_act_state Array with information about active states at qp and qp-1
  @param[in,out] vit_fw_cur_act_states Viterbi values of active states (is_cur_act_state)
  @param[in,out] best_pred_idx_cur_act_states Best predecessor of active states (is_cur_act_state)
****************************************************************************************/
void HMM::calc_act_D_intervals_right_endpoint_last_qp(ProfileHMM *phmm,
				                      int         type,
						      Query       q, 
				                      int         qp,
						      pair<int, int> cur_al_interval,
			                              int         cur_nucl_idx,
			                              int         cur_act_int_left,
			                              bool    ***is_cur_act_state,
			                              pair<mytype, mytype> ***vit_fw_cur_act_states,
			                              int    ***best_pred_idx_cur_act_states) {
  int    cur_col        = cur_act_int_left;
  mytype cur_vit        = -INF;
  mytype cur_fw         = -INF;
  State* cur_best_pred  = 0;
  int    cur_phmm_to_sub_idx = m_phmm_to_sub_idx[phmm];
  SubtypeProfile* cur_sub_prof = phmm->sub_prof();
  
  const int cur_act_array_idx = 1; //! [1] = array with information about active states for current qp.
  
  boost::array<multi_array<vector<ActiveState*>,3>::index,3> cur_act_int_idx = {{qp,cur_phmm_to_sub_idx,type}};

  int cur_act_int_right = m_act_intervals(cur_act_int_idx).second;
  
  //! Deletes: successors of current active Matches (and Deletes).
  //! Deletes in last cons col aren't active.
  int potential_right_endpoint = min(cur_al_interval.second, q.end_in_al());
  
  //! first column to be tested must be first active Match column +1.
  //! ignore here beginning (if blat interval or by beam) -> already set up for M and I states
  //! last column to be tested must be last potential column cur_al_interval or exception:
  
  //! Viterbi recursion for all(!) following Deletes, no beam search for last qp
  do {
    if (phmm->has_state(cur_col, type)) {
      State *cur_state    = phmm->state(cur_col, type);
      int    cur_state_id = cur_state->id();
      vit_and_fw_var(cur_state, qp, cur_nucl_idx, &cur_vit, &cur_fw, &cur_best_pred, is_cur_act_state, vit_fw_cur_act_states, best_pred_idx_cur_act_states);

      (*is_cur_act_state)[cur_act_array_idx][cur_state_id]      = true;
      (*vit_fw_cur_act_states)[cur_act_array_idx][cur_state_id] = make_pair (cur_vit, cur_fw);

      //! Current right endpoint of interval.
      cur_act_int_right = cur_col;
    }
    cur_col  = (cur_col < q.end_in_al()) ? cur_sub_prof->succ_cons_col(cur_col) : m_len;
  }
  while (cur_col <= potential_right_endpoint);
  m_act_intervals(cur_act_int_idx).second = cur_act_int_right;
}

/*********************************************************************//**
  Save active states for each subtype and state type.

  For each state that belongs to a column within the given active interval
  check the Viterbi variable (vit_fw_cur_act_states).
  If the beam search criterion is satisfied an object ActiveState() is created,
  i.e. the state is saved as active state for the current query position,
  and also its forward variable and the best predecessor is saved for 
  Viterbi backtracking and the backward algorithm.
  If the beam search criterion isn't satisfied
  an empty pointer is saved.

  @param[in] phmm Current profile HMM.
  @param[in] type Current state type.
  @param[in,out] cur_act_int Current active interval.
  @param[in,out] is_cur_act_state Array with information about active states at qp and qp-1
  @param[in,out] vit_fw_cur_act_states Viterbi values of active states (is_cur_act_state)
  @param[in,out] best_pred_idx_cur_act_states Best predecessor of active states (is_cur_act_state)
  @param[out] act_states Active states within current active interval for current phmm and state type
*************************************************************************/
vector<ActiveState*> HMM::save_act_states (ProfileHMM*     phmm,
					   int             type, 
                                           pair<int, int> *cur_act_int,
					   mytype          cur_beam_thresh,
					   bool   ***is_cur_act_state,
			                   pair<mytype, mytype> ***vit_fw_cur_act_states,
			                   int    ***best_pred_idx_cur_act_states) {

  vector<ActiveState*> act_states;

  const int cur_act_array_idx = 1; //! [1] = array with information about active states for current qp.
  
  int orig_int_left  = (*cur_act_int).first;
  int orig_int_right = (*cur_act_int).second;
  
  if (orig_int_left < 0) {
    (*cur_act_int) = make_pair(-1,-1);
  }
  else {
    int pred_col = orig_int_left;
    for (int cur_col = orig_int_left; cur_col <= orig_int_right; ++cur_col) {
      if (phmm->has_state(cur_col, type)) {
        State *cur_state    = phmm->state(cur_col, type);
        int    cur_state_id = cur_state->id();
	
        if ((*is_cur_act_state)[cur_act_array_idx][cur_state_id]) {
	  //! State was originally set to be active.
	  //! Check Viterbi variable again, since now the local maximal Viterbi variable of all active states is known.
	  pair<mytype, mytype> cur_vit_fw = (*vit_fw_cur_act_states)[cur_act_array_idx][cur_state_id];
          mytype cur_vit = cur_vit_fw.first;
      
          if (cur_vit >= cur_beam_thresh) {
            //! Current state is active for this qp (beam search criterion satisfied).
	    mytype cur_fw = cur_vit_fw.second;
	    
            //! Create an Active State.
	    ActiveState* cur_act_state = new ActiveState(cur_state, m_idx_to_state[(*best_pred_idx_cur_act_states)[cur_act_array_idx][cur_state_id]], cur_vit, cur_fw, -INF);
            //! Add empty states for preceding non-conserved alignment columns or non-active states.
            if (!act_states.empty()) {
              for (int i=pred_col; i<cur_col; ++i)
	        act_states += (ActiveState*) 0;
	    }
	    act_states += cur_act_state;    
	    pred_col    = cur_col+1;
	  }
	  else { //! set to be inactive.
            (*is_cur_act_state)[cur_act_array_idx][cur_state_id]      = false;
            (*vit_fw_cur_act_states)[cur_act_array_idx][cur_state_id] = make_pair (-1, -1);
            (*best_pred_idx_cur_act_states)[cur_act_array_idx][cur_state_id] = -1;
	  }
        }
      }
    }
  
    if (!act_states.empty()) {
      ActiveState* first_act_state = act_states.front();
      ActiveState* last_act_state  = act_states.back();
      int cur_int_left  = first_act_state->state()->column();
      int cur_int_right = last_act_state->state()->column();
      (*cur_act_int) = make_pair(cur_int_left, cur_int_right);
    }
    else
      (*cur_act_int) = make_pair(-1,-1);
  }

  return act_states;
}

/***************************************************************************//**
  Initialize multi arrays for active states, viterbi & forward values, etc.

  2-dim. arrays:
    [0]: active states at qp-1
    [1]: active states at qp
  
  update in this function:
    [0] <- [1]
    [1] will be reinitialized

  @param[in,out] is_cur_act_state Array with information about active states at qp and qp-1
  @param[in,out] vit_fw_cur_act_states Viterbi values of active states (is_cur_act_state)
  @param[in,out] best_pred_idx_cur_act_states Best predecessor of active states (is_cur_act_state)
*******************************************************************************/
void HMM::reinit_tmp_arrays(bool                 ***is_cur_act_state, 
			    pair<mytype, mytype> ***vit_fw_cur_act_states, 
			    int                  ***best_pred_idx_cur_act_states) {
  
  const int prec_act_array_idx = 0; //! [0] = array with information about active states for preceding qp.
  const int cur_act_array_idx  = 1; //! [1] = array with information about active states for current qp.
  
  bool                 *dummy_is_cur_act_state             = (*is_cur_act_state)[prec_act_array_idx];
  pair<mytype, mytype> *dummy_vit_fw_cur_act_states        = (*vit_fw_cur_act_states)[prec_act_array_idx];
  int                  *dummy_best_pred_idx_cur_act_states = (*best_pred_idx_cur_act_states)[prec_act_array_idx];

  (*is_cur_act_state)[prec_act_array_idx]             = (*is_cur_act_state)[cur_act_array_idx];
  (*is_cur_act_state)[cur_act_array_idx]              = dummy_is_cur_act_state;
  (*vit_fw_cur_act_states)[prec_act_array_idx]        = (*vit_fw_cur_act_states)[cur_act_array_idx];
  (*vit_fw_cur_act_states)[cur_act_array_idx]         = dummy_vit_fw_cur_act_states;
  (*best_pred_idx_cur_act_states)[prec_act_array_idx] = (*best_pred_idx_cur_act_states)[cur_act_array_idx];
  (*best_pred_idx_cur_act_states)[cur_act_array_idx]  = dummy_best_pred_idx_cur_act_states;
  
  //! Reinit. array for succ. qp.
  for (int i=0; i<m_num_states; ++i) {
    (*is_cur_act_state)[cur_act_array_idx][i]             = false;
    (*vit_fw_cur_act_states)[cur_act_array_idx][i]        = make_pair(-1, -1);
    (*best_pred_idx_cur_act_states)[cur_act_array_idx][i] = -1;
  }
}

/**********************************************************************************//**
  Most left active column of M, I and D states at qp.

  @param[in]  qp Current query position.
  @param[out] cur_most_left_act_col Column of most left active state at qp.
**************************************************************************************/
int HMM::most_left_act_col(Query q, int qp) {

  int cur_most_left_act_col = q.end_in_al();

  //! Define most left active column for all subtypes.
  if (qp <= 0) {
    cur_most_left_act_col = q.start_in_al();
  }
  else {
    //! first act. col = most left act col of all state types of all subtypes at qp.
    for (int s=0; s<m_num_phmms; ++s) {
      for (int type = (int)Match; type<=(int)Delete; ++type) {
        int cur_act_int_le = m_act_intervals[qp][s][type].first;
        if (cur_act_int_le > -1)
          cur_most_left_act_col = min(cur_most_left_act_col, cur_act_int_le);
      }
    }
    cur_most_left_act_col = max(cur_most_left_act_col, q.start_in_al());
  }
  return cur_most_left_act_col;
}

/**********************************************************************************//**
  Most right active column of M, I and D states at qp.

  @param[in]  qp Current query position.
  @param[out] cur_most_right_act_col Column of most right active state at qp.
**************************************************************************************/
int HMM::most_right_act_col(Query q, 
			    int   qp) {

  int cur_most_right_act_col = q.start_in_al();

  //! last act. col = most right act col of all state types of all subtypes at qp.
  for (int s=0; s<m_num_phmms; ++s) {
    for (int type = (int)Match; type<=(int)Delete; ++type) {
      int cur_act_int_ri = m_act_intervals[qp][s][type].second;
      if (cur_act_int_ri > -1)
        cur_most_right_act_col = max(cur_most_right_act_col, cur_act_int_ri);
    }
  }
  cur_most_right_act_col = min(cur_most_right_act_col, q.end_in_al());

  return cur_most_right_act_col;
}

/************************************************************************************//**
  Number of active subtypes.
 
  Number of active subtypes (per state type (Match/Insert) for a certain query position.

  @param[in] act_subtypes Array containing information about active subtypes (for different state types).
  @param[in] max_act_type Current maximal active state type (usually Insert, for special qp only Match).
****************************************************************************************/
int HMM::num_active_subtypes(multi_array<bool,2> *act_subtypes,
			     StateType            max_act_type) const {

  int num_act_subtypes = 0;
  //! Ususally MATCH and INSERTs active, but for some qp, only MATCH states are active (= max_act_type).
  for (int type = (int)Match; type<=(int)max_act_type; ++type)
    for (int s=0; s<m_num_phmms; ++s)
      if ((*act_subtypes)[s][type])
        ++num_act_subtypes;

  return num_act_subtypes;
}

/*******************************************************************************//**
  Has active successor.
  
  Test for current state if its successor state (in same profile HMM, same type)
  can be active for the current qp.
  This is the case if a predecessor of the successor in any subtype is active.
  
  @param[in] cur_state Current state.
  @param[in] qp Current query position.
  @param[out] true/false Current state has potential active successor for current qp ?
***********************************************************************************/
bool HMM::has_active_succ(State* cur_state,
                          int    qp) {

  ProfileHMM* cur_phmm = cur_state->phmm();
  SubtypeProfile* cur_sub_prof = cur_phmm->sub_prof();
  int cur_col = cur_state->column();
  
  if (cur_sub_prof->has_succ_cons_col(cur_col)) {
    int cur_succ_col = cur_sub_prof->succ_cons_col(cur_col);

    if (cur_state->is_match()) {
      //! Check MATCH, INSERT, DELETE (at respective preceding col) predecessors in all subtypes.
      for (int p=0; p<m_num_phmms; ++p) {
        int cur_pred_cons_col = m_phmms[p]->sub_prof()->pred_cons_col(cur_succ_col);
        for (int t=(int)Match; t<=(int)Delete; ++t) {
	  if (m_act_intervals[qp-1][p][t].second >= cur_pred_cons_col) {
	    //! Predecessor is within active interval.
	    return true;
	  }
        }
      }
    }
    else if (cur_state->is_insert()) {
      //! Check preceding MATCH state in same subtype for qp-1, 
      if (m_act_intervals[qp-1][m_phmm_to_sub_idx[cur_phmm]][(int)Match].second >= cur_succ_col) {
	//! Predecessor is within active interval.
        return true;
      }
    }
    else {
      //! Check MATCH (at respective preceding col) predecessors in all subtypes.
      for (int p=0; p<m_num_phmms; ++p) {
        if (m_act_intervals[qp][p][(int)Match].second >= m_phmms[p]->sub_prof()->pred_cons_col(cur_succ_col)) {
	  //! Predecessor is within active interval.
	  return true;
        }
      }
    }
  }
  return false;
}

/*************************************************************************//**
  Viterbi backtracking.

  Determination of Viterbi path (= predicted recombination) by backtracking.

  @param[in] al Alignment object.
  @param[in] q Query object. 
*****************************************************************************/
void HMM::vit_back(Alignment *al, 
		   Query      q) {

  const bool OUTPUT_DETAILS = false;
  
  vector<State*> vit_path;             /*! Viterbi state path. */
  vector<int>    vit_path_start_pos;   /*! start positions of segments in predicted recombination. */
  vector<int>    vit_path_end_pos;     /*! end positions of segments in predicted recombination. */
  vector<string> vit_path_subtype;     /*! order of predicted subtypes = predicted recombination pattern. */
  list<string>   pred_subtypes;        /*! list of predicted subtypes. */

  /*! backtracking starts with End state. */
  int          qp            = q.str().length()-1;
  State       *cur_state     = end;
  ActiveState *cur_act_state = m_act_spec_states[qp][m_spec_state_to_idx[cur_state]];

  string       cur_best_st   = m_spec_state_to_str[cur_state];

  /*! last state in viterbi path is End. */
  vit_path += cur_state;

  /*! end position of last segment in predicted recombination is length of query -1. */
  vit_path_end_pos += qp;
  
  bool vit_path_subtype_is_empty = true;
  //! backtracking until Begin state or query position 0 is reached.

  while (!cur_state->is_begin() && qp >= 0) {
    //! change in recombination pattern.
    /*! current best predecessor belongs to different subtype than current predicted subtype. */
    if (vit_path_subtype_is_empty || cur_best_st != vit_path_subtype.back()) {
      if (!cur_state->is_mute()){
        if (!vit_path_subtype_is_empty) {
          /*! add start position of new segment. */
          vit_path_start_pos += qp+1;
          /*! add end position of current segment. */
          vit_path_end_pos   += qp;
        }
        /*! add subtype of new segment. */
        vit_path_subtype   += cur_best_st;
	vit_path_subtype_is_empty = false;
      }
    }
    if (!cur_state->is_mute())
      --qp;

    /*! for current qp determine best predecessor of current state. */
    State* best_pred = cur_act_state->best_pred();
    if (best_pred) {
      if (best_pred->is_special()) {
	cur_best_st   = m_spec_state_to_str[best_pred];
        cur_act_state = m_act_spec_states[qp][m_spec_state_to_idx[best_pred]];
      }
      else if (best_pred->is_jump_state()) {
	if (Const_Def::jump_version() != linear)
          throw Error("jump state type cannot be predecessor if linear jump version is not chosen!");
        StateType     best_pred_type      = best_pred->type();
        JumpStateType best_pred_jump_type = m_type_to_jump_type[best_pred_type];
	cur_best_st   = m_jump_state_to_str[best_pred_jump_type];
	
	pair<int, int> cur_act_int = m_act_jump_intervals[qp][(int)best_pred_jump_type];
        int       cur_act_int_left = cur_act_int.first;

        cur_act_state = m_act_jump_states[qp][(int)best_pred_jump_type][best_pred->column()-cur_act_int_left];
      }
      else {
	ProfileHMM* cur_best_pred_phmm = best_pred->phmm();

        int       cur_best_pred_phmm_idx = m_phmm_to_sub_idx[cur_best_pred_phmm];
	StateType best_pred_type   = best_pred->type();
	pair<int, int> cur_act_int = m_act_intervals[qp][cur_best_pred_phmm_idx][(int)best_pred_type];
        int       cur_act_int_left = cur_act_int.first;

        cur_act_state = m_act_states[qp][cur_best_pred_phmm_idx][(int)best_pred_type][best_pred->column()-cur_act_int_left];
        /*! determine corresponding subtype of best predecessor. */
        cur_best_st   = cur_best_pred_phmm->sub_prof()->sub()->name();
        /*! add current predicted subtype to list of predicted subtypes. */
        pred_subtypes.push_back(cur_best_st);
      }
    }
    else {
      cout << "qp: " << qp << " " << cur_act_state->state()->sub_name() << " " << cur_act_state->state()->type() << " " << cur_act_state->state()->column() << endl;
      throw Error("No best predecessor. Error.");
    }

    cur_state = cur_act_state->state();

    /*! add current state to viterbi path. */
    vit_path += cur_state;

  }
  /*! start position of first segment in predicted recombination is 0. */
  vit_path_start_pos += 0;

  if (Const_Def::genome_type() == circular && q.is_full_length_seq()) {
    prune_ext_circ_genome(q, &vit_path, &vit_path_start_pos, &vit_path_end_pos, &vit_path_subtype, &pred_subtypes);
  }

  /* output of Viterbi path details for debugging. */
  if (OUTPUT_DETAILS)
    output_details(q, vit_path, vit_path_subtype);

  /*! output of predicted recombination in different formats. */
  recomb_output(q, vit_path_start_pos, vit_path_end_pos, vit_path_subtype, pred_subtypes);
  /*! output of msa columns aligned to each query sequence position. */
  alignment_to_msa_output(q.name(), vit_path);
  alignment_to_ref_seq_output(al, q, vit_path);

  

}

/***************************************************************//**
  Prune extended circular genomes:

  For extended circular genomes,
  the Viterbi path was determined for the extended sequence.
  Output: only for the original query sequence
  -> prune the predicted recombination.
    
 @param[in] query_str query sequence.
 @param[in] vit_path  Viterbi state path.
 @param[in] vit_path_subtype Order of predicted subtypes = predicted recombination pattern.
 @param[in] vit_path_start_pos  Start positions of segments in predicted recombination.
 @param[in] vit_path_end_pos  End positions of segments in predicted recombination.
 @param[in] pred_subtypes List of predicted subtypes.
*******************************************************************/
void HMM::prune_ext_circ_genome(Query          q,
			        vector<State*> *vit_path,
			        vector<int>    *vit_path_start_pos,
			        vector<int>    *vit_path_end_pos,
			        vector<string> *vit_path_subtype,
			        list<string>   *pred_subtypes) {

  //! take into account reverse order in vit_path (End, ..., Begin)
  int q_ext = q.ext();
  int q_end_in_ext_q_str = q.end_in_ext_q_str();

  //! Delete states at beginning of Viterbi path, that were predicted for last ext. part in query (except End and C_Delete):
  //! Increment 'q_ext' emitting states (this part in Viterbi path will be deleted)
  int num_emit_states = 0;
  int i = 0;
  while (i < (int)(*vit_path).size() && num_emit_states < q_ext) { 
    if (!(*vit_path)[i]->is_mute())
      ++num_emit_states;
    ++i;
  }
  int cut_pos = i-1; //! position of last state in Viterbi path that will be deleted.
    
  //! Check, whether first state is End state -> should remain in Viterbi path
  //! C_Delete state will be deleted.
  int first_del_pos = 0; //! first position in Viterbi state that must be deleted
  //while (first_del_pos < (int)(*vit_path).size() && ((*vit_path)[first_del_pos]->is_special() && (*vit_path)[first_del_pos]->is_mute()))
  while (first_del_pos < (int)(*vit_path).size() && (*vit_path)[first_del_pos]->is_end())
    ++first_del_pos;

  if (first_del_pos < cut_pos)
    (*vit_path).erase((*vit_path).begin()+first_del_pos,(*vit_path).begin()+cut_pos+1); //! cut part [first_del_pos, cut_pos+1] in vit_path.

   
  //! Delete states at end ofs Viterbi path, that were predicted for first ext. part in query (except Begin and N_Delete):
  //! Decrement 'q_ext' emitting states (this part in Viterbi path will be deleted)
  num_emit_states = 0;
  i = (int)(*vit_path).size()-1;
  while (i >= 0 && num_emit_states < q_ext) {
    if (!(*vit_path)[i]->is_mute())
      ++num_emit_states;
    --i;
  }
  cut_pos = i+1; //! position of first state in Viterbi path that will be deleted.
    
  //! Check, whether last state is Begin state -> should remain in Viterbi path
  //! N_Delete state will be deleted.
  int last_del_pos = (int)(*vit_path).size()-1; //! last position in Viterbi state that must be deleted
  //while (last_del_pos >=0 && (*vit_path)[last_del_pos]->is_special() && (*vit_path)[last_del_pos]->is_mute())
  while (last_del_pos >=0 && (*vit_path)[last_del_pos]->is_end())
    --last_del_pos;
  
  if (cut_pos < last_del_pos)
    (*vit_path).erase((*vit_path).begin()+cut_pos, (*vit_path).begin()+last_del_pos+1); //! cut part [cut_pos, last_del_pos+1] in vit_path.

  //! determine first and last REAL subtype or insertion (not B, N_Delete, C_Delete, E)
  //! notice: viterbi_path_subtype is stored reversely
  int b = 0;
  map<string, State*>::const_iterator it = m_str_to_spec_state.find((*vit_path_subtype)[b]);
  while (b < (int)(*vit_path_subtype).size() && it != m_str_to_spec_state.end() && (*it).second->is_mute()) {
    //! special mute state
    ++b;
    it = m_str_to_spec_state.find((*vit_path_subtype)[b]);
  }
  int e = (*vit_path_subtype).size()-1;
  it = m_str_to_spec_state.find((*vit_path_subtype)[e]);
  while (e > -1 && it != m_str_to_spec_state.end() && (*it).second->is_mute()) {
    --e;
    it = m_str_to_spec_state.find((*vit_path_subtype)[e]);
  }

  for (int v=b; v<=e; ++v) {
    (*vit_path_start_pos)[v] = (*vit_path_start_pos)[v] - q_ext;
    (*vit_path_end_pos)[v]   = (*vit_path_end_pos)[v] - q_ext;
  }
  int cut = e;
  while (cut >=b && (*vit_path_end_pos)[cut]<1)
    --cut;

  (*vit_path_start_pos)[cut] = 0;
  if (cut < e) {
    (*vit_path_start_pos).erase((*vit_path_start_pos).begin()+cut+1,(*vit_path_start_pos).begin()+e+1);
    (*vit_path_end_pos).erase((*vit_path_end_pos).begin()+cut+1,(*vit_path_end_pos).begin()+e+1);
    (*vit_path_subtype).erase((*vit_path_subtype).begin()+cut+1,(*vit_path_subtype).begin()+e+1);
  }

  cut = b;
  while (cut < (int)(*vit_path_start_pos).size() && cut <=e && (*vit_path_start_pos)[cut] > q_end_in_ext_q_str-q_ext)
    cut++;

  (*vit_path_end_pos)[cut] = q_end_in_ext_q_str-q_ext;
  if (cut > b) {
    (*vit_path_start_pos).erase((*vit_path_start_pos).begin()+b,(*vit_path_start_pos).begin()+cut);
    (*vit_path_end_pos).erase((*vit_path_end_pos).begin()+b,(*vit_path_end_pos).begin()+cut);
    (*vit_path_subtype).erase((*vit_path_subtype).begin()+b,(*vit_path_subtype).begin()+cut);
  }

  //! adapt list of predicted subtypes:
  (*pred_subtypes).clear();
  FOREACH (string pred_sub, (*vit_path_subtype)) {
    map<string, State*>::const_iterator it = m_str_to_spec_state.find(pred_sub);
    if (it == m_str_to_spec_state.end()) { //! no special state
      (*pred_subtypes).push_back(pred_sub);
    }
  }
}


/***************************************************************//**
  Backward algorithm.

  Computation of the backward table
  based on the active states determined in the Viterbi algorithm.
  Order of calculation of backward variables is given:
    1. Special states 2. jump state Match, 3. Deletes, 4. jump state Delete, 5. Match and Inserts.
    since backward variable of a certain state at a certain qp
    is calculated by the backward variables of successors (at qp (mute states) or qp+1)
      so, e.g. bw variable for Deletes must be calculated to calculate bw variable for Matches.
  
  @param[in] query Query sequence (string).
*******************************************************************/
void HMM::bw_algo(string query) {
  
  Alphabet *alphabet = Const_Def::alphabet();

  int query_size = (int)query.length();

  //! Backward variable for End (=log(1)).
  int qp = query_size-1;
  mytype cur_bw = -INF;
  m_act_spec_states[qp][m_spec_state_to_idx[end]]->backward() = 0;

  //! Backward variable for C_Delete.
  State* cur_state = c_delete;
  int cur_idx = m_spec_state_to_idx[cur_state];
  ActiveState* cur_act_state = m_act_spec_states[qp][cur_idx];
  if (cur_act_state) {
    bw_rec(cur_state, query, qp, 0, &cur_bw);
    m_act_spec_states[qp][cur_idx]->backward() = cur_bw;
  }

  //! Backward recursion for each query sequence position.
  vector<State*> cur_spec_states;
  cur_spec_states += n_insert, c_insert;

  for (qp=query_size-1; qp>=0; --qp) {
    int cur_nucl_idx = 0;
    if (qp < query_size-1)
      cur_nucl_idx = alphabet->idx(query[qp+1]);
    
    cur_bw = -INF;
    if (Const_Def::jump_version() == linear && qp < query_size-1) {
      //! Backward variable for jump states Match.
      int type = (int)Jump_Match;
      pair<int, int> cur_act_int = m_act_jump_intervals[qp][type];
      int cur_act_int_left  = cur_act_int.first;
      int cur_act_int_right = cur_act_int.second;
      if (cur_act_int_left != -1) {
        int mapped_right = cur_act_int_right - cur_act_int_left;
	for (int i=mapped_right; i>=0; --i) {
	  cur_act_state = m_act_jump_states[qp][type][i];
	  if (cur_act_state) {
	    bw_rec(cur_act_state->state(), query, qp, cur_nucl_idx, &cur_bw);
	    m_act_jump_states[qp][type][i]->backward() = cur_bw;
	  }
	}
      }
    }

    //! Backward variable for Deletes.
    int type = (int)Delete;
    for (int p=0; p<m_num_phmms; ++p) {
      pair<int, int> cur_act_int = m_act_intervals[qp][p][type];
      int cur_act_int_left  = cur_act_int.first;
      if (cur_act_int_left != -1) {
        int cur_act_int_right = cur_act_int.second;
        int mapped_right = cur_act_int_right - cur_act_int_left;
	for (int i=mapped_right; i>=0; --i) {
	  cur_act_state = m_act_states[qp][p][type][i];
	  if (cur_act_state) {
	    bw_rec(cur_act_state->state(), query, qp, cur_nucl_idx, &cur_bw);
	    m_act_states[qp][p][type][i]->backward() = cur_bw;
	  }
        }
      }
    }

    if (Const_Def::jump_version() == linear) {
      //! Backward variable for jump states Delete.
      type = (int)Jump_Delete;
      pair<int, int> cur_act_int = m_act_jump_intervals[qp][type];
      int cur_act_int_left  = cur_act_int.first;
      int cur_act_int_right = cur_act_int.second;
      if (cur_act_int_left != -1) {
	for (int i=cur_act_int_right; i>=cur_act_int_left; --i) {
	  cur_act_state = m_act_jump_states[qp][type][i-cur_act_int_left];
	  if (cur_act_state) {
	    bw_rec(cur_act_state->state(), query, qp, cur_nucl_idx, &cur_bw);
	    m_act_jump_states[qp][type][i-cur_act_int_left]->backward() = cur_bw;
	  }
	}
      }
    }

    //! Backward variable for Inserts and Matches.
    for (int type = (int)Insert; type >=(int)Match; --type) {
      for (int p=0; p<m_num_phmms; ++p) {
	pair<int, int> cur_act_int = m_act_intervals[qp][p][type];
        int cur_act_int_left  = cur_act_int.first;
        if (cur_act_int_left != -1) {
          int cur_act_int_right = cur_act_int.second;
          int mapped_right = cur_act_int_right - cur_act_int_left;
	  for (int i=mapped_right; i>=0; --i) {
            cur_act_state = m_act_states[qp][p][type][i];
            if (cur_act_state) {
              bw_rec(cur_act_state->state(), query, qp, cur_nucl_idx, &cur_bw);
	      m_act_states[qp][p][type][i]->backward() = cur_bw;
	    }
          }
        }
      }
    }
    
    //! Backward variable for special Insert states.
    for (vector<State*>::reverse_iterator rit=cur_spec_states.rbegin(); rit<cur_spec_states.rend(); ++rit) {
      cur_state     = (*rit);
      cur_idx       = m_spec_state_to_idx[cur_state];
      cur_act_state = m_act_spec_states[qp][cur_idx];
      if (cur_act_state) {
        bw_rec(cur_state, query, qp, cur_nucl_idx, &cur_bw);
	m_act_spec_states[qp][cur_idx]->backward() = cur_bw;
      }
    }
  }

  qp     = -1;
  cur_bw = -INF;
  int cur_nucl_idx = alphabet->idx(query[qp+1]);
  if (Const_Def::jump_version() == linear) {
    //! Backward variable for jump states Match.
    int type = (int)Jump_Match;
    pair<int, int> cur_act_int = m_act_jump_intervals[qp][type];
    int cur_act_int_left  = cur_act_int.first;
    int cur_act_int_right = cur_act_int.second;
    if (cur_act_int_left != -1) {
      for (int i=cur_act_int_right; i>=cur_act_int_left; --i) {
	ActiveState* cur_act_state = m_act_jump_states[qp][type][i-cur_act_int_left];
	if (cur_act_state) {
	  bw_rec(cur_act_state->state(), query, qp, cur_nucl_idx, &cur_bw);
	  m_act_jump_states[qp][type][i-cur_act_int_left]->backward() = cur_bw;
	}
      }
    }
  }

  //! Backward variable for DELETEs in front of first query sequence position.
  int type = (int) Delete;
  for (int p=0; p< m_num_phmms; ++p) {
    pair<int, int> cur_act_int = m_act_intervals[qp][p][type];
    int cur_act_int_left  = cur_act_int.first;
    if (cur_act_int_left != -1) {
      int cur_act_int_right = cur_act_int.second;
      int mapped_right = cur_act_int_right - cur_act_int_left;
      for (int i=mapped_right; i>=0; --i) {
        ActiveState* cur_act_state = m_act_states[qp][p][type][i];
        if (cur_act_state) {
          bw_rec(cur_act_state->state(), query, qp, cur_nucl_idx, &cur_bw);
	  m_act_states[qp][p][type][i]->backward() = cur_bw;
	}
      }
    }
  }

  //! Backward variable for N_DELETE and BEGIN in front of first query sequence position.
  cur_spec_states.clear();
  cur_spec_states += begin, n_delete;
  for (vector<State*>::reverse_iterator rit=cur_spec_states.rbegin(); rit<cur_spec_states.rend(); ++rit) {
    State*       cur_state     = (*rit);
    int          cur_idx       = m_spec_state_to_idx[cur_state];
    ActiveState* cur_act_state = m_act_spec_states[qp][cur_idx];
    if (cur_act_state) {
      bw_rec(cur_state, query, qp, cur_nucl_idx, &cur_bw);
      m_act_spec_states[qp][cur_idx]->backward() = cur_bw;
    }
  }
}


/******************************************************//**
 Backward recursion.

  Computation of the backward value for a certain state.
  Recursion over all (active) successors.

  @param[in] cur_state Current state.
  @param[in] query Query sequence (string).
  @param[in] qp Current query position.
  @param[in] cur_nucl_idx Index of nucleotide at current qp in alphabet
  @param[in,out] backward Backward variable of current state.
**********************************************************/
void HMM::bw_rec(State  *cur_state,
                 string  query,
                 int     qp,
		 int     cur_nucl_idx,
                 mytype *backward) {

  int    query_size = query.length();
  mytype log_bw     = -INF;
  //! backward recursion.
  /*!
     computation of backward variable for current state at current position.
     recursion over backward variables of successive states that were active for successive query position.
  */

  /*! recursion over all successors.
      within loop: test if successor was active for successive position.
  */
  FOREACH( Transition* cur_trans, cur_state->succ() ) {
    State* cur_succ = cur_trans->whereto();

    //! successive query position.
    /*!
     for M, I, N_I and C_I: position qp+1.
     for D, N_D, C_D and E: position qp since no emission.
    */
    mytype cur_log_ep  = 0;
    int    cur_succ_qp = qp;
    if (!cur_succ->is_mute()) {
      cur_succ_qp = qp+1;
      cur_log_ep  = cur_succ->log_ep()[cur_nucl_idx];
    }

    /*! determine if successor was active for successive query position. */
    ActiveState* cur_act_succ  = 0;
    if (cur_succ_qp < query_size) {
      if (cur_succ->is_special())
        cur_act_succ = m_act_spec_states[cur_succ_qp][m_spec_state_to_idx[cur_succ]];
      else if (cur_succ->is_jump_state()) {
	if (Const_Def::jump_version() != linear)
          throw Error("jump state type cannot be predecessor if linear jump version is not chosen!");

	int       cur_succ_col           = cur_succ->column();
        StateType     cur_succ_type      = cur_succ->type();
        JumpStateType cur_succ_jump_type = m_type_to_jump_type[cur_succ_type];
	pair<int, int> cur_act_int  = m_act_jump_intervals[cur_succ_qp][(int)cur_succ_jump_type];
	int       cur_act_int_left  = cur_act_int.first;
        int       cur_act_int_right = cur_act_int.second;

        if (cur_succ_col >= cur_act_int_left && cur_succ_col <= cur_act_int_right)
          cur_act_succ = m_act_jump_states[cur_succ_qp][(int)cur_succ_jump_type][cur_succ_col-cur_act_int_left];
      }
      else {
        int       cur_succ_phmm_idx = m_phmm_to_sub_idx[cur_succ->phmm()];
        int       cur_succ_col      = cur_succ->column();
        StateType cur_succ_type     = cur_succ->type();
	pair<int, int> cur_act_int  = m_act_intervals[cur_succ_qp][cur_succ_phmm_idx][(int)cur_succ_type];
	int       cur_act_int_left  = cur_act_int.first;
        int       cur_act_int_right = cur_act_int.second;

        if (cur_succ_col >= cur_act_int_left && cur_succ_col <= cur_act_int_right)
          cur_act_succ = m_act_states[cur_succ_qp][cur_succ_phmm_idx][(int)cur_succ_type][cur_succ_col-cur_act_int_left];
      }
      
      /*! calculation of backward variable if successor was active. */
      if (cur_act_succ) {
        mytype log_b = cur_trans->log_p() + cur_act_succ->backward() + cur_log_ep;
        /*! sum up probabilities for current backward variable. */
        if (log_bw > log_b)
          log_bw = log_bw + log(1 + exp(log_b-log_bw));
        else
          log_bw = log_b + log(1 + exp(log_bw-log_b));
      }
    }
  }
  
  /*! backward variable for current state at current query position. */
  *backward = log_bw;
}

/******************************************************************************************//**
  Posterior probabilities.

  - Computation of posterior probabilities for each subtype at every query sequence position.
  - Determination of subtype (incl. N-/C-Insert) with the highest post prob
    => gives 'posterior recombination path'
   
  @param[in] q Query object.
  @param[in] no_of_query Number of query in list of all query sequences.
**********************************************************************************************/
void HMM::post_probs(Query q) {

  //! \todo {Include option for jpHMM version, that contains profileHMM for whole alignment.}

  int cur_q_start = q.start_in_ext_q_str();
  int cur_q_end   = q.end_in_ext_q_str();

  int query_size = q.str().length();
  int qp         = query_size-1;

  //! forward, backward and viterbi score.
  ActiveState* cur_act_state = m_act_spec_states[qp][m_spec_state_to_idx[end]];
  mytype       fw_score      = cur_act_state->forward();

  //! Posterior probabilities.
  //! matrix containing posterior probabilities for all qp and all subtypes ( 1st entry for N_Insert, last entry for C_Insert)
  int num_entries = m_num_phmms+2; //! +2 for special inserts
  multi_array<mytype, 2> post_probs;
  multi_array<mytype, 2>::extent_gen ext;
  post_probs.resize(ext[query_size][num_entries]);
  for (int qp=0; qp<query_size; ++qp)
    for (int s=0; s<num_entries; ++s)
      post_probs[qp][s] = 0;


  //! Posterior recombination path.
  //! For each query sequence position the subtype with the highest posterior probability is determined.
  //! -> segments saved.
  vector<int>    post_path_start_pos;  /*! first position for subtype. */
  vector<int>    post_path_end_pos;    /*! last position for subtype. */
  vector<string> post_path_subtype;    /*! current subtype mit highest posterior probability. */
  
  //! calculate posterior probabilities 
  //! and path of subtypes with highest posterior probability 
  //! only for original query sequence:
  for (int qp=cur_q_start; qp <=cur_q_end; ++qp) {

    //! posterior probabilities for all subtypes at current query position qp.
    map<string, mytype> cur_post_probs = post_prob_rec(qp, fw_score);

    //! save post probs for current query position in matrix of all post probs
    //! and determine subtype (or N-/C-Insert) with highest posterior prob. for current qp.
    //! N_INSERT.
    string cur_sub_name  = m_spec_state_to_str[n_insert];
    mytype cur_post_prob = cur_post_probs[cur_sub_name];
    string cur_best_st   = cur_sub_name;
    mytype max_post_prob = cur_post_prob;
    post_probs[qp][0]    = cur_post_prob;
    
    //! Subtypes.
    for (int s=0; s<m_num_phmms; ++s) {
      cur_sub_name        = m_phmms[s]->sub_prof()->sub()->name();
      cur_post_prob       = cur_post_probs[cur_sub_name];
      post_probs[qp][s+1] = cur_post_prob;
      if (cur_post_prob > max_post_prob) {
        /*! current subtype is subtype with highest posterior prob. */
	max_post_prob = cur_post_prob;
	cur_best_st   = cur_sub_name;
      }
    }

    //! C_INSERT.
    cur_sub_name  = m_spec_state_to_str[c_insert];
    cur_post_prob = cur_post_probs[cur_sub_name];
    post_probs[qp][num_entries-1] = cur_post_prob;
    if (cur_post_prob > max_post_prob) {
      max_post_prob = cur_post_prob;
      cur_best_st   = cur_sub_name;
    }
    
    //! change in pattern of/new segment in 'posterior path'.
    if (post_path_subtype.empty() || cur_best_st != post_path_subtype.back()) {
      if (!post_path_subtype.empty())
        post_path_end_pos   += qp - 1 - cur_q_start;
      post_path_subtype   += cur_best_st;
      post_path_start_pos += qp - cur_q_start;
    }
  }
  post_path_end_pos += cur_q_end - cur_q_start;

  post_prob_output(q, post_probs);
  post_path_output(post_path_start_pos, post_path_end_pos, post_path_subtype);
}

/****************************************************************************************************************//**
 Posterior probabilities recursion.

 Computation of posterior probabilities for all subtypes at current query sequence position:
   post. prob. of a subtype = sum of post. probs of all Match and Insert states that are active for the current qp.
   
   @param[in] qp Current query position.
   @param[in] state_to_str Mapping state to string.
   @param[in] fw_score Forward score.
********************************************************************************************************************/
map<string, mytype> HMM::post_prob_rec(int qp,
				       mytype fw_score) {

  //! Posterior probability for current qp.
  map<string, mytype> post_probs;
  FOREACH(ProfileHMM* phmm, m_phmms)
    post_probs[phmm->sub_prof()->sub()->name()] = 0;
  post_probs[m_spec_state_to_str[n_insert]] = post_probs[m_spec_state_to_str[c_insert]] = 0;

  //! posterior probability of each subtype.
  for (int s=0; s< m_num_phmms; ++s) {
    string cur_sub_name = m_phmms[s]->sub_prof()->sub()->name();
    for (int type = (int)Match; type <=(int)Insert; ++type) { // no Deletes (do not emit anything)
      pair<int, int> cur_int = m_act_intervals[qp][s][type];
      int left  = cur_int.first;
      int right = cur_int.second;
      if (left != -1 && right != -1) {
	/*! posterior probabilities for every active Match and Insert. */
        for (int i=0; i<=right-left; ++i) { //! start at i=left -> but then i-left -> start at 0
          if (m_act_states[qp][s][type][i]) {
            ActiveState* cur_act_state = m_act_states[qp][s][type][i];

            if (cur_act_state && cur_act_state->forward() != -INF && cur_act_state->backward() != -INF) {
              /*! posterior probability for current state. */      
              mytype percentage = exp(cur_act_state->backward() + cur_act_state->forward() - fw_score);

	      /*! posterior probabilities summed up for current subtype. */
              post_probs[cur_sub_name] += percentage;
            }
          }
        }
      }
    }
  }

  /*! posterior probabilities for N_INSERT and C_INSERT. */
  vector<State*> cur_spec_states;
  cur_spec_states += n_insert, c_insert;

  FOREACH (State* cur_state, cur_spec_states) {
    ActiveState* cur_act_state = m_act_spec_states[qp][m_spec_state_to_idx[cur_state]];
    if (cur_act_state && cur_act_state->forward() != -INF && cur_act_state->backward() != -INF) {
      /*! posterior probability. */
      post_probs[m_spec_state_to_str[cur_state]] += exp(cur_act_state->backward() + cur_act_state->forward() - fw_score);
    }
  }
  return post_probs;
}
