// ~/jpHMM/src/hmm_output.cpp
/**************************************************************
 File name	: hmm_output.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Output of 
  - predicted recombination defined by Viterbi path
  - alignment of query sequence to multiple alignment defined by Viterbi path
  - alignment of query sequence to reference genome defined by Viterbi path
  - posterior probabilities
  - path of subtypes with highest posterior probability for each sequence position

 Test output for HMM: 
  - all states and their successors (or predecessors)
  - active states at each query sequence positoin
  - active intervals (alignment columns) for M, I and D states
  - details about Viterbi path and emission probabilities in tabular form
    
 Declaration of functions in hmm.h
***************************************************************/

#include "std_source.h"
#include "error.h"
#include "const_def.h"
#include "alignment.h"
#include "emTransProfile.h"
#include "query.h"
#include "profileHMM.h"
#include "hmm.h"
#include "html_tab.h"
#include "alphabet.h"

using namespace std;

/**************************************************
 *
 *   Output functions for class HMM.
 *
***************************************************/

/******************************************************************************//**
 Output of recombination.

 Output of predicted recombination in different formats:
   - Predicted recombination pattern in txt-format
   - Predicted recombination pattern in txt-format incl. gaps (not yet finished)
   - List of predicted subtypes
   - Predicted recombination pattern in gff-format
   
 @param[in] query_name Name of query sequence.
 @param[in] vit_path_subtype Order of predicted subtypes = predicted recombination pattern.
 @param[in] vit_path_start_pos  Start positions of segments in predicted recombination.
 @param[in] vit_path_end_pos  End positions of segments in predicted recombination.
 @param[in] pred_subtypes List of predicted subtypes.

  //! \todo {Output of predicted recombination for queries containing gaps. -> take into account "vector<int>& list_of_gaps" and "int num_gaps"}
**********************************************************************************/
void HMM::recomb_output(Query          q,
  			vector<int>    vit_path_start_pos,
  			vector<int>    vit_path_end_pos,
			vector<string> vit_path_subtype,
			list<string>   pred_subtypes) {

  string query_name = q.name();
  mytype cur_beam = q.beam_width();
  
  ofstream RECOMB(Const_Def::recomb_file_name().c_str(), ios::app);
  ofstream RECOMB_WO_POS(Const_Def::pred_subtypes_file_name().c_str(), ios::app);
  ofstream RECOMB_GAPS(Const_Def::recomb_incl_gaps_file_name().c_str(), ios::app);
  ofstream VITERBI_GFF(Const_Def::viterbi_gff_file_name().c_str(), ios::app);  
  ofstream POST_PATH(Const_Def::posterior_path_file_name().c_str(), ios::app);
  ofstream VAR_REGS(Const_Def::variable_regions_file_name().c_str(), ios::app);
  
  RECOMB << ">" << query_name << " (";
  RECOMB_WO_POS << ">" << query_name << "\t" << flush;

  //! additional output of recombination for query sequences with gaps:
  RECOMB_GAPS << ">" << query_name << " (";

  VITERBI_GFF << "#" << endl
              << "# predicted subtypes for sequence '" << query_name << "': " << endl;
  POST_PATH << ">" << query_name << endl;
  VAR_REGS << ">" << query_name << endl;
  
  //! output of chosen parameters and program versions etc.
  if (Const_Def::genome_type() == circular) {
    RECOMB << "circular genome, ";
    RECOMB_WO_POS << "circular genome, ";
    RECOMB_GAPS << "circular genome, ";
    if (q.is_full_length_seq()) {
      RECOMB << "full-length, ";
      RECOMB_WO_POS << "full-length, ";
      RECOMB_GAPS << "full-length, ";
    }
    else {
      RECOMB << "fragment, ";
      RECOMB_WO_POS << "fragment, ";
      RECOMB_GAPS << "fragment, ";
    }
  }
  if (Const_Def::jump_version() == linear) {
    RECOMB << "linear model, ";
    RECOMB_WO_POS << "linear model, ";
    RECOMB_GAPS << "linear model, ";
  }
  if (q.prog_version() == blat) {
    RECOMB << "blat, ";
    RECOMB_WO_POS << "blat, ";
    RECOMB_GAPS << "blat, ";
  }
  //! also for jpHMM based on BLAT alignment beam-search algo applied
  RECOMB << "bw=" << cur_beam << ")" << endl;
  RECOMB_WO_POS << "bw=" << cur_beam << ", jp=" << Const_Def::jump_prob() << "\t" << flush;
  RECOMB_GAPS << "bw=" << cur_beam << ")" << endl;
  
  /*! output of predicted recombination in text and gff format. */
  for (int i=(int)vit_path_start_pos.size()-1; i>-1; --i) { // without first. entry for BEGIN
    int    cur_start = vit_path_start_pos[i]; //! numbering starts at 0
    int    cur_end   = vit_path_end_pos[i];
    string cur_subtype_name = vit_path_subtype[i];
    RECOMB  << cur_start+1 << "\t" << cur_end+1 << "\t" << cur_subtype_name << endl;                                               //! output based on 1
    VITERBI_GFF << query_name << "\t" << "jpHMM\t" << cur_subtype_name << "\t" << cur_start << "\t" << cur_end << "\t.\t+\t." << endl; //! output based on 0
  }

  /*! output of predicted subtypes. */
  pred_subtypes.sort();
  pred_subtypes.unique();
  FOREACH (string pred_sub, pred_subtypes)
    RECOMB_WO_POS << pred_sub;

  RECOMB  << endl;
  VITERBI_GFF << endl;
  RECOMB_WO_POS << endl;

  RECOMB.close(); RECOMB_WO_POS.close(); RECOMB_GAPS.close(); VITERBI_GFF.close(); POST_PATH.close(); VAR_REGS.close();
}

/*********************************************************************************//**
 Output of aligned MSA columns.

 Output of the alignment MSA column for each query position
  defined by the Viterbi path, 
  i.e. the column corresponding to the resp. emitting state

 @param[in] query_name Name of query sequence.
 @param[in] vit_path Viterbi path, starting with End and beginning with Begin state.
*************************************************************************************/
void HMM::alignment_to_msa_output(string          query_name, 
		                  vector<State*>  vit_path) {

  /*! output of aligned MSA columns. */
  ofstream MSA_AL(Const_Def::al_to_msa_file_name().c_str(), ios::app);

  MSA_AL << ">" << query_name << endl;
  for (vector<State*>::reverse_iterator rit=vit_path.rbegin(); rit != vit_path.rend(); ++rit) {
    State* cur_state = (*rit);
    if (cur_state->is_match() || cur_state->is_insert()){
      MSA_AL << cur_state->column() + 1 << " "; //! real numbering
    }
    else if (cur_state->is_n_insert()){
      MSA_AL << 0 << " ";
    }
    else if(cur_state->is_c_insert()){
      MSA_AL << m_len+1 << " "; 
    }
  }
  MSA_AL << endl;
  MSA_AL.close();
}

/*********************************************************************************//**
 Output of pairwise alignment of query to reference sequence

 Output of the pairwise alignment of the query sequence to the reference sequence
  (if the reference sequence is contained in the MSA)
  defined by the Viterbi path
 
 Capital letters: real aligned positions
 Lower case letters: not aligned positions (e.g. qp emitted by insert state and insertion in ref seq aligned by chance)
 '-': gaps, introduced by delete or insert state
 '.': residue in query is the same as the aligned residue in the reference sequence

 @param[in] query_name Name of query sequence.
 @param[in] vit_path Viterbi path, starting with End and beginning with Begin state.
*************************************************************************************/
void HMM::alignment_to_ref_seq_output(Alignment      *al,
                                      Query           q,
		                      vector<State*>  vit_path) {

  string query_str   = q.str();
  int    cur_q_start = q.start_in_ext_q_str();
  int    cur_q_end   = q.end_in_ext_q_str();
  int    orig_q_len  = cur_q_end - cur_q_start + 1;
  
  string orig_query_str = query_str.substr(cur_q_start, orig_q_len);
  string query_name = q.name();

  int prec_aligned_al_col = -1;
  //! determine first and last aligned column
  
  if (al->has_ref_seq()) {
    //! save ref seq and query as they appear in the pairwise alignment, i.e. incl. gaps
    string ref_seq_str_out;
    string query_str_out;
     
    Seq*   ref_seq     = al->ref_seq(); 
    string ref_seq_str = ref_seq->str();

    //! determine first Match state
    State* first_match_state;
    int    num_N_insertions      = 0;

    vector<State*>::reverse_iterator rit=vit_path.rbegin();
    while (rit != vit_path.rend() && !(*rit)->is_match() ) {
      if ((*rit)->is_n_insert())
        ++num_N_insertions;
      ++rit;
    }
    if (rit != vit_path.rend())
      first_match_state = (*rit);
    else
      throw Error("No match state in Viterbi path");

    int col_first_match_state = first_match_state->column();

    int cur_first_aligned_col = col_first_match_state;
    int cur_last_aligned_col  = col_first_match_state;
  
    int cur_col               = col_first_match_state-1; //! insert in front of (in the model) first match state in viterbi path

    if (num_N_insertions > 0) {
      string ref_seq_str_tmp;
      string query_str_tmp;
      for (int i=num_N_insertions-1; i>=0; --i) {
        if (cur_col >= 0) {
          ref_seq_str_tmp += tolower(ref_seq_str[cur_col]);
	  query_str_tmp   += tolower(orig_query_str[i]);
	  cur_first_aligned_col = cur_col;
        }
        else {
          ref_seq_str_tmp += "-";
          query_str_tmp   += toupper(orig_query_str[i]);
        }
        --cur_col;
      }

      //! strings in reverse order:
      for (string::reverse_iterator rit=ref_seq_str_tmp.rbegin() ; rit < ref_seq_str_tmp.rend(); ++rit)
        ref_seq_str_out += *rit;
      for (string::reverse_iterator rit=query_str_tmp.rbegin() ; rit < query_str_tmp.rend(); ++rit)
        query_str_out += *rit;
    }

    int qp            = num_N_insertions-1; //! to start numbering at 0
    int size_vit_path = (int)vit_path.size();

    for (int v=size_vit_path-1; v>=0; --v) {
      State* cur_state = vit_path[v];
      cur_col = cur_state->column();

      if (cur_state->is_match()) {
        ++qp;

        //! check if 'gap' between two successive states
        //! corresponds to insert columns in alignment
        if (cur_state != first_match_state && cur_col > prec_aligned_al_col+1) {
	  //! prec_aligned_al_col is column that was aligned in previous step
	  //! in case of insertions this can be somewhere in between two states
	  //! check, if rest is insertion, i.e. if still part of alignment is missing:
	  //! insert columns in alignment:
	  //! add corresponding residues from ref seq
	  for (int i=prec_aligned_al_col+1; i<cur_col; ++i) {
            ref_seq_str_out += tolower(ref_seq_str[i]);
            query_str_out   += "-";
	  }
	  prec_aligned_al_col = cur_col-1;
        }

        //! add residue of current position:
        ref_seq_str_out += toupper(ref_seq_str[cur_col]);
        query_str_out   += toupper(orig_query_str[qp]);
        cur_last_aligned_col = prec_aligned_al_col = cur_col;
      }
      else if (cur_state->is_insert()) {
        ++qp;
      
        //! if preceding state == M: check whether in alignment as well insert column is following by checking the column of the successive state
        //! if preceding state == I: also check if insert columns in alignment. take into account length of insertion in query
        //! check respective number of columns in alignment.
      
        //! up to successive state columns are considered as insertions in alignment.
       
        //! save last aligned column in alignment to fill reference sequence in alignment correctly.
      
        //! prec_aligned_al_col is last aligned alignment column
        //! prec_aligned_al_col+1 is currently examined column
        //! if column of successive MATCH state > prec_aligned_al_col+1 => prec_aligned_al_col+1 must be considered als insert column
        //! => no gap inserted in reference sequence, but respective nucleotide in prec_aligned_al_col+1

        //! determine succ state in Viterbi path
        //! take into account that viterbi path determined with backtracking => starts with End state -> reverse order
        int    cur_idx = v-1;
        State* cur_succ;
        int    succ_col;
        while (cur_idx >= 0 && ((!vit_path[cur_idx]->is_match() && Const_Def::genome_type() != circular) || (!vit_path[cur_idx]->is_end() && Const_Def::genome_type() == circular)))
	  --cur_idx;

        //! if cur_idx >=0 , vit_path[cur_idx] is succ. match state (or End state for circular genomes)
        if (cur_idx >= 0) {
	  cur_succ = vit_path[cur_idx];
          succ_col = cur_succ->column();
        }
        else { 
	  //! for non-circular genomes:
	  //! succ. state != match state -> not possible, since each insert state has a match state as successor
	  //! for circular genomes:
	  //! due to pruning of extended sequences (and of Viterbi path), insert state in Viterbi path does not need to have an active match state as successor
	  
	  throw Error("Error in Viterbi path");
	}
      
        if (succ_col > prec_aligned_al_col+1) {
	  //! columns in between are considered als insert columns:
          //! instead of introducing a gap into the ref seq,
	  //! align respective column
	  ref_seq_str_out += tolower(ref_seq_str[prec_aligned_al_col+1]); //! tolower since random alignment in these insert columns
          query_str_out   += tolower(orig_query_str[qp]);
          ++prec_aligned_al_col;
	  cur_last_aligned_col = prec_aligned_al_col;
        }
        else {
          ref_seq_str_out += "-";
          query_str_out   += toupper(orig_query_str[qp]);
        }
      }
      else if (cur_state->is_c_insert()) {
        ++qp;
        if (prec_aligned_al_col+1 < m_len) {
	  //! insert columns at end of alignment
          //! instead of introducing a gap into the ref seq,
	  //! align respective column
	  ref_seq_str_out += tolower(ref_seq_str[prec_aligned_al_col+1]); //! tolower since random alignment in these insert columns
	  query_str_out   += tolower(orig_query_str[qp]);
          ++prec_aligned_al_col;
	  cur_last_aligned_col = prec_aligned_al_col;
        }
        else {
          ref_seq_str_out += "-";
          query_str_out   += toupper(orig_query_str[qp]);
        }
      }
      else if (cur_state->is_delete()) {
        //! determine column of previous state: 
        //! if gap between these two states, introduce gaps in query and resp. columns in ref seq from alignment

        if ((int)query_str_out.length() > 0 && qp+1 < (int)orig_query_str.length() ) {
          //! introduce gap in query:
          //! (only if not first or last query sequence position)

          if (cur_col > prec_aligned_al_col+1) {
	    if (cur_col-prec_aligned_al_col > Const_Def::threshold_long_deletion()) {
	      cout << query_name << " contains a deletion of > 50bp" << endl;
	    }
	    for (int i=prec_aligned_al_col+1; i<cur_col; ++i) {
              ref_seq_str_out += tolower(ref_seq_str[i]);
              query_str_out   += "-";
	    }
	    prec_aligned_al_col = cur_col-1;
          }

          //! introduce gap in query:
          ref_seq_str_out += toupper(ref_seq_str[cur_col]);
          query_str_out   += "-";
          prec_aligned_al_col  = cur_col;
          cur_last_aligned_col = prec_aligned_al_col;
        }
      }
    }

    if (ref_seq_str_out.length() != query_str_out.length()) {
      throw Error("Something is wrong with the pairwise alignment of the query to the reference sequence");
    }

    //! compare pairwise alignment:
    //! if same nucleotide in query and ref seq, output '.' in query:

    int cur_pos = 0;
    while (cur_pos < (int) ref_seq_str_out.length()) {
      if (ref_seq_str_out[cur_pos] == query_str_out[cur_pos]) {
	if (ref_seq_str_out[cur_pos] == '-' || ref_seq_str_out[cur_pos] == '*' || ref_seq_str_out[cur_pos] == '.') {
	  //! if both contain a gap at that position delete the gap
	  query_str_out.erase(query_str_out.begin()+cur_pos);
	  ref_seq_str_out.erase(ref_seq_str_out.begin()+cur_pos);
	}
	else {
	  query_str_out.replace(cur_pos,1,".");
	  ++cur_pos;
	}
      }
      else
	++cur_pos;
    }
    
    if (ref_seq_str_out.length() != query_str_out.length()) {
      cout << "size of output strings: " << "query: " << query_str_out.length() << " ref seq: " << ref_seq_str_out.length() << endl;
      throw Error("Something is wrong with the pairwise alignment of the query to the reference sequence");
    }

    /*! output of pairwise alignment. */

    ofstream PW_AL(Const_Def::query_to_ref_seq_al_file_name().c_str(), ios::app);

    //! determine first and last aligned reference sequence position
    int cur_first_ref_seq_pos = 0;
    int cur_last_ref_seq_pos  = 0;
    int len_ref_seq_in_orig_al = 0; //! = 0 if alignment not extended, i.e. for non-circular genomes
    bool al_is_extended = al->is_extended();
    int  orig_al_len    = al->orig_len();

    for (int i=0; i<=cur_last_aligned_col; ++i) {
      if (ref_seq_str[i] != '-' && ref_seq_str[i] != '*' && ref_seq_str[i] != '.') {
	if (i <= cur_first_aligned_col)
	  ++cur_first_ref_seq_pos;
        ++cur_last_ref_seq_pos;
	if (al_is_extended && i<orig_al_len)
	  ++len_ref_seq_in_orig_al;
      }
    }

    if (al_is_extended && cur_last_ref_seq_pos > len_ref_seq_in_orig_al) {
      cur_first_ref_seq_pos -= len_ref_seq_in_orig_al;
      cur_last_ref_seq_pos  -= len_ref_seq_in_orig_al;
    }
    
    PW_AL << ">" << Const_Def::virus_ref_seq_name(Const_Def::virus_type()) << " " << cur_first_ref_seq_pos << " " << cur_last_ref_seq_pos << endl;
    PW_AL << ref_seq_str_out << endl;
    PW_AL << ">" << query_name << endl;
    PW_AL << query_str_out << endl;
    PW_AL.close();
  }
}

/*************************************************************************************//**
  Output of posterior probabilities.

  Output of the posterior probabilities for query positions and all subtypes as a table.
 
  @param[in] q Query object.
  @param[in] post_probs Table of posterior probabilities for each qp and each subtype.
*****************************************************************************************/
void HMM::post_prob_output(Query q,
			   multi_array<mytype, 2> post_probs) {

  //! output file for posterior probabilities: posterior_probabilities_for_seq_x.txt" x = 1,2,...
  const std::string file_name(Const_Def::posterior_prob_file_name().c_str());
  const std::string file_type("txt");
  char  nbr[255];
  sprintf(nbr, "%d", q.no());
  std::string tmp_name = file_name + nbr + '.' + file_type;
  std::ofstream POST_PROB_MATRIX(tmp_name.c_str());

  //! header output.
  POST_PROB_MATRIX << "#" << q.name() << ": " << "5'-Insertion ";
  FOREACH (ProfileHMM* phmm, m_phmms)
    POST_PROB_MATRIX << phmm->sub_prof()->sub()->name() << " ";
  POST_PROB_MATRIX << "3'-Insertion" << endl;

  int cur_q_start = q.start_in_ext_q_str();
  int cur_q_end   = q.end_in_ext_q_str();

  for (int qp=cur_q_start; qp <=cur_q_end; ++qp) {
    POST_PROB_MATRIX << post_probs[qp][0] << " "; //! N_Insert.
    for (int s=0; s<m_num_phmms; ++s)
      POST_PROB_MATRIX << post_probs[qp][s+1] << " ";
    POST_PROB_MATRIX << post_probs[qp][m_num_phmms+1] << endl; //! C_Insert.
  }
}

/******************************************************************************************//**
  Output of posterior recombination path.

  Output of subtypes with the highest posterior probability for each query sequence position.

  @param[in] post_path_start_pos Start positions of segments in posterior recombination path.
  @param[in] post_path_end_pos End positions of segments in posterior recombination path.
  @param[in] post_path_subtype Order of predicted subtypes in posterior recombination path.
**********************************************************************************************/
void HMM::post_path_output(vector<int>    post_path_start_pos,
  			   vector<int>    post_path_end_pos,
			   vector<string> post_path_subtype) {

  //! Output of the subtype 'path' defined by the posterior probs.
  /*! (in the same format as recombination output). */
  ofstream POST_PATH(Const_Def::posterior_path_file_name().c_str(), ios::app);
  
  for (unsigned i=0; i<post_path_start_pos.size(); ++i) {
    POST_PATH << post_path_start_pos[i]+1 << "\t" << post_path_end_pos[i]+1 << "\t" << post_path_subtype[i] << endl;
  }
  POST_PATH << endl;
  POST_PATH.close();
}


/*************************************
 *
 *   Test output for class HMM.
 *
 *************************************/

/*****************************************************//**
 Output of states and successors (or predecessors)

 incl. emission and transition probabilities.
*********************************************************/

void HMM::hmm_out() {

  ofstream ACT_INT_OUT(Const_Def::active_state_intervals_file_name().c_str(), ios::out);

  ACT_INT_OUT << "BEGIN: " << endl;
  for (vector<Transition*>::iterator it=begin->succ().begin(); it!=begin->succ().end(); ++it) {
    if ((*it))
      ACT_INT_OUT << "type: " << (*it)->whereto()->type() << " tp: " << (*it)->trans_prob() << endl;
  }
  ACT_INT_OUT << "N_DELETE: " << endl;
  FOREACH (Transition* t, n_delete->succ()) {
    ACT_INT_OUT << t->whereto()->type() << " ";
    ACT_INT_OUT << t->trans_prob() << endl;
  }
  ACT_INT_OUT << "N_INSERT: " << endl;
  for (int i=0; i<4; ++i)
    ACT_INT_OUT << n_insert->em_prob(i) << " ";
  ACT_INT_OUT << endl;
  FOREACH (Transition* t, n_insert->succ()) {
    ACT_INT_OUT << t->whereto()->type() << " ";
    ACT_INT_OUT << t->trans_prob() << endl;
  }

  ACT_INT_OUT << "MATCH, INSERT AND DELETEs: " << endl;
  for (int i=m_first_cons_col; i<=m_last_cons_col; i++) {
    FOREACH (ProfileHMM* phmm, m_phmms) {
      for (int type=(int)Match; type<=(int)Delete; type++) {
        if (phmm->has_state(i, type)) {
          State* cur_state = phmm->state(i, type);
	  ACT_INT_OUT << "cur_state: " << type << " " << i+1 << " " << phmm->sub_name() << endl;
	  if (cur_state->is_insert() || cur_state->is_match()) {
	    //! emission probabilities:
	    for (int i=0; i<4; ++i)
	      ACT_INT_OUT << exp(cur_state->log_ep()[i]) << " ";
	    ACT_INT_OUT << endl;
	  }
          FOREACH (Transition* t, cur_state->succ()) {
	    //! transition probabilities to successor states:
	    if (t->whereto()->is_match() || t->whereto()->is_insert() || t->whereto()->is_delete())
              ACT_INT_OUT << t->whereto()->type() << " " << t->whereto()->column()+1 << " " << t->whereto()->phmm()->sub_name() << " ";
            else
 	      ACT_INT_OUT << t->whereto()->type() << " ";
	    ACT_INT_OUT << exp(t->log_p()) << endl;
	  }
/*	  FOREACH (Transition* t, cur_state->pred()) {
            //! transition probabilities from predecessors:
            if (t->wherefrom()->is_match()) {
 	      ACT_INT_OUT << t->wherefrom()->type() << " ";
	      if (t->wherefrom()->is_match() || t->wherefrom()->is_insert() || t->wherefrom()->is_delete())
                ACT_INT_OUT << t->wherefrom()->column()+1 << " " << t->wherefrom()->phmm()->sub_prof()->sub()->name() << " ";
	      ACT_INT_OUT << t->log_p() << endl;
            }
	  }*/
        }
      }
    }
	    
    //! jump states
    if (Const_Def::jump_version() == linear) {
      pair<State*, State*> jump_states = m_jump_states[i];
      State* cur_M_jump = jump_states.first;
      State* cur_D_jump = jump_states.second;
      ACT_INT_OUT << "JUMP M: " << endl;
      FOREACH (Transition* t, cur_M_jump->succ()) {
        ACT_INT_OUT << t->whereto()->type() << " " << t->whereto()->column()+1 << " " << t->whereto()->phmm()->sub_prof()->sub()->name() << " ";
        ACT_INT_OUT << exp(t->log_p()) << endl;
      }
      ACT_INT_OUT << "JUMP D: " << endl;
      FOREACH (Transition* t, cur_D_jump->succ()) {
        ACT_INT_OUT << t->whereto()->type() << " " << t->whereto()->column()+1 << " " << t->whereto()->phmm()->sub_prof()->sub()->name() << " ";
        ACT_INT_OUT << exp(t->log_p()) << endl;
      }
    }
  }
  ACT_INT_OUT << "C_INSERT: " << endl;
  for (int i=0; i<4; ++i)
    ACT_INT_OUT << c_insert->em_prob(i) << " ";
  ACT_INT_OUT << endl;
  FOREACH (Transition* t, c_insert->succ()) {
    ACT_INT_OUT << t->whereto()->type() << " ";
    if (!t->whereto()->is_c_insert() && !t->whereto()->is_end())
      ACT_INT_OUT << t->whereto()->column() << " " << t->whereto()->phmm()->sub_prof()->sub()->name() << " ";
    ACT_INT_OUT << t->trans_prob() << endl;
  }
    
  ACT_INT_OUT << "C_DELETE: " << endl;
  FOREACH (Transition* t, c_delete->succ()) {
    ACT_INT_OUT << t->whereto()->type() << " " << t->trans_prob() << endl;
  }

/*        ACT_INT_OUT << "N_INSERT PREDECESSORS: " << endl;
        FOREACH (Transition* t, this->n_insert()->pred()) {
          ACT_INT_OUT << t->wherefrom()->type() << " ";
          if (!t->wherefrom()->is_n_insert())
            ACT_INT_OUT << t->wherefrom()->column() << " ";
          ACT_INT_OUT << t->trans_prob() << endl;
        }*/
/*        ACT_INT_OUT << "C_DELETE PREDECESSORS: " << endl;
        FOREACH (Transition* t, this->c_delete()->pred())
           ACT_INT_OUT << t->wherefrom()->type() << " " << t->wherefrom()->column() << " " << t->wherefrom()->phmm()->sub_prof()->sub()->name() << " " << t->trans_prob() << endl;
        ACT_INT_OUT << "END PREDECESSORS: " << endl;
        FOREACH (Transition* t, this->end()->pred())
           ACT_INT_OUT << t->wherefrom()->type() << " " << t->trans_prob() << endl;*/
}

/*****************************************************************************//**
  Output of active states.

  For each query position and each profile HMM and each state type
  output of the active interval and all active states (incl. forward variable).
*********************************************************************************/
void HMM::act_intervals_out(int    qp, 
			    mytype max_vit, 
			    mytype log_beam) {

    ofstream ACT_INT_OUT(Const_Def::active_state_intervals_file_name().c_str(), ios::app);

    ACT_INT_OUT << "Active intervals for query position " << qp << ":" << endl;
    ACT_INT_OUT << "MAX_VIT: " << max_vit << " BEAM: " << log_beam+max_vit << endl;
    for (int s=0; s<m_num_phmms; ++s) {
      ACT_INT_OUT << "subtype " << m_phmms[s]->sub_prof()->sub()->name() << " (M, I, D):" << endl;
      for (int type = (int)Match; type<= (int)Delete; ++type) {
	if (m_act_intervals[qp][s][type].first != -1)
          ACT_INT_OUT << "Type: " << type << " [" << m_act_intervals[qp][s][type].first-m_first_cons_col << ", " << m_act_intervals[qp][s][type].second-m_first_cons_col << "]\t" << flush;
	else
	  ACT_INT_OUT << "Type: " << type << " [" << m_act_intervals[qp][s][type].first << ", " << m_act_intervals[qp][s][type].second << "]\t" << flush;
        int num_act_states = 0;
        FOREACH (ActiveState* cur_act_state, m_act_states[qp][s][type]) {
          if (cur_act_state) {
            num_act_states++;
          }
        }
        ACT_INT_OUT << "num_act_states: " << num_act_states << endl;
        FOREACH (ActiveState* cur_act_state, m_act_states[qp][s][type]) {
          if (cur_act_state) {
            ACT_INT_OUT << "T: " << cur_act_state->state()->type() << "\t C: " << cur_act_state->state()->column()-m_first_cons_col << "\t VIT: " << cur_act_state->viterbi();
	    if (cur_act_state->best_pred()->is_special())
	      ACT_INT_OUT << "\t PRED: T: " << cur_act_state->best_pred()->type() << endl;
	    else
	      ACT_INT_OUT << "\t PRED: T: " << cur_act_state->best_pred()->type() << "\t C: " << cur_act_state->best_pred()->column()-m_first_cons_col << "\t ST: " << cur_act_state->best_pred()->phmm()->sub_name() << endl;
          }
        }
      }
    }
    ACT_INT_OUT << "Special active states: " << endl;
    FOREACH (ActiveState* cur_act_state, m_act_spec_states[qp]) {
      if (cur_act_state) {
        ACT_INT_OUT << "T: " << cur_act_state->state()->type() << "\t VIT: " << cur_act_state->viterbi();
	if (!cur_act_state->state()->is_begin()) {
          if (cur_act_state->best_pred()->is_special())
	    ACT_INT_OUT << "\t PRED: T: " << cur_act_state->best_pred()->type() << endl;
	  else
            ACT_INT_OUT << "\t PRED: T: " << cur_act_state->best_pred()->type() << "\t C: " << cur_act_state->best_pred()->column()-m_first_cons_col << "\t ST: " << cur_act_state->best_pred()->phmm()->sub_name() << endl;
	}
	else
	  ACT_INT_OUT << endl;
      }
    }
    ACT_INT_OUT << "----------------------------------------------" << endl;
  ACT_INT_OUT.close();
}

/*****************************************************************************//**
  Output of active intervals for M, I and D states
  (to plot them with Matlab).

  For each query position 
  the column of the most left and the most right active M, I or D state are given.

  @param[in] q Query
*********************************************************************************/
void HMM::act_intervals_out_for_plot(Query q) {

  int query_size = (int)q.str().size();

  ofstream ACT_INT_OUT(Const_Def::active_state_intervals_for_plot_file_name().c_str(), ios::out);

  vector<int> query_pos;
  vector<int> left_borders;
  vector<int> right_borders;

  for (int qp=0; qp<query_size; ++qp) {
    int cur_act_int_le = most_left_act_col(q, qp)+1;  //! to start numbering at 1
    int cur_act_int_ri = most_right_act_col(q, qp)+1;

    query_pos += qp+1;
    left_borders += cur_act_int_le;
    right_borders += cur_act_int_ri;
  }

  //! name of query:
  ACT_INT_OUT << "name = " << "\"" << q.name() << "\";" << endl;

  //! output of query positions (x axis).
  ACT_INT_OUT << "q = [";
  for (int qp=0; qp<query_size-1; ++qp) {
    ACT_INT_OUT << query_pos[qp] << ", ";
  }
  ACT_INT_OUT << query_pos.back() << "];" << endl;

  //! output of left borders (y axis).
  ACT_INT_OUT << "le = [";
  for (int qp=0; qp<query_size-1; ++qp) {
    ACT_INT_OUT << left_borders[qp] << ", ";
  }
  ACT_INT_OUT << left_borders.back() << "];" << endl;

  //! output of left borders (y axis).
  ACT_INT_OUT << "ri = [";
  for (int qp=0; qp<query_size-1; ++qp) {
    ACT_INT_OUT << right_borders[qp] << ", ";
  }
  ACT_INT_OUT << right_borders.back() << "];" << endl;

  ACT_INT_OUT << "plot(le, q, 'r')" << endl;
  ACT_INT_OUT << "plot(ri, q, 'r')" << endl;
}


/*****************************************************************************//**
  Output details about Viterbi path and emission probabilities in tabular form
*********************************************************************************/
void HMM::output_details(Query query, vector<State*> vit_path_states, vector<string> vit_path_subs)
{
	Html_Tab tab;
	tab.set_next_cell_empty(4);
	for (unsigned i = 0; i < m_phmms.size(); ++i) {
		tab.set_next_cell(m_phmms[i]->sub_name());
		tab.curr_cell()->set_col_span(NUM_NUCL);
	}
	for (unsigned i = 0; i < m_phmms.size(); ++i) {
		tab.set_next_cell(m_phmms[i]->sub_name());
		tab.curr_cell()->set_col_span(NUM_NUCL);
	}
	tab.next_row();
	
	tab.set_next_cell("Col");
	tab.set_next_cell("Type");
	tab.set_next_cell("Sub");
	tab.set_next_cell("Q");
	tab.set_vert_line(tab.curr_col());	
	
	for (unsigned i = 0; i < m_phmms.size(); ++i) {		
		for (unsigned j = 0; j < NUM_NUCL; ++j)
			tab.set_next_cell(Const_Def::alphabet()->character(j));
		tab.set_vert_line(tab.curr_col());
	}
	for (unsigned i = 0; i < m_phmms.size(); ++i) {		
		for (unsigned j = 0; j < NUM_NUCL; ++j)
			tab.set_next_cell(Const_Def::alphabet()->character(j));
		tab.set_vert_line(tab.curr_col());
	}

	for (unsigned i = 0; i < m_phmms.size(); ++i)
		tab.set_next_cell(m_phmms[i]->sub_name());
	
	tab.set_curr_type(NORM);
	vector<double> v = list_of(0.0)(0.01)(0.025)(0.1)(0.2)(0.3333)(0.6667)(0.8)(0.9)(0.975)(0.99)(1000.0);
	Spectrum freq_spec(v);
	v = list_of(0.0)(1 / 5)(1 / 1.5)(1 / 1.1)(1 / 1.025)(1 / 1.01)(1.01)(1.025)(1.1)(1.5)(5)(numeric_limits<double>::max());
	Spectrum ratio_spec(v);
	unsigned query_pos = 0;
	
	for (unsigned i = 0; i < vit_path_states.size(); ++i) {
		unsigned pos = vit_path_states.size() - 1 - i;
		char state_type;
		State* curr_state = vit_path_states[pos];
		if (curr_state->is_match())
			state_type = 'M';
		else
			if (curr_state->is_insert())
				state_type = 'I';
			else
				if (curr_state->is_delete())
					state_type = 'D';
				else
					state_type = 'O';
		char query_char = curr_state->is_mute() ? ' ' : query.str(query_pos++);
		unsigned col = curr_state->column();
		tab.set_next_cell(format("%4d") % col);
		tab.set_next_cell(state_type);
		tab.set_next_cell(curr_state->sub_name());
		tab.set_next_cell(query_char);
	
		if (curr_state->is_match()) {
			for (unsigned ph = 0; ph < m_phmms.size(); ++ph) {
				for (unsigned j = 0; j < NUM_NUCL; ++j) {
					if (!m_phmms[ph]->has_state(col, Match))
						tab.set_next_cell("-");
					else {
						tab.set_next_cell(format("%3d") % m_phmms[ph]->sub_prof()->col_profs()[col]->frequencies()[j]);
						if (query_char == Const_Def::alphabet()->character(j)) {
							tab.curr_cell()->set_bold();
							tab.curr_cell()->set_large();
						}
						if (m_phmms[ph]->sub_name() == curr_state->sub_name())
							tab.curr_cell()->bg_color() = "#FFE8E8";
						tab.curr_cell()->font_color() = freq_spec.color_bound((double)m_phmms[ph]->sub_prof()->col_profs()[col]->frequencies()[j] / 3);
					}
				}
			}
			
			double vit_ep = 0;
			for (unsigned ph = 0; ph < m_phmms.size(); ++ph) {				
				for (unsigned j = 0; j < NUM_NUCL; ++j) {
					if (!m_phmms[ph]->has_state(pos, Match))
						tab.set_next_cell("-");
					else {
							tab.set_next_cell(format("%1.4f") % m_phmms[ph]->state(pos, Match)->em_prob(j));
						if (query_char == Const_Def::alphabet()->character(j)) {
							tab.curr_cell()->set_bold();
							tab.curr_cell()->set_large();
							if (m_phmms[ph]->sub_name() == curr_state->sub_name())
								vit_ep = m_phmms[ph]->state(pos, Match)->em_prob(Const_Def::alphabet()->idx(query_char));										
						}
						if (m_phmms[ph]->sub_name() == curr_state->sub_name())
							tab.curr_cell()->bg_color() = "#FFE8E8";
						tab.curr_cell()->font_color() = freq_spec.color_bound(m_phmms[ph]->state(pos, Match)->em_prob(j));
					}
				}
			}
			
			for (unsigned ph = 0; ph < m_phmms.size(); ++ph) {
				if (m_phmms[ph]->sub_name() == curr_state->sub_name())
					tab.set_next_cell_empty(1);
				else {
					if (!m_phmms[ph]->has_state(pos, Match))
						tab.set_next_cell("-");
					else {					
						double curr_ep = m_phmms[ph]->state(pos, Match)->em_prob(Const_Def::alphabet()->idx(query_char));
						tab.set_next_cell(format("%1.4f") % (vit_ep / curr_ep));
						tab.curr_cell()->font_color() = ratio_spec.color_bound(vit_ep / curr_ep);
						tab.curr_cell()->set_align_right();
					}
				}
			}
		}
		
//		if (!(count % COLORED_ROWS_LEN))
//			tab.set_row_color(tab.curr_row(), "#FFFFC0");
//		++count;
		
		tab.next_row();
	}
	
	tab.output(Const_Def::hmm_analysis_file_name());
}
