/**
 * \file html_tab.h
 */

#ifndef _HTMLTAB_H
#define _HTMLTAB_H

#include <map>
#include <set>
#include <string>
#include <vector>
#include <boost/shared_ptr.hpp>
#include <boost/multi_array.hpp>
#include "html_cell.h"


/**
 * a vertical line in Html_Tab
 */
struct Vert_Line
{
	Vert_Line() : color("black") { }
	Vert_Line(std::string p_color) : color(p_color) { }
	std::string color;
};


/**
 * color spectrum
 */
class Spectrum
{
public:
	Spectrum(unsigned num_colors);
	Spectrum(std::vector<double> boundaries);
	std::string color(unsigned idx) const { return m_colors[idx]; }
	std::string color_bound(double x) const;
	unsigned num_colors() const { return m_colors.size(); }
private:
	std::vector<std::string> m_colors;
	bool m_has_boundaries;
	std::vector<double> m_boundaries;

	std::vector<std::string> generate_colors(unsigned num_colors);
};


/**
 * stores data in tabular form, can output this data as HTML file
 */
class Html_Tab
{
public:
	Html_Tab();

	boost::shared_ptr<Html_Cell> curr_cell() const { return m_cells[m_curr_type][m_curr_row][m_curr_col]; }
	void set_next_cell_empty(unsigned num);
	template<class T> void set_next_cell(T t);
	template<class T> void set_cell(Cell_Type type, unsigned row, unsigned col, T t);
	void next_row() { m_curr_col = 0; ++m_curr_row; m_is_new_row = true; }
	void hide_cols(unsigned begin, unsigned end);
	bool is_col_hidden(unsigned col) const { return m_hidden_cols.count(col); }

	void set_row_color(unsigned row, std::string color) { m_row_colors[row] = color; }
	bool has_row_color(unsigned row) const { return m_row_colors.count(row); }
	std::string row_color(unsigned row) const;
	void set_curr_type(Cell_Type type) { m_curr_type = type; m_curr_col = 0; m_curr_row = 0; m_is_new_row = true; }
	unsigned curr_col() const { return m_curr_col; }
	unsigned curr_row() const { return m_curr_row; }
	void set_vert_line(unsigned left_col, std::string color = "black");
	bool has_vert_line(unsigned left_col) const { return m_vert_lines.count(left_col); }
	Vert_Line vert_line(unsigned left_col) const;

	void output(std::string filename) const;
private:	
	std::string get_output_cells(Cell_Type type) const;
	
	unsigned m_num_cols; 
	std::vector<unsigned> m_num_rows;
	std::vector< boost::multi_array<boost::shared_ptr<Html_Cell>, 2> > m_cells;
	std::map<unsigned, Vert_Line> m_vert_lines;
	std::map<unsigned, std::string> m_row_colors;
	std::set<unsigned> m_hidden_cols;
	unsigned m_curr_col;
	unsigned m_curr_row;
	Cell_Type m_curr_type;
	bool m_is_new_row;
};

#endif    // _HTMLTAB_H
