#include <vector>
#include <fstream>
#include <iostream>
#include <boost/assign/list_of.hpp>
#include <boost/assign/std/vector.hpp>
#include <boost/assign/std/set.hpp>
#include <boost/lexical_cast.hpp>
#include "html_tab.h"
#include "html_cell.h"

using namespace std;
using namespace boost;
using namespace boost::assign;

Spectrum::Spectrum(unsigned num_colors)
: m_colors(generate_colors(num_colors)), m_has_boundaries(false)
{ }


Spectrum::Spectrum(vector<double> boundaries)
: m_colors(generate_colors(boundaries.size() - 1)), m_has_boundaries(true),
m_boundaries(boundaries)
{
	for (unsigned i = 0; i < num_colors(); ++i)
		if (m_boundaries[i] > m_boundaries[i + 1])
			throw runtime_error("boundaries not ascending");
}


string Spectrum::color_bound(double x) const
{
	if (!m_has_boundaries)
		throw runtime_error("Spectrum has no boundaries");
	for (unsigned i = 0; i < num_colors(); ++i)
		if ((m_boundaries[i] <= x) && (x <= m_boundaries[i + 1]))
			return color(i);
	stringstream ss;
	ss << "x (" << x << ") lies outside the boundaries";
	throw runtime_error(ss.str());
}


vector<string> Spectrum::generate_colors(unsigned num_colors)
{
	vector<string> ret;
	switch (num_colors) {
	case 5:
		ret = list_of("green")("blue")("goldenrod")("orangered")("red");
		break;
	case 11:
		ret = list_of("lawngreen")("darkgreen")("blue")("darkviolet")("deeppink")("goldenrod")("orangered")("red")("darkred")("gray")("black");
		break;
	default:
		throw runtime_error("Spectrum not implemented for this case");
	}
	return ret;
}


/**
 * construct empty table
 */
Html_Tab::Html_Tab() : m_num_cols(0), m_curr_col(0), m_curr_row(0), m_curr_type(HEAD), m_is_new_row(true)
{ 
	m_num_rows = list_of(0)(0);  
	m_cells.resize(2);
}


/**
 * set content of next cell(s) in the row to an empty string
 * @param num number of cells affected
 */
void Html_Tab::set_next_cell_empty(unsigned num)
{
	for (unsigned i = 0; i < num; ++i)
		set_next_cell("");
}


/**
 * set content of next cell in the row
 * @param t the content
 */
template<class T> void Html_Tab::set_next_cell(T t) 
{ 
	if (!m_is_new_row) 
		m_curr_col += curr_cell()->col_span();
	set_cell(m_curr_type, m_curr_row, m_curr_col, t); 
	m_is_new_row = false;
}


/**
 * set content of a cell
 * @param type the cell type
 * @param row the row
 * @param col the column
 * @param t the content
 */
template<class T> void Html_Tab::set_cell(Cell_Type type, unsigned row, unsigned col, T t)
{ 
	vector<bool> is_resize = list_of(false)(false);   // resize header / normal part of table
	unsigned old_num_cols = m_num_cols;
	vector<unsigned> old_num_rows = m_num_rows;
	
	if (col >= m_num_cols) {
		is_resize = list_of(true)(true);
		m_num_cols = col + 1;
	}
		
	if (row >= m_num_rows[type]) {
		is_resize[type] = true;
		m_num_rows[type] = row + 1;
	}
	
	for (unsigned ctp = HEAD; ctp <= NORM; ++ctp) {
		Cell_Type curr_type = (Cell_Type)ctp;
		multi_array<Html_Cell, 2>::extent_gen ext;
		if (is_resize[curr_type]) {
			m_cells[curr_type].resize(ext[m_num_rows[curr_type]][m_num_cols]);
			for (unsigned i = old_num_rows[curr_type]; i < m_num_rows[curr_type]; ++i)
				for (unsigned j = 0; j < m_num_cols; ++j) {
					if (!m_cells[curr_type][i][j])
						m_cells[curr_type][i][j].reset(new Html_Cell(curr_type, i, j));
				}
			for (unsigned i = 0; i < m_num_rows[curr_type]; ++i)
				for (unsigned j = old_num_cols; j < m_num_cols; ++j)
					if (!m_cells[curr_type][i][j])
						m_cells[curr_type][i][j].reset(new Html_Cell(curr_type, i, j));
		}
	}
	
	m_cells[type][row][col]->str() = lexical_cast<string>(t);
}


/**
 * hide given interval of columns
 * @param begin most left column to hide
 * @param end most right column to hide
 */
void Html_Tab::hide_cols(unsigned begin, unsigned end)
{
	for (unsigned i = begin; i <= end; ++i)
		m_hidden_cols += i;
}


/// get vertical line
Vert_Line Html_Tab::vert_line(unsigned left_col) const
{
	if (!m_vert_lines.count(left_col))
		throw runtime_error("no vertical line defined");
	return m_vert_lines.find(left_col)->second;
}


/// get row color
string Html_Tab::row_color(unsigned row) const
{
	if (!m_row_colors.count(row))
		throw runtime_error("no row color defined");
	return m_row_colors.find(row)->second;
}


/**
 * output table to HTML file
 * @param filename the output file
 */
void Html_Tab::output(string filename) const
{
	ofstream out(filename.c_str());
	out << "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">" << endl;
	out << "<html>" << endl;
	out << "<style type=\"text/css\">" << endl;
	out << "body {font-family: monospace;}" << endl;
	out << "table {empty-cells: show; border-spacing: 0px}" << endl;
	out << "td {padding: 2px; margin: 0px;}" << endl;
//	.rightbar { border-right: 1px solid black; }
//	.right { text-align: right; }
//	.range_0 { color: lawngreen; }
	out << "</style>" << endl;

	out << "<body>" << endl;
	out << "<table border=\"0\">" << endl;
	
	out << "<thead>" << endl;
	out << get_output_cells(HEAD);
	out << "</thead>" << endl;
	out << get_output_cells(NORM);
	
	out << "</table>" << endl;
	out << "</body>" << endl;
	out << "</html>" << endl;
}


/**
 * insert vertical line in table
 * @param left_col the number of the column left to the vertical line
 * @param color the color of the line
 */
void Html_Tab::set_vert_line(unsigned left_col, string color)
{
	m_vert_lines[left_col] = Vert_Line(color);
}


/**
 * get output of cells of given type (for HTML file)
 * @param filename the output file
 * @param type the cell type
 */
string Html_Tab::get_output_cells(Cell_Type type) const
{
	stringstream ss;
	for (unsigned row = 0; row < m_num_rows[type]; ++row) {
		ss << "<tr>" << endl;
		unsigned col = 0;
		while (col < m_num_cols) {
			if (!is_col_hidden(col))
				ss << m_cells[type][row][col]->output(*this);
			col += m_cells[type][row][col]->col_span();
		}
		ss << endl << "</tr>" << endl;
	}
	return ss.str();
}


