// ~/jpHMM/src/jpHMM.cpp
/**************************************************************
 File name	: jpHMM.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 - Call of scripts to extend full-length circular genomes
 - Call of BLAT to prealign query sequences to multiple sequence alignment
 - Read in of query sequences
 - Construction of the model (jpHMM)
 - Call of all algorithms (Viterbi, Forward, etc.)
***************************************************************/

#include "std_header.h"
#include "std_source.h"
#include "error.h"
#include "const_def.h"
#include "alphabet.h"
#include "alignment.h"
#include "emissionPriors.h"
#include "transitionPriors.h"
#include "emTransProfile.h"
#include "query.h"
#include "profileHMM.h"
#include "hmm.h"
#include "extEmTrProbs.h"

using namespace std;

const bool OUTPUT_DEBUG = false;


/***************************************************************//**
  Read in query sequences.

   @param[in] query_file Name of file containing query sequences.
   @param[in] al Alignment object.
   @param[in] hmm HMM object.
   @param[out]query_seqs Vector of query objects.
********************************************************************/
vector<Query*> read_queries(string     query_file,
		            Alignment *al,
		            HMM       *hmm) {

          //! query sequence file in format similar to FASTA
	  //! for circular genomes, this is the file with the extended sequences;
	  //!                       for full-length sequences, information about the original start and end position in the extended sequence are given in the file
	  ifstream queryfile(query_file.c_str(), ios::in);
	  if(!queryfile)
	    throw Error("failure of the query sequence input file");

	  vector<Query*> query_seqs;

	  string str;
	  string cur_q_str;
	  int    cur_no_of_q = 0;
	  string cur_q_name;
	  bool   has_cur_q = false;
	  int    cur_q_start_in_al = 1;                    //! most left column in alignment that query can be aligned to
	  int    cur_q_end_in_al   = al->subs()[0]->len(); //! most right column in alignment that query can be aligned to
          bool   cur_q_is_full_length_seq = true;          //! for circular sequences: sequence is full_length sequence (or fragment)?
	  int    cur_q_start_in_ext_q = 0;                 //! for circular sequences, the query is extended: orig_q_start_in_ext_q = start of orig seq in extended seq
	  int    cur_q_end_in_ext_q   = 0;                 //! for circular sequences, the query is extended: orig_q_end_in_ext_q = start of orig seq in extended seq
	  int    cur_q_ext        = 0;

	  while (!queryfile.eof()) {
	    getline(queryfile, str);
	    
	    //! erase Windows line break:
	    size_t pos = str.find_first_of('\x0D');
	    if (pos != string::npos)
	      str.erase(pos, 1);
	    
	    if (str.substr(0, 1) == ">") {
	      //! init.:
              cur_q_start_in_ext_q = 1;
	      cur_q_end_in_ext_q   = 1;
	      cur_q_ext            = 0;
	      cur_q_is_full_length_seq = true;

	      has_cur_q  = true;

	      string s = str.substr(1);
	      
	      //! replace special characters in query sequence name: '|' and '.' by '_'
	      string char_replace = "_";
	      vector<string> special_chars;
	      special_chars.push_back("|");
	      special_chars.push_back(".");
	      
	      for (vector<string>::iterator it=special_chars.begin(); it<special_chars.end();++it) {
		size_t found = s.find(*it);
		while (found != string::npos) {
		  s.replace(s.find(*it), 1, char_replace);
		  found = s.find(*it);
		}
	      }
	        
	      //! read in query name and - if given - start and end position of the query in the alignment
	      stringstream ss(s);
	      ss >> cur_q_name;
	      
	      //! remove last character in query name if it is a special character
	      unsigned found = cur_q_name.find_last_of(char_replace);
	      if (found == cur_q_name.size()-1)
		cur_q_name.erase(cur_q_name.size()-1);

      
	      vector<int> cur_q_positions_in_al;
	      cur_q_positions_in_al.push_back(1);
	      cur_q_positions_in_al.push_back(hmm->len());
	      
	      //! read in rest of string in '>' line if available
	      //! check if start and end position in the reference genome/alignment are given
	      bool pos_in_al_given = true; //! check if two positions in alignment are given
	      string cur_substr;
	      vector<string> cur_substrings;

	      while (!ss.eof()) {
		ss >> cur_substr;
		cur_substrings.push_back(cur_substr);
	      }
	      
	      if (cur_substrings.size() > 1) {
		//! name of sequence contains at least two more words
		//! check if two last entries are integers, 
		//! these should be the numbers defining the start and end position in the alignment
		
		for (int j=2; j>0; j--){
		  if (pos_in_al_given == true) {
		    int cur_idx = cur_substrings.size()-j;
		    cur_substr = cur_substrings[cur_idx];
		    pos_in_al_given = true;

		    unsigned int i;
		    try {
		      i = lexical_cast<unsigned>(cur_substr); 
		    }
		    catch (bad_lexical_cast const&) {
		      //! last position is not an integer, thus it is part of the sequence name (e.g. Genbank entry) and not positions in the genome
		      pos_in_al_given = false;
		    }
		    int cur_pos = 0;
		    stringstream as(cur_substr);
		    as >> cur_pos;
		    cur_q_positions_in_al[2-j] = cur_pos;
		  }
		}
	      }
	      else {
		//check which size, e.g. if only start position given etc. (see below)
	      }
	      
	      if (pos_in_al_given == true) {
		if (cur_q_positions_in_al[0] < cur_q_positions_in_al[1]) {
		  if (Const_Def::genome_type() == circular) {
		    cur_q_start_in_ext_q = cur_q_positions_in_al[0];
		    cur_q_end_in_ext_q   = cur_q_positions_in_al[1];
		  }
		  else {
		    cur_q_start_in_al = cur_q_positions_in_al[0];
		    cur_q_end_in_al   = cur_q_positions_in_al[1];
		    if (al->has_ref_seq()) {
		      cur_q_start_in_al = max(1, al->ref_seq_pos_to_al_col(cur_q_start_in_al)-MSA_EXT);
		      cur_q_end_in_al   = min(hmm->len(), al->ref_seq_pos_to_al_col(cur_q_end_in_al)+MSA_EXT);
		    }
		  }
		}
		else {
		  pos_in_al_given = false;
		}
	      }
	      
	      cur_q_str = "";
	    }
	    else {
	      if (!has_cur_q)
	        throw Error("Query sequence input file in illegal state. No sequence name given.");
	      else {
		//! erase gaps: "-"
	        size_t found_gap = str.find_first_of("-");
                while (found_gap != string::npos) {
                  str.erase(found_gap, 1);
                  found_gap = str.find_first_of("-");
                }

                cur_q_str += str;
	        bool end_of_q_reached = false;
	        if (!queryfile.eof()) {
	          char c;
	          queryfile >> c;
	          if (queryfile.eof() || c == '>')
	            end_of_q_reached = true;
		  if (!queryfile.eof())
	            queryfile.unget();
	        }
		else
		  end_of_q_reached = true;
		if (end_of_q_reached) {
		  //! subtract -1 for the given positions in the alignment to start counting at 0
		  //! and assure that start and end position in alignment are within consensus columns
                  cur_q_start_in_al = max(cur_q_start_in_al-1, hmm->first_cons_col());
                  cur_q_end_in_al   = min(cur_q_end_in_al-1, hmm->last_cons_col());
		  
		  ++cur_no_of_q;
		  
	          if (Const_Def::genome_type() == circular && (cur_q_start_in_ext_q > 1 || cur_q_end_in_ext_q < (int)cur_q_str.size())) {
		    //! if one of the values is = 0, the sequence is not a full-length sequence but a fragment -> not extended by script
		    cur_q_is_full_length_seq = true;
                    cur_q_ext = cur_q_start_in_ext_q-1;
		  }
		  else {
		    //! sequence hasn't been extended
		    cur_q_is_full_length_seq = false;
                    cur_q_start_in_ext_q = 1;
		    cur_q_end_in_ext_q   = (int)cur_q_str.size();
		  }
                  cur_q_start_in_ext_q--; //! to start at pos. 0 instead of 1.
		  cur_q_end_in_ext_q--;
		  
		  //! if not blat version
		  //! extend start in al and end in al +- 500 (siehe alte jpHMM Version)
		  
	          Query* cur_query = new Query(cur_q_name, cur_no_of_q, cur_q_str, cur_q_start_in_al, cur_q_end_in_al, cur_q_is_full_length_seq, cur_q_ext, cur_q_start_in_ext_q, cur_q_end_in_ext_q);

		  query_seqs += cur_query;
		}
	      }
	    }
	  }
	  return query_seqs;

}

/**********************************************************************************//**
  Call extension script.
  
  Call perl script to extend query sequences and multiple sequence alignment
  for circular genomes.

   @param[in] infile Name of file containing the multiple sequence alignment.
   @param[in,out] orig_al_length Length of original non-extended input alignment.
   @param[in] query_file Name of file containing query sequences.
   @param[out] ext_file_names Names of files containing the extended queries and MSA.
***************************************************************************************/
pair<string, string> call_ext_script(string infile,
				     int   *orig_al_length,
	                             string query_file) {

  string ext_script_name  = Const_Def::ext_script_name();
  pair<string, string> ext_file_names;
  
  //! Call script that extends MSA (2xMSA+certain part) and queries at both ends
  char buffer[1024];
  sprintf(buffer, "[ -x %s ]", ext_script_name.c_str()); 
  int ext_script_is_exec = system(buffer);

  if (ext_script_is_exec == 0) {
    //! script for extension (for circular genomes) is executable
    sprintf(buffer, "%s %s %s %f %d", ext_script_name.c_str(), infile.c_str(), query_file.c_str(), Const_Def::msa_ext(), Const_Def::query_ext());

    //! for LINUX only (for WINDOWS: _popen ...):
    FILE *p = popen(buffer, "r");
    
    if (p != NULL) {
      char s [100];
      char* c = fgets(s, 100, p);
      fclose(p);
      c = 0;
      
      stringstream ss(s);
      int length = 0;

      ss >> length;
      *orig_al_length = length;

      //! Script for extension ran successfully.
      //! => assign new file names
      string str = ".ext";
      infile.assign((infile.c_str()+str).c_str());
      query_file.assign((query_file.c_str()+str).c_str());
  
      //! change name for query and MSA files for rest of program:
      ext_file_names = make_pair(infile, query_file);
    }
  }

  return ext_file_names;
}

/*****************************************************************************************//**
  Call BLAT.
  
  Call perl script that calls BLAT to align query sequences to MSA
  and evaluated the aligned columns.

   @param[in] infile Name of file containing the multiple sequence alignment.
   @param[in] query_file Name of file containing query sequences.
   @param[out] act_al_intervals Active alignment intervals for each query sequence position.
**********************************************************************************************/
vector<vector<pair<int, int> > > call_blat(string infile,
			                   string query_file) {

  vector<vector<pair<int, int> > > act_al_intervals;

  string blat_script_name  = Const_Def::blat_script_name();
  string blat_program_name = Const_Def::blat_program_name();
  string active_alignment_columns_file_name = Const_Def::active_alignment_columns_file_name();
    
  //! Call script that calls BLAT and evaluates the resulting alignments for all query sequences:
  char buffer[1024];
  sprintf(buffer, "[ -x %s ]", blat_script_name.c_str()); 
  int blat_script_is_exec = system(buffer);
  sprintf(buffer, "[ -x %s ]", blat_program_name.c_str()); 
  int blat_prog_is_exec = system(buffer);

  if (blat_script_is_exec == 0 && blat_prog_is_exec == 0) {
  
    sprintf(buffer, "%s %s %s %s %s", blat_script_name.c_str(), blat_program_name.c_str(), infile.c_str(), query_file.c_str(), active_alignment_columns_file_name.c_str());
    int i = system(buffer);

    if (i == 0) {
      //! Blat ran successfully
    
      //! Read in columns in alignment to which queries are aligned (assume queries sorted in same order as in input file):
      int   peek;
      string str;

      ifstream ACT_AL_COLUMNS(active_alignment_columns_file_name.c_str(), ios::in);
      if (!ACT_AL_COLUMNS)
        throw Error("failure of the file containing active alignment columns for each query sequence.");

      while ((peek = ACT_AL_COLUMNS.peek()) != EOF) {
        if (peek == '>')
          getline(ACT_AL_COLUMNS, str);
        else {
          getline(ACT_AL_COLUMNS, str);
      
          tokenizer<> tok(str);
          vector<int> cur_act_al_columns;

          for (tokenizer<>::iterator it=tok.begin(); it!=tok.end(); ++it) {
	    stringstream ss(*it);
	    int test;
	    ss >> test;
	    cur_act_al_columns += test;
          }
          //! create intervals (pairs) for each qp
          //! and save them in current vector of pairs.
          vector<pair<int, int> > cur_act_al_intervals;
          vector<int>::const_iterator it = cur_act_al_columns.begin();
          while (it != cur_act_al_columns.end()) {
	    int cur_al_int_start = (*it);
	    ++it;

	    int cur_al_int_end = (*it);
            pair<int, int> cur_al_int = make_pair (cur_al_int_start, cur_al_int_end);
	    cur_act_al_intervals += cur_al_int;
	    ++it;
          }
          act_al_intervals += cur_act_al_intervals;
        }
      }
    }
  }
  return act_al_intervals;
}

/*******************************************************************************//**
  Determine beam width.
  
  Determine beam width for beam-search algorithm included in Viterbi algorithm
  depending on length of current query sequence.

   @param[in] beam_df Default beam width.
   @param[in] query Query object.
   @param[out] beam Beam width for current query sequence.
************************************************************************************/
mytype determine_beam_width(mytype beam_df,
			    Query  q) {

  mytype beam = beam_df;
  int cur_query_size = q.str().length();
  
  if (q.prog_version() == blat && Const_Def::virus_type() == HBV)
    beam = 1e-30;
  else {
    if (cur_query_size <= 500)
      beam = 1e-45;
    else if (500 < cur_query_size && cur_query_size <= 1200)
      beam = 1e-40;
    else if (1200 < cur_query_size && cur_query_size <= 1500)
      beam = 1e-35;
    else if (1500 < cur_query_size && cur_query_size <= 3200)
      beam = 1e-30;
    else if (3200 < cur_query_size && cur_query_size <= 5000)
      beam = 1e-25;
    // else: beam_df
  }

  return beam;
}

/***********************************************************************************************************************//**
  jpHMM.
  
  Build jpHMM.
  Call all jpHMM algorithms, e.g. Viterbi, Forward.

   @param[in] emission_priors_file Name of file containing emission priors.
   @param[in] transition_priors_file Name of file containing transition priors.
   @param[in] infile Name of file containing the multiple sequence alignment.
   @param[in] query_file Name of file containing query sequences.
   @param[in] beam_is_given is beam width given by user?.
   @param[in] use_ext_em_tr_probs use external emission and transition probabilities?.
   @param[in] ep_file_name emission probabilities file if external em. and trans. probs are used.
   @param[in] tp_file_name transition probabilities file if external em. and trans. probs are used.
   @param[in] cons_file_name information about consensus columns in the subtypes if external em. and trans. probs are used.
*****************************************************************************************************************************/
void jpHMM(string emission_priors_file,
	   string transition_priors_file,
	   string infile,
	   string query_file,
	   bool   beam_is_given, 
	   bool   use_ext_em_tr_probs,
	   string ep_file_name, 
	   string tp_file_name, 
	   string cons_file_name) {

  /*****************************************************************
      recombination prediction for each query sequence with jpHMM:
   *****************************************************************/

  /*****************************************************************
   ***        1.) sequence(s) input                              ***
   ***        2.) HMM for each sequence                          ***
   ***             -> depending on HXB2 positions                ***
   *****************************************************************/

  /* constant parameters:        */
  Const_Def::init_out_files();
  EmissionPriors   *e  = new EmissionPriors(emission_priors_file);
  TransitionPriors *t  = new TransitionPriors(transition_priors_file);
  
  bool al_is_extended = false;
  int orig_al_length = 0;

  //! Call Script for extending queries and MSA (for circular genomes)
  if (Const_Def::genome_type() == circular) {
    pair<string, string > ext_file_names = call_ext_script(infile, &orig_al_length, query_file);
    if (!ext_file_names.first.empty()) {
      //! query_file_name and msa_file_name are changed!
      infile     = ext_file_names.first;
      query_file = ext_file_names.second;
      al_is_extended = true;
    }
    else {
      Const_Def::genome_type() = non_circular;
    }
  }
  
  Alignment *al = new Alignment(infile, al_is_extended, orig_al_length);

  ExtEmTrProbs* ext_em_tr_probs = 0;
  if (use_ext_em_tr_probs)
          ext_em_tr_probs = new ExtEmTrProbs(ep_file_name, tp_file_name, cons_file_name);
  EmTransProfile *etpr = use_ext_em_tr_probs ? new EmTransProfile(al, true, ext_em_tr_probs) : new EmTransProfile(al, false);

  //! Build HMM
  HMM *hmm = use_ext_em_tr_probs ? new HMM(al, *e, *t, *etpr, true, ext_em_tr_probs) : new HMM(al, *e, *t, *etpr, false);

  //! Read in query sequences.
  vector<Query*> query_seqs = read_queries(query_file, al, hmm);

  //! Call BLAT for alignment of queries to MSA sequences
  vector<vector<pair<int, int> > > act_al_intervals;
  if (Const_Def::prog_version() == blat) {
    act_al_intervals = call_blat(infile, query_file);

    //! beam search version, if BLAT alignment not successful
    if (act_al_intervals.empty())
      Const_Def::prog_version() = beam_search;
  }
  
  //! run jpHMM for each query sequence

  for (vector<Query*>::iterator it=query_seqs.begin(); it!=query_seqs.end(); ++it) {
    Query* cur_query = (*it);
    string cur_query_name  = cur_query->name();
    string cur_query_str   = cur_query->str();
    int    cur_query_size  = cur_query_str.length();
    int    cur_query_start = cur_query->start_in_al();
    int    cur_query_end   = cur_query->end_in_al();
    cur_query->prog_version() = Const_Def::prog_version(); //! could have been changed if blat run not successful
    
    //int    cur_num_gaps    = cur_query->num_gaps(cur_query_size-1);

    /***********************************************************
       Output of query sequence info:
    ************************************************************/
    //! beam width for current query.
    mytype beam_df = Const_Def::beam_width();
    cur_query->beam_width() = beam_df;

    if (!beam_is_given) {
      cur_query->beam_width() = determine_beam_width(beam_df, *cur_query);
    }
    
    vector<pair<int, int> > cur_act_al_intervals;    
    
    if (Const_Def::prog_version() == blat) {
      int cur_query_no = cur_query->no();
      cur_act_al_intervals = act_al_intervals[cur_query_no-1];
      cur_query_start = max(hmm->first_cons_col(), cur_act_al_intervals[0].first);
      cur_query_end   = min(hmm->last_cons_col(), cur_act_al_intervals.back().second);
      cur_query->start_in_al() = cur_query_start;
      cur_query->end_in_al()   = cur_query_end;
    }

    bool is_recomb_pred_possible = true;
    if (hmm->first_cons_col() > cur_query_end || hmm->last_cons_col() < cur_query_start || hmm->first_cons_col() > hmm->last_cons_col()) {
      //! no recombination prediction possible.
      is_recomb_pred_possible = false;
    }
    else {
      /********************************************************//**
         Recombination prediction.
      ************************************************************/
      HMM* cur_hmm;
      bool is_local_hmm = false;
      if (hmm->first_cons_col() != cur_query->start_in_al() || hmm->last_cons_col() != cur_query->end_in_al()) {
	is_local_hmm = true;
        cur_hmm = new HMM(*hmm, cur_query->start_in_al(), cur_query->end_in_al());
	cur_hmm->local_hmm(*e, *etpr);
      }
      else {
        cur_hmm = hmm;
      }
      //! Viterbi algorithm and backtracking (-> recombination prediction) for current query.
      bool successf_1stvit_run = cur_hmm->vit_and_fw_algo(cur_query, cur_act_al_intervals);

      if (!successf_1stvit_run) {
	if (Const_Def::prog_version() == blat) {
	  cur_hmm->clear_act_states(*cur_query);
	  
	  //! restart algorithm without blat alignment
	  cur_query->start_in_al() = cur_hmm->first_cons_col();
          cur_query->end_in_al()   = cur_hmm->last_cons_col();

          cur_query->prog_version() = beam_search;
	  //! adjust beam width:
          cur_query->beam_width() = beam_df;
          if (!beam_is_given) {
            cur_query->beam_width() = determine_beam_width(beam_df, *cur_query);
          }
	  bool successf_2ndvit_run = cur_hmm->vit_and_fw_algo(cur_query, cur_act_al_intervals);
	  if (!successf_2ndvit_run) {
	    //! no prediction possible due to beam search or anything else.
	    is_recomb_pred_possible = false;
	  }
	}
	else {
	  //! no prediction possible due to beam search or anything else.
	  is_recomb_pred_possible = false;
	}
      }
      if (is_recomb_pred_possible) {
        //! Backtracking Viterbi path
	cur_hmm->vit_back(al, *cur_query);
	//! Backward algorithm for current query.
        cur_hmm->bw_algo(cur_query_str);
	//! Calculation of posterior probabilities for current query.
        cur_hmm->post_probs(*cur_query);
      }
      cur_hmm->clear_act_states(*cur_query);

      if (is_local_hmm) {
        delete cur_hmm;
      }
    }
    if (!is_recomb_pred_possible) {
      //! recombination prediction was not possible, due to different reasons
      //! => output: N/A to full-length sequence
      ofstream RECOMB(Const_Def::recomb_file_name().c_str(), ios::app);
      ofstream RECOMB_WO_POS(Const_Def::pred_subtypes_file_name().c_str(), ios::app);
      ofstream RECOMB_GAPS(Const_Def::recomb_incl_gaps_file_name().c_str(), ios::app);
      ofstream VITERBI_GFF(Const_Def::viterbi_gff_file_name().c_str(), ios::app);  
      ofstream POST_PATH(Const_Def::posterior_path_file_name().c_str(), ios::app);
      ofstream VAR_REGS(Const_Def::variable_regions_file_name().c_str(), ios::app);

      RECOMB << ">" << cur_query_name << " (";
      RECOMB_WO_POS << ">" << cur_query_name << "\t" << flush;
      RECOMB_GAPS << ">" << cur_query_name << " (";
      VITERBI_GFF << "#" << endl
                  << "# predicted subtypes for sequence '" << cur_query_name << "': " << endl;
      POST_PATH << ">" << cur_query_name << endl;
      VAR_REGS << ">" << cur_query_name << endl;
  
      //! output of chosen parameters and program versions etc.
      if (Const_Def::genome_type() == circular) {
        RECOMB << "circular genome, ";
        RECOMB_WO_POS << "circular genome, ";
        RECOMB_GAPS << "circular genome, ";
        if (cur_query->is_full_length_seq()) {
          RECOMB << "full-length, ";
          RECOMB_WO_POS << "full-length, ";
          RECOMB_GAPS << "full-length, ";
        }
        else {
          RECOMB << "fragment, ";
          RECOMB_WO_POS << "fragment, ";
          RECOMB_GAPS << "fragment, ";
        }
      }
      if (Const_Def::jump_version() == linear) {
        RECOMB << "linear model, ";
        RECOMB_WO_POS << "linear model, ";
        RECOMB_GAPS << "linear model, ";
      }
      if (cur_query->prog_version() == blat) {
        RECOMB << "blat, ";
        RECOMB_WO_POS << "blat, ";
        RECOMB_GAPS << "blat, ";
      }
      //! also for jpHMM based on BLAT alignment beam-search algo applied
      RECOMB << "bw=" << cur_query->beam_width() << ")" << endl;
      RECOMB_WO_POS << "bw=" << cur_query->beam_width() << ", jp=" << Const_Def::jump_prob() << "\t" << flush;
      RECOMB_GAPS << "bw=" << cur_query->beam_width() << ")" << endl;
      
      RECOMB << "1\t" << cur_query_size << "\tN/A" << endl << endl;
      RECOMB_WO_POS << "N/A" << endl;
      RECOMB_GAPS << "1\t" << cur_query_size << "\tN/A" << endl << endl; // anpassen!
      VITERBI_GFF << cur_query_name << "\t" << "jpHMM\t" << "N/A" << "\t" << "1\t"
                  << cur_query_size << "\t.\t+\t." << endl << endl;
      POST_PATH << "1\t" << cur_query_size << "\tN/A" << endl << endl;
      VAR_REGS << ">" << cur_query_name << endl << endl;

      RECOMB.close(); RECOMB_WO_POS.close(); RECOMB_GAPS.close(); VITERBI_GFF.close(); POST_PATH.close(); VAR_REGS.close();
    }
  }

  for (vector<Query*>::iterator it=query_seqs.begin(); it!=query_seqs.end(); ++it) {
    delete (*it);
  }
  delete e;
  delete t;
  delete al;
  delete etpr;
  delete hmm;
}

