#!/usr/bin/perl

use warnings;
use strict;
use Cwd;
use List::Util qw[min max];
####################################
# convert jpHMM recombination prediction output (recombination.txt)
# in Circos readable format
# to plot the predicted recombination in a circle
####################################
# output (recomb prediction and post probs)
# is mapped to reference sequence AM282986
####################################
# breakpoint e.g.:
# 1	830	A
# 831	3000	B
# -> 830 is saved as BP position
####################################
# TODO: mapping of qp to ref seq for post probs -> ok?
# TODO: also plot exact BP position in addition to mid point of BPI?
####################################


if (scalar(@ARGV) != 6) {
  print "Enter [dir_for_circos_input_files] and [recomb_pred_file_name] and [recomb_incl_UR_and_BPI_pred_file_name] and [query_alignment_to_msa_file_name] and [msa2refSeq_file_name] and [output_directory]\n\n";
  print "Example for calling the script in the jpHMM output directory:\n";
  die "~/Programs/jpHMM/src/jphmm_out_to_circos.pl ~/Programs/jpHMM/src/circos_input_files/ recombination.txt recombination_incl_UR_and_BPR.txt alignment_to_msa.txt ~/Programs/jpHMM/input/hbv_msa_to_refseq.txt ./\n";
}


####################################
# directory with circos executable
my $circosDir = "~/Circos/circos-0.52/bin/circos";
####################################

####################################
# input:
####################################
my $inputDir          = $ARGV[0];
my $predRecomb        = $ARGV[1]; # recombination.txt
my $predRecombInclBPI = $ARGV[2]; # recombination_incl_UR_and_BPI.txt
my $alToMSA           = $ARGV[3]; # query sequence positions aligned to msa columns
my $msaToRefSeq       = $ARGV[4]; # msa positions mapped to ref seq
my $outputDir         = $ARGV[5]; # directory for Circos output files

if (! (-e $outputDir)) {
  my $command0 = "mkdir $outputDir";
  system($command0);
}
####################################

my $workingDir = getcwd;

####################################
# file names:
####################################
# txt:
my $txt_script_ext = ".txt";

# script name:
my $scriptName     = "recomb_prediction.conf";

# posterior probabilities:
my $post_probs_in_file_name  = "posterior_probabilities_for_seq_";
my $post_probs_out_file_name = "post_probs_";

# Input files for Circos:
# 1) config files;
my $colors_conf   = "etc/colors.conf";
my $fonts_conf    = "etc/fonts.conf";
my $ideogram_conf = "ideogram.conf";
my $ticks_conf    = "ticks.conf";

# 2) genes
my $genes          = "genes.formatted_inner.txt";
my $subtype_colors = "subtypes.formatted.txt";

# 3) labelling for plotted genes.
my $text_genes     = $inputDir."text.genes*";
my $text_genes_P   = "text.genes.txt.P";
my $text_genes_C_S = "text.genes.txt.C_S";
my $text_genes_X   = "text.genes.txt.X";
my $text_genes_DR  = "text.genes.txt.DR";
my $text_subtypes  = "text.subtypes.txt";
my $text_legend    = "text.legend.txt";

# Output file names of Circos:
my $circos_out_file_name     = "recombination_prediction_";
my $circos_out_file_name_png_ext = ".png";
my $circos_out_file_name_svg_ext = ".svg";


# Output of jpHMM for single sequences in Circos format (read in by Circos program)
# 1) karyotype
my $karyotype_file_name     = "karyotype"; # different for each sequence/recomb. prediction
my $karyotype_file_name_ext = ".hbv.txt";

# 2) breakpoint interval positions
my $bp_out_file_name     = "text.breakpoints";

# 3) uncertainty regions positions
my $ur_out_file_name     = "text.uncertainty";

# TODO: also predicted BP?

# File names for 
# predicted recombination mapped to reference sequence:
my $mapped_recomb_out_file_name = $predRecomb;
my @dummy = split('\.', $predRecomb);
if (scalar(@dummy)>0) {
  $mapped_recomb_out_file_name = $dummy[0];
}
# predicted recombination incl. UR and BPI mapped to reference sequence:
my $mapped_recomb_incl_BPI_UR_out_file_name = $predRecombInclBPI;
@dummy = split('\.', $predRecombInclBPI);
if (scalar(@dummy)>0) {
  $mapped_recomb_incl_BPI_UR_out_file_name = $dummy[0];
}
# predicted UR and BPI mapped to reference sequence:
my $mapped_BPI_UR_out_file_name = "uncertainty_and_BP_regions";
my $mapped_recomb_out_file_name_ext = ".mapped.txt";
####################################

####################################
# Definition of colors for subtypes:
# I = Insertion, UR = uncertainty region
my $color_insert = "grey";
my %colors = (
  'A' => 'red4',
  'B' => 'dblue',
  'C' => 'green4',
  'D' => 'yellow',
  'E' => 'orange3',
  'F' => 'lpurple',
  'G' => 'violet',
  'H' => 'lgreen',
  '5\'-Insertion' => 'grey', # black
  '3\'-Insertion' => 'grey', # black
  'N\/A' => 'grey', # black
  'I'  => 'grey', # black
  'UR' => 'grey');
####################################


# TODO: if no prediction exists for certain seq -> what happens with numbering in post probs?

#----------------------------------------------------------------------------------------
# Read in predicted recombination incl. BPI and UR
#----------------------------------------------------------------------------------------
my $numSeqs    = 0;
my $curSeqName = "";
my @seqNames   = ();
my %predBP;
my %pattern;
my @curPredBP  = ();
my @curPattern = ();

open(PRED_RECOMB_BPI, "<$predRecombInclBPI") || die "$predRecombInclBPI not available\n";
while (<PRED_RECOMB_BPI>){
  chomp();
  if (m/^#/ || (m/^\s*$/ && scalar(@seqNames) == 0)) { # blank line above 1st seq
    next;
  }
  elsif (m/^\>/i) { # seq name
    @curPredBP  = ();
    @curPattern = ();
    $numSeqs++;
    my @seqName = split(" ", $_);
    $curSeqName = $seqName[0];
    $curSeqName =~ s/>//;
    push (@seqNames, $curSeqName);
    push (@curPredBP, 1); # for mapping of query to reference genome
  }
  elsif (m/^[0-9]/i) {
    # predicted recombination
    # format: begin    end   pred_subtype
    my @curSeg    = split("\t", $_);
    my $curPredBP = $curSeg[1];
    my $curST     = $curSeg[2];
    # Insertions saved as N/A
    $curST =~ s/5\'-Insertion/N\/A/g;
    $curST =~ s/3\'-Insertion/N\/A/g;
    push (@curPredBP, $curPredBP); # can be = seq_len
    push (@curPattern, $curST);    # pattern
    next;
  }
  else {
    my @predBP = @curPredBP;
    $predBP{$curSeqName} = \@predBP;
    my @pattern = @curPattern;
    $pattern{$curSeqName} = \@pattern;
  }
}
close(PRED_RECOMB_BPI);

#----------------------------------------------------------------------------------------
# Read in predicted recombination
#----------------------------------------------------------------------------------------
$numSeqs    = 0;
$curSeqName = "";
my %predPreciseBP;
my %rawPattern;
my @curPredPreciseBP = ();
my @curRawPattern    = ();

open(PRED_RECOMB, "<$predRecomb") || die "$predRecomb not available\n";
while (<PRED_RECOMB>){
  chomp();
  if (m/^#/ || (m/^\s*$/ && $numSeqs == 0)) { # blank line above 1st seq
    next;
  }
  elsif (m/^\>/i) { # seq name
    @curPredPreciseBP  = ();
    @curRawPattern = ();
    $numSeqs++;
    my @seqName = split(" ", $_);
    $curSeqName = $seqName[0];
    $curSeqName =~ s/>//;

    push (@curPredPreciseBP, 1); # for mapping of query to reference genome
  }
  elsif (m/^[0-9]/i) {
    # predicted recombination
    # format: begin    end   pred_subtype
    my @curSeg           = split("\t", $_);
    my $curPredPreciseBP = $curSeg[1];
    my $curST            = $curSeg[2];
    # Insertions saved as N/A
    $curST =~ s/5\'-Insertion/N\/A/g;
    $curST =~ s/3\'-Insertion/N\/A/g;
    push (@curPredPreciseBP, $curPredPreciseBP); # can be = seq_len
    push (@curRawPattern, $curST);    # pattern
    next;
  }
  else {
    my @predPreciseBP = @curPredPreciseBP;
    $predPreciseBP{$curSeqName} = \@predPreciseBP;
    my @rawPattern = @curRawPattern;
    $rawPattern{$curSeqName} = \@rawPattern;
  }
}
close(PRED_RECOMB);

#----------------------------------------------------------------------------------------
# read in alignment of query sequences to MSA
#----------------------------------------------------------------------------------------
my %ali;
my @alignedSeqNames = ();

open(ALI, "<$alToMSA") || die "$alToMSA not available\n";
while (<ALI>){
  chomp();
  if (m/^#/ || (m/^\s*$/)) { # blank line above 1st seq
    next;
  }
  elsif (m/^\>/i) { # seq name
    my @seqName = split(" ", $_);
    $curSeqName = $seqName[0];
    $curSeqName =~ s/>//;
    push(@alignedSeqNames, $curSeqName);

    next;
  }
  elsif (m/^[0-9]/i) { # aligned MSA positions
    my @curAli    = split(" ", $_);
    $ali{$curSeqName} = \@curAli;
    next;
  }
  else {
  }
}
close(ALI);

if (scalar(@seqNames) != scalar(@alignedSeqNames)) {
  print scalar(@seqNames), " ", scalar(@alignedSeqNames), "\n";
  die "recombination prediction file and alignment file do not contain the same number of sequences\n";
}

#----------------------------------------------------------------------------------------
# Read in MSA columns mapped to ref seq TODO: directly in jpHMM code
#----------------------------------------------------------------------------------------
open(REFSEQ, "<$msaToRefSeq") || die "mapping of msa columns to ref seq not available\n";

my @msa2refSeq = ();
while (<REFSEQ>){
  chomp();
  @msa2refSeq = split(" ", $_);
}
my $refSeqLen = $msa2refSeq[-1];
my $msaLen = scalar(@msa2refSeq);
close(REFSEQ);


#----------------------------------------------------------------------------------------
# Map all "BPs" in recombination incl UR and BPI (recombination_incl_UR_and_BPR.txt) to reference sequence positions
#  e.g. 1	300	A
#       301	350	A/B, etc.
# => map 300, 350, etc.
#----------------------------------------------------------------------------------------
my %mappedPredBP;

foreach my $seqName (@seqNames) {
  my @curPredBP = @{$predBP{$seqName}};

  if (exists($ali{$seqName})) {
    my @ali             = @{$ali{$seqName}};
    my @curMappedPredBP = ();  

    for (my $i=0; $i<scalar(@curPredBP); $i++) {
      my $predBP           = $curPredBP[$i];
      my $curMappedBPToMSA = $ali[$predBP-1]%$msaLen; # (1=1st position in file)
      # $ali[...] = 0 / msa_len+1 for Insertions
      my $curMappedBPToRef = $msa2refSeq[$curMappedBPToMSA-1];
    
      #print "i: ", $i, " BP: ", $predBP, " mapped to full MSA: ", $ali[$predBP-1], " mapped to MSA: ", $curMappedBPToMSA, " mapped to Ref: ", $curMappedBPToRef, "\n";
      #print "ref seq len: ", $refSeqLen, "\n";
      if ($i> 1 && $i == scalar(@curPredBP)-1) {
        if ($curMappedPredBP[0] <= $curMappedBPToRef && $curMappedBPToRef <= $curMappedPredBP[1]) {
          #overlapping at sequence ends
          $curMappedBPToRef = ($curMappedPredBP[0]-1)%$refSeqLen;
          if ($curMappedBPToRef == 0) {
            $curMappedBPToRef = $refSeqLen;
          }
        }
      }
      if ($i>1 && $i < scalar(@curPredBP)-1 && $curMappedBPToRef == $curMappedPredBP[-1]) {
        # BP within genome identical with preceding BP -> end;
        # except 1st segment
        last;
      }
      else {
        if (scalar(@curPredBP) == 2 && $i>0 && $curMappedBPToRef == $curMappedPredBP[-1]) {
          # first and last position of query mapped to same reference sequence position (can happen in case of insertion)
          if ($curMappedBPToRef == 1) {
            $curMappedBPToRef = $refSeqLen;
          }
          else {
            $curMappedBPToRef = $curMappedPredBP[-1]-1;
          }
        }
        push(@curMappedPredBP, $curMappedBPToRef);
      }
    }
    $mappedPredBP{$seqName} = \@curMappedPredBP;
  }
  else {
    die "something's wrong with the predicted alignment positions\n";
  }
}

#----------------------------------------------------------------------------------------
# Map all BPs in predicted recombination (recombination.txt) to reference sequence positions
#----------------------------------------------------------------------------------------
my %mappedPredPreciseBP;

foreach my $seqName (@seqNames) {
  my @curPredBP = @{$predPreciseBP{$seqName}};

  if (exists($ali{$seqName})) {
    my @ali             = @{$ali{$seqName}};
    my @curMappedPredBP = ();  

    for (my $i=0; $i<scalar(@curPredBP); $i++) {
      my $predBP           = $curPredBP[$i];
      my $curMappedBPToMSA = $ali[$predBP-1]%$msaLen; # (1=1st position in file)
      # $ali[...] = 0 / msa_len+1 for Insertions
      my $curMappedBPToRef = $msa2refSeq[$curMappedBPToMSA-1];

      if ($i> 1 && $i == scalar(@curPredBP)-1) {
        if ($curMappedPredBP[0] <= $curMappedBPToRef && $curMappedBPToRef <= $curMappedPredBP[1]) {
          #overlapping at sequence ends
          $curMappedBPToRef = ($curMappedPredBP[0]-1)%$refSeqLen;
          if ($curMappedBPToRef == 0) {
            $curMappedBPToRef = $refSeqLen;
          }
        }
      }
      if ($i>1 && $i < scalar(@curPredBP)-1 && $curMappedBPToRef == $curMappedPredBP[-1]) {
        # BP identical with preceding BP within genome -> end;
        # except 1st segment
        last;
      }
      else {
        if (scalar(@curPredBP) == 2 && $i>0 && $curMappedBPToRef == $curMappedPredBP[-1]) {
        # first and last position of query mapped to same reference sequence position (can happen in case of insertion)
        if ($curMappedBPToRef == 1) {
            $curMappedBPToRef = $refSeqLen;
          }
          else {
            $curMappedBPToRef = $curMappedPredBP[-1]-1;
          }
        }
        push(@curMappedPredBP, $curMappedBPToRef);
      }
    }
    $mappedPredPreciseBP{$seqName} = \@curMappedPredBP;
  }
  else {
    die "something's wrong with the predicted alignment positions\n";
  }
}

#----------------------------------------------------------------------------------------
# output for each sequence:
# 1) karyotype file for Circos
# 2) output of BPI and UR in Circos-readable format
# 3) output of predicted recombination based on ref seq numbering
#----------------------------------------------------------------------------------------
my %outDirNames;
for (my $i=0; $i<scalar(@seqNames); $i++) {
  my $seqName                 = $seqNames[$i];

  ###########################################################################
  # output of karyoptype file for Circos
  # output of BPI and UR in Circos-readable format
  ###########################################################################

  my $cur_karyotype_file_name = $karyotype_file_name."_".$seqName.$karyotype_file_name_ext;
  my $cur_bp_out_file_name = $bp_out_file_name."_".$seqName.$txt_script_ext;
  my $cur_ur_out_file_name = $ur_out_file_name."_".$seqName.$txt_script_ext;

  &circos_infiles_output($seqName, $cur_karyotype_file_name, $cur_bp_out_file_name, $cur_ur_out_file_name);

  ###########################################################################
  # output of predicted recombination (incl. BPI and UR) mapped on ref seq:
  # 1) map predicted recombination to reference sequence 
  # 2) save in correct order (starting at pos 1 in reference sequence)
  ###########################################################################

  my $cur_mapped_recomb_out_file_name             = $mapped_recomb_out_file_name."_".$seqName.$mapped_recomb_out_file_name_ext;
  my $cur_mapped_recomb_incl_BPI_UR_out_file_name = $mapped_recomb_incl_BPI_UR_out_file_name."_".$seqName.$mapped_recomb_out_file_name_ext;
  my $cur_mapped_BPI_UR_out_file_name             = $mapped_BPI_UR_out_file_name."_".$seqName.$mapped_recomb_out_file_name_ext;

  &mapped_recomb_incl_UR_BPI_out($seqName, $cur_mapped_recomb_incl_BPI_UR_out_file_name, $cur_mapped_BPI_UR_out_file_name);
  &mapped_recomb_out($seqName, $cur_mapped_recomb_out_file_name);

  ##########################################################
  # current output directory:
  ##########################################################
  my $curOutputDir = $outputDir."/".$seqName;
  if (exists($outDirNames{$curOutputDir})) {
    # two sequences with same name -> create 2 different subdirectories
    $curOutputDir               = $curOutputDir."_".$outDirNames{$curOutputDir};
    $outDirNames{$curOutputDir} = $outDirNames{$curOutputDir}+1;
  }
  else {
    $outDirNames{$curOutputDir} = 1;
  }

  ##########################################################
  # posterior probabilities in Circos output
  # for each sequence and each subtype:
  ##########################################################
  my $curSeqNo                    = $i+1;
  my $cur_post_probs_in_file_name = $post_probs_in_file_name.$curSeqNo.$txt_script_ext;

  my @subtypeNames = &post_prob_out($seqName, $curSeqNo, $cur_post_probs_in_file_name);

  ##########################################################
  # create circos config file for current query sequence:
  # and call Circos program
  ##########################################################
  my @scriptName    = split('\.', $scriptName);

  my $curScriptName_png = $scriptName[0]."_".$seqName.".png.conf"; # ordner anlegen fuer jede Sequenz?
  my $curScriptName_svg = $scriptName[0]."_".$seqName.".svg.conf"; # ordner anlegen fuer jede Sequenz?

  &circos_config_file("png", $seqName, $cur_karyotype_file_name, $cur_bp_out_file_name, $cur_ur_out_file_name , $curScriptName_png, $curOutputDir, @subtypeNames);
  &circos_config_file("svg", $seqName, $cur_karyotype_file_name, $cur_bp_out_file_name, $cur_ur_out_file_name , $curScriptName_svg, $curOutputDir, @subtypeNames);

  if (! (-e $curOutputDir)) {
    my $command1 = "mkdir $curOutputDir";
    system($command1);
  }

  my $allPostProbs = $post_probs_out_file_name."*";

  my $command2  = "mv $cur_karyotype_file_name $curOutputDir";
  my $command3  = "mv $cur_bp_out_file_name $curOutputDir";
  my $command4  = "mv $cur_ur_out_file_name $curOutputDir";
  my $command5  = "mv $allPostProbs $curOutputDir";
  my $command6  = "mv $cur_mapped_recomb_out_file_name $curOutputDir";
  my $command7  = "mv $cur_mapped_recomb_incl_BPI_UR_out_file_name $curOutputDir";
  my $command8  = "mv $cur_mapped_BPI_UR_out_file_name $curOutputDir";

  my $command9  = "cp $inputDir$ideogram_conf $curOutputDir";
  my $command10 = "cp $inputDir$ticks_conf $curOutputDir";
  my $command11 = "cp $inputDir$genes $curOutputDir";
  my $command12 = "cp $inputDir$subtype_colors $curOutputDir";
  my $command13 = "cp $text_genes $curOutputDir";
  my $command14 = "cp $inputDir$text_subtypes $curOutputDir";
  my $command15 = "cp $inputDir$text_legend $curOutputDir";

  my $command16 = "mv $curScriptName_png $curOutputDir";
  my $command17 = "mv $curScriptName_svg $curOutputDir";

  system($command2);
  system($command3);
  system($command4);
  system($command5);
  system($command6);
  system($command7);
  system($command8);
  system($command9);
  system($command10);
  system($command11);
  system($command12);
  system($command13);
  system($command14);
  system($command15);
  system($command16);
  system($command17);

  my $callCircos_png = "$circosDir -conf $curScriptName_png";
  my $callCircos_svg = "$circosDir -conf $curScriptName_svg";
  chdir $curOutputDir;

  qx($callCircos_png);
  qx($callCircos_svg);
  chdir $workingDir; 
  
}


#################################################################
#
# SUBROUTINES
#
#################################################################

##########################################################
# center_bpi():
# center of a breakpoint interval
#  [in]:  breakpoint interval
#  [out]: string with BPI information for output
##########################################################
sub center_bpi {
  my @bpi = @_;
  my $bpi_start = $bpi[0];
  my $bpi_end   = $bpi[1];

  my $center = 0;
  my $ext    = 0;
  my $mid    = 0;
  if ($bpi_start <= $bpi_end) {
    $center = ($bpi_start+$bpi_end)/2;
  }
  else {
    # end of ref seq within BPI, e.g. bpi = [3200, 5]
    my $int_ext = $refSeqLen-$bpi_start+1 + $bpi_end; # interval extension incl seg ends
    $center = ($bpi_start + ($int_ext/2))%$refSeqLen;
    if ($center == 0) {
      $center = $refSeqLen;
    }
  }

  if ($center%1 == 0) {
    # e.g.: bpi = [4,6] => center = (4+6)/2 = 5, ext = 5-4 = 1
    $mid = $center;
    if ($center >= $bpi_start) {
      $ext = $center-$bpi_start;
    }
    else {
      # end of ref seq within BPI, e.g. bpi = [3200, 5]
      $ext = $refSeqLen-$bpi_start-1 + $center-0;
    }
  }
  else {
    # e.g.: bpi = [3,6] => center = (3+6)/2 = 4.5, ext = 4.5-3 = 1.5
    my $mid1 = $center-0.5;
    my $mid2 = $center+0.5;

    if ($mid1 < 1) {
      $mid1 = $refSeqLen-$mid1;
    }
    if ($mid2 > $refSeqLen) {
      $mid2 = $mid2-$refSeqLen;
    }
    $mid = $mid1."/".$mid2;
    if ($mid1 >= $bpi_start) {
      $ext = $mid1-$bpi_start;
    }
    else {
      # end of ref seq within BPI, e.g. bpi = [3200, 5]
      $ext = $refSeqLen-$bpi_start-1 + $mid1-0;
    }
  }
  my $outString = $mid;
  if ($ext > 0) {
     $outString .= "+-".$ext;
  }
  return $outString;
}

###########################################################################
# circos_infiles_output():
# output of karyoptype file for Circos
# output of BPI and UR in Circos-readable format
###########################################################################
sub circos_infiles_output {
  my $seqName = shift;
  my $cur_karyotype_file_name = shift;
  my $cur_bp_out_file_name = shift;
  my $cur_ur_out_file_name = shift; 

  my @curPattern = @{$pattern{$seqName}};
  my @predBP     = @{$mappedPredBP{$seqName}};

  open (KAR, ">$cur_karyotype_file_name");
  print KAR "# HBV genome\n";
  print KAR "chr - genome1 - 1 ", $refSeqLen, " grey\n";

  open (BP_OUT, ">$cur_bp_out_file_name");
  open (UR_OUT, ">$cur_ur_out_file_name");

  # 1st BP = 1 (= start of a segment, else end of a segment!)!
  for (my $j=0; $j<scalar(@curPattern); $j++) {
    my $curSubtype = $curPattern[$j];
    my $segStart   = $predBP[$j]+1; # end of preceding segment+1

    if ($j == 0) { # for first segment
      $segStart = $predBP[$j];
    }
    if ($predBP[$j] eq $refSeqLen) {
      $segStart = 1;
    }
    
    my $segEnd = $predBP[$j+1]; # end of segment

    if ($curSubtype =~ /\// && $curSubtype !~ /N\/A/) {
      # BPI or UR
      #print "TEST IF: ", $curSubtype, "\n";
      my @predSubtypes = split("/", $curSubtype);
      my @curInt       = ($segStart, $segEnd);
      my $plotSegStart = $segStart;
      my $plotSegEnd   = $segEnd;
      my $outputString = "";

      if ($j == 0 && scalar(@curPattern) > 1 && $curSubtype eq $curPattern[-1]) {
        # BPI or UR at beginning and end of seq = ONE/SAME BPI
        $plotSegStart = $predBP[-2]+1;
        my @curPlotInt = ($plotSegStart, $plotSegEnd);
        # plot information now or for last segment
        $outputString = &center_bpi(@curPlotInt);

        if ($plotSegStart > $plotSegEnd) {
          # interval around $refSeqLen/1
          if ($plotSegEnd > $refSeqLen-$plotSegStart+1) {
            $plotSegStart = 1;
            $plotSegEnd   = ($plotSegEnd - ($refSeqLen-$plotSegStart+1))%$refSeqLen;
          }
          elsif ($plotSegEnd < $refSeqLen-$plotSegStart+1) {
            $plotSegStart = $plotSegStart + $plotSegEnd;
            $plotSegEnd   = $refSeqLen;
          }
          else { # equal distance
            $plotSegStart = 0;
            $plotSegEnd   = 1; 
          }
        }
      }
      else {
        $outputString = &center_bpi(@curInt);
      }

      if ($curSubtype =~ /\?/) {
        #UR:
        if ($j != scalar(@curPattern)-1 || scalar(@curPattern) <= 1 || $curSubtype ne $curPattern[0]) {
          print UR_OUT "genome1 ", $plotSegStart, " ", $plotSegEnd, " ", $outputString, "\n";
        }

        if ($predSubtypes[0] =~ /\?/) {
           $predSubtypes[0] = "UR";
        }
        else {
           $predSubtypes[1] = "UR";
        }
      }
      else {
        #BPI:
        if ($j != scalar(@curPattern)-1 || scalar(@curPattern) <= 1 || $curSubtype ne $curPattern[0]) {
          print BP_OUT "genome1 ", $plotSegStart, " ", $plotSegEnd, " ", $outputString, "\n";
        }
      }
      my $k=$segStart;
      while ($k<$segEnd) {
        # plot UR or BPI, interfingering of colors for the 2 predicted subtypes (or for UR: color of predicted subtype and grey (for UR)
        my $curColor = $colors{$predSubtypes[($k/2)%2]};
        my $endPoint = min($k+2, $segEnd);

        $endPoint = min($endPoint, $refSeqLen);

        print KAR "band genome1 - - ", $k, " ", $endPoint, " ", $curColor, "\n";
        $k += 2;
      }
    }
    else {
      # pure subtype or N/A
      if ($curSubtype !~ /N\/A/) {
        my $curColor = $colors{$curSubtype};

        if ($segEnd < $segStart) {
          # predicted segment around len/1
          print KAR "band genome1 - - ", $segStart, " ", $refSeqLen, " ", $curColor, "\n";
          print KAR "band genome1 - - ", 1, " ", $segEnd, " ", $curColor, "\n";
        }
        else {
          if ($segStart eq $segEnd && $j eq scalar(@curPattern)-1) {
            $segEnd = $segEnd + 1;
          }
          print KAR "band genome1 - - ", $segStart, " ", $segEnd, " ", $curColor, "\n";
        }
      }
    }
  }
  close(KAR);
  close(BP_OUT);
  close(UR_OUT);
}

###########################################################################
# mapped_recomb_incl_UR_BPI_out():
# output of predicted recombination (incl. BPI and UR) mapped on ref seq:
# 1) map predicted recombination to reference sequence 
# 2) save in correct order (starting at pos 1 in reference sequence)
###########################################################################
sub mapped_recomb_incl_UR_BPI_out {
  my $seqName = shift;
  my $cur_mapped_recomb_incl_BPI_UR_out_file_name = shift;
  my $cur_mapped_BPI_UR_out_file_name             = shift;

  my %predMappedPattern;

  # 1st BP = 1 (= start of a segment, else end of a segment!)!
  my @predBP     = @{$mappedPredBP{$seqName}};
  my @curPattern = @{$pattern{$seqName}};
  # save start positions for output  
  my @startPositions = ();

  for (my $j=0; $j<scalar(@curPattern); $j++) {
    my $curSubtype = $curPattern[$j];
    my $segStart   = $predBP[$j]+1; # end of preceding segment+1
    if ($j == 0) { # for first segment
      $segStart = $predBP[$j];
    }

    if ($j>0 && $predBP[$j] eq $refSeqLen) {
      $segStart = 1;
    }
    my $segEnd = $predBP[$j+1]; # end of segment
    if ($segEnd < $segStart) {
      # Segment around origin
      my $segStart1 = $segStart;
      my $segEnd1   = $refSeqLen;
      my $segStart2 = 1;
      my $segEnd2   = $segEnd;
      $predMappedPattern{$segStart1}->{"end"} = $segEnd1;
      $predMappedPattern{$segStart1}->{"ST"}  = $curSubtype;
      push(@startPositions, $segStart1);
      $predMappedPattern{$segStart2}->{"end"} = $segEnd2;
      $predMappedPattern{$segStart2}->{"ST"}  = $curSubtype;
      push(@startPositions, $segStart2);
    }
    else {
      $predMappedPattern{$segStart}->{"end"} = $segEnd;
      $predMappedPattern{$segStart}->{"ST"}  = $curSubtype;
      push(@startPositions, $segStart);
    }
  }

  # output:
  open (MAPPED_RECOMB_INCL_BPI_UR_OUT, ">$cur_mapped_recomb_incl_BPI_UR_out_file_name");
  open (MAPPED_BPI_UR_OUT, ">$cur_mapped_BPI_UR_out_file_name");

  
  # sort list with start positions so that it starts with 1
  my @sortedStartPositions = sort { $a <=> $b } @startPositions;
  @startPositions = @sortedStartPositions;

  # check for successive segments with same pred. ST (due to sequence cutoff wrt ref seq origin)
  # save the respective start positions
  my @mergedStartPositions = ();

  my $prevStartPos = $startPositions[0];
  push(@mergedStartPositions, $prevStartPos);

  for (my $i=1; $i<scalar(@startPositions); $i++) {
    my $curStartPos = $startPositions[$i];
    my $curST       = $predMappedPattern{$curStartPos}->{"ST"};
    my $curEndPos   = $predMappedPattern{$curStartPos}->{"end"};

    if ($curST eq $predMappedPattern{$prevStartPos}->{"ST"}) {
      # cur segment same ST as prev. segment -> merge and delete current start pos:
      $predMappedPattern{$prevStartPos}->{"end"} = $curEndPos;
      delete($predMappedPattern{$curStartPos});
    }
    else {
      $prevStartPos = $curStartPos;

      push(@mergedStartPositions, $curStartPos);
    }
  }

  my %predMappedUR;
  my %predMappedBPI;
  my @startPositionsUR = ();
  my @startPositionsBPI = ();

  foreach my $curStartPos (@mergedStartPositions) {
    my $endPos = $predMappedPattern{$curStartPos}->{"end"};
    my $subtype = $predMappedPattern{$curStartPos}->{"ST"};
    print MAPPED_RECOMB_INCL_BPI_UR_OUT $curStartPos, "\t", $endPos, "\t", $subtype, "\n";

    $predMappedPattern{$curStartPos}->{"ST"} = $subtype;
    
    # save BPI and UR (see Webserver output)
    if ($subtype =~ /\// && $subtype !~ /N\/A/) {
      if ($subtype =~ /\?/) {
        # UR
        $predMappedUR{$curStartPos}->{"end"} = $predMappedPattern{$curStartPos}->{"end"};
        $predMappedUR{$curStartPos}->{"ST"} = $predMappedPattern{$curStartPos}->{"ST"};
        push (@startPositionsUR, $curStartPos);
      }
      else {
        # BPI
        $predMappedBPI{$curStartPos}->{"end"} = $predMappedPattern{$curStartPos}->{"end"};
        $predMappedBPI{$curStartPos}->{"ST"} = $predMappedPattern{$curStartPos}->{"ST"};
        push (@startPositionsBPI, $curStartPos);
      }
    }
  }

  close(MAPPED_RECOMB_INCL_BPI_UR_OUT);

  print MAPPED_BPI_UR_OUT "# uncertainty regions:\n";

  # breakpoint regions:
  foreach my $curStartPos (@startPositionsUR) {
    my $endPos = $predMappedUR{$curStartPos}->{"end"};
    my $subtype = $predMappedUR{$curStartPos}->{"ST"};

    print MAPPED_BPI_UR_OUT $curStartPos, "\t", $endPos, "\t", $subtype, "\n";
  }

  print MAPPED_BPI_UR_OUT "# breakpoint regions:\n";

  foreach my $curStartPos (@startPositionsBPI) {
    my $endPos = $predMappedBPI{$curStartPos}->{"end"};
    my $subtype = $predMappedBPI{$curStartPos}->{"ST"};
    
    print MAPPED_BPI_UR_OUT $curStartPos, "\t", $endPos, "\t", $subtype, "\n";
  }

  close(MAPPED_BPI_UR_OUT);
}

###########################################################################
# mapped_recomb_out():
# output of predicted recombination (incl. BPI and UR) mapped on ref seq:
# 1) map predicted recombination to reference sequence 
# 2) save in correct order (starting at pos 1 in reference sequence)
###########################################################################
sub mapped_recomb_out {

  # output of recomb without UR/BPI
  # necessary for table output (beachte NAMEN fuer Webserver)

  my $seqName = shift;
  my $cur_mapped_recomb_out_file_name = shift;

  my %predMappedRawPattern;

  # 1st BP = 1 (= start of a segment, else end of a segment!)!
  my @predPreciseBP = @{$mappedPredPreciseBP{$seqName}};

  my @curRawPattern = @{$rawPattern{$seqName}};

  # save start positions for output
  my @startPositions = ();

  for (my $j=0; $j<scalar(@curRawPattern); $j++) {
    my $curSubtype = $curRawPattern[$j];
    my $segStart   = $predPreciseBP[$j]+1; # end of preceding segment+1
    if ($j == 0) { # for first segment
      $segStart = $predPreciseBP[$j];
    }
    if ($j>0 && $predPreciseBP[$j] eq $refSeqLen) {
      $segStart = 1;
    }
    my $segEnd = $predPreciseBP[$j+1]; # end of segment

    if ($segEnd < $segStart) {
      # Segment around origin
      my $segStart1 = $segStart;
      my $segEnd1   = $refSeqLen;
      my $segStart2 = 1;
      my $segEnd2   = $segEnd;
      $predMappedRawPattern{$segStart1}->{"end"} = $segEnd1;
      $predMappedRawPattern{$segStart1}->{"ST"}  = $curSubtype;
      push(@startPositions, $segStart1);
      $predMappedRawPattern{$segStart2}->{"end"} = $segEnd2;
      $predMappedRawPattern{$segStart2}->{"ST"}  = $curSubtype;
      push(@startPositions, $segStart2);
    }
    else {
      $predMappedRawPattern{$segStart}->{"end"} = $segEnd;
      $predMappedRawPattern{$segStart}->{"ST"}  = $curSubtype;
      push(@startPositions, $segStart);
    }
  }

  # output:
  open (MAPPED_RECOMB_OUT, ">$cur_mapped_recomb_out_file_name");

  # sort list with start positions so that it starts with 1
  my @sortedStartPositions = sort { $a <=> $b } @startPositions;
  @startPositions = @sortedStartPositions;

  # check for successive segments with same pred. ST (due to sequence cutoff wrt ref seq origin)
  my @mergedStartPositions = ();

  my $prevStartPos = $startPositions[0];
  push(@mergedStartPositions, $prevStartPos);

  for (my $i=1; $i<scalar(@startPositions); $i++) {
    my $curStartPos = $startPositions[$i];
    my $curST       = $predMappedRawPattern{$curStartPos}->{"ST"};
    my $curEndPos   = $predMappedRawPattern{$curStartPos}->{"end"};

    if ($curST eq $predMappedRawPattern{$prevStartPos}->{"ST"}) {
      # cur segment same ST as prev. segment -> merge and delete current start pos:
      $predMappedRawPattern{$prevStartPos}->{"end"} = $curEndPos;
      delete($predMappedRawPattern{$curStartPos});
    }
    else {
      $prevStartPos = $curStartPos;
      push(@mergedStartPositions, $curStartPos);
    }
  }

  foreach my $curStartPos (@mergedStartPositions) {
    my $endPos = $predMappedRawPattern{$curStartPos}->{"end"};
    my $subtype = $predMappedRawPattern{$curStartPos}->{"ST"};
    
    print MAPPED_RECOMB_OUT $curStartPos, "\t", $endPos, "\t", $subtype, "\n";
  }

  close(MAPPED_RECOMB_OUT);
}

##########################################################
# post_prob_out():
# posterior probabilities in Circos output
# for each sequence and each subtype:
# - mapped to ref seq positions
##########################################################
sub post_prob_out {
  my $seqName                     = shift;
  my $curSeqNo                    = shift;
  my $cur_post_probs_in_file_name = shift;

  open (POST_PROBS_IN, "<$cur_post_probs_in_file_name") || die "file for posterior probabilities $cur_post_probs_in_file_name doesn't exist\n";
  my @postProbs = <POST_PROBS_IN>;
  close (POST_PROBS_IN);

  my $header = $postProbs[0];
  my @header = split(" ", $header);
  $curSeqName = $header[0];
  $curSeqName =~ s/\#//g;
  $curSeqName =~ s/\://g;

  if ($curSeqName ne $seqName) { # TODO:
     print $curSeqName, " ", $seqName, "\n";
     die "error: seq names in post probs and recomb pred do not match\n";
  }

  my @subtypeNames = ();
  foreach (my $j=1; $j<scalar(@header); $j++) {
     my $curSubtype = $header[$j];
     push(@subtypeNames, $curSubtype);
  }
  
  my %postProbs;
  for (my $j=1; $j<scalar(@postProbs); $j++) {
    my @curPostProbs = split(" ", $postProbs[$j]);
    for (my $k=0; $k<scalar(@curPostProbs); $k++) {
       $postProbs{$subtypeNames[$k]}->{$j} = $curPostProbs[$k];
    }
  }

  my @ali = @{$ali{$seqName}};

  foreach my $subtypeName (@subtypeNames) {
    my $cur_post_probs_out_file_name = $post_probs_out_file_name.$subtypeName.".txt";

    open (POST_PROBS_OUT, ">$cur_post_probs_out_file_name");

    my %tmpPostProbs;
    for (my $j=1; $j<scalar(@postProbs); $j++) {
      my $curMsaPos       = $ali[$j-1]%$msaLen;
      my $curMappedSeqPos = $msa2refSeq[$curMsaPos-1];

      $tmpPostProbs{$curMappedSeqPos} = $postProbs{$subtypeName}->{$j};

      my @keys = keys(%postProbs);
      print POST_PROBS_OUT "genome1 ", $curMappedSeqPos, " ", $curMappedSeqPos, " ", $postProbs{$subtypeName}->{$j}, "\n";
    }

    close(POST_PROBS_OUT);
  }

  return @subtypeNames;
}

##########################################################
# circos_config_file():
# create circos config file for current query sequence:
##########################################################
sub circos_config_file {
  my $format = shift;
  my $seqName = shift;
  my $cur_karyotype_file_name = shift;
  my $cur_bp_out_file_name = shift;
  my $cur_ur_out_file_name = shift;
  my $curScriptName = shift;
  my $curOutputDir = shift;
  my @subtypeNames = @_;

  my @predBP     = @{$mappedPredBP{$seqName}};

  $seqName =~ s/>//;
  my $cur_circos_out_file_name = $circos_out_file_name."$seqName".".".$format;

  open(OUT, ">$curScriptName"); # TODO: variable script name

  print OUT "<colors>\n";
  print OUT "<<include $colors_conf>>\n";
  print OUT "</colors>\n\n";

  print OUT "<fonts>\n";
  print OUT "<<include $fonts_conf>>\n";
  print OUT "</fonts>\n\n";

  print OUT "<<include $ideogram_conf>>\n";
  print OUT "<<include $ticks_conf>>\n\n";

  print OUT "karyotype = $cur_karyotype_file_name\n\n";

  print OUT "<image>\n";
  print OUT "dir = .\n";
  print OUT "$format = yes\n";
  print OUT "file = $cur_circos_out_file_name\n";
  print OUT "# radius of inscribed circle in image\n";
  print OUT "radius         = 440p\n"; # 1500p
  print OUT "background     = white\n";
  print OUT "# by default angle=0 is at 3 o'clock position\n";
  print OUT "angle_offset   = -90\n";
  print OUT "</image>\n\n";

  print OUT "chromosomes_units = 1\n";
  print OUT "chromosomes       = genome1\n";
  print OUT "chromosomes_display_default = no\n\n";

  print OUT "<plots>\n\n";

  my $cur_color = "black";
  &text_output($text_subtypes, $cur_color);
  
  my $cur_r0 = "0.28r";
  my $cur_r1 = "100p";
  &radius_output($cur_r0, $cur_r1);

  my $cur_label_size = "11p";
  &label_output($cur_label_size);

  &text_output($text_legend, $cur_color);
  $cur_r0 = "0.05r";
  &radius_output($cur_r0, $cur_r1);

  &label_output($cur_label_size);

  &text_output($cur_bp_out_file_name, $cur_color);
  $cur_r0 = "1r";
  &radius_output($cur_r0, $cur_r1);

  print OUT "show_links     = yes\n";
  print OUT "link_dims      = 0p,2p,25p,25p,5p\n";
  print OUT "link_thickness = 2p\n";
  print OUT "link_color     = dgrey\n\n";

  print OUT "label_snuggle        = yes\n";
  print OUT "max_snuggle_distance = 1r\n";
  print OUT "snuggle_tolerance    = 0.25r\n";
  print OUT "snuggle_sampling     = 2\n";
  print OUT "snuggle_refine       = yes\n\n";

  $cur_label_size = "12p";
  &label_output($cur_label_size);

  &text_output($cur_ur_out_file_name, $cur_color);
  &radius_output($cur_r0, $cur_r1);

  print OUT "show_links     = yes\n";
  print OUT "link_dims      = 0p,2p,25p,25p,5p\n";
  print OUT "link_thickness = 2p\n";
  print OUT "link_color     = dgrey\n\n";

  print OUT "label_snuggle        = yes\n";
  print OUT "max_snuggle_distance = 1r\n";
  print OUT "snuggle_tolerance    = 0.25r\n";
  print OUT "snuggle_sampling     = 2\n";
  print OUT "snuggle_refine       = yes\n\n";

  &label_output($cur_label_size);

  # posterior probabilities
  my $subtypeName                  = $subtypeNames[0];
  my $cur_post_probs_out_file_name = $post_probs_out_file_name.$subtypeName.".txt";
  my $curColor                     = $colors{$subtypeName};

  print OUT "<plot>\n";
  print OUT "show    = yes\n";
  print OUT "type    = line\n";
  print OUT "max_gap = 100\n"; # print OUT "max_gap = 100000\n";
  print OUT "file    = $cur_post_probs_out_file_name\n";
  print OUT "color   = $curColor\n";
  print OUT "thickness = 3\n";
  print OUT "min   = 0\n";
  print OUT "max   = 1\n";
  print OUT "r0    = 0.7r\n";
  print OUT "r1    = 0.95r\n\n";

  print OUT "background       = yes\n";
  print OUT "background_color = vvlgrey\n";
  print OUT "background_stroke_color = vvlgrey\n";
  print OUT "background_stroke_thickness = 1\n\n";

  print OUT "axis           = yes\n";
  print OUT "axis_color     = lgrey\n";
  print OUT "axis_thickness = 1\n";
  print OUT "axis_spacing   = 0.1\n";
  print OUT "</plot>\n\n";

  for (my $i=1; $i<scalar(@subtypeNames); $i++) {
    $subtypeName                  = $subtypeNames[$i];
    $cur_post_probs_out_file_name = $post_probs_out_file_name.$subtypeName.".txt";
    $curColor                     = $colors{$subtypeName};

    print OUT "<plot>\n";
    print OUT "show    = yes\n";
    print OUT "type    = line\n";
    print OUT "max_gap = 100\n"; # print OUT "max_gap = 100000\n";
    print OUT "file    = $cur_post_probs_out_file_name\n";
    print OUT "color   = $curColor\n";
    print OUT "thickness = 3\n";
    print OUT "min   = 0\n";
    print OUT "max   = 1\n";
    print OUT "r0    = 0.7r\n";
    print OUT "r1    = 0.95r\n\n";
    print OUT "</plot>\n\n";
  }

  &text_output($text_genes_P, $cur_color);

  print OUT "# on tick scale\n";
  $cur_r0 = "0.51r";
  $cur_r1 = "50p";
  &radius_output($cur_r0, $cur_r1);

  $cur_label_size = "10p";
  &label_output($cur_label_size);

  &text_output($text_genes_C_S, $cur_color);

  print OUT "# on tick scale\n";
  $cur_r0 = "0.57r";
  $cur_r1 = "100p";
  &radius_output($cur_r0, $cur_r1);

  &label_output($cur_label_size);

  &text_output($text_genes_X, $cur_color);

  print OUT "# on tick scale\n";
  $cur_r0 = "0.62r";
  $cur_r1 = "50p";
  &radius_output($cur_r0, $cur_r1);

  &label_output($cur_label_size);


  &text_output($text_genes_DR, $cur_color);

  print OUT "# on tick scale\n";
  $cur_r0 = "0.35r";
  $cur_r1 = "100p";
  &radius_output($cur_r0, $cur_r1);

  &label_output($cur_label_size);

  print OUT "</plots>\n\n";

  print OUT "<highlights>\n\n";

  print OUT "z=0\n";
  print OUT "fill_color = green\n\n";

  print OUT "<highlight>\n";
  print OUT "# subtype colors -> names of subtypes above in text...\n";

  print OUT "file       = $subtype_colors\n";
  print OUT "r0         = 1.1r\n";
  print OUT "r1         = 1.1r+100p\n";
  print OUT "</highlight>\n\n";

  print OUT "<highlight>\n";
  print OUT "# genes -> names of genes above in text...\n";
  print OUT "file       = $genes\n";
  print OUT "r0         = 1.1r\n";
  print OUT "r1         = 1.1r+100p\n";
  print OUT "</highlight>\n\n";

  print OUT "</highlights>\n\n";

  print OUT "anglestep       = 0.5\n";
  print OUT "minslicestep    = 10\n";
  print OUT "beziersamples   = 40\n";
  print OUT "debug           = no\n";
  print OUT "warnings        = no\n";
  print OUT "imagemap        = no\n\n";

  print OUT "# don't touch!\n";
  print OUT "units_ok        = bupr\n";
  print OUT "units_nounit    = n\n";

  close(OUT);
  
}


##########################################################
# short output parts of circos config file
##########################################################

##########################################################
# label_output():
##########################################################
sub label_output {
  my $cur_label_size = shift;
  print OUT "label_size = $cur_label_size\n";
  print OUT "label_font = condensedbold\n";
  print OUT "padding    = 0p\n";
  print OUT "rpadding   = 0p\n\n";
  print OUT "</plot>\n\n";
}

##########################################################
# text_output():
##########################################################
sub text_output {
  my $cur_file = shift;
  my $cur_color = shift;

  print OUT "<plot>\n";
  print OUT "type  = text\n";
  print OUT "color = $cur_color\n";
  print OUT "file  = ", $cur_file, "\n\n";
}

##########################################################
# radius_output():
##########################################################

sub radius_output {
  my $cur_r0 = shift;
  my $cur_r1 = shift;

  print OUT "r0 = $cur_r0\n";
  print OUT "r1 = $cur_r0+$cur_r1\n\n";
}
