// ~/jpHMM/src/main.cpp
/**************************************************************
 File name	: main.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 main program
 program call: ./jpHMM
 
 - Read in of input arguments/parameters
    e.g. determination of virus type
 - Definition of jpHMM variables by calling const_def
 - Call of function jpHMM() for building the model and recombination prediction
***************************************************************/

#include "std_header.h"
#include "std_source.h"
#include "error.h"
#include "const_def.h"
#include "alphabet.h"
#include "alignment.h"
#include "emissionPriors.h"
#include "transitionPriors.h"
#include "emTransProfile.h"
#include "query.h"
#include "profileHMM.h"
#include "hmm.h"
#include "jpHMM.h"


using namespace std;

//! usage
void usage (char *prog,
	    string errstr) {
  if (errstr.length())
    cerr << errstr << endl << endl;
  cerr << "usage: " << prog
       << "\n\t[ -s  file containing the query sequence(s) ] "
       << "\n\t[ -v  type of the given sequences (HIV or HCV)]\n"
       << "\n\toptional arguments:\n"
       << "\n\t[ -a  (opt.) file containing the priors for the estimation of the emission probabilities (include path to file here or with -P)] "
       << "\n\t[ -b  (opt.) file containing the priors for the estimation of the transition probabilities] "
       << "\n\t[ -i  (opt.) file containing the given multiple sequence alignment  (include path to file here or with -I)] "
       //<< "\n\t[ -g  (opt.) gap initialization prob. (= probability of a jump from MATCH to C_DELETE taking into account the corresponding consensus column of the MATCH)    (default: 0.01) ] "
       //<< "\n\t[ -d  (opt.) gap extension prob. (= probability of extending deletion (in N_DELETE or C_DELETE) by one consensus column)    (default: 0.99) ] "
       //<< "\n\t[ -f  (opt.) (1 - f) = transition probability from N_INSERT to N_INSERT and from C_INSERT to C_INSERT    (default: 1 - 0.01) ] "
       << "\n\t[ -j  (opt.) probability of a jump from one subtype to any other subtype (default: HIV: 1e-9, HBV: 1e-7) ]"
       << "\n\t[ -B  (opt.) beam-width for the beam algorithm    (default: 1e-20) ]"
       << "\n\t[ -P  (opt.) input directory for emission priors (default: ./priors/, if not -a chosen) ]"
       << "\n\t[ -I  (opt.) input directory for alignment file (default: ./input/, if not -i chosen) ]"
       << "\n\t[ -o  (opt.) output directory for result files  (default: ./output/) ]"
       << "\n\t[ -Q  (opt.) algorithm to speed up Viterbi algorithm (beam_search or blat) (default: beam_search) ]"
       << "\n\t[ -C  (opt.) circular genome type: if chosen: circular genomes    (default: non-circular genomes) ]"
       //<< "\n\t[ -A  (only testing): if chosen: jump probabilities in jpHMM are adapted to jump prob in linear jump version to compare results]\n"
       << "\n\t[ -J  (opt.) kind of jumps in HMM (quadratic (from each ST to each ST) or linear (via a mute jump state)) (default: quadratic) ]\n"
       << "\n\tfor usage of model probabilities that have been determined externally:\n"
       << "\n\t[ -e  (opt.) file containing the (external) emission probabilities ]"
       << "\n\t[ -t  (opt.) file containing the (external) transition probabilities ]"
       << "\n\t[ -c  (opt.) file containing the (external) consensus columns ]\n\n";
       // OUTPUT UND PRIORS DIR NAME
  exit(1);
}


//! Main program.
int main(int argc, char *argv[]) {

  /*********************************************************
     Set arguments/parameters/variables & Initialization:
  **********************************************************/

  if (argc == 1)
    usage(argv[0], "no arguments given");

  //! Flags for options.
  bool has_a = false, has_g = false;
  bool has_d = false, has_f = false;
  bool has_i = false, has_j = false;
  bool has_s = false, has_v = false;
  bool has_B = false, has_C = false;
  bool has_I = false, has_P = false;
  bool has_Q = false, has_J = false;
  bool has_x = false;
  bool has_e = false;
  bool has_t = false;
  bool has_c = false;
  bool has_o = false;  
  bool has_b = false;  
  bool has_A = false;
  
  //! Variables for option parameters.
  string infile, query_file, emission_priors_file, transition_priors_file, input_dir_name, priors_dir_name, output_dir_name, virus_type_input, program_version, jump_version;
  string ep_file_name, tp_file_name, cons_file_name;

  int argument;

  //! Arguments set to 0. if available -> set to 1 (if appear twice -> error):
  while (-1 != (argument = getopt(argc, argv, "a:b:g:d:f:i:j:s:v:B:I:P:O:Q:J:x:e:t:c:o:CA"))) {
    switch (argument) {
    case 'j':
      if (has_j)
	cerr << "multiple option -j given,"
	     << " ignore further declaration\n";
      else
	Const_Def::jump_prob() = atof(optarg);
      has_j = true;
      break;
    case 'g':
      if (has_g)
	cerr << "multiple option -g given,"
	     << " ignore further declaration\n";
      else
	Const_Def::gap_init() = atof(optarg);
      has_g = true;
      break;
    case 'd':
      if (has_d)
	cerr << "multiple option -d given,"
	     << " ignore further declaration\n";
      else
	Const_Def::gap_ext() = atof(optarg);
      has_d = true;
      break;
    case 'f':
      if (has_f)
        cerr << "multiple option -f given,"
             << " ignore further declaration\n";
      else
        Const_Def::epsilon() = atof(optarg);
      has_d = true;
      break;
    case 'v':
      if (has_v)
	cerr << "multiple option -v given,"
	     << " ignore further declaration\n";
      else
	virus_type_input.assign(optarg);
      has_v = true;
      break;
    case 's':
      if (has_s)
	cerr << "multiple option -s given,"
	     << " ignore further declaration\n";
      else
	query_file.assign(optarg);
      has_s = true;
      break;
    case 'B':
      if (has_B)
	cerr << "multiple option -B given,"
	     << " ignore further declaration\n";
      else
        Const_Def::beam_width() = atof(optarg);
      has_B = true;
      break;
    case 'I':
      if (has_I)
	cerr << "multiple option -I given,"
	     << " ignore further declaration\n";
      else
	Const_Def::input_dir_name() = optarg;
      has_I = true;
      break;
    case 'P':
      if (has_P)
	cerr << "multiple option -P given,"
	     << " ignore further declaration\n";
      else
	Const_Def::priors_dir_name() = optarg;
      has_P = true;
      break;
    case 'i':
      if (has_i)
	cerr << "multiple option -i given,"
	     << " ignore further declaration\n";
      else
	infile.assign(optarg);
      has_i = true;
      break;
    case 'a':
      if (has_a)
	cerr << "multiple option -a given,"
	     << " ignore further\n";
      else
	emission_priors_file.assign(optarg);
      has_a = true;
      break;
    case 'C':
      if (has_C)
	cerr << "multiple option -C given,"
	     << " ignore further declaration\n";
      else
	Const_Def::genome_type() = circular;
      has_C = true;
      break;
    case 'Q':
      if (has_Q)
	cerr << "multiple option -Q given,"
	     << " ignore further declaration\n";
      else
	program_version.assign(optarg);
      has_Q = true;
      break;
    case 'J':
      if (has_J)
	cerr << "multiple option -J given,"
	     << " ignore further declaration\n";
      else
	jump_version.assign(optarg);
      has_J = true;
      break;
    case 'b':
      if (has_b)
	cerr << "multiple option -b given,"
	     << " ignore further\n";
      else
	transition_priors_file.assign(optarg);
      has_b = true;
      break;
    case 'x':
      if (has_x)
	cerr << "multiple option -x given,"
	     << " ignore further declaration\n";
      else {
    	 Const_Def::stamp() = optarg;
    	 Const_Def::has_stamp() = true;
      }   
      has_x = true;
      break;
    case 'e':
      if (has_e)
	cerr << "multiple option -e given,"
	     << " ignore further declaration\n";
      else
    	 ep_file_name = optarg;
      has_e = true;
      break;
    case 'c':
      if (has_c)
	cerr << "multiple option -c given,"
	     << " ignore further declaration\n";
      else
    	 cons_file_name = optarg;
      has_c = true;
      break;
    case 't':
      if (has_t)
	cerr << "multiple option -t given,"
	     << " ignore further declaration\n";
      else
    	  tp_file_name = optarg;
      has_t = true;
      break;
    case 'o':
      if (has_o)
	cerr << "multiple option -o given,"
	     << " ignore further declaration\n";
      else
    	  Const_Def::output_dir_name() = optarg;
      has_o = true;
      break;
    case 'A':
      if (has_A)
	cerr << "multiple option -A given,"
	     << " ignore further declaration\n";
      else
    	 Const_Def::adapt_jump_probs() = true;
      has_A = true;
      break;
    default:
      usage(argv[0], (string)"Invalid Option");

    } // switch
  } // while (-1 ...

  if (optind < argc) {
    usage(argv[0], (string)"Invalid Option: " + argv[optind]);
  }

  //! Input parameters.
  if (!has_s)
    usage(argv[0], "no query sequence input file given (-s)\n");
  if (!has_v)
    usage(argv[0], "no virus type of the given sequences given (HIV, HCV or HBV) (-v)\n");

  //! Output directory.
  if (Const_Def::output_dir_name() != ".") {
    string command2 = "test -d " + Const_Def::output_dir_name();
    if (system(command2.c_str()) ) {
      string command1 = "mkdir "+ Const_Def::output_dir_name();
      int i = system( command1.c_str() );
      i=0;
    }
  }
  
  //! Priors input directory:
  if (!has_P && has_a)
    Const_Def::priors_dir_name() = ".";
  //! input directory:
  if (!has_I && has_i)
    Const_Def::input_dir_name() = ".";

  //! Type of given virus
  //! and priors (pseudo counts) for emission probabilities:
  std::transform(virus_type_input.begin(), virus_type_input.end(), virus_type_input.begin(), ::toupper);

  if (virus_type_input == "HIV") {
    Const_Def::virus_type() = HIV;
    if (!has_i)
      infile.assign(Const_Def::hiv_msa_file_name());
    if (!has_a)
      emission_priors_file.assign(Const_Def::hiv_priors_file_name());
  }
  else if (virus_type_input == "HCV") {
    Const_Def::virus_type() = HCV;
    if (!has_i)
      infile.assign(Const_Def::hcv_msa_file_name());
    if (!has_a)
      emission_priors_file.assign(Const_Def::hcv_priors_file_name());
  }
  else if (virus_type_input == "HBV") {
    Const_Def::virus_type() = HBV;
    if (!has_i)
      infile.assign(Const_Def::hbv_msa_file_name());
    if (!has_a)
      emission_priors_file.assign(Const_Def::hbv_priors_file_name());
  }
  else {
    stringstream ss;
    ss << "failure: given type of sequences (virus) (";
    ss <<  virus_type_input;
    ss << ") is different from HIV, HCV and HBV";
    throw Error(ss.str());
  }
  
  //! Priors (pseudo counts) for transition probabilities:
  if (!has_b) {
    transition_priors_file.assign(Const_Def::transition_priors_file_name());
    //usage(argv[0], "no transition priors file given (-b)\n");
  }

  //! Jump probabilitiy for given virus type
  if (!has_j) {
    Const_Def::jump_prob(Const_Def::virus_type());
  }

  //! Determine program version (blat or beam search algorithm for Viterbi algorithm).
  if (Const_Def::genome_type() == circular)  {// circular genomes: pre-alignment necessary
    program_version = "blat";
  }
  
  std::transform(program_version.begin(), program_version.end(), program_version.begin(), ::tolower);
  if (program_version == "blat")
    Const_Def::prog_version() = blat;
  else
    Const_Def::prog_version() = beam_search;

  //! Jump version: linear number of jumps (by mute jump states) or quadratic number of jumps per column (original jpHMM version)
  std::transform(jump_version.begin(), jump_version.end(), jump_version.begin(), ::tolower);
  if (jump_version == "linear") {
    Const_Def::jump_version() = linear;
  }
  else
    Const_Def::jump_version() = quadratic;

  //! Usage of external emission and transition probabilities:
  bool use_ext_em_tr_probs = has_e && has_t && has_c;
  if (!use_ext_em_tr_probs && (has_e || has_t || has_c))
	  throw Error("emission and transition probabilities and consensus columns have to be given");

  //!-----------------------------------------------------
  //! Call jpHMM:
  //! Recombination prediction for given query sequences.
  //!-----------------------------------------------------
  jpHMM(emission_priors_file, transition_priors_file, infile, query_file, has_B, use_ext_em_tr_probs, ep_file_name, tp_file_name, cons_file_name);

  return 0;
}

// EMISSION PRIORS FILE
