// ~/jpHMM/src/profileHMM.cpp
/**************************************************************
 File name	: profileHMM.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 - Construction of profile HMM for a certain subtype
    incl. definition of states and transitions
 - Determination of emission and transition probabilities
***************************************************************/

#include "std_source.h"
#include "error.h"
#include "alphabet.h"
#include "const_def.h"
#include "alignment.h"
#include "emissionPriors.h"
#include "transitionPriors.h"
#include "emissionProfile.h"
#include "transitionProfile.h"
#include "emTransProfile.h"
#include "profileHMM.h"
#include "extEmTrProbs.h"


using namespace std;

/******************************************************
 *                                                    *
 *                   class profileHMM                 *
 *                                                    *
 \******************************************************/

/**************************************************************************************//**
   ProfileHMM.

   Create profile HMM for certain subtype.

   @param[in] sub_prof Profile of subtype.
   @param[in] t Transition priors.
   @param[in] e Emission priors.
   @param[in] t_al Transition alphabet, i.e. possible state transitions.
   @param[in] prob_spec_chars Emission probability of special characters.
   @param[in] use_eetp Use external emission and transition probabilities?.
   @param[in] eetp External emission and transition probabilities.
******************************************************************************************/
ProfileHMM::ProfileHMM(SubtypeProfile* sub_prof, 
		       const TransitionPriors& t, 
		       const EmissionPriors& e, 
		       const TransAlphabet& t_al, 
		       mytype prob_spec_chars,
		       bool  use_eetp, 
		       const ExtEmTrProbs* eetp) {

  m_sub_prof = sub_prof;
  m_len      = m_sub_prof->sub()->len();
  map<int, ConsColumnProfile*> cons_col_profs = m_sub_prof->cons_col_profs();

  vector<mytype> trans_priors     = t.make_vector();
  vector<mytype> match_em_priors  = e.match_priors();
  vector<mytype> insert_em_priors = e.insert_priors();

  multi_array<State*, 2>::extent_gen extents;
  m_states.resize(extents[m_len][NUM_STATES]);

  mytype prob_non_spec_chars = (mytype)1 - prob_spec_chars;
  
  for(map<int, ConsColumnProfile*>::const_iterator it=cons_col_profs.begin(); it != --cons_col_profs.end(); ++it) {
    for (int type = (int)Match; type<=(int)Delete; ++type) {
      m_states[it->first][type] = new State(this, it->first, (StateType)type);
      State* cur_state = m_states[it->first][type];
      if (!cur_state->is_mute()) {
    	vector<mytype> cur_ep;
    	if (!use_eetp || (type != (int)Match)) {
		    EmissionProfile* em_prof   = (it->second)->em_prof();
		    vector<mytype>   cur_priors;
		    vector<int>      cur_ef;
		    if (cur_state->is_match()) {
		      cur_priors = match_em_priors;
		      for (unsigned i=0; i<cur_priors.size(); ++i)
		        cur_ef += em_prof->match_frequency(i);
		    } 
		    else {
		      cur_priors = insert_em_priors;
		      for (unsigned i=0; i<cur_priors.size(); ++i)
		        cur_ef += em_prof->insert_frequency(i);
		    }
		    cur_ep = emission_probs(cur_priors, cur_ef);
    	}
    	else {
    		cur_ep = eetp->em_probs(sub_prof->sub()->name(), it->first);
    	}
    	
        for (unsigned i=0; i<cur_ep.size(); ++i)
          cur_ep[i] = prob_non_spec_chars*cur_ep[i];
        cur_ep += prob_spec_chars;

        cur_state->log_ep()  = log_of_prob(cur_ep);
      }
    }
    //! transitions of preceding consensus column to this consensus column
    if (it != cons_col_profs.begin()) {
      map<int, ConsColumnProfile*>::const_iterator jt = --it;
      ++it;

      TransitionProb* tp;
      if (use_eetp) 
    	  tp = new TransitionProb(eetp->tr_probs(sub_prof->sub()->name(), it->first));
      else {
	      TransitionProfile* trans_prof = (jt->second)->trans_prof();
	      vector<int>        cur_tf     = trans_prof->frequencies();
	      tp = new TransitionProb(trans_priors, cur_tf);
      }

      for (int type=(int)Match; type<=(int)Delete; ++type) {
        State* cur_state = m_states[jt->first][type];
        transitions(t_al, tp, cur_state, it->first);
      }
      delete tp;
    }
  }
  //! last column: no INSERT because of general C_INSERT
  map<int, ConsColumnProfile*>::const_iterator it = (--cons_col_profs.end());
  m_states[it->first][(int) Delete] = new State(this, it->first, Delete);
  m_states[it->first][(int) Match]  = new State(this, it->first, Match);

  State*           cur_state  = m_states[it->first][(int) Match];
  EmissionProfile* em_prof    = (it->second)->em_prof();
  vector<mytype>   cur_priors = match_em_priors;
  vector<int>      cur_ef;

  for (unsigned i=0; i<cur_priors.size(); ++i)
    cur_ef += em_prof->match_frequency(i);

  vector<mytype> cur_ep  = emission_probs(cur_priors, cur_ef);

  for (unsigned i=0; i<cur_ep.size(); ++i)
    cur_ep[i] = prob_non_spec_chars*cur_ep[i];
  cur_ep += prob_spec_chars;

  cur_state->log_ep()  = log_of_prob(cur_ep);

  TransitionProfile* trans_prof = ((--it)->second)->trans_prof();
  ++it;

  vector<int>        cur_tf     = trans_prof->frequencies();
  TransitionProb*    tp         = new TransitionProb(trans_priors, cur_tf);

  for (int type=(int)Match; type<=(int)Delete; ++type) {
    cur_state = m_states[(--it)->first][type];
    ++it;
    transitions(t_al, tp, cur_state, it->first);
  }
  delete tp;
}

/*****************************************************************************************************************//**
   ProfileHMM copy constructor.

   Creates a copy of the profile HMM for a local profile HMM.
   @param[in] phmm Original (global) profile HMM for whole subtype.
   @param[in] first_cons_col First consensus column in (local) subtype = start of profile HMM.
   @param[in] last_cons_col Last consensus column in (local) subtype = end of profile HMM.
   @param[in] state_to_state Mapping of state in original profile HMM to corresponding state in copy and vice versa.
**********************************************************************************************************************/
ProfileHMM::ProfileHMM(const ProfileHMM& phmm, 
		       int first_local_cons_col, 
		       int last_local_cons_col, 
		       map<State*, State*> *state_to_state) {

  m_sub_prof = phmm.m_sub_prof;
  m_len      = phmm.m_len;

  multi_array<State*, 2>::extent_gen extents;
  m_states.resize(extents[m_len][NUM_STATES]);

  //! copy only states within currently used region in alignment: [first_cons_col, last_cons_col]
  for (int i=first_local_cons_col; i<last_local_cons_col; ++i) {
    for (int type = (int)Match; type<=(int)Delete; ++type) {
      if (phmm.has_state(i, type)){
        m_states[i][type] = new State(*phmm.m_states[i][type]);
        (*state_to_state)[m_states[i][type]]      = phmm.m_states[i][type];
        (*state_to_state)[phmm.m_states[i][type]] = m_states[i][type];
      }
    }
  }
    //! last cons col: no Insert active
  if (phmm.has_state(last_local_cons_col, (int)Match)) {
    //! match state:
    m_states[last_local_cons_col][(int)Match] = new State(*phmm.m_states[last_local_cons_col][(int)Match]);
    (*state_to_state)[m_states[last_local_cons_col][(int)Match]]      = phmm.m_states[last_local_cons_col][(int)Match];
    (*state_to_state)[phmm.m_states[last_local_cons_col][(int)Match]] = m_states[last_local_cons_col][(int)Match];
    //! delete state:
    m_states[last_local_cons_col][(int)Delete] = new State(*phmm.m_states[last_local_cons_col][(int)Delete]);
    (*state_to_state)[m_states[last_local_cons_col][(int)Delete]]      = phmm.m_states[last_local_cons_col][(int)Delete];
    (*state_to_state)[phmm.m_states[last_local_cons_col][(int)Delete]] = m_states[last_local_cons_col][(int)Delete];
  }
}


/****************************//**
   ProfileHMM destructor.
*********************************/
ProfileHMM::~ProfileHMM() {

  multi_array<State*, 2>::size_type const* ext = this->m_states.shape();
  int num_rows = ext[0];
  int num_cols = ext[1];
  for (int i=0; i<num_rows; ++i)
    for (int j=0; j<num_cols; ++j) {
      if (this->m_states[i][j])
        delete this->m_states[i][j];
  }
}


/***********************//**
   Check.
***************************/
void ProfileHMM::check() {

  multi_array<State*, 2>::size_type const* ext = m_states.shape();
  int num_rows = ext[0];
  int num_cols = ext[1];

  if (!m_sub_prof || !m_len || !num_rows || !num_cols)
   throw Error("Profile in illegal state");

  for (int i=0; i<num_rows; ++i)
    for (int j=0; j<num_cols; ++j)
      m_states[i][j]->check();
}


/******************************************************************************//**
   Transitions.
   
   Transitions for each state in profile HMM.

   @param[in] trans_alphabet Transition alphabet, i.e. allowed state transitions.
   @param[in] tp Transition probability
   @param[in] cur_state Current state.
   @param[in] succ_cons_col Successive consensus column of current state/column.
***********************************************************************************/
void ProfileHMM::transitions(const TransAlphabet& trans_alphabet, 
			     TransitionProb *tp, 
			     State          *cur_state, 
			     int             succ_cons_col) {

  StateType cur_type = cur_state->type();
  bool is_match = false;
  if (cur_state->is_match())
    is_match = true;
  int  cur_col  = cur_state->column();

  for (int succ_type = (int)Match; succ_type <= (int)Delete; ++succ_type) {
    if (cur_state->is_match() || succ_type == (int)Match || succ_type == (int) cur_type) {
      int cur_succ_col = succ_cons_col;
      if (succ_type == (int) Insert)
        cur_succ_col = cur_col;

      Transition* cur_transition = new Transition(cur_state, m_states[cur_succ_col][succ_type]);

      cur_transition->log_p() = log(tp->value(trans_alphabet, cur_type, (StateType) succ_type));
      cur_state->succ().push_back(cur_transition);
      m_states[cur_succ_col][succ_type]->pred().push_back(cur_transition);
    }
  }
}


/******************************************************************************//**
   Emission probabilities.
   
   Emission probabilities of a certain state.

   @param[in] priors Emission priors.
   @param[in] frequencies Observed nucleotide frequencies for current state.
   @param[out] probs Emission probabilities of current state.
***********************************************************************************/
vector<mytype> ProfileHMM::emission_probs(vector<mytype> &priors, vector<int> &frequencies) {

  if (priors.size() != frequencies.size())
    throw Error("number of emissions could not be determined");

  vector<mytype> probs;
  int            sum_freqs  = 0;
  mytype         sum_priors = 0;

  FOREACH (int f, frequencies)
    sum_freqs += f;
  FOREACH (mytype p, priors)
    sum_priors += p;

  mytype sum = (mytype)sum_freqs + sum_priors;
  for (unsigned i=0; i<frequencies.size(); ++i)
    probs += ((mytype)frequencies[i] + priors[i])/sum;

  return probs;
}


/******************************************************************************//**
   Logarithm of probabilities.
   
   Logarithm of all probabilities in a given vector.

   @param[in] probs Current probabilities, emission or transition probs.
   @param[out] log_p Logarithm of all given probabilities. 
***********************************************************************************/
vector<mytype> ProfileHMM::log_of_prob(vector<mytype> probs) {

  vector<mytype> log_p;

  FOREACH (mytype prob, probs) {
    try {
      log_p += log(prob);
    }
    catch (out_of_range) {
      throw Error ("log of value <= 0");
    }
  }

  return log_p;
}

/******************************************************************************//**
   Has state.
   
   Current column has a state of the given state type?.

   @param[in] column Current column.
   @param[in] type State type.
   @param[out] false/true Current column has such a state?. 
***********************************************************************************/
bool ProfileHMM::has_state(int column, int type) const {
  if (column < 0 || column >= m_len || type < (int)Match || type > (int)Delete || m_states[column][type] == 0) {
    return false;
  }
  else
    return true;
}

/******************************************************************************//**
   State.
   
   Return state for given column and of given type.

   @param[in] column Current column.
   @param[in] type State type.
   @param[out] m_states[column][type] Return state. 
***********************************************************************************/
State* ProfileHMM::state(int column, int type) const { 

  if (!has_state(column, type)) {
    cout << "col: " << column << " type " << type << endl;
    throw Error("State doesn't exist");
  }
  return m_states[column][type];
}

/******************************//**
   Output:
***********************************/
ostream& operator<<(ostream& s, const ProfileHMM& p) {

  s << "Profile HMM for subtype " << p.m_sub_prof->sub()->name() << ":" << endl;
  s << "States: " << "(subtype column type)" << endl;
  multi_array<State*, 2>::size_type const* ext = p.states().shape();
  int num_rows = ext[0];
  int num_cols = ext[1];

  for (int i=0; i<num_rows; ++i)
    for (int j=0; j<num_cols; ++j)
      if (p.state(i,j) != 0)
        s << *p.state(i,j) << endl;

  return s;
}


/******************************************************
 *                                                    *
 *                   class Transition                 *
 *                                                    *
 \******************************************************/

/**************************************************************************************//**
   Transition.

   Create transition for certain pair of states.

   @param[in] wherefrom State where the transition starts.
   @param[in] whereto State where the transition end.
******************************************************************************************/
Transition::Transition(State *wherefrom, State *whereto) {
  m_wherefrom = wherefrom;
  m_whereto   = whereto;
}


/*****************************************************************************************************************//**
   Transition copy constructor.

   Creates a copy of a transition.
   @param[in] t Original transition.
**********************************************************************************************************************/
Transition::Transition(const Transition& t) {

  *m_wherefrom = *t.m_wherefrom;
  *m_whereto   = *t.m_whereto;
  m_log_p = t.m_log_p;
}


/****************************************************************//**
   Transition destructor.
********************************************************************/
Transition::~Transition() {

}

/***********************//**
   Check.
***************************/
void Transition::check() {
  if (!m_wherefrom || !m_whereto)
    throw Error("Transition in illegal state");
}


/**********************************************************************//**
   Trans prob.

   Return transition probability for a certain pair of states.
   
   @param[in] wherefrom State where the transition starts.
   @param[in] whereto State where the transition end.
   @param[in] trans_prob Probability of current transition.
**************************************************************************/
mytype Transition::trans_prob(State *wherefrom, State *whereto) {

  vector<Transition*> cur_succ = wherefrom->succ();
  vector<Transition*>::const_iterator it = cur_succ.begin();

  while (it != cur_succ.end() && (*it)->whereto() != whereto) {
    ++it;
  }

  if (it == cur_succ.end()) {
    // cout << "wherefrom: " << wherefrom->sub_name() << " " << wherefrom->column() << " " << wherefrom->type() << " whereto: " << whereto->sub_name() << " " << whereto->column() << " " << whereto->type() << endl;
    throw Error("Transition doesn't exist");
  }
  return (*it)->trans_prob();
}

/******************************//**
   Output:
***********************************/
ostream& operator<<(ostream& s, const Transition& t) {

  s << "wherefrom: " << t.wherefrom()->phmm()->sub_prof()->sub()->name() << " " << t.wherefrom()->type() << " " << t.wherefrom()->column() << endl;
  s << "whereto: "   << t.whereto()->phmm()->sub_prof()->sub()->name() << " " << t.whereto()->type() << " " << t.whereto()->column() << endl;
  s << "probability: " << t.trans_prob() << endl;

  return s;
}


/******************************************************
 *                                                    *
 *                   class State                      *
 *                                                    *
 \******************************************************/

/************************************************//**
   State.

   Create state.

   @param[in] phmm Current profile HMM.
   @param[in] column Current column.
   @param[in] type Current state type.
*****************************************************/
State::State(ProfileHMM* phmm, 
	     int column, 
	     StateType type) {

  m_phmm   = phmm;
  m_column = column;
  m_type   = type;

}

/*****************************************************************************************************************//**
   State copy constructor.

   Creates a copy of a state.
   @param[in] st Original state.
**********************************************************************************************************************/
State::State(const State& st) {

  m_phmm   = st.m_phmm;
  m_column = st.m_column;
  m_type   = st.m_type;
  m_log_ep = st.m_log_ep;
}

/****************************************************************//**
   ProfileHMM destructor.
********************************************************************/
State::~State() {
  for (int i=0; i<(int)m_succ.size(); ++i)
     delete m_succ[i];
}

/***********************//**
   Check.
***************************/
void State::check() {

  if (!m_phmm || (m_pred.size()==0 && m_succ.size()==0))
    throw Error("State in illegal state");

  FOREACH (Transition* t, m_succ)
    t->check();
}

/*****************************************************//**
   Type.

   Type of current state.
   @param[out] m_type Type of current state.
**********************************************************/
StateType State::type()   const { 
  return m_type; 
}

/*****************************************************//**
   Is begin.

   Current state is begin state?.
   @param[out] yes/no Current state is begin state?.
**********************************************************/
bool State::is_begin() const {
  if (m_type == Begin)
    return true;
  return false;
}

/*****************************************************//**
   Is n_delete.

   Current state is n_delete state?.
   @param[out] yes/no Current state is n_delete state?.
**********************************************************/
bool State::is_n_delete() const {
  if (m_type == N_Delete)
    return true;
  return false;
}

/*****************************************************//**
   Is n_insert.

   Current state is n_insert state?.
   @param[out] yes/no Current state is n_insert state?.
**********************************************************/
bool State::is_n_insert() const {
  if (m_type == N_Insert)
    return true;
  return false;
}

/*****************************************************//**
   Is match.

   Current state is match state?.
   @param[out] yes/no Current state is match state?.
**********************************************************/
bool State::is_match() const {
  if (m_type == Match)
    return true;
  return false;
}

/*****************************************************//**
   Is insert.

   Current state is insert state?.
   @param[out] yes/no Current state is insert state?.
**********************************************************/
bool State::is_insert() const {
  if (m_type == Insert)
    return true;
  return false;
}

/*****************************************************//**
   Is delete.

   Current state is delete state?.
   @param[out] yes/no Current state is delete state.
**********************************************************/
bool State::is_delete() const {
  if (m_type == Delete)
    return true;
  return false;
}

/*****************************************************//**
   Is c_insert.

   Current state is c_insert state?.
   @param[out] yes/no Current state is c_insert state.
**********************************************************/
bool State::is_c_insert() const {
  if (m_type == C_Insert)
    return true;
  return false;
}

/*****************************************************//**
   Is c_delete.

   Current state is c_delete state?.
   @param[out] yes/no Current state is c_delete state.
**********************************************************/
bool State::is_c_delete() const {
  if (m_type == C_Delete)
    return true;
  return false;
}

/*****************************************************//**
   Is end.

   Current state is end state?.
   @param[out] yes/no Current state is end state.
**********************************************************/
bool State::is_end() const {
  if (m_type == End)
    return true;
  return false;
}

/*****************************************************//**
   Is special.

   Current state is special state 
   (begin, n_insert, n_delete, c_insert, c_delete, end)?.
   @param[out] yes/no Current state is special state.
**********************************************************/
bool State::is_special() const {
  if (m_type == Match || m_type == Insert || m_type == Delete || m_type == Jump_State_Match || m_type == Jump_State_Delete)
    return false;
  return true;
}

/*****************************************************//**
   Is mute.

   Current state is mute state (doesn't emit any symbol)?.
   @param[out] yes/no Current state is mute state.
**********************************************************/
bool State::is_mute() const {
  if (m_type == Match || m_type == Insert || m_type == N_Insert || m_type == C_Insert)
    return false;
  return true;
}

/*****************************************************//**
   Is jump state match.

   Current state is jump state to match states?.
   @param[out] yes/no Current state is jump state to match states.
**********************************************************/
bool State::is_jump_state_match() const {
  if (m_type == Jump_State_Match)
    return true;
  else
    return false;
}

/*****************************************************//**
   Is jump state delete.

   Current state is jump state to delete states?.
   @param[out] yes/no Current state is jump state to delete states.
**********************************************************/
bool State::is_jump_state_delete() const {
  if (m_type == Jump_State_Delete)
    return true;
  else
    return false;
}

/*****************************************************//**
   Is jump state.

   Current state is jump state?.
   @param[out] yes/no Current state is jump state.
**********************************************************/
bool State::is_jump_state() const {
  if (m_type == Jump_State_Match || m_type == Jump_State_Delete)
    return true;
  else
    return false;
}

/*****************************************************//**
   Succ standardization.

   Standardization of transition probabilities 
   to all successors of a certain state.
   @param[in] first_cons_col First consensus column in the model.
   @param[in] last_cons_col Last consensus column in the model.
**********************************************************/
void State::succ_standardization(int first_cons_col, int last_cons_col) {
  mytype sum_trans_prob = 0;

  FOREACH (Transition* t, this->succ()) {
    sum_trans_prob += t->trans_prob();
  }
  if (sum_trans_prob > 0) {
    FOREACH (Transition* t, this->succ()) {
      mytype trans_prob = (t->trans_prob())/sum_trans_prob;
      t->log_p() = log(trans_prob);
    }
  }
  //check for each state (except B and E) if it has successors AND predecessors
  if (!this->is_begin() && !this->is_end()) {
    if (this->succ().size() == 0 || this->pred().size() == 0)
      cout << this->type() << " " << this->column() << " has no predecessor or successor " << this->succ().size() << " " << this->pred().size() << endl;
  }
}

/******************************//**
   Output:
***********************************/
ostream& operator<<(ostream& s, const State& st) {

  s << "STATE: " << st.m_phmm->sub_prof()->sub()->name() << "\t" << st.m_column << "\t" << st.m_type << endl;
  s << " em_prob: ";
  FOREACH (mytype l, st.m_log_ep)
    s << exp(l) << " ";

  s << endl << " successors: (subtype column type trans_prob)" << endl;
  FOREACH (Transition* t, st.m_succ) {
    State* cur_succ = t->whereto();
    s << "  " << cur_succ->phmm()->sub_prof()->sub()->name() << "\t" << cur_succ->column() << "\t" << cur_succ->type() << "\t" << t->trans_prob() << endl;
  }
  return s;
}


/******************************************************
 *                                                    *
 *               class TransitionProb                 *
 *                                                    *
 \******************************************************/

/**************************************************************************************//**
   TransitionProb.

   Calculation of all transition probabilities for a certain column in a subtype.

   @param[in] priors Transition priors.
   @param[in] frequencies Observed transition frequencies for a certain column.
******************************************************************************************/
TransitionProb::TransitionProb(vector<mytype> priors, vector<int> frequencies) {

  if (priors.size() != frequencies.size())
    throw Error("number of transitions could not be determined");

  vector<mytype> probs;

  vector<int> poss_trans_per_state;
  poss_trans_per_state += 3,2,2;

  int start = 0;
  FOREACH (int num_trans, poss_trans_per_state) {
    int    sum_freqs  = 0;
    mytype sum_priors = 0;

    int cur_end = start+num_trans;
    for (int i=start; i<cur_end; ++i) {
      sum_freqs  += frequencies[i];
      sum_priors += priors[i];
    }
    mytype sum = (mytype)sum_freqs + sum_priors;
    for (int i=start; i<cur_end; ++i){
      probs += ((mytype)frequencies[i] + priors[i])/sum;
    }
    start += num_trans;
  }

  m_trans_probs = probs;
}

/***********************//**
   Check.
***************************/
void TransitionProb::check() {

}

/**************************************************************************************//**
   Value.

   Transition probability for a certain pair of state types.

   @param[in] t_al Transition alphabet, i.e. possible state transitions.
   @param[in] wherefrom Type of state where the transition starts.
   @param[in] whereto Type of state where the transition end.
   @param[out] m_trans_probs[t_al.idx(wherefrom, whereto)] Current transition probability.
******************************************************************************************/
mytype TransitionProb::value(const TransAlphabet& t_al, 
			     StateType wherefrom, 
			     StateType whereto) {
  pair<StateType, StateType> p = make_pair(wherefrom, whereto);
  return m_trans_probs[t_al.idx(wherefrom, whereto)];
}

/******************************//**
   Output:
***********************************/
ostream& operator<<(ostream& s, const TransitionProb& tp) {

  s << tp.m_trans_probs << endl;

  return s;
}
