// ~/jpHMM/src/profileHMM.h
/**************************************************************
 File name	: profileHMM.h
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Classes: ProfileHMM
          Transition
          State
          TransitionProb

 - Construction of profile HMM for a certain subtype
    incl. definition of states and transitions
 - Determination of emission and transition probabilities
***************************************************************/

#ifndef _PROFILEHMM_H
#define _PROFILEHMM_H

#include "std_header.h"

using namespace std;

class TransAlphabet;
class EmissionPriors;
class TransitionPriors;
class TransitionProb;
class SubtypeProfile;
class State;
class Transition;
class ExtEmTrProbs;


/******************************************************\
 *                class ProfileHMM                    *
\******************************************************/

class ProfileHMM {

  friend ostream& operator<<(ostream& s, const ProfileHMM& p);
  public:
    ProfileHMM(){};
    ProfileHMM(SubtypeProfile* sub_prof, const TransitionPriors& t, const EmissionPriors& e, const TransAlphabet& t_al, mytype prob_spec_chars, bool use_eetp, const ExtEmTrProbs* eetp = 0);
    ProfileHMM(const ProfileHMM &phmm, int first_local_cons_col, int last_local_cons_col, map<State*, State*> *state_to_state); //! copy constructor
    ~ProfileHMM();
    void                  check();
    void                  transitions(const TransAlphabet& t_al, TransitionProb *tp, State *cur_state, int succ_cons_column);
    static vector<mytype> emission_probs(vector<mytype> &priors, vector<int> &frequencies);
    static vector<mytype> log_of_prob(vector<mytype> probs);
    bool                  has_state(int column, int type) const;
    State*                state(int column, int type)     const;
    SubtypeProfile*       sub_prof()                      const { return m_sub_prof; };
    string                sub_name()                      const { return sub_prof()->sub()->name(); }
    int                   len()                           const { return m_len; };
    multi_array<State*,2> states()                        const { return m_states; };

  private:
    SubtypeProfile*       m_sub_prof;
    int                   m_len;
    multi_array<State*,2> m_states;
};


/******************************************************\
 *                 class Transition                   *
\******************************************************/

class Transition{

  friend ostream& operator<<(ostream& s, const Transition& t);
  public:

    //Transition() {};
    Transition(State* wherefrom, State* whereto);
    Transition(const Transition& t); //! copy constructor
    ~Transition();
    void check();
    mytype trans_prob(State *wherefrom, State *whereto);

    State*  wherefrom()  const { return m_wherefrom; };
    State*  whereto()    const { return m_whereto; };
    mytype  trans_prob() const { return exp(m_log_p); };
    mytype  log_p()      const { return m_log_p; };
    mytype &log_p()            { return m_log_p; };

  private:

    State* m_wherefrom;   //! transition between 2 states: start of transition
    State* m_whereto;     //!                              end of transition
    mytype m_log_p;
};


/******************************************************\
 *                   class State                      *
\******************************************************/

class State{

  friend ostream& operator<<(ostream& s, const State& st);
  public:

    State() {};
    State(ProfileHMM* phmm, int cur_column, StateType type);
    State(const State& st); // copy constructor
    ~State();
    void check();
    bool is_begin()    const;
    bool is_n_delete() const;
    bool is_n_insert() const;
    bool is_match()    const;
    bool is_insert()   const;
    bool is_delete()   const;
    bool is_c_insert() const;
    bool is_c_delete() const;
    bool is_end()      const;
    bool is_special()  const;
    bool is_mute()     const;
    
    bool is_jump_state()        const;
    bool is_jump_state_match()  const;
    bool is_jump_state_delete() const;
    
    void succ_standardization(int first_cons_col, int last_cons_col);

    StateType            type()         const;
    ProfileHMM*          phmm()         const { return m_phmm; };
    string               sub_name()     const { return phmm() ? phmm()->sub_prof()->sub()->name() : "-"; }
    int                  column()       const { return m_column; };
    int                  id()           const { return m_id; };
    int                 &id()                 { return m_id; };
    vector<Transition*>  pred()         const { return m_pred; };
    vector<Transition*> &pred()               { return m_pred; };
    vector<Transition*>  succ()         const { return m_succ; };
    vector<Transition*> &succ()               { return m_succ; };
    mytype               em_prob(int i) const { return exp(m_log_ep[i]); };
    vector<mytype>       log_ep()       const { return m_log_ep; };
    vector<mytype>      &log_ep()             { return m_log_ep; };

  private:
    ProfileHMM*         m_phmm;
    int                 m_column;
    StateType           m_type;            //! type of State (B, M, D, I, I_C, I_N, D_C, D_N, E)
    int                 m_id;              //! id in list of all HMM states
    vector<Transition*> m_pred;            //! predecessors of a state
    vector<Transition*> m_succ;            //! successors of a state
    vector<mytype>      m_log_ep;          //! log of emission probs
};

/******************************************************\
 *                class TransitionProb                *
\******************************************************/

class TransitionProb {

  friend ostream& operator<<(ostream& s, const TransitionProb& tp);
  public:
    TransitionProb(vector<mytype> priors, vector<int> frequencies);
    TransitionProb(vector<double> probs) : m_trans_probs(probs) { }
    ~TransitionProb(){};
    void   check(); // TODO
    mytype value(const TransAlphabet& t_al, StateType wherefrom, StateType whereto);

  private:
    vector<mytype> m_trans_probs;
};

ostream& operator<<(ostream& s, const ProfileHMM& p);
ostream& operator<<(ostream& s, const Transition& t);
ostream& operator<<(ostream& s, const State& st);
#endif        // _PROFILEHMM_H
