// ~/jpHMM/src/transitionProfile.cpp
/**************************************************************
 File name	: transitionProfile.cpp
 Author		: Anne-Kathrin Schultz
 Version	: 2.0
 Date		: Oct 2010

Content:

 Determination of observed transition frequencies 
  between different state types in the given sequences in the alignment
***************************************************************/

#include "std_source.h"
#include "error.h"
#include "alphabet.h"
#include "const_def.h"
#include "alignment.h"
#include "emTransProfile.h"
#include "transitionProfile.h"


using namespace std;


/*******************************************************\

 *                 class TransitionProfile               *

\*******************************************************/

TransitionProfile::TransitionProfile(ConsColumnProfile* cons_prof, const TransAlphabet& t_al) {

  m_column    = cons_prof->column();

  Subtype*      cur_sub       = cons_prof->sub_prof()->sub();
  vector<Seq*>  cur_seqs      = cur_sub->seqs();
  int           sub_len       = cur_sub->len();
  int succ_cons_col = cons_prof->succ_column();

  m_frequencies.assign(t_al.num_trans(), 0);

  FOREACH (Seq* s, cur_seqs) {
    string cur_str = s->str();

    int count_inserts_per_seq = 0;
    for (int col=m_column+1; col<succ_cons_col; ++col) {
      if (cur_str[col] != '-' && cur_str[col] != '*') { //! or other gap characters
        ++m_frequencies[t_al.idx(Insert, Insert)];
	++count_inserts_per_seq;
      }
    }

    if (count_inserts_per_seq != 0) { //! at least 1 INSERT in this gap
      --m_frequencies[t_al.idx(Insert, Insert)]; //! for gap opening
      if (m_column == -1 || (m_column >= 0 && cur_str[m_column] != '-' && cur_str[m_column] != '*'))
        ++m_frequencies[t_al.idx(Match, Insert)]; //! M->I OR B->I (cur_col = -1)
      if (succ_cons_col == sub_len || (succ_cons_col < sub_len && (cur_str[succ_cons_col] != '-' && cur_str[succ_cons_col] != '*')))
        ++m_frequencies[t_al.idx(Insert, Match)]; //! M->M OR M->E (succ_cons_col = sub_len)
    } 
    else {
      if (m_column >=0 && (cur_str[m_column] == '-' || cur_str[m_column] == '*')) { //! Delete in cur_col
	if (succ_cons_col == sub_len || (succ_cons_col < sub_len && (cur_str[succ_cons_col] != '-' && cur_str[succ_cons_col] != '*')))
          ++m_frequencies[t_al.idx(Delete, Match)];
	else
          ++m_frequencies[t_al.idx(Delete, Delete)];
      }
      else { //! MATCH OR BEGIN
        if (succ_cons_col == sub_len || (succ_cons_col < sub_len && (cur_str[succ_cons_col] != '-' && cur_str[succ_cons_col] != '*')))
          ++m_frequencies[t_al.idx(Match, Match)];
        else
          ++m_frequencies[t_al.idx(Match, Delete)];
      }
    }
  }
  check();
}

//------------------------------------------------------------------------------
// void check():
//------------------------------------------------------------------------------
void TransitionProfile::check() {
  if (!m_frequencies.size())
    throw Error("Column transition profile in illegal state");
}

//------------------------------------------------------------------------------
// output:
//------------------------------------------------------------------------------
ostream& operator<<(ostream& s, const TransitionProfile& t) {
  s << "transition frequencies (MM,MI,MD,IM,II,DM,DD): " << t.m_frequencies;
  return s;
}

